/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bfilter.h"
#include <string>
#include <vector>
#include <algorithm>
#include <iterator>

#ifdef WIN32

#include <winsock2.h>

#else

#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#ifdef HAVE_SYS_FCNTL_H
#include <sys/fcntl.h>
#endif

#endif

#include "network.h"
#include "tselect.h"
#include "httperrormessages.h"
#include "sessionfinishexception.h"
#include "syscall.h"
#include "debug.h"

#define SERVER_READ_TIMEOUT  500
#define CLIENT_READ_TIMEOUT  100
#define SERVER_WRITE_TIMEOUT 100
#define CLIENT_WRITE_TIMEOUT 100
#define IDLE_TIMEOUT         40

BFilter::BFilter(Network::Socket csock, State* state)
:	m_state(state), m_csock(csock), m_ssock(Network::INVALID_SOCK), m_server_port(-1),
	m_html_processor(this), m_request_stream(this), m_response_stream(this),
	m_numPendingRequests(0)
{
}

void
BFilter::run()
{
	try {
		runUnsafe();
		DEBUGLOG("Session finishes normally");
	} catch (SessionFinishException& e) {
#ifdef DEBUG
		if (e.isNormalFinish()) {
			DEBUGLOG("Session finishes normally");
		} else {
			DEBUGLOG("Abnormal session finish (dbginfo: "+e.getDebugInfo());
		}
#endif
	}
	flushClientData();
	if (Network::isValidSocket(m_ssock)) {
		Network::closeSocket(m_ssock);
	}
	// client socket will be closed by the creator of this object
}

void
BFilter::runUnsafe()
{
	fd_set rfds, wfds;
	Network::Socket maxfd;
	TSelect tselect;
	struct timeval sread_timeout;
	sread_timeout.tv_sec = SERVER_READ_TIMEOUT;
	sread_timeout.tv_usec = 0;
	struct timeval cread_timeout;
	cread_timeout.tv_sec = CLIENT_READ_TIMEOUT;
	cread_timeout.tv_usec = 0;
	struct timeval swrite_timeout;
	swrite_timeout.tv_sec = SERVER_WRITE_TIMEOUT;
	swrite_timeout.tv_usec = 0;
	struct timeval cwrite_timeout;
	cwrite_timeout.tv_sec = CLIENT_WRITE_TIMEOUT;
	cwrite_timeout.tv_usec = 0;
	struct timeval idle_timeout;
	idle_timeout.tv_sec = IDLE_TIMEOUT;
	idle_timeout.tv_usec = 0;
	int sread_timer = 0;
	int cread_timer = 0;
	int swrite_timer = 0;
	int cwrite_timer = 0;
	int idle_timer = 0;
	uint8_t recv_buf[8192];
	enum { IOV_ELEMENTS = 4 };
	Network::IoVec iov[IOV_ELEMENTS];
	
	while (true) {
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_SET(m_csock, &rfds);
		maxfd = m_csock;
		if (requestStream().isInitialState()) {
			if (cread_timer) {
				tselect.removeTimer(cread_timer);
				cread_timer = 0;
			}
		} else {
			if (!cread_timer)
				cread_timer = tselect.addTimer(cread_timeout);
		}
		if (requestStream().isInitialState() &&
		    responseStream().isInitialState() &&
		    m_numPendingRequests == 0) {
			if (!idle_timer)
				idle_timer = tselect.addTimer(idle_timeout);
		} else if (idle_timer) {
			tselect.removeTimer(idle_timer);
			idle_timer = 0;
		}
		if (!m_dataForClient.isEmpty()) {
			FD_SET(m_csock, &wfds);
			if (!cwrite_timer)
				cwrite_timer = tselect.addTimer(cwrite_timeout);
		}
		if (!m_dataForServer.isEmpty()) {
			if (!Network::isValidSocket(m_ssock)) {
				makeServerConnection();
			}
			FD_SET(m_ssock, &wfds);
			if (m_ssock > maxfd)
				maxfd = m_ssock;
			if (!swrite_timer)
				swrite_timer = tselect.addTimer(swrite_timeout);
		}
		if (Network::isValidSocket(m_ssock)) {
			FD_SET(m_ssock, &rfds);
			if (m_ssock > maxfd)
				maxfd = m_ssock;
			if (!sread_timer && !responseStream().isInitialState())
				sread_timer = tselect.addTimer(sread_timeout);
		} else if (sread_timer) {
			tselect.removeTimer(sread_timer);
			sread_timer = 0;
		}
		bool finish = false;
		bool can_finish = (requestStream().isInitialState() && responseStream().isInitialState());
		int res = -1;
		do {
			if (can_finish && m_state->isMasterDead()) {
				finish = true;
				break;
			}
			res = tselect.selectNoRetry(maxfd+1, &rfds, &wfds, NULL);
#ifdef WIN32
		} while (false);
#else
		} while (res == -1 && errno == EINTR);
#endif
		if (finish) {
			flushClientData();
			break;
		}
		
		if (res < 0) {
			// error
			throw SessionFinishException(false, "select() error");
		} else if (res == 0) {
			std::set<int> timers = tselect.getTriggeredTimers();
			if (timers.find(cread_timer) != timers.end()) {
				throw SessionFinishException(false, "Client read timout");
			} else if (timers.find(cwrite_timer) != timers.end()) {
				throw SessionFinishException(false, "Client write timeout");
			} else if (timers.find(idle_timer) != timers.end()) {
				throw SessionFinishException(true);
			} else {
				// server timeout
				closeServerConnection(false);
				responseStream().resetState();
				continue;
			}
		}
		if (FD_ISSET(m_csock, &rfds)) {
			if (cread_timer) {
				tselect.removeTimer(cread_timer);
				cread_timer = 0;
			}
			if (idle_timer) {
				tselect.removeTimer(idle_timer);
				idle_timer = 0;
			}
			int len = -1;
			SYSCALL(len = recv(m_csock, (char*)recv_buf, sizeof(recv_buf), 0));
			if (len < 0) {
				// error
				throw SessionFinishException(false, "Client read error: "+Network::getLastError());
			} else if (len == 0) {
				// client eof
				DEBUGLOG("client EOF");
				requestStream().consumeEOF();
				throw SessionFinishException(true);
			}
			LOG_TRAFFIC_FROM_CLIENT(recv_buf, recv_buf+len);
			requestStream().consumeDataChunk(recv_buf, recv_buf+len);
		}
		if (Network::isValidSocket(m_ssock) && FD_ISSET(m_ssock, &rfds)) {
			if (sread_timer) {
				tselect.removeTimer(sread_timer);
				sread_timer = 0;
			}
			if (idle_timer) {
				tselect.removeTimer(idle_timer);
				idle_timer = 0;
			}
			int len = -1;
			SYSCALL(len = recv(m_ssock, (char*)recv_buf, sizeof(recv_buf), 0));
			if (len < 0) {
				// error
				DEBUGLOG("Error reading from server: " << Network::getLastError());
				responseStream().consumeEOF();
				closeServerConnection(true);
				responseStream().resetState();
			} else if (len == 0) {
				// server eof
				DEBUGLOG("Server closes connection");
				responseStream().consumeEOF();
				closeServerConnection(true);
				responseStream().resetState();
			} else {
				LOG_TRAFFIC_FROM_SERVER(recv_buf, recv_buf+len);
				responseStream().consumeDataChunk(recv_buf, recv_buf+len);
			}
		}
		if (FD_ISSET(m_csock, &wfds) && !m_dataForClient.isEmpty()) {
			if (cwrite_timer) {
				tselect.removeTimer(cwrite_timer);
				cwrite_timer = 0;
			}
			int n_chunks = 0;
			Buffer::RangeIterator iter(m_dataForClient.begin(), m_dataForClient.end());
			for (const Buffer::Range* range; (range = iter.next()); ) {
				iov[n_chunks].iov_base = reinterpret_cast<char*>(range->begin());
				iov[n_chunks].iov_len = range->end() - range->begin();
				if (++n_chunks == IOV_ELEMENTS) {
					break;
				}
			}
			int len = Network::writev(m_csock, iov, n_chunks);
			if (len < 0) {
				throw SessionFinishException(false, "Error writing to client: "+Network::getLastError());
			}
#ifdef DEBUG
			std::vector<uint8_t> traf;
			Buffer::appendToByteVector(traf, m_dataForClient.begin(), m_dataForClient.begin() + len);
			LOG_TRAFFIC_TO_CLIENT(&traf[0], &traf[traf.size()]);
#endif
			m_dataForClient.shiftBy(len);
		}
		if (Network::isValidSocket(m_ssock) && FD_ISSET(m_ssock, &wfds) && !m_dataForServer.isEmpty()) {
			if (swrite_timer) {
				tselect.removeTimer(swrite_timer);
				swrite_timer = 0;
			}
			int n_chunks = 0;
			Buffer::RangeIterator iter(m_dataForServer.begin(), m_dataForServer.end());
			for (const Buffer::Range* range; (range = iter.next()); ) {
				iov[n_chunks].iov_base = reinterpret_cast<char*>(range->begin());
				iov[n_chunks].iov_len = range->end() - range->begin();
				if (++n_chunks == IOV_ELEMENTS) {
					break;
				}
			}
			int len = Network::writev(m_ssock, iov, n_chunks);
			if (len < 0) {
				DEBUGLOG("Error writing to server: " << Network::getLastError());
				if (responseStream().isForwardingInProgress()) {
					throw SessionFinishException(false, "Error writing to server: "+Network::getLastError());
				} else {
					DEBUGLOG("Closing the server connection but don't finishing session");
					closeServerConnection(true);
					responseStream().resetState();
				}
			}
#ifdef DEBUG
			std::vector<uint8_t> traf;
			Buffer::appendToByteVector(traf, m_dataForServer.begin(), m_dataForServer.begin() + len);
			LOG_TRAFFIC_TO_SERVER(&traf[0], &traf[traf.size()]);
#endif
			m_dataForServer.shiftBy(len);
		}
	}
}

void
BFilter::moreDataForServer(const uint8_t* begin, const uint8_t* end)
{
#ifndef MSG_DONTWAIT
	m_dataForServer.appendData(begin, end);
#else
	if (end - begin < 300 || !m_dataForServer.isEmpty() || !Network::isValidSocket(m_ssock)) {
		m_dataForServer.appendData(begin, end);
		return;
	}
	int len;
	SYSCALL(len = send(m_ssock, begin, end - begin, MSG_DONTWAIT));
	if (len > 0) {
#ifdef DEBUG
		LOG_TRAFFIC_TO_SERVER(begin, begin + len);
#endif
		if ((begin += len) == end) {
			return;
		} else {
			m_dataForServer.appendData(begin, end);
		}
	} else {
		m_dataForServer.appendData(begin, end);
	}
#endif
}

void
BFilter::moreDataForServer(const std::string& str)
{
	moreDataForServer(
		reinterpret_cast<const uint8_t*>(str.data()),
		reinterpret_cast<const uint8_t*>(str.data()+str.length())
	);
}

void
BFilter::moreDataForClient(const uint8_t* begin, const uint8_t* end)
{
#ifndef MSG_DONTWAIT
	m_dataForClient.appendData(begin, end);
#else
	if (end - begin < 300 || !m_dataForClient.isEmpty()) {
		m_dataForClient.appendData(begin, end);
		return;
	}
	int len;
	SYSCALL(len = send(m_csock, begin, end - begin, MSG_DONTWAIT));
	if (len > 0) {
#ifdef DEBUG
		LOG_TRAFFIC_TO_CLIENT(begin, begin + len);
#endif
		if ((begin += len) == end) {
			return;
		} else {
			m_dataForClient.appendData(begin, end);
		}
	} else {
		m_dataForClient.appendData(begin, end);
	}
#endif
}

void
BFilter::moreDataForClient(const std::string& str)
{
	moreDataForClient(
		reinterpret_cast<const uint8_t*>(str.data()),
		reinterpret_cast<const uint8_t*>(str.data()+str.length())
	);
}

namespace {
	struct RequestCantBeRetried
	{
		bool operator()(IntrusivePtr<HttpClientRequest>& request) {
			return !request->isIdempotent();
		}
	};
}

void
BFilter::closeServerConnection(bool servers_initiative)
{
	DEBUGLOG("closeServerConnection()");
	if (!Network::isValidSocket(m_ssock)) {
		DEBUGLOG("connection already closed");
		return;
	}
	Network::closeSocket(m_ssock);
	m_ssock = Network::INVALID_SOCK;
	m_server_host.erase();
	m_server_port = 0;
	if (responseStream().isForwardingInProgress()) {
		throw SessionFinishException(false, "Server connection was closed in the middle of response.");
	}
	if (servers_initiative) {
		if (serverInfo().getServedRequestsCount() == 0) {
			throw SessionFinishException(false, "Server has closed connection without serving a single request");
		}
	}
	m_dataForServer.clear();
	if (m_numPendingRequests) {
		RequestQueue::iterator begin = requestsToBeServed().begin();
		RequestQueue::iterator end = begin;
		std::advance(end, m_numPendingRequests);
		if (std::find_if(begin, end, RequestCantBeRetried()) != end) {
			throw SessionFinishException(false, "Server has closed connection, "
				"and some of the unanswered requests can't be retried");
		}
	}
	m_numPendingRequests = 0;
	if (!requestsToBeServed().empty())
		requestStream().dispatchMoreRequests();
}

void
BFilter::switchToServer(const std::string& host, int port)
{
	DEBUGLOG("switchToServer(" << host << ":" << port << ")");
#ifdef DEBUG
	if (m_numPendingRequests) {
		DEBUGLOG("WARNING: have some pending requests in switchToServer()");
	}
#endif
	if (m_server_host == host && m_server_port == port) {
		return;
	}
	if (Network::isValidSocket(m_ssock)) {
		Network::closeSocket(m_ssock);
		m_ssock = Network::INVALID_SOCK;
	}
	m_server_host = host;
	m_server_port = port;
}

void
BFilter::makeServerConnection()
{
	DEBUGLOG("makeServerConnection(" << m_server_host << ":" << m_server_port << ")");
	serverInfo().reset();
	if (Network::isValidSocket(m_ssock)) {
		Network::closeSocket(m_ssock);
		m_ssock = Network::INVALID_SOCK;
	}
	std::string host;
	int port;
	if (getState().getConfig().willGoThroughProxy(m_server_host)) {
		host = getState().getConfig().getProxyHost();
		port = getState().getConfig().getProxyPort();
		DEBUGLOG("connecting to proxy: " << host << ":" << port);
	} else {
		host = m_server_host;
		port = m_server_port;
	}
	NetworkError err;
	m_ssock = Network::tcpConnect(host.c_str(), port, &err);
	if (Network::isValidSocket(m_ssock)) {
		DEBUGLOG("connected");
		return;
	} else {
		if (!responseStream().isForwardingInProgress()) {
			std::string response;
			switch (err.getErrorCode()) {
			case NetworkError::ERR_UNKNOWN_HOST:
				response = HttpErrorMessages::errDNS(
					requestsToBeServed().front()->getOrigMetadata().requestLine(), host);
				break;
			case NetworkError::ERR_CONNECTION_REFUSED:
				response = HttpErrorMessages::errConnRefused(
					requestsToBeServed().front()->getOrigMetadata().requestLine(), host, port);
				break;
			case NetworkError::ERR_CONNECTION_TIMEOUT:
				response = HttpErrorMessages::errConnTimeout(
					requestsToBeServed().front()->getOrigMetadata().requestLine(), host, port);
				break;
			case NetworkError::ERR_NETWORK_UNREACHABLE:
				response = HttpErrorMessages::errNetworkUnreachable(
					requestsToBeServed().front()->getOrigMetadata().requestLine(), host, port);
				break;
			case NetworkError::ERR_HOST_UNREACHABLE:
				response = HttpErrorMessages::errHostUnreachable(
					requestsToBeServed().front()->getOrigMetadata().requestLine(), host, port);
				break;
			default:
				response = HttpErrorMessages::errInternal(
					requestsToBeServed().front()->getOrigMetadata().requestLine(), err.getErrorMessage());
			}
			moreDataForClient(response);
			LOG_ERROR_OUTGOING_RESPONSE(response);
		}
		throw SessionFinishException(false, "Could not connect to server");
	}
}

void
BFilter::flushClientData()
{
	fd_set rfds, wfds;
	TSelect tselect;
	struct timeval tv;
	char buf[4096];
	tv.tv_sec = CLIENT_WRITE_TIMEOUT;
	tv.tv_usec = 0;
	int tid = tselect.addTimer(tv);
	bool got_eof = false;
	while (!m_dataForClient.isEmpty()) {
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		if (!got_eof)
			FD_SET(m_csock, &rfds);
		FD_SET(m_csock, &wfds);
		int res = tselect.select(m_csock+1, &rfds, &wfds, NULL);
		if (res <= 0) {
			// timeout or error
			break;
		}
		if (!got_eof && FD_ISSET(m_csock, &rfds)) {
			int len = -1;
			SYSCALL(len = recv(m_csock, (char*)buf, sizeof(buf), 0));
			if (len < 0) {
				// client read error
				break;
			} else if (len == 0) {
				// client eof
				got_eof = true;
			}
		}
		if (FD_ISSET(m_csock, &wfds)) {
			Buffer::RangeIterator iter(m_dataForClient.begin(), m_dataForClient.end());
			const Buffer::Range* range = iter.next();
			int len;
			SYSCALL(len = send(m_csock, (const char*)range->begin(), range->end() - range->begin(), 0));
			if (len < 0) {
#ifdef WIN32
				if (WSAGetLastError() == WSAEWOULDBLOCK) {
					// a bug in Win9x
					continue;
				}
#endif
				break;
			}
			LOG_TRAFFIC_TO_CLIENT(range->begin(), range->begin() + len);
			m_dataForClient.shiftBy(len);
			tselect.resetTimer(tid);
		}
	}
	m_dataForClient.clear();
	DEBUGLOG("flushClientData() finishes");
}

void
BFilter::appendRequest(HttpClientRequest* request)
{
	requestsToBeServed().push_back(RequestPtr(request));
}

void
BFilter::shiftRequests()
{
#ifdef DEBUG
	if (requestsToBeServed().empty()) {
		DEBUGLOG("No requests to shift in BFilter::shiftRequests()");
		return;
	}
#endif
	requestsToBeServed().pop_front();
	if (m_numPendingRequests) {
		--m_numPendingRequests;
	} // or else it was a substitution request
}

