/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "buffer.h"
#include <cstring>
#include <cstdlib>
#include <algorithm>

Buffer::Buffer()
{
}

Buffer::Buffer(const Buffer& other)
{
	Block* block = other.m_impl.m_leftSentinel.m_pNext;
	for (; block != &other.m_impl.m_rightSentinel; block = block->m_pNext) {
		appendData(block->m_pBegin, block->m_pEnd);
	}
}

Buffer&
Buffer::operator=(const Buffer& rhs)
{
	clear();
	Block* block = rhs.m_impl.m_leftSentinel.m_pNext;
	for (; block != &rhs.m_impl.m_rightSentinel; block = block->m_pNext) {
		appendData(block->m_pBegin, block->m_pEnd);
	}
	return *this;
}

void
Buffer::clear()
{
	for (Block* block = m_impl.m_rightSentinel.m_pPrev; block != &m_impl.m_leftSentinel; ) {
		// we go from right to left because blocks may borrow memory from their left neighbour
		Block* tmp = block;
		block = block->m_pPrev;
		if (tmp->isOwner()) {
			std::free(tmp->m_pMemBegin);
		}
	}
	m_impl.m_allocSize = 0;
	m_impl.m_numAssimilatedBlocks = 0;
	m_impl.m_foreignBlocks.clear();
	m_impl.m_leftSentinel.m_pPrev = m_impl.m_rightSentinel.m_pPrev = &m_impl.m_leftSentinel;
	m_impl.m_leftSentinel.m_pNext = m_impl.m_rightSentinel.m_pNext = &m_impl.m_rightSentinel;
}

void
Buffer::appendToString(std::string& str, const ConstIterator& begin, const ConstIterator& end)
{
	str.reserve(end - begin + str.size());
	ConstRangeIterator iter(begin, end);
	for (const ConstRange* range; (range = iter.next()); ) {
		str.append(reinterpret_cast<const char*>(range->begin()),
			reinterpret_cast<const char*>(range->end()));
	}
}

void
Buffer::appendToByteVector(std::vector<uint8_t>& vec, const ConstIterator& begin, const ConstIterator& end)
{
	std::size_t pos = vec.size();
	vec.resize(end - begin + pos);
	ConstRangeIterator iter(begin, end);
	for (const ConstRange* range; (range = iter.next()); ) {
		std::size_t len = range->end() - range->begin();
		std::memcpy(&vec[pos], range->begin(), len);
		pos += len;
	}
}

void
Buffer::shiftBy(std::size_t amount)
{
	for (Block* block = m_impl.m_leftSentinel.m_pNext;; ) {
		std::size_t block_size = block->m_pEnd - block->m_pBegin;
		if (amount < block_size) {
			block->m_pBegin += amount;
			break;
		}
		amount -= block_size;
		block = block->m_pNext;
		removeFirstBlock();
	}
}

void
Buffer::shiftTo(const ConstIterator& iter)
{
	Block* stop_block = iter.m_pBlock;
	for (Block* block = m_impl.m_leftSentinel.m_pNext; block != stop_block; ) {
		block = block->m_pNext;
		removeFirstBlock();
	}
	stop_block->m_pBegin = stop_block->m_pIterBase + iter.m_offset;
}

void
Buffer::trimBy(std::size_t amount)
{
	for (Block* block = m_impl.m_rightSentinel.m_pPrev;; ) {
		std::size_t block_size = block->m_pEnd - block->m_pBegin;
		if (amount < block_size) {
			block->m_pEnd -= amount;
			break;
		}
		amount -= block_size;
		block = block->m_pPrev;
		removeLastBlock();
	}
}

void
Buffer::trimTo(const ConstIterator& iter)
{
	Block* stop_block = iter.m_pBlock;
	for (Block* block = m_impl.m_rightSentinel.m_pPrev; block != stop_block; ) {
		block = block->m_pPrev;
		removeLastBlock();
	}
	stop_block->m_pEnd = stop_block->m_pIterBase + iter.m_offset;
}

void
Buffer::appendData(const uint8_t* begin, const uint8_t* end)
{
	if ((begin += appendToBlock(m_impl.m_rightSentinel.m_pPrev, begin, end)) == end) {
		// this also protects from begin == end situation
		return;
	}
	insertBlockBefore(createBlock(begin, end), &m_impl.m_rightSentinel);
}

void
Buffer::appendData(const ConstIterator& begin, const ConstIterator& end)
{
	ConstRangeIterator iter(begin, end);
	for (const ConstRange* range; (range = iter.next()); ) {
		appendData(range->begin(), range->end());
	}
}

void
Buffer::appendData(const std::string& data)
{
	appendData(
		reinterpret_cast<const uint8_t*>(data.data()),
		reinterpret_cast<const uint8_t*>(data.data()+data.length())
	);
}

void
Buffer::appendForeignData(uint8_t* begin, uint8_t* end)
{
	if (begin == end) {
		return;
	}
	Block block;
	block.m_memOwnership = Block::MEM_FOREIGN;
	block.m_pBegin = block.m_pMemBegin = block.m_pIterBase = begin;
	block.m_pEnd = block.m_pMemEnd = end;
	m_impl.m_foreignBlocks.push_back(block);
	insertBlockBefore(&m_impl.m_foreignBlocks.back(), &m_impl.m_rightSentinel);
}

void
Buffer::assimilateForeignData()
{
	typedef std::deque<Block>::iterator Iter;
	Iter cur = m_impl.m_foreignBlocks.begin();
	std::advance(cur, m_impl.m_numAssimilatedBlocks);
	Iter end = m_impl.m_foreignBlocks.end();
	for (; cur != end; ++cur) {
		Block* block = &*cur;
		if (block->m_memOwnership != Block::MEM_FOREIGN) {
			// already assimilated
			continue;
		}
		++m_impl.m_numAssimilatedBlocks;
		std::ptrdiff_t bsize = block->m_pEnd - block->m_pBegin;
		Block* prev = block->m_pPrev;
		if (bsize <= prev->m_pMemEnd - prev->m_pEnd) {
			block->m_memOwnership = Block::MEM_BORROWING_FROM_PREV;
			uint8_t* new_begin = prev->m_pEnd;
			std::memcpy(new_begin, block->m_pBegin, bsize);
			block->m_pIterBase = new_begin - (block->m_pBegin - block->m_pIterBase);
			block->m_pBegin = new_begin;
			block->m_pEnd = new_begin + bsize;
			block->m_pMemBegin = prev->m_pMemBegin;
			block->m_pMemEnd = prev->m_pMemEnd;
		} else {
			block->m_memOwnership = Block::MEM_CURRENT_OWNER;
			std::size_t new_size = adviseBlockSize(bsize);
			uint8_t* new_begin = reinterpret_cast<uint8_t*>(std::malloc(new_size));
			m_impl.m_allocSize += new_size;
			std::memcpy(new_begin, block->m_pBegin, bsize);
			block->m_pIterBase = new_begin - (block->m_pBegin - block->m_pIterBase);
			block->m_pBegin = block->m_pMemBegin = new_begin;
			block->m_pEnd = new_begin + bsize;
			block->m_pMemEnd = new_begin + new_size;
		}
	}
}

Buffer::Block*
Buffer::createBlock(const uint8_t* begin, const uint8_t* end)
{
	std::size_t requested_size = end - begin;
	std::size_t alloc_size = adviseBlockSize(requested_size) + sizeof(Block);
	uint8_t* buf = reinterpret_cast<uint8_t*>(std::malloc(alloc_size));
	m_impl.m_allocSize += alloc_size;
	Block* block = new(buf) Block;
	block->m_memOwnership = Block::MEM_ORIGINAL_OWNER;
	block->m_pMemBegin = block->m_pIterBase = buf;
	block->m_pBegin = buf + sizeof(Block);
	block->m_pMemEnd = buf + alloc_size;
	block->m_pEnd = block->m_pBegin + requested_size;
	std::memcpy(block->m_pBegin, begin, requested_size);
	return block;
}

void
Buffer::freeBlock(Block* block)
{
	Block* prev = block->m_pPrev;
	Block* next = block->m_pNext;
	if (block->isOwner()) {
		if (next->m_memOwnership == Block::MEM_BORROWING_FROM_PREV) {
			next->m_memOwnership = Block::MEM_CURRENT_OWNER;
		} else {
			m_impl.m_allocSize -= block->m_pMemEnd - block->m_pMemBegin;
			std::free(block->m_pMemBegin);
		}
	}
	prev->m_pNext = next;
	next->m_pPrev = prev;
}

std::size_t
Buffer::appendToBlock(Block* block, const uint8_t* begin, const uint8_t* end)
{
	std::ptrdiff_t append_size = std::min(end - begin, block->m_pMemEnd - block->m_pEnd);
	std::memcpy(block->m_pEnd, begin, append_size);
	block->m_pEnd += append_size;
	return append_size;
}

void
Buffer::removeFirstBlock()
{
	Block* block = m_impl.m_leftSentinel.m_pNext;
	Block::MemoryOwnership ownership = block->m_memOwnership;
	freeBlock(block);
	if (!m_impl.m_foreignBlocks.empty() && block == &m_impl.m_foreignBlocks.front()) {
		if (ownership != Block::MEM_FOREIGN) {
			--m_impl.m_numAssimilatedBlocks;
		}
		m_impl.m_foreignBlocks.pop_front();
	}
}

void
Buffer::removeLastBlock()
{
	Block* block = m_impl.m_rightSentinel.m_pPrev;
	Block::MemoryOwnership ownership = block->m_memOwnership;
	freeBlock(block);
	if (!m_impl.m_foreignBlocks.empty() && block == &m_impl.m_foreignBlocks.back()) {
		if (ownership != Block::MEM_FOREIGN) {
			--m_impl.m_numAssimilatedBlocks;
		}
		m_impl.m_foreignBlocks.pop_back();
	}
}

Buffer::ConstIterator&
Buffer::ConstIterator::operator++()
{
	if (m_pConstBlock->m_pIterBase + ++m_offset == m_pConstBlock->m_pEnd) {
		m_pConstBlock = m_pConstBlock->m_pNext;
		m_offset = m_pConstBlock->m_pBegin - m_pConstBlock->m_pIterBase;
	}
	return *this;
}

Buffer::ConstIterator&
Buffer::ConstIterator::operator--()
{
	if (m_pConstBlock->m_pIterBase + m_offset-- == m_pConstBlock->m_pBegin) {
		m_pConstBlock = m_pConstBlock->m_pPrev;
		m_offset = m_pConstBlock->m_pEnd - m_pConstBlock->m_pIterBase - 1;
	}
	return *this;
}

Buffer::ConstIterator&
Buffer::ConstIterator::operator+=(std::size_t inc)
{
	const Block* block = m_pConstBlock;
	std::ptrdiff_t offset = m_offset;
	std::size_t remaining_space = block->m_pEnd - (block->m_pIterBase + offset);
	while (true) {
		if (remaining_space > inc) {
			m_pConstBlock = block;
			m_offset = offset + inc;
			return *this;
		}
		block = block->m_pNext;
		inc -= remaining_space;
		remaining_space = block->m_pEnd - block->m_pBegin;
		offset = block->m_pBegin - block->m_pIterBase;
	}
}

Buffer::ConstIterator&
Buffer::ConstIterator::operator-=(std::size_t dec)
{
	const Block* block = m_pConstBlock;
	std::ptrdiff_t offset = m_offset;
	std::size_t remaining_space = block->m_pIterBase + offset - block->m_pBegin + 1;
	while (true) {
		if (remaining_space > dec) {
			m_pConstBlock = block;
			m_offset = offset - dec;
			return *this;
		}
		block = block->m_pPrev;
		dec -= remaining_space;
		remaining_space = block->m_pEnd - block->m_pBegin;
		offset = block->m_pEnd - block->m_pIterBase - 1;
	}
}

std::size_t
operator-(const Buffer::ConstIterator& to, const Buffer::ConstIterator& from)
{
	if (from.m_pConstBlock == to.m_pConstBlock) {
		return to.m_offset - from.m_offset;
	}
	const Buffer::Block* block = from.m_pConstBlock;
	std::size_t size = block->m_pEnd - (block->m_pIterBase + from.m_offset);
	while ((block = block->m_pNext) != to.m_pConstBlock) {
		size += block->m_pEnd - block->m_pBegin;
	}
	size += block->m_pIterBase + to.m_offset - block->m_pBegin;
	return size;
}


Buffer::RangeIterator::RangeIterator(const Iterator& begin, const Iterator& end)
:	m_begin(begin), m_end(end)
{
}

const Buffer::Range*
Buffer::RangeIterator::next()
{
	Block* block;
	uint8_t* begin;
	uint8_t* end;
	
	if (!m_range.m_impl.m_pBlock) {
		block = m_begin.m_pBlock;
		begin = block->m_pIterBase + m_begin.m_offset;
	} else if (m_range.m_impl.m_pBlock != m_end.m_pBlock) {
		block = m_range.m_impl.m_pBlock->m_pNext;
		begin = block->m_pBegin;
	} else {
		return 0;
	}
	
	if (block == m_end.m_pBlock) {
		end = block->m_pIterBase + m_end.m_offset;
	} else {
		end = block->m_pEnd;
	}
	
	if (begin == end) {
		return 0;
	}
	
	m_range.m_impl.m_pBlock = block;
	m_range.m_impl.m_pBegin = begin;
	m_range.m_impl.m_pEnd = end;
	return &m_range;
}


Buffer::ConstRangeIterator::ConstRangeIterator(const ConstIterator& begin, const ConstIterator& end)
:	m_begin(begin), m_end(end)
{
}

const Buffer::ConstRange*
Buffer::ConstRangeIterator::next()
{
	const Block* block;
	const uint8_t* begin;
	const uint8_t* end;
	
	if (!m_range.m_constImpl.m_pBlock) {
		block = m_begin.m_pConstBlock;
		begin = block->m_pIterBase + m_begin.m_offset;
	} else if (m_range.m_impl.m_pBlock != m_end.m_pConstBlock) {
		block = m_range.m_impl.m_pBlock->m_pNext;
		begin = block->m_pBegin;
	} else {
		return 0;
	}
	
	if (block == m_end.m_pConstBlock) {
		end = block->m_pIterBase + m_end.m_offset;
	} else {
		end = block->m_pEnd;
	}
	
	if (begin == end) {
		return 0;
	}
	
	m_range.m_constImpl.m_pBlock = block;
	m_range.m_constImpl.m_pBegin = begin;
	m_range.m_constImpl.m_pEnd = end;
	return &m_range;
}


Buffer::Impl::Impl()
{
	static uint8_t* const s_begin = 0;
	static uint8_t* const s_end = s_begin + 1;
	m_leftSentinel.m_memOwnership = m_rightSentinel.m_memOwnership = Block::MEM_NA;
	m_leftSentinel.m_pIterBase = m_rightSentinel.m_pIterBase = s_begin;
	m_leftSentinel.m_pBegin = m_leftSentinel.m_pMemBegin = s_begin;
	m_rightSentinel.m_pBegin = m_rightSentinel.m_pMemBegin = s_begin;
	m_leftSentinel.m_pEnd = m_leftSentinel.m_pMemEnd = s_end;
	m_rightSentinel.m_pEnd = m_rightSentinel.m_pMemEnd = s_end;
	m_leftSentinel.m_pPrev = m_rightSentinel.m_pPrev = &m_leftSentinel;
	m_leftSentinel.m_pNext = m_rightSentinel.m_pNext = &m_rightSentinel;
	m_allocSize = 0;
	m_numAssimilatedBlocks = 0;
}

Buffer::Impl::~Impl()
{
	for (Block* block = m_rightSentinel.m_pPrev; block != &m_leftSentinel; ) {
		// we go from right to left because blocks may borrow memory from their left neighbour
		Block* tmp = block;
		block = block->m_pPrev;
		if (tmp->isOwner()) {
			std::free(tmp->m_pMemBegin);
		}
	}
}

