/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef BUFFER_H_
#define BUFFER_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "types.h"
#include <cstddef>
#include <string>
#include <vector>
#include <deque>
#include <iterator>

class Buffer
{
public:
	Buffer();
	
	Buffer(const Buffer& other);
	
	virtual ~Buffer() {}
	
	Buffer& operator=(const Buffer& rhs);
private:
	struct Block
	{
	public:
		enum MemoryOwnership {
			MEM_NA                  = 1, // for sentinel blocks
			MEM_ORIGINAL_OWNER      = 2,
			MEM_CURRENT_OWNER       = 4,
			MEM_BORROWING_FROM_PREV = 8,
			MEM_FOREIGN             = 16
		};
		
		bool isOwner() const { return m_memOwnership & 6; }
		
		Block* m_pPrev;
		Block* m_pNext;
		uint8_t* m_pBegin;
		uint8_t* m_pEnd;
		uint8_t* m_pMemBegin; /* Begin of the allocated memory block
		Equals to m_pBegin if foreign. If MEM_ORIGINAL_OWNER, the object itself is located there. */
		uint8_t* m_pMemEnd;   // End of the allocated memory block. Equals to m_pEnd if foreign.
		uint8_t* m_pIterBase; /* Base data pointer for iterators.
		Initially set to m_pMemBegin, adjusted when assimilating a foreign block. */
		MemoryOwnership m_memOwnership;
	};
private:
	struct IteratorImpl
	{
		union
		{
			Block* m_pBlock;
			const Block* m_pConstBlock;
		};
		std::ptrdiff_t m_offset;
	};
public:
	class RangeIterator;
	class ConstRangeIterator;
	
	class ConstIterator : protected IteratorImpl
	{
		friend class Buffer;
		friend class Buffer::ConstRangeIterator;
		friend std::size_t operator-(const ConstIterator&, const ConstIterator&);
	public:
		typedef std::random_access_iterator_tag iterator_category;
		typedef const uint8_t                   value_type;
		typedef std::size_t                     difference_type;
		typedef const uint8_t*                  pointer;
		typedef const uint8_t&                  reference;
		
		ConstIterator() { m_pConstBlock = 0; }
		bool isOutOfBounds() const { return !m_pConstBlock->m_pBegin; }
		bool isAtEnd() const { return m_pConstBlock->m_pNext == m_pConstBlock; }
		bool isAtREnd() const { return m_pConstBlock->m_pPrev == m_pConstBlock; }
		uint8_t operator*() const { return m_pConstBlock->m_pIterBase[m_offset]; }
		ConstIterator& operator++();
		ConstIterator& operator--();
		const ConstIterator operator++(int);
		const ConstIterator operator--(int);
		ConstIterator& operator+=(std::size_t inc);
		ConstIterator& operator-=(std::size_t dec);
		bool operator==(const ConstIterator& other) const;
		bool operator!=(const ConstIterator& other) const { return !(*this==other); }
	protected:
		ConstIterator(const Block* b, std::ptrdiff_t off) {
			m_pConstBlock = b;
			m_offset = off;
		}
	};
	
	class Iterator : public ConstIterator
	{
		friend class Buffer;
		friend class Buffer::RangeIterator;
	public:
		typedef std::random_access_iterator_tag iterator_category;
		typedef uint8_t                         value_type;
		typedef std::size_t                     difference_type;
		typedef uint8_t*                        pointer;
		typedef uint8_t&                        reference;
		
		Iterator() { m_pBlock = 0; }
		uint8_t& operator*() { return m_pBlock->m_pIterBase[m_offset]; }
		Iterator& operator++() {
			return reinterpret_cast<Iterator&>(++*static_cast<ConstIterator*>(this));
		}
		Iterator& operator--() {
			return reinterpret_cast<Iterator&>(--*static_cast<ConstIterator*>(this));
		}
		const Iterator operator++(int);
		const Iterator operator--(int);
		Iterator& operator+=(std::size_t inc) {
			return reinterpret_cast<Iterator&>(*static_cast<ConstIterator*>(this) += inc);
		}
		Iterator& operator-=(std::size_t dec) {
			return reinterpret_cast<Iterator&>(*static_cast<ConstIterator*>(this) -= dec);
		}
	private:
		Iterator(Block* b, std::ptrdiff_t off)
		: ConstIterator(b, off) {}
	};
	
	friend std::size_t operator-(const ConstIterator&, const ConstIterator&);
private:
	class RangeImpl
	{
		friend class Buffer::RangeIterator;
		friend class Buffer::ConstRangeIterator;
	protected:
		struct Impl
		{
			uint8_t* m_pBegin;
			uint8_t* m_pEnd;
			Block* m_pBlock;
		};
		struct ConstImpl
		{
			const uint8_t* m_pBegin;
			const uint8_t* m_pEnd;
			const Block* m_pBlock;
		};
		union
		{
			Impl m_impl;
			ConstImpl m_constImpl;
		};
	};
public:
	class ConstRange : protected RangeImpl
	{
		friend class Buffer::ConstRangeIterator;
	public:
		ConstRange() { m_constImpl.m_pBlock = 0; }
		const uint8_t* begin() const { return m_constImpl.m_pBegin; }
		const uint8_t* end() const { return m_constImpl.m_pEnd; }
	};
	
	class Range : protected ConstRange
	{
		friend class Buffer::RangeIterator;
	public:
		Range() { m_impl.m_pBlock = 0; }
		uint8_t* begin() const { return m_impl.m_pBegin; }
		uint8_t* end() const { return m_impl.m_pEnd; }
	};
	
	class RangeIterator
	{
	public:
		RangeIterator(const Iterator& begin, const Iterator& end);
		const Range* next();
		void reset() { m_range.m_impl.m_pBlock = 0; }
	private:
		Iterator m_begin;
		Iterator m_end;
		Range m_range;
	};
	
	class ConstRangeIterator
	{
	public:
		ConstRangeIterator(const ConstIterator& begin, const ConstIterator& end);
		const ConstRange* next();
		void reset() { m_range.m_constImpl.m_pBlock = 0; }
	private:
		ConstIterator m_begin;
		ConstIterator m_end;
		ConstRange m_range;
	};
public:
	Iterator begin();
	ConstIterator begin() const;
	Iterator end();
	ConstIterator end() const;
	Iterator rbegin();
	ConstIterator rbegin() const;
	Iterator rend();
	ConstIterator rend() const;
	Iterator at(std::size_t pos) { return begin()+=pos; }
	ConstIterator at(std::size_t pos) const { return begin()+=pos; }
	
	static void appendToString(std::string& str, const ConstIterator& begin, const ConstIterator& end);
	static void appendToByteVector(std::vector<uint8_t>& vec, const ConstIterator& begin, const ConstIterator& end);
	
	bool isEmpty() const;
	void clear();
	void shiftBy(std::size_t amount);
	void shiftTo(const ConstIterator& iter);
	void trimBy(std::size_t amount);
	void trimTo(const ConstIterator& iter);
	void appendData(const uint8_t* begin, const uint8_t* end);
	void appendData(const std::string& data);
	void appendData(const ConstIterator& begin, const ConstIterator& end);
	void appendForeignData(uint8_t* begin, uint8_t* end);
	void assimilateForeignData();
private:
	std::size_t adviseBlockSize(std::size_t requested_size);
	Block* createBlock(const uint8_t* begin, const uint8_t* end);
	void freeBlock(Block* block);
	static void insertBlockBefore(Block* new_block, Block* before_this);
	static std::size_t appendToBlock(Block* block, const uint8_t* begin, const uint8_t* end);
	void removeFirstBlock();
	void removeLastBlock();
private:
	class Impl
	{
	public:
		Impl();
		~Impl();
	private:
		Impl(const Impl&);            //
		Impl& operator=(const Impl&); // forbidden
	public:
		Block m_leftSentinel;
		Block m_rightSentinel;
		std::size_t m_allocSize;
		std::deque<Block> m_foreignBlocks;
		std::size_t m_numAssimilatedBlocks;
	};
	Impl m_impl;
};


inline bool
Buffer::isEmpty() const
{
	return m_impl.m_leftSentinel.m_pNext == &m_impl.m_rightSentinel;
}

inline Buffer::Iterator
Buffer::begin()
{
	Block* block = m_impl.m_leftSentinel.m_pNext;
	return Iterator(block, block->m_pBegin - block->m_pIterBase);
}

inline Buffer::ConstIterator
Buffer::begin() const
{
	const Block* block = m_impl.m_leftSentinel.m_pNext;
	return ConstIterator(block, block->m_pBegin - block->m_pIterBase);
}

inline Buffer::Iterator
Buffer::end()
{
	return Iterator(&m_impl.m_rightSentinel, 0);
}

inline Buffer::ConstIterator
Buffer::end() const
{
	return ConstIterator(&m_impl.m_rightSentinel, 0);
}

inline Buffer::Iterator
Buffer::rbegin()
{
	Block* block = m_impl.m_rightSentinel.m_pPrev;
	return Iterator(block, block->m_pEnd - block->m_pIterBase - 1);
}

inline Buffer::ConstIterator
Buffer::rbegin() const
{
	const Block* block = m_impl.m_rightSentinel.m_pPrev;
	return ConstIterator(block, block->m_pEnd - block->m_pIterBase - 1);
}

inline Buffer::Iterator
Buffer::rend()
{
	return Iterator(&m_impl.m_leftSentinel, 0);
}

inline Buffer::ConstIterator
Buffer::rend() const
{
	return ConstIterator(&m_impl.m_leftSentinel, 0);
}

inline std::size_t
Buffer::adviseBlockSize(std::size_t requested_size)
{
	return ((requested_size + ((m_impl.m_allocSize + requested_size) >> 1)) + 32) & ~std::size_t(31);
}

inline void
Buffer::insertBlockBefore(Block* new_block, Block* before_this)
{
	new_block->m_pNext = before_this;
	new_block->m_pPrev = before_this->m_pPrev;
	before_this->m_pPrev->m_pNext = new_block;
	before_this->m_pPrev = new_block;
}

inline const Buffer::ConstIterator
Buffer::ConstIterator::operator++(int)
{
	ConstIterator old(*this);
	++*this;
	return old;
}

inline const Buffer::ConstIterator
Buffer::ConstIterator::operator--(int)
{
	ConstIterator old(*this);
	--*this;
	return old;
}

inline const Buffer::Iterator
Buffer::Iterator::operator++(int)
{
	Iterator old(*this);
	++*this;
	return old;
}

inline const Buffer::Iterator
Buffer::Iterator::operator--(int)
{
	Iterator old(*this);
	--*this;
	return old;
}

inline bool
Buffer::ConstIterator::operator==(const ConstIterator& other) const
{
	return (m_pConstBlock == other.m_pConstBlock && m_offset == other.m_offset);
}

inline const Buffer::ConstIterator
operator+(const Buffer::ConstIterator& lhs, std::size_t rhs)
{
	Buffer::ConstIterator iter(lhs);
	iter += rhs;
	return iter;
}

inline const Buffer::Iterator
operator+(const Buffer::Iterator& lhs, std::size_t rhs)
{
	Buffer::Iterator iter(lhs);
	iter += rhs;
	return iter;
}

inline const Buffer::ConstIterator
operator-(const Buffer::ConstIterator& lhs, std::size_t rhs)
{
	Buffer::ConstIterator iter(lhs);
	iter -= rhs;
	return iter;
}

inline const Buffer::Iterator
operator-(const Buffer::Iterator& lhs, std::size_t rhs)
{
	Buffer::Iterator iter(lhs);
	iter -= rhs;
	return iter;
}

std::size_t
operator-(const Buffer::ConstIterator& to, const Buffer::ConstIterator& from);

#endif

