/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "clientrequeststream.h"
#include "bfilter.h"
#include "httperrormessages.h"
#include "sessionfinishexception.h"
#include "stringutils.h"
#include "debug.h"
#include <algorithm>

ClientRequestStream::ClientRequestStream(BFilter* owner)
:	m_pOwner(owner)
{
}

void
ClientRequestStream::eventError(int ecode, const char *info)
{
	if (!info) {
		info = "";
	}
	DEBUGLOG("ClientRequestStream::eventError(ecode = " << ecode << ")");
	
	if (m_pOwner->requestsToBeServed().empty() || m_pOwner->requestsToBeServed().back()->isComplete()) {
		// the request which generated this error, has not yet emerged
		m_pOwner->appendRequest(new HttpClientRequest(
			HttpRequestMetadata(
				HttpRequestLine(
					"GET", URI("http://dummy.request-original.was.broken/bf.intercept-request/error_substitute"),
					HttpVersion::HTTP_1_0
				)
			)
		));
	}
	
	const HttpRequestLine& request_line = m_pOwner->requestsToBeServed().back()->getOrigMetadata().requestLine();
	
	std::string error_message;
	switch (ecode) {
	case ERR_PARSING_REQUEST_LINE:
		error_message = HttpErrorMessages::errMalformedRequestLine(request_line);
		break;
	case ERR_PARSING_HEADER:
		error_message = HttpErrorMessages::errMalformedRequestHeader(request_line);
		break;
	case ERR_PARSING_CHUNK_HEADER:
		error_message = HttpErrorMessages::errMalformedChunkInRequest(request_line);
		break;
	case ERR_LENGTH_REQUIRED:
		error_message = HttpErrorMessages::errRequestBodyLengthRequired(request_line);
		break;
	case ERR_UNEXPECTED_EOF:
		throw SessionFinishException(false, "Unexpected client EOF");
		break;
	case ERR2_BAD_REQUEST_URI:
		error_message = HttpErrorMessages::errBadRequestURI(request_line, info);
		break;
	case ERR2_UNSUPPORTED_METHOD:
		error_message = HttpErrorMessages::errUnsupportedRequestMethod(request_line);
		break;
	case ERR2_UNSUPPORTED_EXPECTATION:
		error_message = HttpErrorMessages::errUnsupportedExpectation(request_line, info);
		break;
	default:
		error_message = HttpErrorMessages::errInternal(request_line,
			"Unhandled error in ClientRequestStream: "+StringUtils::fromNumber(ecode));
	}
	
	BFilter::RequestPtr req(m_pOwner->requestsToBeServed().back());
	req->setInterceptHandler(HttpClientRequest::IH_ERROR_SUBSTITUTE);
	req->setErrorResponse(error_message);
	req->setComplete();
	if (m_pOwner->getNumPendingRequests() == m_pOwner->requestsToBeServed().size()) {
		// This means the request that generated this error has already been
		// sent to server.  We just pretend it hasn't.  This is not a problem
		// as we are going to finish this session after generating an error response.
		m_pOwner->decNumPendingRequests();
	}
	dispatchMoreRequests();
}

void
ClientRequestStream::eventGotMetadata(HttpRequestMetadata& metadata)
{
	LOG_INCOMING_REQUEST_STARTS(metadata.requestLine().toString(), metadata.headers().toString());
	
	if (metadata.requestLine().getHttpVersion() > m_pOwner->clientInfo().getProtocolCompability()) {
		m_pOwner->clientInfo().setProtocolCompability(metadata.requestLine().getHttpVersion());
	}
	
	DEBUGLOG("Request begins: " << metadata.requestLine().getURI().toString());
	
	m_pOwner->appendRequest(new HttpClientRequest(metadata));
	m_ptrCurrentRequest = m_pOwner->requestsToBeServed().back();
	prepareRequestForForwarding(m_ptrCurrentRequest->metadata());
	
	std::string method = metadata.requestLine().getMethod();
	if (method != "GET" && method != "POST" && method != "PUT" && method != "HEAD") {
		eventError(ERR2_UNSUPPORTED_METHOD, method.c_str());
	}
	if (!metadata.requestLine().getURI().isAbsolute()) {
		eventError(ERR2_BAD_REQUEST_URI, "URI must be absolute");
	} else if (metadata.requestLine().getURI().getScheme() != "http") {
		eventError(ERR2_BAD_REQUEST_URI, "Scheme not supported");
	} else {
		HttpHeader* expect_hdr = metadata.headers().getHeaderPtr("Expect");
		if (expect_hdr) {
			std::list<HttpHeaderElement> elements = expect_hdr->elements().getElements();
			std::list<HttpHeaderElement>::iterator iter = elements.begin();
			for (; iter != elements.end(); ++iter) {
				if (iter->getLowercaseName() != "100-continue") {
					eventError(ERR2_UNSUPPORTED_EXPECTATION, iter->getName().c_str());
					break;
				}
			}
		}
	}
	
	dispatchMoreRequests();
}

void
ClientRequestStream::eventGotContentPart(const uint8_t* begin, const uint8_t* end)
{
	m_ptrCurrentRequest->appendBodyData(begin, end);
	if (m_ptrStreamWriter.get()) {
		m_ptrStreamWriter->write(begin, end);
	}
}

void
ClientRequestStream::eventMessageEnds(bool connection_needs_closing)
{
	LOG_INCOMING_REQUEST_FINISHES();
	DEBUGLOG("ClientRequestStream::eventMessageEnds()");
	m_pOwner->serverInfo().registerRequest(); // FIXME: this does not mean the request has been sent
	m_ptrCurrentRequest->setComplete();
	m_ptrCurrentRequest.reset(0);
	if (m_ptrStreamWriter.get()) {
		LOG_OUTGOING_REQUEST_FINISHES();
		m_ptrStreamWriter->end();
	}
	dispatchMoreRequests();
}

void
ClientRequestStream::moreData(const uint8_t* begin, const uint8_t* end)
{
	m_pOwner->moreDataForServer(begin, end);
}

void
ClientRequestStream::done(bool conn_close)
{
	m_ptrStreamWriter.reset(0);
#ifdef DEBUG
	if (conn_close) {
		DEBUGLOG2("BUG");
	}
#endif
}

void
ClientRequestStream::prepareRequestForForwarding(HttpRequestMetadata& metadata) const
{
	HttpHeader connection_hdr = metadata.headers().getHeader("Connection");
	connection_hdr.elements().appendElements(metadata.headers().getHeader("Proxy-Connection").elements());
	std::list<HttpHeaderElement> hop_by_hop_headers = connection_hdr.elements().getElements();
	std::list<HttpHeaderElement>::iterator iter = hop_by_hop_headers.begin();
	for (; iter != hop_by_hop_headers.end(); ++iter) {
		metadata.headers().removeHeader(iter->getName());
	}
	metadata.requestLine().setHttpVersion(HttpVersion::HTTP_1_1);
	metadata.headers().removeHeader("Keep-Alive");
	metadata.headers().removeHeader("Proxy-Connection");
	metadata.headers().removeHeader("Upgrade");
	metadata.headers().setHeader(HttpHeader("Connection", "te"));
	metadata.headers().setHeader(HttpHeader("Accept-Encoding", "gzip, deflate"));
	metadata.headers().setHeader(HttpHeader("TE", "gzip, deflate"));
	if (!metadata.headers().hasHeader("Host")) {
		std::string host = metadata.requestLine().getURI().getHost();
		int port = metadata.requestLine().getURI().getPort();
		if (port >= 0) {
			host += ":"+StringUtils::fromNumber(port);
		}
		metadata.headers().setHeader(HttpHeader("Host", host));
	}
	if (!m_pOwner->getState().getConfig().willGoThroughProxy(metadata.requestLine().getURI().getHost())) {
		URI uri = metadata.requestLine().getURI();
		uri.makeRelative();
		metadata.requestLine().setURI(uri);
	}
}

void
ClientRequestStream::dispatchMoreRequests()
{
	DEBUGLOG("dispatchMoreRequests()");
	DEBUGLOG("REQUESTS_TO_BE_SERVED: " << m_pOwner->requestsToBeServed().size());
	DEBUGLOG("FIRST_QUEUED_REQUEST: " << m_pOwner->m_numPendingRequests);
	
	while (!m_pOwner->requestsToBeServed().empty()) {
		// in this loop we are going to serve the substitution requests at the beginning of the list
		HttpClientRequest* request = m_pOwner->requestsToBeServed().front().get();
		if (request->isToBeSent()) {
			break;
		}
		if (!request->isComplete()) {
			return; // wait for it to finish
		}
		m_pOwner->responseStream().processSubstitutionRequest(request);
		m_pOwner->shiftRequests();
	}
	
	HttpClientRequest* last_request_sent_to_server = 0;
	if (std::size_t npr = m_pOwner->getNumPendingRequests()) {
		last_request_sent_to_server = m_pOwner->requestsToBeServed()[npr-1].get();
	}
	
	BFilter::RequestQueue::iterator iter = m_pOwner->requestsToBeServed().begin();
	std::advance(iter, m_pOwner->getNumPendingRequests());
	BFilter::RequestQueue::iterator end = m_pOwner->requestsToBeServed().end();
	for (; iter != end; ++iter) {
		HttpClientRequest* request = iter->get();
		bool send_it_now = true;
		if (last_request_sent_to_server) {
			if (!last_request_sent_to_server->isIdempotent() || !request->isIdempotent()) {
				send_it_now = false;
			} else if (!isSameServer(request->getOrigMetadata().requestLine().getURI(),
			           last_request_sent_to_server->getOrigMetadata().requestLine().getURI())) {
				send_it_now = false;
			} else if (m_pOwner->serverInfo().getPipeliningCapability() == ServerInfo::P_UNKNOWN
			           && m_pOwner->serverInfo().getServedRequestsCount() > 0) {
				m_pOwner->serverInfo().setPipeliningCapability(ServerInfo::P_TESTING);
				// send_it_now = true;
				DEBUGLOG("Pipelining a single request to test the pipelining capability of the server");
			} else if (m_pOwner->serverInfo().getPipeliningCapability() != ServerInfo::P_OK) {
				send_it_now = false;
			}
		} else {
			m_pOwner->switchToServer(request->getOrigMetadata().requestLine().getURI().getHost(),
				request->getOrigMetadata().requestLine().getURI().guessPort());
		}
		if (!send_it_now) {
			return;
		}
		last_request_sent_to_server = request;
		if (request->isComplete()) {
			LOG_OUTGOING_REQUEST_STARTS(request->metadata().requestLine().toString(),
				request->metadata().headers().toString());
			m_pOwner->incNumPendingRequests();
			m_ptrStreamWriter.reset(new HttpStreamWriter(this, request->metadata(), true, false));
			m_ptrStreamWriter->write(request->getBodyData());
			m_ptrStreamWriter->end();
			LOG_OUTGOING_REQUEST_FINISHES();
		} else {
			LOG_OUTGOING_REQUEST_STARTS(request->metadata().requestLine().toString(),
				request->metadata().headers().toString());
			m_pOwner->incNumPendingRequests();
			m_ptrStreamWriter.reset(new HttpStreamWriter(this, request->metadata(), true, false));
			m_ptrStreamWriter->write(request->getBodyData());
			return;
		}
	}
}

bool
ClientRequestStream::isSameServer(const URI& one, const URI& other)
{
	return StringUtils::toLowerStr(one.getHost()) == StringUtils::toLowerStr(other.getHost())
		&& one.guessPort() == other.guessPort();
}

