/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "compressorresponsefilter.h"
#include "gzipcompressor.h"
#include "deflatecompressor.h"
#include "httpheaderelement.h"

AbstractResponseFilter*
CompressorResponseFilter::create(NextFilterProvider* follower,
	HttpResponseMetadata* metadata, const AcceptEncoding& aenc)
{
	static const char* compressors[] = { "gzip", "deflate" };
	static const char** compressors_end = compressors+sizeof(compressors)/sizeof(*compressors);
	const char** enc_ptr = aenc.bestAccepted(compressors, compressors_end);
	if (enc_ptr == compressors_end) {
		return 0;
	}
	if (metadata->headers().hasHeader("Transfer-Encoding")) {
		metadata->headers().getHeaderPtr("Transfer-Encoding")
			->elements().appendElement(HttpHeaderElement(*enc_ptr));
	} else {
		if (!metadata->headers().hasHeader("Content-Encoding")) {
			metadata->headers().setHeader(HttpHeader("Content-Encoding", *enc_ptr));
		} else {
			metadata->headers().getHeaderPtr("Content-Encoding")
				->elements().appendElement(HttpHeaderElement(*enc_ptr));
		}
	}
	metadata->setBodyStatus(HttpMessageMetadata::BODY_SIZE_UNKNOWN);
	AbstractCompressor* delegate = (enc_ptr == compressors)
		? static_cast<AbstractCompressor*>(new GzipCompressor)
		: static_cast<AbstractCompressor*>(new DeflateCompressor(true, true));
#ifdef DEBUG
	metadata->headers().addHeader("X-filtered-by", "CompressorResponseFilter ["+std::string(*enc_ptr)+"]");
#endif
	return new CompressorResponseFilter((*follower)(metadata), delegate);
}

CompressorResponseFilter::CompressorResponseFilter(AbstractResponseFilter* follower, AbstractCompressor* delegate)
:	m_pFollower(follower), m_pDelegate(delegate)
{
	delegate->onDataOutput().connect(SigC::slot(*this, &CompressorResponseFilter::processOutputData));
}

CompressorResponseFilter::~CompressorResponseFilter()
{
	delete m_pDelegate;
	m_pFollower->destroy();
}

void
CompressorResponseFilter::consumeDataChunk(const uint8_t* begin, const uint8_t* end)
{
	m_pDelegate->consumeDataChunk(begin, end);
}

void
CompressorResponseFilter::consumeEOF()
{
	m_pDelegate->consumeEOF();
	m_pFollower->consumeEOF();
}

void
CompressorResponseFilter::processOutputData(const uint8_t* begin, const uint8_t* end)
{
	m_pFollower->consumeDataChunk(begin, end);
}

