/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "conf.h"
#include <map>
#include <algorithm>
#include <string>
#include <cstring>
#include <sstream>
#include <regex.h>
#include "util.h"

#define DEFAULT_LISTEN_ADDR	"127.0.0.1"
#define DEFAULT_LISTEN_PORT	8080

Config::Config()
{
	m_listen_addr = DEFAULT_LISTEN_ADDR;
	m_listen_port = DEFAULT_LISTEN_PORT;
	m_use_proxy = false;
	//m_proxy_host = "";
	m_proxy_port = -1;
	//m_no_proxy_for = "";
	m_client_compression = false;
	m_draw_borders = true;
	m_border_color[0] = 0x67;
	m_border_color[1] = 0x67;
	m_border_color[2] = 0x67;
	m_no_flash = false;
}

Config::Config(const Config& other)
{
	*this = other;
}

Config&
Config::operator=(const Config& other)
{
	if (&other == this)
		return *this;
	m_listen_addr = other.m_listen_addr;
	m_listen_port = other.m_listen_port;
	m_use_proxy = other.m_use_proxy;
	m_proxy_host = other.m_proxy_host;
	m_proxy_port = other.m_proxy_port;
	setNoProxyFor(other.m_no_proxy_for);
	m_client_compression = other.m_client_compression;
	m_draw_borders = other.m_draw_borders;
	memcpy(m_border_color, other.m_border_color, sizeof(m_border_color));
	m_no_flash = other.m_no_flash;
	return *this;
}

std::string
Config::serialize() const
{
	std::map<std::string, std::string> m;
	m["listen_addr"] = m_listen_addr;
	m["listen_port"] = num2str(m_listen_port);
	m["use_proxy"] = (m_use_proxy ? "1" : "0");
	m["proxy_host"] = m_proxy_host;
	m["proxy_port"] = num2str(m_proxy_port);
	m["no_proxy_for"] = ListSerializer<std::string>::serialize(m_no_proxy_for);
	m["client_compression"] = (m_client_compression ? "1" : "0");
	m["draw_borders"] = (m_draw_borders ? "1" : "0");
	m["border_color"] = std::string(m_border_color, m_border_color+3);
	m["no_flash"] = (m_no_flash ? "1" : "0");
	return MapSerializer<std::string, std::string>::serialize(m);
}

void
Config::unserialize(const std::string& data)
{
	std::map<std::string, std::string> m;
	MapUnserializer<std::string, std::string>::unserialize(data, m);
	std::map<std::string, std::string>::iterator iter = m.begin();
	for (; iter != m.end(); ++iter) {
		const std::string& key = iter->first;
		const std::string& val = iter->second;
		if (key == "listen_addr") {
			m_listen_addr = val;
		} else if (key == "listen_port") {
			m_listen_port = str2num<int>(val, m_listen_port);
		} else if (key == "use_proxy") {
			m_use_proxy = (val == "1");
		} else if (key == "proxy_host") {
			m_proxy_host = val;
		} else if (key == "proxy_port") {
			m_proxy_port = str2num<int>(val, m_proxy_port);
		} else if (key == "no_proxy_for") {
			ListUnserializer<std::string>::unserialize(val, m_no_proxy_for);
			setNoProxyFor(m_no_proxy_for);
		} else if (key == "client_compression") {
			m_client_compression = (val == "1");
		} else if (key == "draw_borders") {
			m_draw_borders = (val == "1");
		} else if (key == "border_color") {
			memcpy(m_border_color, val.data(), std::min<std::string::size_type>(val.length(), 3));
		} else if (key == "no_flash") {
			m_no_flash = (val == "1");
		}
	}
}

bool
Config::operator==(const Config& other) const
{
	if (m_listen_addr != other.m_listen_addr)
		return false;
	if (m_listen_port != other.m_listen_port)
		return false;
	if (m_use_proxy != other.m_use_proxy)
		return false;
	if (m_use_proxy) {
		if (m_proxy_host != other.m_proxy_host)
			return false;
		if (m_proxy_port != other.m_proxy_port)
			return false;
		if (m_no_proxy_for != other.m_no_proxy_for)
			return false;
	}
	if (m_client_compression != other.m_client_compression)
		return false;
	if (m_draw_borders != other.m_draw_borders)
		return false;
	if (m_draw_borders) {
		if (memcmp(m_border_color, other.m_border_color, sizeof(m_border_color)))
			return false;
	}
	if (m_no_flash != other.m_no_flash) {
		return false;
	}
	return true;
}

void
Config::setNoProxyFor(const std::list<std::string>& no_proxy_for)
{
	m_no_proxy_for = no_proxy_for;
	std::list<std::string>::iterator iter = m_no_proxy_for.begin();
	bool first = true;
	std::string composite_regex = "^";
	while (iter != m_no_proxy_for.end()) {
		if (iter->empty()) {
			m_no_proxy_for.erase(iter++);
			continue;
		}
		if (!first) {
			composite_regex += "|";
		} else {
			first = false;
		}
		bool starts_with_dot = (iter->c_str()[0] == '.');
		bool ends_with_dot = (iter->c_str()[iter->length()-1] == '.');
		composite_regex += (starts_with_dot ? "(.*" : "(")+regQuote(*iter)+(ends_with_dot ? ".*)" : ")");
		++iter;
	}
	composite_regex += "$";
	if (m_no_proxy_for.empty()) {
		m_ptrReg_no_proxy_for.reset(0);
		return;
	}
	m_ptrReg_no_proxy_for.reset(new RegExp(composite_regex, REG_EXTENDED|REG_ICASE|REG_NOSUB));
}

bool
Config::willGoThroughProxy(const std::string& host) const
{
	if (!m_use_proxy)
		return false;
	if (!m_ptrReg_no_proxy_for.get())
		return true;
	return !m_ptrReg_no_proxy_for->matches(host);
}

std::string
Config::regQuote(const std::string& str)
{
	static const char *metachars = ".\\+*?[^]$(){}=!<>|:";
	std::string quoted;
	std::string::size_type pos = 0, pos1;
	while ((pos1 = str.find_first_of(metachars, pos)) != std::string::npos) {
		quoted.append(str, pos, pos1-pos);
		quoted += '\\';
		quoted += str[pos1];
		pos = pos1+1;
	}
	quoted.append(str, pos, std::string::npos);
	return quoted;
}

