/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "configio.h"
#include <sstream>
#include <iterator>
#include <cctype>

using namespace std;

ConfigIO&
ConfigIO::operator=(const ConfigIO& other)
{
	if (&other == this)
		return *this;
	m_status = other.m_status;
	m_config = other.m_config;
	m_elements = other.m_elements;
	m_cur_section = other.m_cur_section;
	m_sections.clear();
	map<string, SectionPointer>::const_iterator iter = other.m_sections.begin();
	for (; iter != other.m_sections.end(); ++iter) {
		SectionPointer& sptr = m_sections[iter->first];
		sptr.m_section_element = iter->second.m_section_element;
		list<ElementPointer>::const_iterator iter2 = iter->second.m_kv_elements.begin();
		for (; iter2 != iter->second.m_kv_elements.end(); ++iter2) {
			list<Element>::iterator it = m_elements.begin();
			typedef list<Element>::const_iterator ConstIter;
			advance(it, distance<ConstIter>(other.m_elements.begin(), *iter2));
			sptr.m_kv_elements.push_back(it);
		}
	}
	return *this;
}

void
ConfigIO::reset()
{
	m_status.status = Status::OK;
	m_status.errors.clear();
	m_config = Config();
	m_elements.clear();
	m_cur_section.erase();
	m_sections.clear();
}

void
ConfigIO::setConfig(const Config& config)
{
	m_config = config;
	string section = "global";
	map<string, SectionPointer>::iterator sec_it = m_sections.find(section);
	if (sec_it == m_sections.end()) {
		m_elements.push_front(Element(Element::SECTION, section));
		m_sections[section].m_section_element = m_elements.begin();
		sec_it = m_sections.find(section);
	}
	
	ostringstream listen_address;
	listen_address << m_config.getListenAddr() << ":" << m_config.getListenPort();
	assignParam(sec_it->second, "listen_address", listen_address.str());
	assignParam(sec_it->second, "client_compression", m_config.getClientCompression() ? "yes" : "no");
	ostringstream ad_border;
	if (m_config.getDrawBorders()) {
		ad_border.fill('0');
		ad_border.width(2);
		ad_border << hex << (unsigned)m_config.getBorderColor()[0]
			<< (unsigned)m_config.getBorderColor()[1] << (unsigned)m_config.getBorderColor()[2];
	} else {
		ad_border << "none";
	}
	assignParam(sec_it->second, "ad_border", ad_border.str());
	
	section = "forwarding";
	sec_it = m_sections.find(section);
	if (sec_it == m_sections.end()) {
		m_elements.push_back(Element(Element::SECTION, section));
		m_sections[section].m_section_element = m_elements.begin();
		sec_it = m_sections.find(section);
	}
	assignParam(sec_it->second, "use_proxy", m_config.getUseProxy() ? "yes" : "no");
	assignParam(sec_it->second, "proxy_host", m_config.getProxyHost());
	ostringstream proxy_port;
	if (m_config.getProxyPort() >= 0)
		proxy_port << m_config.getProxyPort();
	assignParam(sec_it->second, "proxy_port", proxy_port.str());
	string no_proxy_for;
	list<string>::const_iterator iter = config.m_no_proxy_for.begin();
	for (bool first = true; iter != config.m_no_proxy_for.end(); ++iter) {
		if (!first) {
			no_proxy_for += ", ";
		} else {
			first = false;
		}
		no_proxy_for += *iter;
	}
	assignParam(sec_it->second, "no_proxy_for", no_proxy_for);
}

bool
ConfigIO::processElement(const Element& element, int line)
{
	m_elements.push_back(element);
	if (element.getType() == Element::SECTION) {
		m_cur_section = element.getText();
		if (m_cur_section.empty()) {
			// backwards compability
			m_cur_section = "global";
			m_elements.push_front(Element(Element::SECTION, m_cur_section));
			m_elements.back().setText(m_cur_section);
		}
		if (!validateSection(m_cur_section, line)) {
			m_elements.pop_back();
			return false;
		}
		m_sections[m_cur_section].m_section_element = --m_elements.end();
	} else if (element.getType() == Element::KEY_VALUE) {
		SectionPointer& sec = m_sections[m_cur_section];
		if (sec.findParam(element.getKey(), m_elements.end()) != m_elements.end()) {
			processError(Error("Parameter is already defined: "+element.getKey(), line));
			m_elements.pop_back();
			return false;
		}
		if (!processKeyValue(m_cur_section, element.getKey(), element.getValue(), line)) {
			return false;
		}
		sec.m_kv_elements.push_back(--m_elements.end());
	}
	return true;
}

void
ConfigIO::processError(const Error& error)
{
	m_status.errors.push_back(error);
	if (m_status.status != Status::ERRORS)
		m_status.status = (error.getType() == Error::T_ERROR ? Status::ERRORS : Status::WARNINGS);
}

bool
ConfigIO::validateSection(const string& section, int line)
{
	if (m_sections.find(section) != m_sections.end()) {
		processError(Error("Section already defined: "+section, line));
		return false;
	}
	if (section != "global" && section != "forwarding") {
		processError(Error(Error::T_WARNING, "Unknown section: "+section, line));
	}
	return true;
}

bool
ConfigIO::processKeyValue(const string& section, const string& key, const string& value, int line)
{
	if (section == "global") {
		if (key == "listen_address") {
			string::size_type pos = value.find(":");
			m_config.m_listen_addr = value.substr(0, pos);
			if (pos != string::npos) {
				istringstream strm(value.substr(pos+1));
				strm >> m_config.m_listen_port;
			}
		} else if (key == "client_compression") {
			m_config.m_client_compression = (value == "yes");
		} else if (key == "ad_border") {
			if (value == "none") {
				m_config.m_draw_borders = false;
			} else {
				m_config.m_draw_borders = true;
				if (value.length() >= 6) {
					istringstream strm1(value.substr(0, 2));
					istringstream strm2(value.substr(2, 2));
					istringstream strm3(value.substr(4, 2));
					unsigned int r = m_config.m_border_color[0];
					unsigned int g = m_config.m_border_color[1];
					unsigned int b = m_config.m_border_color[2];
					strm1 >> hex >> r;
					strm2 >> hex >> g;
					strm3 >> hex >> b;
					m_config.m_border_color[0] = r;
					m_config.m_border_color[1] = g;
					m_config.m_border_color[2] = b;
				}
			}
		} else if (key == "no_flash") {
			m_config.m_no_flash = (value == "yes");
		} else {
			processError(Error(Error::T_WARNING, "Unknown paramenter: "+key, line));
		}
	} else if (section == "forwarding") {
		if (key == "use_proxy") {
			if (value == "yes")
				m_config.m_use_proxy = true;
		} else if (key == "proxy_host") {
			m_config.m_proxy_host = value;
		} else if (key == "proxy_port") {
			istringstream strm(value);
			strm >> m_config.m_proxy_port;
		} else if (key == "no_proxy_for") {
			m_config.m_no_proxy_for.clear();
			string buf = value;
			buf += ",";
			string::size_type pos = 0, pos1;
			while ((pos1 = buf.find_first_of(",;", pos)) != string::npos) {
				string item = trim(buf.substr(pos, pos1-pos));
				if (!item.empty()) {
					m_config.m_no_proxy_for.push_back(item);
				}
				pos = pos1+1;
			}
			m_config.setNoProxyFor(m_config.m_no_proxy_for);
		} else {
			processError(Error(Error::T_WARNING, "Unknown paramenter: "+key, line));
		}
	}
	return true;
}

void
ConfigIO::assignParam(SectionPointer& sp, const string& key, const string& value)
{
	
	ElementPointer pkey = sp.findParam(key, m_elements.end());
	if (pkey == m_elements.end()) {
		Element el(Element::KEY_VALUE, key, value);
		if (sp.m_kv_elements.empty()) {
			sp.m_kv_elements.push_back(m_elements.insert(m_elements.end(), el));
		} else {
			ElementPointer last = sp.m_kv_elements.back();
			sp.m_kv_elements.push_back(m_elements.insert(++last, el));
		}
	} else {
		pkey->setValue(value);
	}
}

ConfigIO::ElementPointer
ConfigIO::SectionPointer::findParam(const string& key, ElementPointer dflt)
{
	list<ElementPointer>::iterator iter = m_kv_elements.begin();
	for (; iter != m_kv_elements.end(); ++iter) {
		if ((*iter)->getKey() == key) {
			return *iter;
		}
	}
	return dflt;
}

string
ConfigIO::trim(const string& str)
{
	if (str.empty())
		return "";
	const char* data = str.c_str();
	string::size_type len = str.length();
	string::size_type i = 0;
	for (; i < len && isspace(data[i]); ++i);
	string::size_type j = len-1;
	do {
		if (j <= i || !isspace(data[j]))
			break;
	} while (j--);
	return str.substr(i, j-i+1);
}
