/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "confio.h"
#include "conflexerdefinitions.h"
#include <cstring>
#include <cstddef>
#include <list>

ConfIO::ConfIO()
:	ConfLexerCore(this),
	m_lineNumber(1),
	m_lineStartPos(0),
	m_pCharSafetyTable(getCharSafetyTable()),
	m_pCharEscapeTable(getCharEscapeTable()),
	m_pCharUnescapeTable(getCharUnescapeTable())
{
}

void
ConfIO::read(std::istream& strm)
{
	uint8_t buf[4096];
	while (strm && !isStopped()) {
		strm.read(reinterpret_cast<char*>(buf), sizeof(buf));
		consumeDataChunk(buf, buf+strm.gcount());
	}
	if (!isStopped()) {
		consumeEOF();
		switch (scond()) {
			case ConfLexerDefinitions::INITIAL:
			case ConfLexerDefinitions::COMMENT:
			case ConfLexerDefinitions::AFTER_SECTION_OR_VALUE:
			case ConfLexerDefinitions::RECOVERY:
			break;
			default:
			processError(Error("Unexpected end of file"));
		}
	}
	reset();
	m_lineNumber = 1;
	m_lineStartPos = 0;
}

void
ConfIO::write(std::ostream& strm, const std::list<Element>& elements)
{
	std::list<Element>::const_iterator iter = elements.begin();
	for (; iter != elements.end(); ++iter) {
		switch (iter->getType()) {
			case Element::BLANK_LINE:
				strm << std::endl;
				break;
			case Element::COMMENT:
				strm << ';' << iter->getText() << std::endl;
				break;
			case Element::SECTION:
				strm << '[' << iter->getText() << ']' << std::endl;
				break;
			case Element::KEY_VALUE:
				strm << iter->getKey() << '=' << (valueRequiresQuoting(iter->getValue())
					? quoteValue(iter->getValue()) : iter->getValue()) << std::endl;
				break;
		}
	}
}

void
ConfIO::newLine(const ConstIterator& pos)
{
	++m_lineNumber;
	m_lineStartPos = tokenPosition(pos);
}

void
ConfIO::processBlankLine()
{
	if (!processElement(Element(Element::BLANK_LINE), m_lineNumber)) {
		stop();
	}
}

void
ConfIO::processBrokenSection()
{
	processError(Error("Invalid section declaration", m_lineNumber));
	stop();
}

void
ConfIO::processUnexpectedCharacter(const ConstIterator& pos)
{
	processError(Error("Unexpected character", m_lineNumber, tokenPosition(pos)-m_lineStartPos+1));
}

void
ConfIO::processComment(const ConstIterator& begin, const ConstIterator& end)
{
	if (!processElement(Element(Element::COMMENT,
	    tokenAsString(begin, end)), m_lineNumber)) {
		stop();
	}
}

void
ConfIO::processSection(const ConstIterator& begin, const ConstIterator& end)
{
	if (!processElement(Element(Element::SECTION,
	     tokenAsString(begin, end)), m_lineNumber)) {
		stop();
	}
}

void
ConfIO::processKey(const ConstIterator& begin, const ConstIterator& end)
{
	m_key = tokenAsString(begin, end);
	m_value = "";
}

void
ConfIO::processValue()
{
	if (!processElement(Element(Element::KEY_VALUE, m_key, m_value), m_lineNumber)) {
		stop();
	}
}

void
ConfIO::processUnclosedQuote()
{
	processError(Error("Unclosed quote", m_lineNumber));
}

bool
ConfIO::isAtLineStart(const ConstIterator& pos)
{
	return tokenPosition(pos) == m_lineStartPos;
}

void
ConfIO::valueAppend(const ConstIterator& begin, const ConstIterator& end)
{
	Buffer::appendToString(m_value, begin, end);
}

void
ConfIO::valueAppendEscaped(uint8_t ch)
{
	m_value.append(1, m_pCharUnescapeTable[ch]);
}

bool
ConfIO::valueRequiresQuoting(const std::string& value)
{
	const char *data = value.data();
	std::size_t length = value.length();
	for (std::size_t i = 0; i < length; i++) {
		if (!m_pCharSafetyTable[static_cast<unsigned char>(data[i])]) {
			return true;
		}
	}
	return false;
}

std::string
ConfIO::quoteValue(const std::string& value)
{
	std::string res = "\"";
	const char *data = value.data();
	std::size_t length = value.length();
	for (std::size_t i = 0; i < length; i++) {
		unsigned char ch = static_cast<unsigned char>(data[i]);
		if (ch == '"' || ch == '\\' || m_pCharEscapeTable[ch] != ch) {
			res.append(1, '\\');
			res.append(1, m_pCharEscapeTable[ch]);
		} else {
			res.append(1, ch);
		}
	}
	res += "\"";
	return res;
}

const char*
ConfIO::getCharSafetyTable()
{
	struct Table
	{
		char t[256];
		Table() {
			std::memset(t, 0, sizeof(t));
			std::memset(t+32, 1, 128-32);
			static const unsigned char more_unsafe[] = " <>\"{}|\\^~[]`#%\177";
			for (unsigned int i = 0; i < sizeof(more_unsafe); i++) {
				t[more_unsafe[i]] = 0;
			}
		}
	};
	static const Table table;
	return table.t;
}

const char*
ConfIO::getCharEscapeTable()
{
	struct Table
	{
		char t[256];
		Table() {
			for (unsigned int i = 0; i < 256; i++) {
				t[i] = static_cast<unsigned char>(i);
			}
			e('\a', 'a');
			e('\b', 'b');
			e('\f', 'f');
			e('\n', 'n');
			e('\r', 'r');
			e('\t', 't');
			e('\v', 'v');
		}
		void e(unsigned char key, char val) {
			t[key] = val;
		}
	};
	static const Table table;
	return table.t;
}

const char*
ConfIO::getCharUnescapeTable()
{
	struct Table
	{
		char t[256];
		Table() {
			for (unsigned int i = 0; i < 256; i++) {
				t[i] = static_cast<unsigned char>(i);
			}
			e('a', '\a');
			e('b', '\b');
			e('f', '\f');
			e('n', '\n');
			e('r', '\r');
			e('t', '\t');
			e('v', '\v');
		}
		void e(unsigned char key, char val) {
			t[key] = val;
		}
	};
	static const Table table;
	return table.t;
}

