/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef CONFIO_H_
#define CONFIO_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "types.h"
#include "conflexercore.h"
#include "buffer.h"
#include <iostream>
#include <string>
#include <list>

class ConfIO : private ConfLexerCore
{
	friend class ConfLexerCore::Actions;
public:
	class Error
	{
	public:
		enum Type { T_ERROR, T_WARNING };
		Error(Type type, const std::string& msg, int line=0, int col=0)
		: m_type(type), m_msg(msg), m_line(line), m_col(col) {}
		Error(const std::string& msg, int line=0, int col=0)
		: m_type(T_ERROR), m_msg(msg), m_line(line), m_col(col) {}
		Type getType() const { return m_type; }
		const std::string& getMessage() const { return m_msg; }
		int getLine() const { return m_line; }
		int getCol() const { return m_col; }
	private:
		Type m_type;
		std::string m_msg;
		int m_line;
		int m_col;
	};
	typedef std::list<Error> ErrorList;
	
	ConfIO();
protected:
	class Element
	{
	public:
		enum Type { BLANK_LINE, COMMENT, SECTION, KEY_VALUE };
		Element(Type type) : m_type(type) {}
		Element(Type type, const std::string& text) : m_type(type), m_value(text) {}
		Element(Type type, const std::string& key, const std::string& value)
		: m_type(type), m_key(key), m_value(value) {}
		Type getType() const { return m_type; }
		const std::string& getText() const { return m_value; }
		void setText(const std::string& text) { m_value = text; }
		const std::string& getKey() const { return m_key; }
		void setKey(const std::string& key) { m_key = key; }
		const std::string& getValue() const { return m_value; }
		void setValue(const std::string& value) { m_value = value; }
	private:
		Type m_type;
		std::string m_key;
		std::string m_value;
	};
	
	virtual ~ConfIO() {}
	
	virtual bool processElement(const Element& element, int line) = 0;
	virtual void processError(const Error& error) = 0;
	void read(std::istream& strm);
	void write(std::ostream& strm, const std::list<Element>& elements);
private:
	void newLine(const ConstIterator& pos);
	void processBlankLine();
	void processBrokenSection();
	void processUnexpectedCharacter(const ConstIterator& pos);
	void processComment(const ConstIterator& begin, const ConstIterator& end);
	void processSection(const ConstIterator& begin, const ConstIterator& end);
	void processKey(const ConstIterator& begin, const ConstIterator& end);
	void processValue();
	void processUnclosedQuote();
	bool isAtLineStart(const ConstIterator& pos);
	void valueAppend(const ConstIterator& begin, const ConstIterator& end);
	void valueAppendEscaped(uint8_t ch);
	bool valueRequiresQuoting(const std::string& value);
	std::string quoteValue(const std::string& value);
	
	static const char* getCharSafetyTable();
	static const char* getCharEscapeTable();
	static const char* getCharUnescapeTable();
	
	std::string m_key;
	std::string m_value;
	int m_lineNumber;
	std::size_t m_lineStartPos;
	const char *m_pCharSafetyTable;
	const char *m_pCharEscapeTable;
	const char *m_pCharUnescapeTable;
};

#endif

