/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <fstream>
#include "lexgen.h"
#include "nfachar.h"
#include "nfacharclass.h"
#include "nfanegatedcharclass.h"
#include "nfaconcatenation.h"
#include "nfaclosure.h"
#include "nfastring.h"
#include "nfaistring.h"
#include "nfaunion.h"
#include "nfaeof.h"

#include "conflexerdefinitions.h"

int main(int argc, char** argv)
{
	if (argc < 8) {
		std::cerr << "Usage: makelexer <OutputClass> <out.h> <out.cpp> "
			"<Definitions> <definitions.h> <Subclass> <subclass.h>" << std::endl;
		return 1;
	}
	const char* out_class = argv[1];
	const char* out_header = argv[2];
	const char* out_impl = argv[3];
	const char* def_class = argv[4];
	const char* def_header = argv[5];
	const char* subclass = argv[6];
	const char* subclass_header = argv[7];
	
	std::ofstream header_strm(out_header);
	if (header_strm.fail()) {
		std::cerr << "Failed opening " << out_header << " for writing" << std::endl;
	}
	
	std::ofstream impl_strm(out_impl);
	if (impl_strm.fail()) {
		std::cerr << "Failed opening " << out_impl << " for writing" << std::endl;
	}
	
	LexGen lexgen(out_class, subclass);
	lexgen.enableTrackingTokenPosition();
	
	NfaCharClass space(" \t");
	NfaCharClass comm_begin(";#");
	NfaUnion newline(NfaString("\r\n"), NfaChar('\n'));
	NfaCharClass alpha('a', 'z');
	alpha.addChars('A', 'Z');
	NfaCharClass key_chars = alpha;
	key_chars.addChars('0', '9');
	key_chars.addChars("-_");
	NfaNegatedCharClass any;
	NfaEof eof;
	
	typedef ConfLexerDefinitions Defs;

	lexgen.addRule(Defs::INITIAL, NfaClosure(space), "/* skip spaces */");
	lexgen.addRule(Defs::INITIAL, newline, "obj->processBlankLine(); obj->newLine(tok_end);");
	lexgen.addRule(Defs::INITIAL, comm_begin, "BEGIN(COMMENT);");
	lexgen.addRule(Defs::INITIAL, alpha, "BEGIN(KEY); CONSUME_NOTHING();");
	lexgen.addRule(Defs::INITIAL, NfaChar('['),
		"if (obj->isAtLineStart(tok_begin))\n\t"
		"BEGIN(SECTION);\n\t"
		"else { BEGIN(RECOVERY); obj->processBrokenSection(); }"
	);
	lexgen.addRule(Defs::INITIAL, any,
		"BEGIN(RECOVERY); CONSUME_NOTHING(); obj->processUnexpectedCharacter(tok_begin);");
	lexgen.addRule(Defs::COMMENT, NfaConcatenation(NfaClosure(any), NfaString("\r\n")),
		"BEGIN(INITIAL); obj->processComment(tok_begin, tok_end-2); obj->newLine(tok_end);"
	).setLazy();
	lexgen.addRule(Defs::COMMENT, NfaConcatenation(NfaClosure(any), NfaChar('\n')),
		"BEGIN(INITIAL); obj->processComment(tok_begin, tok_end-1); obj->newLine(tok_end);"
	).setLazy();
	lexgen.addRule(Defs::COMMENT, NfaConcatenation(NfaClosure(any)),
		"BEGIN(INITIAL); obj->processComment(tok_begin, tok_end);"
	);
	lexgen.addRule(Defs::SECTION, NfaConcatenation(NfaClosure(NfaNegatedCharClass("\n")), NfaChar(']')),
		"BEGIN(AFTER_SECTION_OR_VALUE); obj->processSection(tok_begin, tok_end-1);"
	);
	lexgen.addRule(Defs::SECTION, any, "BEGIN(RECOVERY); CONSUME_NOTHING(); obj->processBrokenSection();");
	lexgen.addRule(Defs::KEY, NfaConcatenation(alpha, NfaClosure(key_chars)),
		"BEGIN(EQUAL); obj->processKey(tok_begin, tok_end);"
	);
	lexgen.addRule(Defs::EQUAL,
		NfaConcatenation(NfaClosure(space), NfaChar('=')).addComponent(NfaClosure(space)),
		"BEGIN(VALUE);"
	);
	lexgen.addRule(Defs::EQUAL, any,
		"BEGIN(RECOVERY); CONSUME_NOTHING(); obj->processUnexpectedCharacter(tok_begin);"
	);
	lexgen.addRule(Defs::VALUE, NfaChar('"'), "BEGIN(VALUE_Q_NE);");
	lexgen.addRule(Defs::VALUE, NfaUnion(any, eof), "BEGIN(VALUE_NQ); CONSUME_NOTHING();");
	lexgen.addRule(Defs::VALUE_NQ, NfaClosure(space), NfaUnion(newline, eof),
		"BEGIN(AFTER_SECTION_OR_VALUE);\n\t"
		"obj->valueAppend(tok_begin, tok_end); obj->processValue();"
	);
	lexgen.addRule(Defs::VALUE_NQ, any, "MORE();");
	lexgen.addRule(Defs::VALUE_Q_NE, NfaClosure(NfaNegatedCharClass("\"\n\\")),
		"obj->valueAppend(tok_begin, tok_end);"
	);
	lexgen.addRule(Defs::VALUE_Q_NE, NfaChar('\\'), "BEGIN(VALUE_Q_E);");
	lexgen.addRule(Defs::VALUE_Q_NE, NfaChar('"'),
		"BEGIN(AFTER_SECTION_OR_VALUE); obj->processValue();"
	);
	lexgen.addRule(Defs::VALUE_Q_NE, any,
		"BEGIN(RECOVERY); CONSUME_NOTHING(); obj->processUnclosedQuote();"
	);
	lexgen.addRule(Defs::VALUE_Q_E, any, "BEGIN(VALUE_Q_NE); obj->valueAppendEscaped(*tok_begin);");
	lexgen.addRule(Defs::AFTER_SECTION_OR_VALUE, newline, "BEGIN(INITIAL); obj->newLine(tok_end);");
	lexgen.addRule(Defs::AFTER_SECTION_OR_VALUE, NfaClosure(space), "/* skip spaces */");
	lexgen.addRule(Defs::AFTER_SECTION_OR_VALUE, any,
		"BEGIN(RECOVERY); CONSUME_NOTHING(); obj->processUnexpectedCharacter(tok_begin);"
	);
	lexgen.addRule(Defs::RECOVERY, NfaConcatenation(NfaClosure(any), newline),
		"BEGIN(INITIAL); obj->newLine(tok_end);"
	).setLazy();
	
	lexgen.writeHeader(header_strm, def_class, def_header);
	lexgen.writeCode(impl_strm, subclass_header);
	return 0;
}

