/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "rules.h"
#include "util.h"
#include <string>
#include <sstream>
#include <map>

Rules::Rules()
{
	m_default_rule.m_filter = 1;
	m_default_rule.m_ad = 0;
	m_default_rule.m_scripts = 3;
	m_default_rule.m_jsengine = 1;
	m_default_rule.m_target_blank = 0;
	m_default_global_rule = m_default_rule;
	m_default_local_rule = m_default_rule;
}

std::string
Rules::serialize() const
{
	std::map<std::string, std::string> m;
	m["default_rule"] = m_default_rule.serialize();
	m["default_global_rule"] = m_default_global_rule.serialize();
	m["default_local_rule"] = m_default_local_rule.serialize();
	m["global_re_by_filter"] = serializeRegexMap(m_global_re_by_filter);
	m["local_re_by_filter"] = serializeRegexMap(m_local_re_by_filter);
	m["global_re_by_ad"] = serializeRegexMap(m_global_re_by_ad);
	m["local_re_by_ad"] = serializeRegexMap(m_local_re_by_ad);
	m["global_re_by_scripts"] = serializeRegexMap(m_global_re_by_scripts);
	m["local_re_by_scripts"] = serializeRegexMap(m_local_re_by_scripts);
	m["global_re_by_jsengine"] = serializeRegexMap(m_global_re_by_jsengine);
	m["local_re_by_jsengine"] = serializeRegexMap(m_local_re_by_jsengine);
	m["global_re_by_target_blank"] = serializeRegexMap(m_global_re_by_target_blank);
	m["local_re_by_target_blank"] = serializeRegexMap(m_local_re_by_target_blank);
	return MapSerializer<std::string, std::string>::serialize(m);
}

void
Rules::unserialize(const std::string& data)
{
	std::map<std::string, std::string> m;
	MapUnserializer<std::string, std::string>::unserialize(data, m);
	std::map<std::string, std::string>::iterator iter = m.begin();
	for (; iter != m.end(); ++iter) {
		const std::string& key = iter->first;
		const std::string& val = iter->second;
		if (key == "default_rule") {
			m_default_rule.unserialize(val);
		} else if (key == "default_global_rule") {
			m_default_global_rule.unserialize(val);
		} else if (key == "default_local_rule") {
			m_default_local_rule.unserialize(val);
		} else if (key == "global_re_by_filter") {
			unserializeRegexMap(val, m_global_re_by_filter);
		} else if (key == "local_re_by_filter") {
			unserializeRegexMap(val, m_local_re_by_filter);
		} else if (key == "global_re_by_ad") {
			unserializeRegexMap(val, m_global_re_by_ad);
		} else if (key == "local_re_by_ad") {
			unserializeRegexMap(val, m_local_re_by_ad);
		} else if (key == "global_re_by_scripts") {
			unserializeRegexMap(val, m_global_re_by_scripts);
		} else if (key == "local_re_by_scripts") {
			unserializeRegexMap(val, m_local_re_by_scripts);
		} else if (key == "global_re_by_jsengine") {
			unserializeRegexMap(val, m_global_re_by_jsengine);
		} else if (key == "local_re_by_jsengine") {
			unserializeRegexMap(val, m_local_re_by_jsengine);
		} else if (key == "global_re_by_target_blank") {
			unserializeRegexMap(val, m_global_re_by_target_blank);
		} else if (key == "local_re_by_target_blank") {
			unserializeRegexMap(val, m_local_re_by_target_blank);
		}
	}
}

int
Rules::getFilterParam(const std::string& url) const
{
	std::map<int, RegexPtr>::const_iterator iter = m_local_re_by_filter.begin();
	for (; iter != m_local_re_by_filter.end(); ++iter) {
		if (iter->second->matches(url)) {
			return iter->first;
		}
	}
	iter = m_global_re_by_filter.begin();
	for (; iter != m_global_re_by_filter.end(); ++iter) {
		if (iter->second->matches(url)) {
			return iter->first;
		}
	}
	return m_default_local_rule.m_filter;
}

int
Rules::getAdParam(const std::string& url) const
{
	std::map<int, RegexPtr>::const_iterator iter = m_local_re_by_ad.begin();
	for (; iter != m_local_re_by_ad.end(); ++iter) {
		if (iter->second->matches(url)) {
			return iter->first;
		}
	}
	iter = m_global_re_by_ad.begin();
	for (; iter != m_global_re_by_ad.end(); ++iter) {
		if (iter->second->matches(url)) {
			return iter->first;
		}
	}
	return m_default_local_rule.m_ad;
}

int
Rules::getScriptsParam(const std::string& url) const
{
	std::map<int, RegexPtr>::const_iterator iter = m_local_re_by_scripts.begin();
	for (; iter != m_local_re_by_scripts.end(); ++iter) {
		if (iter->second->matches(url)) {
			return iter->first;
		}
	}
	iter = m_global_re_by_scripts.begin();
	for (; iter != m_global_re_by_scripts.end(); ++iter) {
		if (iter->second->matches(url)) {
			return iter->first;
		}
	}
	return m_default_local_rule.m_scripts;
}

int
Rules::getJSengineParam(const std::string& url) const
{
	std::map<int, RegexPtr>::const_iterator iter = m_local_re_by_jsengine.begin();
	for (; iter != m_local_re_by_jsengine.end(); ++iter) {
		if (iter->second->matches(url)) {
			return iter->first;
		}
	}
	iter = m_global_re_by_jsengine.begin();
	for (; iter != m_global_re_by_jsengine.end(); ++iter) {
		if (iter->second->matches(url)) {
			return iter->first;
		}
	}
	return m_default_local_rule.m_jsengine;
}

int
Rules::getTargetBlankParam(const std::string& url) const
{
	std::map<int, RegexPtr>::const_iterator iter = m_local_re_by_target_blank.begin();
	for (; iter != m_local_re_by_target_blank.end(); ++iter) {
		if (iter->second->matches(url)) {
			return iter->first;
		}
	}
	iter = m_global_re_by_target_blank.begin();
	for (; iter != m_global_re_by_target_blank.end(); ++iter) {
		if (iter->second->matches(url)) {
			return iter->first;
		}
	}
	return m_default_local_rule.m_target_blank;
}

std::string
Rules::serializeRegexMap(const std::map<int, RegexPtr>& m)
{
	return MapSerializer<int, RegexPtr, IntToString, RegexToString>::serialize(m);
}

void
Rules::unserializeRegexMap(const std::string& data, std::map<int, RegexPtr>& m)
{
	MapUnserializer<int, RegexPtr, StringToInt, StringToRegex>::unserialize(data, m);
}

std::string
Rules::Rule::serialize() const
{
	std::ostringstream strm;
	strm << m_filter << ' ' << m_ad << ' ' << m_scripts << ' ' << m_jsengine << ' ' << m_target_blank;
	return strm.str();
}

void
Rules::Rule::unserialize(const std::string& data)
{
	std::istringstream strm(data);
	strm >> m_filter >> m_ad >> m_scripts >> m_jsengine >> m_target_blank;
}

std::string
Rules::IntToString::convert(int val)
{
	return num2str(val);
}

std::string
Rules::RegexToString::convert(const RegexPtr& regex)
{
	return regex->serialize();
}

int
Rules::StringToInt::convert(const std::string& str)
{
	return str2num<int>(str);
}

Rules::RegexPtr
Rules::StringToRegex::convert(const std::string& str)
{
	RegexPtr r(new RegExp);
	r->unserialize(str);
	return r;
}

