/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "rulesreader.h"
#include "util.h"
#include <string>
#include <regex.h>

#define MAX_FILTER       1
#define MAX_AD           2
#define MAX_SCRIPTS      7
#define MAX_JSENGINE     1
#define MAX_TARGET_BLANK 1

using namespace std;

void
RulesReader::reset()
{
	m_status.status = Status::OK;
	m_status.errors_global.clear();
	m_status.errors_local.clear();
	m_rules = Rules();
	m_cur_section.erase();
	m_local = false;
	m_global_rules.clear();
	m_local_rules.clear();
}

RulesReader::Status
RulesReader::readRules(istream& strm_global, istream& strm_local)
{
	reset();
	read(strm_global);
	processDefaultRule(m_rules.m_default_global_rule, m_rules.m_default_rule, m_global_rules);
	
	m_local = true;
	m_cur_section.resize(0);
	read(strm_local);
	processDefaultRule(m_rules.m_default_local_rule, m_rules.m_default_global_rule, m_local_rules);
	
	groupRules();
	
	m_global_rules.clear();
	m_local_rules.clear();
	
	return m_status;
}

bool
RulesReader::processElement(const Element& element, int line)
{
	switch (element.getType()) {
		case Element::SECTION:
			m_cur_section = element.getText();
			processSection(m_cur_section, line);
			break;
		case Element::KEY_VALUE:
			processParameter(m_cur_section, element.getKey(), element.getValue(), line);
			break;
		default:
			break;
	}
	return true;
}

void
RulesReader::processError(const Error& error)
{
	if (m_local) {
		m_status.errors_local.push_back(error);
	} else {
		m_status.errors_global.push_back(error);
	}
	if (m_status.status != Status::ERRORS)
		m_status.status = (error.getType() == Error::T_ERROR ? Status::ERRORS : Status::WARNINGS);
}

void
RulesReader::processSection(const string& section, int line)
{
	// try compiling the regex just to check the syntax
	regex_t regex;
	string pattern = "^http://"+section+"$";
	int res = regcomp(&regex, pattern.c_str(), REG_EXTENDED|REG_ICASE|REG_NOSUB);
	if (res) {
		processError(Error("Error compiling regular expression", line));
	} else {
		regfree(&regex);
	}
}

void
RulesReader::processParameter(const string& section, const string& key, const string& value, int line)
{
	int val = validateParameter(key, value, line);
	if (val < 0)
		return;
	RuleStorage& storage = (m_local ? m_local_rules : m_global_rules);
	storage[section][key] = val;
}

int
RulesReader::validateParameter(const string& key, const string& value, int line)
{
	int range_min = 0;
	int range_max;
	if (key == "filter") {
		range_max = MAX_FILTER;
	} else if (key == "ad") {
		range_max = MAX_AD;
	} else if (key == "scripts") {
		range_max = MAX_SCRIPTS;
	} else if (key == "jsengine") {
		range_max = MAX_JSENGINE;
	} else if (key == "target_blank") {
		range_max = MAX_TARGET_BLANK;
	} else {
		range_max = -1; // prevent compiler warning
		processError(Error(Error::T_WARNING, "Unknown key: "+key, line));
		return -1;
	}
	int val = str2num<int>(value);
	if (value != num2str(val)) {
		processError(Error(Error::T_WARNING, "Value must be integer", line));
		return -1;
	}
	if (val < range_min || val > range_max) {
		processError(Error(Error::T_WARNING, "value outside of range", line));
		return -1;
	}
	return val;
}

void
RulesReader::processDefaultRule(Rules::Rule& result, const Rules::Rule& base, RuleStorage& rules)
{
	RuleStorage::iterator fit = rules.find("");
	if (fit == rules.end()) {
		result = base;
		return;
	}
	map<string, int>::const_iterator fit1 = fit->second.find("filter");
	if (fit1 == fit->second.end()) {
		result.m_filter = base.m_filter;
	} else {
		result.m_filter = fit1->second;
	}
	fit1 = fit->second.find("ad");
	if (fit1 == fit->second.end()) {
		result.m_ad = base.m_ad;
	} else {
		result.m_ad = fit1->second;
	}
	fit1 = fit->second.find("scripts");
	if (fit1 == fit->second.end()) {
		result.m_scripts = base.m_scripts;
	} else {
		result.m_scripts = fit1->second;
	}
	fit1 = fit->second.find("jsengine");
	if (fit1 == fit->second.end()) {
		result.m_jsengine = base.m_jsengine;
	} else {
		result.m_jsengine = fit1->second;
	}
	fit1 = fit->second.find("target_blank");
	if (fit1 == fit->second.end()) {
		result.m_target_blank = base.m_target_blank;
	} else {
		result.m_target_blank = fit1->second;
	}
	rules.erase(fit);
}

void
RulesReader::groupRules()
{
	typedef map<string, map<int, list<string> > > storage_type; // [key => [value => [regexes]]]
	storage_type global_storage, local_storage;
	bool local = false;
	do {
		storage_type& storage = (local ? local_storage : global_storage);
		RuleStorage& source = (local ? m_local_rules : m_global_rules);
		const Rules::Rule& def = (local ? m_rules.m_default_local_rule : m_rules.m_default_global_rule);
		map<string, int> def_map;
		def_map["filter"] = def.m_filter;
		def_map["ad"] = def.m_ad;
		def_map["scripts"] = def.m_scripts;
		def_map["jsengine"] = def.m_jsengine;
		def_map["target_blank"] = def.m_target_blank;
		RuleStorage::const_iterator iter1 = source.begin();
		for (; iter1 != source.end(); ++iter1) {
			map<string, int>::const_iterator iter2 = iter1->second.begin();
			for (; iter2 != iter1->second.end(); ++iter2) {
				if (def_map[iter2->first] == iter2->second) {
					/*
					this is an optimisation:
					if the group-by value is equal to default value,
					then there is no need to associate a regex to it
					because it will fall back to default without it
					*/
					continue;
				}
				storage[iter2->first][iter2->second].push_back(iter1->first);
			}
		}
		local = !local;
	} while (local == true);
	
	groupBy("filter", global_storage["filter"], m_rules.m_global_re_by_filter);
	groupBy("filter", local_storage["filter"], m_rules.m_local_re_by_filter);
	groupBy("ad", global_storage["ad"], m_rules.m_global_re_by_ad);
	groupBy("ad", local_storage["ad"], m_rules.m_local_re_by_ad);
	groupBy("scripts", global_storage["scripts"], m_rules.m_global_re_by_scripts);
	groupBy("scripts", local_storage["scripts"], m_rules.m_local_re_by_scripts);
	groupBy("jsengine", global_storage["jsengine"], m_rules.m_global_re_by_jsengine);
	groupBy("jsengine", local_storage["jsengine"], m_rules.m_local_re_by_jsengine);
	groupBy("target_blank", global_storage["target_blank"], m_rules.m_global_re_by_target_blank);
	groupBy("target_blank", local_storage["target_blank"], m_rules.m_local_re_by_target_blank);
}

void
RulesReader::groupBy(const string& key, const map<int, list<string> >& src, map<int, Rules::RegexPtr>& dst)
{
	map<int, list<string> >::const_iterator iter1 = src.begin();
	for (; iter1 != src.end(); ++iter1) {
		int value = iter1->first;
		if (iter1->second.empty()) {
			dst[value] = Rules::RegexPtr(new RegExp);
			continue;
		}
		string composite_regex;
		list<string>::const_iterator iter2 = iter1->second.begin();
		for (; iter2 != iter1->second.end(); ++iter2) {
			if (iter2 != iter1->second.begin()) {
				composite_regex += "|";
			}
			composite_regex += "(^http://"+*iter2+"$)";
		}
		dst[value] = Rules::RegexPtr(new RegExp(composite_regex, REG_EXTENDED|REG_ICASE|REG_NOSUB));
		if (dst[value]->isNull()) {
			processError(Error("Could not compile composite regex for "+key+"="+num2str(value)));
		}
	}
}

