/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef UTIL_H_
#define UTIL_H_

#include <string>
#include <sstream>
#include <list>
#include <map>

template<typename NumType>
std::string num2str(NumType arg)
{
	std::ostringstream strm;
	strm << arg;
	return strm.str();
}

template<typename NumType>
NumType str2num(const std::string& arg, NumType dflt=0)
{
	std::istringstream strm(arg);
	NumType res = dflt;
	strm >> res;
	return res;
}

template<typename FromType, typename ToType>
class AssignTypeConvertor
{
public:
	static ToType convert(const FromType& from) { return ToType(from); }
};

template<
typename KeyType, typename ValType,
class KeyToString=AssignTypeConvertor<KeyType, std::string>,
class ValToString=AssignTypeConvertor<ValType, std::string>
>
class MapSerializer
{
public:
	typedef std::map<KeyType, ValType> MapType;
	static std::string serialize(const MapType& m)
	{
		std::ostringstream strm;
		typename MapType::const_iterator iter = m.begin();
		for (; iter != m.end(); ++iter) {
			std::string key = KeyToString::convert(iter->first);
			std::string val = ValToString::convert(iter->second);
			strm << ':' << key.length() << ':' << key;
			strm << '=' << val.length() << ':' << val << '\n';
		}
		strm << '\0';
		return strm.str();
	}
};

template<
typename KeyType, typename ValType,
class StringToKey=AssignTypeConvertor<std::string, KeyType>,
class StringToVal=AssignTypeConvertor<std::string, ValType>
>
class MapUnserializer
{
public:
	typedef std::map<KeyType, ValType> MapType;
	static void unserialize(const std::string& data, MapType& m)
	{
		m.clear();
		std::istringstream strm(data);
		while (true) {
			char ch = strm.get();
			if (ch != ':')
				break;
			std::string::size_type keylen = 0;
			strm >> keylen;
			ch = strm.get();
			if (ch != ':')
				break;
			std::string key;
			key.resize(keylen);
			strm.read(const_cast<char *>(key.data()), keylen);
			key.resize(strm.gcount());
			ch = strm.get();
			if (ch != '=')
				break;
			std::string::size_type vlen = 0;
			strm >> vlen;
			ch = strm.get();
			if (ch != ':')
				break;
			std::string val;
			val.resize(vlen);
			strm.read(const_cast<char *>(val.data()), vlen);
			val.resize(strm.gcount());
			ch = strm.get();
			if (ch != '\n')
				break;
			m[StringToKey::convert(key)] = StringToVal::convert(val);
		}
	}
};

template<
typename ValType,
class ValToString=AssignTypeConvertor<ValType, std::string>
>
class ListSerializer
{
public:
	typedef std::list<ValType> ListType;
	static std::string serialize(const ListType& l)
	{
		std::ostringstream strm;
		typename ListType::const_iterator iter = l.begin();
		for (; iter != l.end(); ++iter) {
			std::string val = ValToString::convert(*iter);
			strm << ':' << val.length() << ':' << val << '\n';
		}
		strm << '\0';
		return strm.str();
	}
};

template<
typename ValType,
class StringToVal=AssignTypeConvertor<std::string, ValType>
>
class ListUnserializer
{
public:
	typedef std::list<ValType> ListType;
	static void unserialize(const std::string& data, ListType& l)
	{
		l.clear();
		std::istringstream strm(data);
		while (true) {
			char ch = strm.get();
			if (ch != ':')
				break;
			std::string::size_type len = 0;
			strm >> len;
			ch = strm.get();
			if (ch != ':')
				break;
			std::string item;
			item.resize(len);
			strm.read(const_cast<char *>(item.data()), len);
			item.resize(strm.gcount());
			ch = strm.get();
			if (ch != '\n')
				break;
			l.push_back(StringToVal::convert(item));
		}
	}
};

#endif

