/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "decompressorresponsefilter.h"
#include "gzipdecompressor.h"
#include "deflatedecompressor.h"
#include "httpresponsemetadata.h"
#include "responsefilterchain.h"
#include <sigc++/sigc++.h>
#include <algorithm>

AbstractResponseFilter*
DecompressorResponseFilter::create(NextFilterProvider* follower, HttpResponseMetadata* metadata)
{
	bool ok = true;
	NextFilterProvider constructor = makeConstructor(follower, metadata, ok);
	if (!ok) {
		return 0;
	} else {
		return constructor(metadata);
	}
}

AbstractResponseFilter*
DecompressorResponseFilter::createStack(NextFilterProvider* follower,
	HttpResponseMetadata* metadata, int max_items)
{
	bool ok = true;
	HttpResponseMetadata new_metadata(*metadata);
	NextFilterProvider constructor = makeRecursiveConstructor(follower, &new_metadata, max_items, ok);
	if (!ok) {
		return 0;
	} else {
		return constructor(&new_metadata);
	}
}

DecompressorResponseFilter::DecompressorResponseFilter(AbstractResponseFilter* follower, AbstractDecompressor* delegate)
:	m_pFollower(follower), m_pDelegate(delegate)
{
	delegate->onDataOutput().connect(SigC::slot(*this, &DecompressorResponseFilter::processOutputData));
}

DecompressorResponseFilter::~DecompressorResponseFilter()
{
	delete m_pDelegate;
	m_pFollower->destroy();
}

void
DecompressorResponseFilter::consumeDataChunk(const uint8_t* begin, const uint8_t* end)
{
	m_pDelegate->consumeDataChunk(begin, end);
}

void
DecompressorResponseFilter::consumeEOF()
{
	m_pDelegate->consumeEOF();
	m_pFollower->consumeEOF();
}

void
DecompressorResponseFilter::processOutputData(const uint8_t* begin, const uint8_t* end)
{
	m_pFollower->consumeDataChunk(begin, end);
}

AbstractResponseFilter::NextFilterProvider
DecompressorResponseFilter::makeConstructor(
	NextFilterProvider* follower, HttpResponseMetadata* metadata, bool& ok)
{
	static const char* decompressors[] = { "gzip", "deflate" };
	static const char** decompressors_end = decompressors+sizeof(decompressors)/sizeof(*decompressors);
	HttpHeader *teh = metadata->headers().getHeaderPtr("Transfer-Encoding");
	HttpHeader *ceh = metadata->headers().getHeaderPtr("Content-Encoding");
	const char** enc_ptr;
	if (teh && !teh->elements().empty()) {
		std::string enc = teh->elements().getElements().back().getLowercaseName();
		enc_ptr = std::find(decompressors, decompressors_end, enc);
		if (enc_ptr == decompressors_end) {
			ok = false;
			return *follower;
		}
		teh->elements().removeLastElement();
		if (teh->elements().empty()) {
			metadata->headers().removeHeader("Transfer-Encoding");
		}
	} else if (ceh && !ceh->elements().empty()) {
		std::string enc = ceh->elements().getElements().back().getLowercaseName();
		enc_ptr = std::find(decompressors, decompressors_end, enc);
		if (enc_ptr == decompressors_end) {
			ok = false;
			return *follower;
		}
		ceh->elements().removeLastElement();
		if (ceh->elements().empty()) {
			metadata->headers().removeHeader("Content-Encoding");
		}
	} else {
		ok = false;
		return *follower;
	}
	metadata->setBodyStatus(HttpMessageMetadata::BODY_SIZE_UNKNOWN);
	AbstractDecompressor* delegate = (enc_ptr == decompressors)
		? static_cast<AbstractDecompressor*>(new GzipDecompressor)
		: static_cast<AbstractDecompressor*>(new DeflateDecompressor(true));
	ok = true;
	return SigC::bind(SigC::bind(SigC::slot(&DecompressorResponseFilter::construct), *enc_ptr), follower, delegate);
}

AbstractResponseFilter::NextFilterProvider
DecompressorResponseFilter::makeRecursiveConstructor(
	NextFilterProvider* follower, HttpResponseMetadata* metadata, int max_items, bool& ok)
{
	if (max_items <= 0) {
		ok = false;
		return *follower;
	}
	NextFilterProvider nfp = makeConstructor(follower, metadata, ok);
	if (ok == false) {
		return *follower;
	}
	if (!ResponseFilterChain::isBodyEncoded(*metadata)) {
		ok = true;
		return nfp;
	}
	return makeRecursiveConstructor(&nfp, metadata, max_items-1, ok);
}

AbstractResponseFilter*
DecompressorResponseFilter::construct(HttpResponseMetadata* metadata,
	NextFilterProvider* follower, AbstractDecompressor* delegate, const char* encoding)
{
#ifdef DEBUG
	metadata->headers().addHeader("X-filtered-by", "DecompressorResponseFilter ["+std::string(encoding)+"]");
#endif
	return new DecompressorResponseFilter((*follower)(metadata), delegate);
}

