/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "deflatecompressor.h"

DeflateCompressor::DeflateCompressor(bool no_zlib_header, bool add_stream_end_mark)
:	m_noZlibHeader(no_zlib_header),
	m_addStreamEndMark(no_zlib_header && add_stream_end_mark),
	m_isError(false),
	m_isEOF(false),
	m_unsyncedSize(0)
{
	initStream();
}

DeflateCompressor::~DeflateCompressor()
{
	deflateEnd(&m_strm);
}

void
DeflateCompressor::reset()
{
	deflateReset(&m_strm);
	m_isError = false;
	m_isEOF = false;
	m_unsyncedSize = 0;
}

void
DeflateCompressor::consumeDataChunk(const uint8_t* begin, const uint8_t* end)
{
	if (begin == end) {
		return;
	}
	if (m_isError || m_isEOF) {
		m_isError = true;
		m_onErrorSignal();
		return;
	}
	m_strm.next_in = const_cast<Bytef*>(begin);
	m_strm.avail_in = end-begin;
	int flush = 0;
	if ((m_unsyncedSize += m_strm.avail_in) >= SYNC_INTERVAL) {
		flush = Z_SYNC_FLUSH;
		m_unsyncedSize = 0;
	}
	int res;
	do {
		m_strm.next_out = m_outBuf;
		m_strm.avail_out = sizeof(m_outBuf);
		res = deflate(&m_strm, flush);
		if (m_strm.next_out != m_outBuf) {
			m_onDataOutputSignal(m_outBuf, m_strm.next_out);
		}
	} while (res == Z_OK && m_strm.avail_out == 0);
	if (res != Z_OK && res != Z_BUF_ERROR) {
		m_isError = true;
		m_onErrorSignal();
	}
}

void
DeflateCompressor::consumeEOF()
{
	if (m_isEOF) {
		return;
	} else {
		m_isEOF = true;
	}
	if (m_isError) {
		m_onErrorSignal();
		return;
	}
	m_strm.next_in = 0;
	m_strm.avail_in = 0;
	int res;
	do {
		m_strm.next_out = m_outBuf;
		m_strm.avail_out = sizeof(m_outBuf);
		res = deflate(&m_strm, Z_FINISH);
		if (res == Z_STREAM_END && m_addStreamEndMark) {
			// we have to append a dummy byte after the compressed stream,
			// or the decompression will never return Z_STREAM_END
			if (m_strm.avail_out != 0) {
				*m_strm.next_out = 0;
				m_onDataOutputSignal(m_outBuf, m_outBuf+sizeof(m_outBuf)-m_strm.avail_out+1);
			} else {
				m_onDataOutputSignal(m_outBuf, m_outBuf+sizeof(m_outBuf));
				m_outBuf[0] = 0;
				m_onDataOutputSignal(m_outBuf, m_outBuf+1);
			}
			break;
		}
		if (m_strm.avail_out != sizeof(m_outBuf)) {
			m_onDataOutputSignal(m_outBuf, m_outBuf+sizeof(m_outBuf)-m_strm.avail_out);
		}
	} while (res == Z_OK && m_strm.avail_out == 0);
	if (res != Z_STREAM_END) {
		m_isError = true;
		m_onErrorSignal();
	}
}

void
DeflateCompressor::initStream()
{
	m_strm.zalloc = Z_NULL;
	m_strm.zfree = Z_NULL;
	m_strm.opaque = Z_NULL;
	m_strm.next_in = Z_NULL;
	if (m_noZlibHeader) {
		deflateInit2(&m_strm, Z_DEFAULT_COMPRESSION, Z_DEFLATED, -MAX_WBITS, 8, Z_DEFAULT_STRATEGY);
	} else {
		deflateInit(&m_strm, Z_DEFAULT_COMPRESSION);
	}
}

