/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "flashinspector.h"
#include <cstring>
#include <cstddef>
#include <algorithm>
#include <vector>

const int FlashInspector::INSPECTION_FRAME_LIMIT;
const std::size_t FlashInspector::INSPECTION_SIZE_LIMIT;
const double FlashInspector::BUTTON_AREA_THRESHOLD = 0.8;

FlashInspector::FlashInspector()
:	m_isInProgress(true),
	m_isFailed(false),
	m_isAd(false),
	m_state(ST_HEADER),
	m_pos(0),
	m_curFrame(0)
{
	m_data.reserve(10*1024);
}

void
FlashInspector::reset()
{
	m_isInProgress = true;
	m_isFailed = false;
	m_isAd = false;
	m_state = ST_HEADER;
	m_pos = 0;
	m_curFrame = 0;
	if (m_data.capacity() > 20*1024) {
		std::vector<uint8_t>().swap(m_data);
	} else {
		m_data.clear();
	}
	if (m_uncompressedData.capacity() > 50*1024) {
		std::vector<uint8_t>().swap(m_uncompressedData);
	} else {
		m_uncompressedData.clear();
	}
}

void
FlashInspector::consumeDataChunk(const uint8_t* begin, const uint8_t* end)
{
	if (begin == end) {
		return;
	}
	
	std::size_t old_size = m_data.size();
	m_data.resize(end-begin+old_size);
	std::memcpy(&m_data[old_size], begin, end-begin);
	
	if (!m_isInProgress) {
		return;
	}
	if (m_data.size() > INSPECTION_SIZE_LIMIT) {
		m_isInProgress = false;
		return;
	}
	
	if (m_pDecomp.get()) {
		std::size_t size = m_uncompressedData.size();
		m_pDecomp->consumeDataChunk(begin, end);
		if (!m_isInProgress) {
			return;
		}
		if (size == m_uncompressedData.size()) {
			return;
		}
		processMoreData(&m_uncompressedData[m_pos], &m_uncompressedData[m_uncompressedData.size()]);
	} else {
		processMoreData(&m_data[m_pos], &m_data[m_data.size()]);
	}
}

void
FlashInspector::consumeEOF()
{
	if (!m_isInProgress) {
		return;
	}
	if (m_pDecomp.get()) {
		std::size_t size = m_uncompressedData.size();
		m_pDecomp->consumeEOF();
		if (!m_isInProgress || m_uncompressedData.size() == size) {
			return;
		}
		processMoreData(&m_uncompressedData[m_pos], &m_uncompressedData[m_uncompressedData.size()]);
	}
	m_isInProgress = false;
}

void
FlashInspector::processUncompressedData(const uint8_t* begin, const uint8_t* end)
{
	std::size_t size = m_uncompressedData.size();
	if (end-begin+size > INSPECTION_SIZE_LIMIT) {
		m_isInProgress = false;
		return;
	}
	m_uncompressedData.resize(end-begin+size);
	std::memcpy(&m_uncompressedData[size], begin, end-begin);
}

void
FlashInspector::processDecompressionError()
{
	m_isInProgress = false;
	m_isFailed = true;
}

void
FlashInspector::processMoreData(const uint8_t* cur, const uint8_t* end)
try {
	switch (m_state) {
		case ST_TAG_BODY:
		case_ST_TAG_BODY: {
			if (cur+m_tagLength > end) {
				return;
			}
			const uint8_t* tag_end = cur+m_tagLength;
			try {
				if (m_tagCode == 7) {
					processButton(cur, tag_end);
				} else if (m_tagCode == 34) {
					processButton2(cur, tag_end);
				} else if (m_tagCode == 4) {
					processPlaceObject(cur, tag_end);
					if (!m_isInProgress) {
						return;
					}
				} else if (m_tagCode == 26) {
					processPlaceObject2(cur, tag_end);
					if (!m_isInProgress) {
						return;
					}
				} else if (m_tagCode == 2 || m_tagCode == 22 || m_tagCode == 32 || m_tagCode == 46) {
					uint16_t char_id = readU16(cur, tag_end);
					Rect bounds;
					readRect(bounds, cur+2, tag_end, 0);
					m_shapeDimensions[char_id] = Dimensions(bounds.getWidth(), bounds.getHeight());
				} else if (m_tagCode == 1) {
					if (++m_curFrame >= INSPECTION_FRAME_LIMIT) {
						m_isInProgress = false;
						return;
					}
				}
			} catch (OutOfBoundsException e) { // broken file
				m_isInProgress = false;
				m_isFailed = true;
				return;
			}
			cur = tag_end;
			m_pos += m_tagLength;
			m_state = ST_TAG_HEADER;
		}
		case ST_TAG_HEADER:
		case_ST_TAG_HEADER: {
			const uint8_t* start = cur;
			uint16_t code_and_length = readU16(cur, end);
			cur += 2;
			m_tagCode = code_and_length >> 6;
			m_tagLength = code_and_length & 0x3F;
			if (m_tagLength == 0x3F) {
				m_tagLength = readU32(cur, end);
				cur += 4;
			}
			if (code_and_length == 0) {
				m_isInProgress = false;
				return;
			}
			m_pos += cur-start;
			m_state = ST_TAG_BODY;
			goto case_ST_TAG_BODY;
		}
		case ST_HEADER: {
			if (cur+8 > end) {
				return;
			}
			if (!std::memcmp(cur, "FWS", 3)) {
				cur += 8;
				m_pos += 8;
			} else if (!std::memcmp(cur, "FWC", 3) || !std::memcmp(cur, "CWS", 3)) {
				m_uncompressedData.reserve(30*1024);
				m_pDecomp.reset(new DeflateDecompressor(false));
				m_pDecomp->onDataOutput().connect(SigC::slot(*this, &FlashInspector::processUncompressedData));
				m_pDecomp->onError().connect(SigC::slot(*this, &FlashInspector::processDecompressionError));
				m_pDecomp->consumeDataChunk(cur+8, end);
				if (!m_isInProgress) {
					return;
				}
				m_pos = 0;
				cur = &m_uncompressedData[0];
				end = &m_uncompressedData[m_uncompressedData.size()];
			} else {
				// not a flash file
				m_isInProgress = false;
				m_isFailed = true;
				return;
			}
			m_state = ST_HEADER2;
		}
		case ST_HEADER2: {
			const uint8_t* start = cur;
			cur += readRect(m_frameRect, cur, end, 0)+7 >> 3;
			m_frameArea = double(m_frameRect.getWidth())*double(m_frameRect.getHeight());
			/*uint16_t frame_rate =*/ readU16(cur, end);
			/*uint16_t frame_count =*/ readU16(cur+2, end);
			cur += 4;
			m_pos += cur-start;
			m_state = ST_TAG_HEADER;
			goto case_ST_TAG_HEADER;
		}
	}
} catch (OutOfBoundsException e) {
	// more data is needed
}

void
FlashInspector::processButton(const uint8_t* data, const uint8_t* end)
{
	uint16_t char_id = readU16(data, end);
	data = processButtonRecords(m_buttonDimensions[char_id], data+2, end, false);
	bool url_action_found = false;
	while (true) {
		uint8_t action_code = readU8(data++, end);
		if (action_code == 0)
			break;
		if (!(action_code & 0x80)) {
			// short actions are of no interest to us
			continue;
		}
		std::size_t length = readU16(data, end);
		data += 2;
		if (action_code == 0x83 || action_code == 0x9a) {
			url_action_found = processURLAction(data, std::min(end, data+length), action_code);
			if (url_action_found)
				break;
		}
		data += length;
	}
	if (!url_action_found) {
		m_buttonDimensions.erase(char_id);
	}
}

void
FlashInspector::processButton2(const uint8_t* data, const uint8_t* end)
{
	uint16_t char_id = readU16(data, end);
	if (readU8(data+2, end) & 1) { // menu button
		return;
	}
	data += 3;
	const uint8_t* characters = data+2;
	const uint8_t* button_actions = data+readU16(data, end);
	data = button_actions;
	bool url_action_found = false;
	while (true) {
		std::size_t action_size = readU16(data, end);
		const uint8_t* cur = data+4;
		const uint8_t* record_end = action_size ? data+action_size : end;
		if (record_end > end)
			throw OOB;
		while (cur < record_end) {
			uint8_t action_code = readU8(cur++, record_end);
			if (action_code == 0)
				break;
			if (!(action_code & 0x80)) {
				// short actions are of no interest to us
				continue;
			}
			std::size_t length = readU16(cur, record_end);
			cur += 2;
			if (action_code == 0x83 || action_code == 0x9a) {
				url_action_found = processURLAction(cur, std::min(end, cur+length), action_code);
				if (url_action_found)
					break;
			}
			cur += length;
		}
		if (url_action_found) {
			// Some encoders combine the ActionOffset and CharacterEndFlag when there are no button records.
			// In that case, the following condition will be false.
			if (characters < button_actions) {
				processButtonRecords(m_buttonDimensions[char_id], characters, button_actions, true);
			}
			break;
		}
		if (action_size == 0) {
			break;
		}
		data += action_size;
		if (data >= end) {
			throw OOB;
		}
	}
}

void
FlashInspector::processPlaceObject(const uint8_t* data, const uint8_t* end)
{
	uint16_t char_id = readU16(data, end);
	Matrix mat;
	readMatrix(mat, data+4, end);
	std::map<uint16_t, Dimensions>::iterator fit = m_buttonDimensions.find(char_id);
	if (fit == m_buttonDimensions.end())
		return;
	double area = (((1.0/65536)*mat.xScale)*fit->second.width)
		* (((1.0/65536)*mat.yScale)*fit->second.height);
	if (area >= BUTTON_AREA_THRESHOLD*m_frameArea) {
		m_isInProgress = false;
		m_isAd = true;
	}
}

void
FlashInspector::processPlaceObject2(const uint8_t* data, const uint8_t* end)
{
	uint8_t flags = readU8(data, end);
	if (!(flags & 2))
		return;
	uint16_t char_id = readU16(data+3, end);
	std::map<uint16_t, Dimensions>::iterator fit = m_buttonDimensions.find(char_id);
	if (fit == m_buttonDimensions.end())
		return;
	double width = fit->second.width;
	double height = fit->second.height;
	if (flags & 4) {
		Matrix mat;
		readMatrix(mat, data+5, end);
		width *= (1.0/65536)*mat.xScale;
		height *= (1.0/65536)*mat.yScale;
	}
	if (width*height >= BUTTON_AREA_THRESHOLD*m_frameArea) {
		m_isInProgress = false;
		m_isAd = true;
	}
}

bool
FlashInspector::processURLAction(const uint8_t* data, const uint8_t* end, uint8_t code)
{
	if (code == 0x9a) {
		return true;
	} else if (code == 0x83) {
		const uint8_t* url_end = reinterpret_cast<const uint8_t*>(std::memchr(data, 0, end-data));
		if (!url_end)
			throw OOB;
		++url_end;
		const uint8_t* target_end = reinterpret_cast<const uint8_t*>(std::memchr(url_end, 0, end-url_end));
		if (!target_end)
			throw OOB;
		return std::strncmp(reinterpret_cast<const char*>(url_end), "_level", 6);
	}
	return false;
}

namespace {
	struct ButtonRecordInfo {
		uint8_t flags;
		uint16_t char_id;
		int32_t xScale;
		int32_t yScale;
	};
}

const uint8_t*
FlashInspector::processButtonRecords(Dimensions& hit_area, const uint8_t* data, const uint8_t* end, bool has_cxform)
{
	std::vector<ButtonRecordInfo> chars;
	chars.reserve(10);
	while (true) {
		uint8_t flags = readU8(data++, end);
		if (!flags) {
			break;
		}
		uint16_t char_id = readU16(data, end);
		data += 4;
		Matrix mat;
		data += readMatrix(mat, data, end)+7 >> 3;
		if (has_cxform) {
			data += skipColorTransformAlpha(data, end)+7 >> 3;
		}
		chars.resize(chars.size()+1);
		ButtonRecordInfo& r = chars.back();
		r.flags = flags;
		r.char_id = char_id;
		r.xScale = mat.xScale;
		r.yScale = mat.yScale;
	}
	
	struct AreaCalculator {
		static bool getLargestShape(Dimensions& area, uint8_t mask,
			std::vector<ButtonRecordInfo>& list, std::map<uint16_t, Dimensions>& size_map) {
			bool found = false;
			double max_width = 0.0, max_height = 0.0, max_area = 0.0;
			for (unsigned int i = 0; i < list.size(); ++i) {
				if (list[i].flags & mask) {
					std::map<uint16_t, Dimensions>::iterator fit = size_map.find(list[i].char_id);
					if (fit == size_map.end()) {
						continue;
					}
					double width = ((1.0/65536)*list[i].xScale)*fit->second.width;
					double height = ((1.0/65536)*list[i].yScale)*fit->second.height;
					double area = width*height;
					if (!found || area > max_area) {
						found = true;
						max_area = area;
						max_width = width;
						max_height = height;
					}
				}
			}
			if (found) {
				area.width = static_cast<uint32_t>(max_width);
				area.height = static_cast<uint32_t>(max_height);
				return true;
			}
			return false;
		}
	};
	
	if (AreaCalculator::getLargestShape(hit_area, 8, chars, m_shapeDimensions)) { // hit state
		return data;
	}
	if (AreaCalculator::getLargestShape(hit_area, 4, chars, m_shapeDimensions)) { // down state
		return data;
	}
	if (AreaCalculator::getLargestShape(hit_area, 2, chars, m_shapeDimensions)) { // over state
		return data;
	}
	if (AreaCalculator::getLargestShape(hit_area, 1, chars, m_shapeDimensions)) { // up state
		return data;
	}
	hit_area.width = 0;
	hit_area.height = 0;
	return data;
}

std::size_t
FlashInspector::readRect(Rect& rect, const uint8_t* data, const uint8_t* end, std::size_t bitoffset)
{
	uint32_t nbits = readUBits(data, end, 5, bitoffset);
	bitoffset += 5;
	rect.left = readSBits(data, end, nbits, bitoffset);
	bitoffset += nbits;
	rect.right = readSBits(data, end, nbits, bitoffset);
	bitoffset += nbits;
	rect.top = readSBits(data, end, nbits, bitoffset);
	bitoffset += nbits;
	rect.bottom = readSBits(data, end, nbits, bitoffset);
	return bitoffset+nbits;
}

std::size_t
FlashInspector::readMatrix(Matrix& mat, const uint8_t* data, const uint8_t* end)
{
	std::size_t bitoffset = 1;
	if (readU8(data, end) & 0x80) {
		std::size_t nScaleBits = readUBits(data, end, 5, bitoffset);
		bitoffset += 5;
		mat.xScale = readSBits(data, end, nScaleBits, bitoffset);
		bitoffset += nScaleBits;
		mat.yScale = readSBits(data, end, nScaleBits, bitoffset);
		bitoffset += nScaleBits;
	} else {
		mat.xScale = mat.yScale = 0x10000;
	}
	if (readUBits(data, end, 1, bitoffset++)) {
		std::size_t nRotateBits = readUBits(data, end, 5, bitoffset);
		bitoffset += 5;
		mat.xRotate = readSBits(data, end, nRotateBits, bitoffset);
		bitoffset += nRotateBits;
		mat.yRotate = readSBits(data, end, nRotateBits, bitoffset);
		bitoffset += nRotateBits;
	} else {
		mat.xRotate = mat.yRotate = 0x10000;
	}
	std::size_t nTranslateBits = readUBits(data, end, 5, bitoffset);
	bitoffset += 5;
	mat.xTranslate = readSBits(data, end, nTranslateBits, bitoffset);
	bitoffset += nTranslateBits;
	mat.yTranslate = readSBits(data, end, nTranslateBits, bitoffset);
	return bitoffset+nTranslateBits;
}

std::size_t
FlashInspector::skipColorTransformAlpha(const uint8_t* data, const uint8_t* end)
{
	uint8_t first_byte = readU8(data, end);
	std::size_t nbits = (first_byte >> 6) & 0x0f;
	std::size_t bitoffset = 6;
	if (first_byte & (1<<7))
		bitoffset += nbits<<2;
	if (first_byte & (1<<6))
		bitoffset += nbits<<2;
	return bitoffset;
}

uint32_t
FlashInspector::readUBits(const uint8_t* data, const uint8_t* end, std::size_t nbits)
{
	if (data+(nbits+7>>3) > end)
		throw OOB;
	uint32_t res = 0;
	while (nbits > 0) {
		uint32_t dta = *data;
		if (nbits < 8) {
			return res | (dta >> 8-nbits);
		} else {
			nbits -= 8;
			res |= dta << nbits;
			++data;
		}
	}
	return res;
}

uint32_t
FlashInspector::readUBits(const uint8_t* data, const uint8_t* end, std::size_t nbits, std::size_t bitoffset)
{
	if (bitoffset > 8) {
		std::size_t shift = bitoffset >> 3;
		data += shift;
		bitoffset -= shift << 3;
	}
	if (data >= end)
		throw OOB;
	std::size_t bits_avail = 8-bitoffset;
	uint32_t dta = *data & ((1<<bits_avail)-1);
	if (nbits <= bits_avail) {
		return dta >> bits_avail-nbits;
	} else {
		nbits -= bits_avail;
		return (dta << nbits) | readUBits(++data, end, nbits);
	}
}

