/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef FLASHINSPECTOR_H_
#define FLASHINSPECTOR_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "types.h"
#include "deflatedecompressor.h"
#include <sigc++/sigc++.h>
#include <cstddef>
#include <vector>
#include <map>
#include <memory>
#include <utility>

class FlashInspector : public SigC::Object
{
public:
	FlashInspector();
	void consumeDataChunk(const uint8_t* begin, const uint8_t* end);
	void consumeEOF();
	void reset();
	bool isInspectionInProgress() const { return m_isInProgress; }
	bool isInspectionFailed() const { return m_isFailed; }
	bool isAd() const { return m_isAd; }
	std::pair<const uint8_t*, const uint8_t*> getAccumulatedData() const {
		return std::pair<const uint8_t*, const uint8_t*>(&m_data[0], &m_data[m_data.size()]);
	}
private:
	static const int INSPECTION_FRAME_LIMIT = 3;
	static const std::size_t INSPECTION_SIZE_LIMIT = 50*1024;
	static const double BUTTON_AREA_THRESHOLD;
	enum OutOfBoundsException { OOB };
	enum State {
		ST_HEADER,
		ST_HEADER2,
		ST_TAG_HEADER,
		ST_TAG_BODY
	};
	struct Rect {
		int32_t left, right, top, bottom;
		uint32_t getWidth() const { return right-left; }
		uint32_t getHeight() const { return bottom-top; }
	};
	struct Dimensions {
		Dimensions() : width(0), height(0) {}
		Dimensions(uint32_t w, uint32_t h) : width(w), height(h) {}
		uint32_t width, height;
	};
	struct Matrix {
		int32_t xScale, yScale;
		int32_t xRotate, yRotate;
		int32_t xTranslate, yTranslate;
	};
private:
	void processUncompressedData(const uint8_t* begin, const uint8_t* end);
	void processDecompressionError();
	void processMoreData(const uint8_t* begin, const uint8_t* end);
	void processButton(const uint8_t* data, const uint8_t* end);
	void processButton2(const uint8_t* data, const uint8_t* end);
	void processPlaceObject(const uint8_t* data, const uint8_t* end);
	void processPlaceObject2(const uint8_t* data, const uint8_t* end);
	bool processURLAction(const uint8_t* data, const uint8_t* end, uint8_t code);
	const uint8_t* processButtonRecords(Dimensions& hit_area, const uint8_t* data, const uint8_t* end, bool has_cxform);
	static std::size_t readRect(Rect& rect, const uint8_t* data, const uint8_t* end, std::size_t bitoffset);
	static std::size_t readMatrix(Matrix& mat, const uint8_t* data, const uint8_t* end);
	static std::size_t skipColorTransformAlpha(const uint8_t* data, const uint8_t* end);
	static uint8_t readU8(const uint8_t* data, const uint8_t* end);
	static uint16_t readU16(const uint8_t* data, const uint8_t* end);
	static uint32_t readU32(const uint8_t* data, const uint8_t* end);
	static uint32_t readUBits(const uint8_t* data, const uint8_t* end, std::size_t nbits);
	static uint32_t readUBits(const uint8_t* data, const uint8_t* end, std::size_t nbits, std::size_t bitoffset);
	static int32_t readSBits(const uint8_t* data, const uint8_t* end, std::size_t nbits);
	static int32_t readSBits(const uint8_t* data, const uint8_t* end, std::size_t nbits, std::size_t bitoffset);
private:
	bool m_isInProgress;
	bool m_isFailed;
	bool m_isAd;
	State m_state;
	std::vector<uint8_t> m_data;
	std::vector<uint8_t> m_uncompressedData;
	std::size_t m_pos;
	std::auto_ptr<DeflateDecompressor> m_pDecomp;
	std::map<uint16_t, Dimensions> m_shapeDimensions;
	std::map<uint16_t, Dimensions> m_buttonDimensions;
	Rect m_frameRect;
	double m_frameArea;
	uint16_t m_tagCode;
	int m_curFrame;
	std::size_t m_tagLength;
};

inline uint8_t
FlashInspector::readU8(const uint8_t* data, const uint8_t* end)
{
	if (data >= end)
		throw OOB;
	return *data;
}

inline uint16_t
FlashInspector::readU16(const uint8_t* data, const uint8_t* end)
{
	if (data+2 > end)
		throw OOB;
	return (uint16_t)*data | ((uint16_t)*(data+1) << 8);
}

inline uint32_t
FlashInspector::readU32(const uint8_t* data, const uint8_t* end)
{
	if (data+4 > end)
		throw OOB;
	return (uint32_t)*data | ((uint32_t)*(data+1) << 8)
		| ((uint32_t)*(data+2) << 16) | ((uint32_t)*(data+3) << 24);
}

inline int32_t
FlashInspector::readSBits(const uint8_t* data, const uint8_t* end, std::size_t nbits)
{
	uint32_t dta = readUBits(data, end, nbits);
	if (dta & (1 << (nbits-1))) {
		return (~((int32_t(1) << nbits)-1) | dta);
	}
	return dta;
}

inline int32_t
FlashInspector::readSBits(const uint8_t* data, const uint8_t* end, std::size_t nbits, std::size_t bitoffset)
{
	uint32_t dta = readUBits(data, end, nbits, bitoffset);
	if (dta & (1 << nbits-1)) {
		return (~((int32_t(1) << nbits)-1) | dta);
	}
	return dta;
}

#endif

