/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "flashresponsefilter.h"
#include "responsefilterchain.h"
#include "decompressorresponsefilter.h"
#include "httpheader.h"
#include "replacementflash.h"
#include "conf.h"

AbstractResponseFilter*
FlashResponseFilter::create(NextFilterProvider* follower,
	HttpResponseMetadata* metadata, const Params& params)
{
	int status_code = metadata->statusLine().getCode();
	if (status_code != 200) {
		return 0;
	}
	if (metadata->getBodyStatus() == HttpResponseMetadata::NO_BODY_BUT_PRESERVE_HEADERS ||
	    (metadata->getBodyStatus() == HttpResponseMetadata::BODY_SIZE_KNOWN && metadata->getBodySize() == 0)) {
		// no body in response
		return 0;
	}
	
	if (ResponseFilterChain::isBodyEncoded(*metadata)) {
		NextFilterProvider provider =
			SigC::bind(SigC::slot(&FlashResponseFilter::construct), follower, &params);
		return DecompressorResponseFilter::createStack(&provider, metadata, 2);
	} else {
		return construct(metadata, follower, &params);
	}
}

FlashResponseFilter::FlashResponseFilter(NextFilterProvider* follower,
	HttpResponseMetadata* metadata, const Params* params)
:	m_followerProvider(*follower), m_metadata(*metadata), m_params(*params),
	m_state(ST_INSPECTING), m_pFollower(0)
{
}

FlashResponseFilter::~FlashResponseFilter()
{
	if (m_pFollower) {
		m_pFollower->destroy();
	}
}

void
FlashResponseFilter::consumeDataChunk(const uint8_t* begin, const uint8_t* end)
{
	if (m_state == ST_INSPECTING) {
		m_inspector.consumeDataChunk(begin, end);
		if (!m_inspector.isInspectionInProgress()) {
			if (m_inspector.isAd()) {
				m_state = ST_SUBSTITUTING;
				performSubstitution();
			} else {
				m_state = ST_FORWARDING;
				m_pFollower = m_followerProvider(&m_metadata);
				m_pFollower->consumeDataChunk(
					m_inspector.getAccumulatedData().first,
					m_inspector.getAccumulatedData().second
				);
			}
		}
	} else if (m_state == ST_FORWARDING) {
		m_pFollower->consumeDataChunk(begin, end);
	}
	// do nothing if m_state == ST_SUBSTITUTING
}

void
FlashResponseFilter::consumeEOF()
{
	if (m_state == ST_SUBSTITUTING) {
		return;
	}
	if (m_state == ST_INSPECTING) {
		m_inspector.consumeEOF();
		if (m_inspector.isAd()) {
			m_state = ST_SUBSTITUTING;
			performSubstitution();
			return;
		} else {
			m_state = ST_FORWARDING;
			m_pFollower = m_followerProvider(&m_metadata);
			m_pFollower->consumeDataChunk(
				m_inspector.getAccumulatedData().first,
				m_inspector.getAccumulatedData().second
			);
		}
	}
	m_pFollower->consumeEOF();
}

void
FlashResponseFilter::performSubstitution()
{
	ReplacementFlash flash(m_params.width, m_params.height, m_params.url.c_str(),
		m_params.config->getDrawBorders(), m_params.config->getBorderColor());
	m_metadata.headers().setHeader(HttpHeader("Content-Type", flash.getContentType()));
	m_metadata.setBodyStatus(HttpMessageMetadata::BODY_SIZE_KNOWN);
	m_metadata.setBodySize(flash.getBinaryData().second-flash.getBinaryData().first);
	m_pFollower = m_followerProvider(&m_metadata);
	m_pFollower->consumeDataChunk(flash.getBinaryData().first, flash.getBinaryData().second);
	m_pFollower->consumeEOF();
}

AbstractResponseFilter*
FlashResponseFilter::construct(HttpResponseMetadata* metadata,
	NextFilterProvider* follower, const Params* params)
{
#ifdef DEBUG
	metadata->headers().addHeader("X-filtered-by", "FlashResponseFilter");
#endif
	return new FlashResponseFilter(follower, metadata, params);
}

