/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gifimageinspector.h"
#include <cstring>

const uint8_t GIFImageInspector::SIGNATURE[3] = { 'G', 'I', 'F' };

GIFImageInspector::GIFImageInspector()
:	m_state(ST_SIGNATURE), m_delayState(DST_DELAY_NOT_ENCOUNTERED),
	m_isInProgress(true), m_isFailed(false),
	m_width(0), m_height(0), m_isAnimated(false),
	m_curPos(0)
{
}

void
GIFImageInspector::consumeDataChunk(const uint8_t* begin, const uint8_t* end)
{
	std::size_t size = m_data.size();
	m_data.resize(end-begin+size);
	std::memcpy(&m_data[size], begin, end-begin);
	size = m_data.size();;
	if (!m_isInProgress) {
		return;
	}
	switch (m_state) {
		case ST_SIGNATURE: {
			if (size < sizeof(SIGNATURE)) {
				return;
			}
			if (std::memcmp(&m_data[0], SIGNATURE, sizeof(SIGNATURE))) {
				m_isInProgress = false;
				m_isFailed = true;
				return;
			}
			m_state = ST_HEADER;
		}
		case ST_HEADER: {
			if (size < 13) {
				return;
			}
			const uint8_t* bytes = &m_data[0];
			m_width = (unsigned)bytes[6]+((unsigned)bytes[7]<<8);
			m_height = (unsigned)bytes[8]+((unsigned)bytes[9]<<8);
			uint8_t flag = bytes[10];
			if (flag & 0x80) {
				m_state = ST_GLOBAL_COLOR_TABLE;
				m_curBlockSize = 3*(1<<((flag&7)+1));
			} else {
				m_state = ST_BLOCK_BEGIN;
			}
			m_curPos = 13;
			if (!std::memcmp(bytes+3, "87a", 3)) {
				// this version can't be animated
				m_isInProgress = false;
				return;
			}
			if (m_state == ST_BLOCK_BEGIN) {
				goto case_ST_BLOCK_BEGIN;
			}
		}
		case ST_GLOBAL_COLOR_TABLE: {
			if (size < m_curPos+m_curBlockSize) {
				return;
			}
			m_curPos += m_curBlockSize;
			m_state = ST_BLOCK_BEGIN;
		}
		case ST_BLOCK_BEGIN:
		case_ST_BLOCK_BEGIN: {
			if (size <= m_curPos) {
				return;
			}
			uint8_t block_id = m_data[m_curPos];
			switch (block_id) {
				case 0x2c: {
					if (m_delayState == DST_WAITING_FOR_GRAPHIC_BLOCK) {
						m_delayState = DST_WAITING_FOR_ANOTHER_GRAPHIC_BLOCK;
					} else if (m_delayState == DST_WAITING_FOR_ANOTHER_GRAPHIC_BLOCK) {
						m_isInProgress = false;
						m_isAnimated = true;
						return;
					}
					m_state = ST_IMAGE_DESCRIPTOR;
					break;
				}
				case 0x21: {
					m_state = ST_EXTENSION_BLOCK_BEGIN;
					goto case_ST_EXTENSION_BLOCK_BEGIN;
				}
				case 0x3b: {
					m_isInProgress = false;
					return;
				}
				default: {
					m_isInProgress = false;
					m_isFailed = true;
					return;
				}
			}
		}
		case ST_IMAGE_DESCRIPTOR: {
			if (size < m_curPos+10) {
				return;
			}
			uint8_t flag = m_data[m_curPos+9];
			m_curPos += 10;
			if (flag & 0x80) {
				m_state = ST_LOCAL_COLOR_TABLE;
				m_curBlockSize = 3*(1<<((flag&7)+1));
			} else {
				m_state = ST_IMAGE_DATA;
				goto case_ST_IMAGE_DATA;
			}
		}
		case ST_LOCAL_COLOR_TABLE: {
			if (size < m_curPos+m_curBlockSize) {
				return;
			}
			m_curPos += m_curBlockSize;
			m_state = ST_IMAGE_DATA;
		}
		case ST_IMAGE_DATA:
		case_ST_IMAGE_DATA: {
			if (size <= m_curPos) {
				return;
			}
			m_state = ST_IMAGE_DATA_SUBBLOCK_OR_TERMINATOR;
			++m_curPos; // skipping the LZW minimum code size
		}
		case ST_IMAGE_DATA_SUBBLOCK_OR_TERMINATOR:
		case_ST_IMAGE_DATA_SUBBLOCK_OR_TERMINATOR: {
			if (size <= m_curPos) {
				return;
			}
			uint8_t byte = m_data[m_curPos++];
			if (byte == 0x00) {
				m_state = ST_BLOCK_BEGIN;
				goto case_ST_BLOCK_BEGIN;
			} else {
				m_state = ST_IMAGE_DATA_SUBBLOCK;
				m_curBlockSize = byte;
			}
		}
		case ST_IMAGE_DATA_SUBBLOCK: {
			if (size < m_curPos+m_curBlockSize) {
				return;
			}
			m_curPos += m_curBlockSize;
			m_state = ST_IMAGE_DATA_SUBBLOCK_OR_TERMINATOR;
			goto case_ST_IMAGE_DATA_SUBBLOCK_OR_TERMINATOR;
		}
		case ST_EXTENSION_BLOCK_BEGIN:
		case_ST_EXTENSION_BLOCK_BEGIN: {
			if (size < m_curPos+2) {
				return;
			}
			uint8_t ext_byte = m_data[m_curPos+1];
			m_curPos += 2;
			if (ext_byte == 0xf9) {
				m_state = ST_GRAPHIC_CONTROL_EXTENSION_SUBBLOCK_OR_TERMINATOR;
			} else {
				if (ext_byte == 0x01) { // plain text extension
					if (m_delayState == DST_WAITING_FOR_GRAPHIC_BLOCK) {
						m_delayState = DST_WAITING_FOR_ANOTHER_GRAPHIC_BLOCK;
					} else if (m_delayState == DST_WAITING_FOR_ANOTHER_GRAPHIC_BLOCK) {
						m_isInProgress = false;
						m_isAnimated = true;
						return;
					}
				}
				m_state = ST_OTHER_EXTENSION_SUBBLOCK_OR_TERMINATOR;
				goto case_ST_OTHER_EXTENSION_SUBBLOCK_OR_TERMINATOR;
			}
		}
		case ST_GRAPHIC_CONTROL_EXTENSION_SUBBLOCK_OR_TERMINATOR:
		case_ST_GRAPHIC_CONTROL_EXTENSION_SUBBLOCK_OR_TERMINATOR: {
			if (size <= m_curPos) {
				return;
			}
			m_curBlockSize = m_data[m_curPos++];
			if (m_curBlockSize == 0) {
				m_state = ST_BLOCK_BEGIN;
				goto case_ST_BLOCK_BEGIN;
			} else {
				m_state = ST_GRAPHIC_CONTROL_EXTENSION_SUBBLOCK;
				if (m_curBlockSize != 4) {
					m_isInProgress = false;
					m_isFailed = true;
					return;
				}
			}
		}
		case ST_GRAPHIC_CONTROL_EXTENSION_SUBBLOCK: {
			if (size < m_curPos+m_curBlockSize) {
				return;
			}
			if (m_data[m_curPos+1] != 0x00 || m_data[m_curPos+2] != 0x00) {
				if (m_delayState == DST_DELAY_NOT_ENCOUNTERED) {
					m_delayState = DST_WAITING_FOR_GRAPHIC_BLOCK;
				}
			}
			m_curPos += m_curBlockSize;
			m_state = ST_GRAPHIC_CONTROL_EXTENSION_SUBBLOCK_OR_TERMINATOR;
			goto case_ST_GRAPHIC_CONTROL_EXTENSION_SUBBLOCK_OR_TERMINATOR;
		}
		case ST_OTHER_EXTENSION_SUBBLOCK_OR_TERMINATOR:
		case_ST_OTHER_EXTENSION_SUBBLOCK_OR_TERMINATOR: {
			if (size <= m_curPos) {
				return;
			}
			m_curBlockSize = m_data[m_curPos++];
			if (m_curBlockSize == 0) {
				m_state = ST_BLOCK_BEGIN;
				goto case_ST_BLOCK_BEGIN;
			} else {
				m_state = ST_OTHER_EXTENSION_SUBBLOCK;
			}
		}
		case ST_OTHER_EXTENSION_SUBBLOCK: {
			if (size < m_curPos+m_curBlockSize) {
				return;
			}
			m_curPos += m_curBlockSize;
			m_state = ST_OTHER_EXTENSION_SUBBLOCK_OR_TERMINATOR;
			goto case_ST_OTHER_EXTENSION_SUBBLOCK_OR_TERMINATOR;
		}
	}
}

void
GIFImageInspector::consumeEOF()
{
	if (m_isInProgress) {
		m_isInProgress = false;
		m_isFailed = true;
	}
}

void
GIFImageInspector::reset()
{
	m_state = ST_SIGNATURE;
	m_delayState = DST_DELAY_NOT_ENCOUNTERED;
	std::vector<uint8_t>().swap(m_data);
	m_isInProgress = true;
	m_isFailed = false;
	m_width = 0;
	m_height = 0;
	m_isAnimated = false;
	m_curPos = 0;
}

std::pair<const uint8_t*, const uint8_t*>
GIFImageInspector::getAccumulatedData() const
{
	return std::pair<const uint8_t*, const uint8_t*>(&m_data[0], &m_data[m_data.size()]);
}

