/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "application.h"
#include "masterstate.h"
#include "statusdocklet.h"
#include "log.h"
#include "logdialog.h"
#include <glibmm.h>
#include <gtkmm/messagedialog.h>
#include <string>
#include <cstring>
#include <sstream>
#include <fstream>
#include <cstdio>
#include <list>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>

#define NLISTEN 5

Application* Application::m_spInstance = 0;

Application::Application(int argc, char* argv[])
:	m_gtkMain(argc, argv), m_serverSocket(-1)
{
	m_spInstance = this;
	m_confDir = Glib::build_filename(Glib::get_home_dir(), ".bfilter");
	m_backendExecutable = Glib::build_filename(BINDIR, "bfilter");
}

Application::~Application()
{
	std::set<BackendLink*>::iterator iter = m_backends.begin();
	for (; iter != m_backends.end(); ++iter) {
		kill((*iter)->getPid(), SIGUSR1);
		delete *iter;
	}
}

void
Application::run()
{
	m_trayMenu.init();
	if (!checkBackend()) {
		return;
	}
	try {
		if (!checkConfDir()) {
			throw true;
		}
		if (!readConfig()) {
			throw true;
		}
		readConfigDefault();
		if (!readGlobalRules()) {
			throw true;
		}
		if (!readLocalRules()) {
			throw true;
		}
		if (!processConfig(m_config, m_configIO)) {
			throw true;
		}
		if (!processRules(m_globalRules, m_localRules, m_rulesReader)) {
			throw true;
		}
		m_state.setConfig(m_configIO.getConfig());
		m_state.setRules(m_rulesReader.getRules());
		std::string bind_addr = m_state.getConfig().getListenAddr();
		int bind_port = m_state.getConfig().getListenPort();
		m_serverSocket = bindTo(bind_addr, bind_port);
		if (m_serverSocket < 0) {
			throw true;
		}
		m_pServerSocketConnection = Glib::signal_io().connect(
			SigC::slot(*this, &Application::onConnect),
			m_serverSocket, Glib::IO_IN
		);
	} catch (bool e) {
		m_statusDocklet.setOperational(false);
		showLogDialog();
	}
	m_gtkMain.run();
}

void
Application::showTrayMenu(guint button, guint32 activate_time)
{
	m_trayMenu.popup(button, activate_time);
}

void
Application::showLogDialog()
{
	m_logDialog.present();
}

void
Application::showBasicConfigDialog()
{
	if (!m_basicConfigDialog.is_visible()) {
		m_basicConfigDialog.loadConfig(m_state.getConfig());
	}
	m_basicConfigDialog.present();
}

void
Application::showAdvancedConfigWindow()
{
	if (!m_advConfigWindow.is_visible()) {
		m_advConfigWindow.loadConfig(m_config);
		m_advConfigWindow.loadConfigDefault(m_configDefault);
		m_advConfigWindow.loadRules(m_globalRules);
		m_advConfigWindow.loadRulesLocal(m_localRules);
		m_advConfigWindow.setNotModified();
	}
	m_advConfigWindow.present();
}

#ifdef DEBUG
void
Application::showDebugWindow()
{
	m_debugWindow.present();
}
#endif

void
Application::quit()
{
	m_gtkMain.quit();
}

void
Application::updateFilteringState()
{
	std::set<BackendLink*>::iterator iter = m_backends.begin();
	for (; iter != m_backends.end(); ++iter) {
		(*iter)->updateFilteringState();
	}
}

bool
Application::applyNewConfig(const Config& config)
{
	Log::getInstance().appendMessage("Applying the new config ...");
	if (config.getListenPort() < 1 || config.getListenPort() > 65535) {
		Log::getInstance().appendError("Listen port is out of range [1, 65535]");
		return false;
	}
	if (config.getUseProxy() && (config.getProxyPort() < 1 || config.getProxyPort() > 65535)) {
		Log::getInstance().appendError("Proxy port is out of range [1, 65535]");
		return false;
	}
	bool was_bound = (m_serverSocket >= 0);
	if (was_bound && m_configIO.getConfig().getListenAddr() == config.getListenAddr()
	    && m_configIO.getConfig().getListenPort() == config.getListenPort()) {
		Log::getInstance().appendMessage("Rebinding is not necessary");
	} else {
		if (was_bound) {
			m_pServerSocketConnection.disconnect();
			close(m_serverSocket);
			m_serverSocket = -1;
			std::string addr = m_configIO.getConfig().getListenAddr();
			std::ostringstream strm;
			strm << m_configIO.getConfig().getListenPort();
			std::string port_str = strm.str();
			Log::getInstance().appendMessage(Glib::ustring("Unbinding from ")+addr+":"+port_str+" ... [OK]");
		}
		m_serverSocket = bindTo(config.getListenAddr(), config.getListenPort());
		if (m_serverSocket >= 0) {
			m_pServerSocketConnection = Glib::signal_io().connect(
				SigC::slot(*this, &Application::onConnect),
				m_serverSocket, Glib::IO_IN
			);
			if (!was_bound) {
				m_statusDocklet.setOperational(true);
			}
		} else {
			if (was_bound) {
				Log::getInstance().appendMessage("Applying the new config failed, rollbacking changes ...");
				m_serverSocket = bindTo(m_configIO.getConfig().getListenAddr(),
					m_configIO.getConfig().getListenPort());
				if (m_serverSocket >= 0) {
					m_pServerSocketConnection = Glib::signal_io().connect(
						SigC::slot(*this, &Application::onConnect),
						m_serverSocket, Glib::IO_IN
					);
				} else {
					m_statusDocklet.setOperational(false);
				}
			}
			return false;
		}
	}
	m_configIO.setConfig(config);
	m_state.setConfig(config);
	std::ostringstream strm;
	m_configIO.writeConfig(strm);
	m_config = strm.str();
	for (std::set<BackendLink*>::iterator iter = m_backends.begin(); iter != m_backends.end(); ++iter) {
		(*iter)->updateConfig();
	}
	return true;
}

bool
Application::applyNewConfigAndRules(const ConfigIO& cio, const Rules& rules,
	const std::string& config_source, const std::string& rules_source, const std::string& rules_local_source)
{
	if (!applyNewConfig(cio.getConfig()))
		return false;
	m_configIO = cio;
	m_rulesReader.setRules(rules);
	m_state.setRules(rules);
	Log::getInstance().appendMessage("Applying the new rules ... [OK]");
	m_config = config_source;
	m_globalRules = rules_source;
	m_localRules = rules_local_source;
	for (std::set<BackendLink*>::iterator iter = m_backends.begin(); iter != m_backends.end(); ++iter) {
		(*iter)->updateRules();
	}
	return true;
}

bool
Application::writeConfig()
{
	Log::getInstance().appendMessage("Writing config ... ");
	std::string fn = Glib::build_filename(m_confDir, "config");
	std::ofstream fstrm(fn.c_str());
	if (!fstrm.good()) {
		Log::getInstance().appendFailed();
		Log::getInstance().appendError("Could not open "+fn+" for writing");
		return false;
	}
	fstrm.write(m_config.data(), m_config.length());
	fstrm << std::flush;
	if (fstrm.bad()) {
		Log::getInstance().appendFailed();
		Log::getInstance().appendError("Error writing to "+fn);
		return false;
	}
	Log::getInstance().appendOK();
	return true;
}

bool
Application::writeLocalRules()
{
	Log::getInstance().appendMessage("Writing local rules ... ");
	std::string fn = Glib::build_filename(m_confDir, "rules.local");
	std::ofstream fstrm(fn.c_str());
	if (!fstrm.good()) {
		Log::getInstance().appendFailed();
		Log::getInstance().appendError("Could not open "+fn+" for writing");
		return false;
	}
	fstrm.write(m_localRules.data(), m_localRules.length());
	fstrm << std::flush;
	if (fstrm.bad()) {
		Log::getInstance().appendFailed();
		Log::getInstance().appendError("Error writing to "+fn);
		return false;
	}
	Log::getInstance().appendOK();
	return true;
}

bool
Application::onConnect(Glib::IOCondition evt)
{
	struct sockaddr_in client;
	socklen_t addrlen = sizeof(struct sockaddr_in);
	int csock = accept(m_serverSocket, (struct sockaddr *)&client, &addrlen);
	if (csock < 0) {
		return true;
	}
	std::list<std::string> argv;
	argv.push_back(m_backendExecutable);
	argv.push_back("--backend");
	try {
		int stdin = -1;
		int stdout = -1;
		int pid = 0;
		Glib::spawn_async_with_pipes(Glib::get_current_dir(), argv, Glib::SPAWN_DO_NOT_REAP_CHILD,
			SigC::bind(SigC::slot(*this, &Application::prepareToLaunchBackend),
			csock), &pid, &stdin, &stdout, 0);
		addBackend(pid, stdout, stdin);
	} catch (Glib::SpawnError& e) {
		Log::getInstance().appendWarning(Glib::ustring("Could not spawn the backend: ")+e.what());
	}
	close(csock);
	return true;
}

void
Application::addBackend(pid_t pid, int rfd, int wfd)
{
	m_backends.insert(new BackendLink(this, pid, rfd, wfd));
}

void
Application::removeBackend(BackendLink* backend)
{
	m_backends.erase(backend);
	delete backend;
}

void
Application::prepareToLaunchBackend(int csock)
{
	close(3);
	dup2(csock, 3);
}

bool
Application::checkBackend()
{
	Log::getInstance().appendMessage("Checking backend version ... ");
	std::string stdout;
	try {
		std::list<std::string> argv;
		argv.push_back(m_backendExecutable);
		argv.push_back("--version");
		Glib::spawn_sync(Glib::get_current_dir(), argv, Glib::SpawnFlags(0),
			SigC::Slot0<void>(), &stdout, 0, 0);
	} catch (Glib::SpawnError& e) {
		Log::getInstance().appendFailed();
		Gtk::MessageDialog d(Glib::ustring("Could not launch the backend:\n")+e.what(),
			Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
		d.set_title("BFilter Error");
		d.run();
		return false;
	}
	std::istringstream strm(stdout);
	std::string version;
	strm >> version;
	if (version.empty() || version.length() > 40) {
		Log::getInstance().appendFailed();
		Gtk::MessageDialog d(Glib::ustring("Could not get the backend version.\nIs the backend too old?"),
			Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
		d.set_title("BFilter Error");
		d.run();
		return false;
	}
	if (version != VERSION) {
		Log::getInstance().appendFailed();
		Gtk::MessageDialog d(Glib::ustring("Wrong backend version\nExpected: ")+VERSION+"\nReported: "+version,
			Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
		d.set_title("BFilter Error");
		d.run();
		return false;
	}
	Log::getInstance().appendMessage(version, false);
	return true;
}

bool
Application::checkConfDir()
{
	DIR *dir = opendir(m_confDir.c_str());
	if (dir) {
		closedir(dir);
	} else {
		if (errno == ENOENT) {
			Log::getInstance().appendMessage("Creating $HOME/.bfilter ... ");
			if (mkdir(m_confDir.c_str(), 0700) < 0) {
				Log::getInstance().appendError(
					std::string("Could not create $HOME/.bfilter: ")+strerror(errno));
				return false;
			}
			Log::getInstance().appendOK();
		} else {
			Log::getInstance().appendError(
				std::string("Could not open $HOME./bfilter: ")+strerror(errno));
			return false;
		}
	}
	return true;
}

bool
Application::readConfig()
{
	Log::getInstance().appendMessage("Reading config ... ");
	std::string fname = Glib::build_filename(m_confDir, "config");
	FILE *fp = fopen(fname.c_str(), "r");
	if (!fp) {
		if (errno != ENOENT) {
			Log::getInstance().appendFailed();
			Log::getInstance().appendError(fname+": "+strerror(errno));
			return false;
		} else {
			Log::getInstance().appendMessage("[FILE NOT FOUND]", false);
			Log::getInstance().appendMessage("Copying the system-wide config file ... ");
			std::string gconf_fname = Glib::build_filename(CONFDIR, "config");
			fp = fopen(gconf_fname.c_str(), "r");
			if (!fp) {
				Log::getInstance().appendFailed();
				Log::getInstance().appendError(gconf_fname+": "+strerror(errno));
				return false;
			}
			m_config.erase();
			char buf[4096];
			while (true) {
				size_t nread = fread(buf, 1, sizeof(buf), fp);
				if (nread <= 0)
					break;
				m_config.append(buf, buf+nread);
			}
			fclose(fp);
			fp = fopen(fname.c_str(), "w");
			if (!fp) {
				Log::getInstance().appendFailed();
				Log::getInstance().appendError(fname+": "+strerror(errno));
				return false;
			}
			fwrite(m_config.data(), 1, m_config.length(), fp);
			fclose(fp);
			Log::getInstance().appendOK();
		}
	} else {
		m_config.erase();
		char buf[4096];
		while (true) {
			size_t nread = fread(buf, 1, sizeof(buf), fp);
			if (nread <= 0)
				break;
			m_config.append(buf, buf+nread);
		}
		fclose(fp);
		Log::getInstance().appendOK();
	}
	return true;
}

bool
Application::readConfigDefault()
{
	Log::getInstance().appendMessage("Reading default config ... ");
	std::string fname = Glib::build_filename(CONFDIR, "config.default");
	FILE *fp = fopen(fname.c_str(), "r");
	if (!fp) {
		Log::getInstance().appendFailed();
		Log::getInstance().appendError(fname+" :"+strerror(errno));
		return false;
	}
	m_configDefault.erase();
	char buf[4096];
	while (true) {
		size_t nread = fread(buf, 1, sizeof(buf), fp);
		if (nread <= 0)
			break;
		m_configDefault.append(buf, buf+nread);
	}
	fclose(fp);
	Log::getInstance().appendOK();
	return true;
}

bool
Application::readGlobalRules()
{
	Log::getInstance().appendMessage("Reading global rules ... ");
	std::string fname = Glib::build_filename(CONFDIR, "rules");
	FILE *fp = fopen(fname.c_str(), "r");
	if (!fp) {
		Log::getInstance().appendFailed();
		Log::getInstance().appendError(fname+": "+strerror(errno));
		return false;
	}
	m_globalRules.erase();
	char buf[4096];
	while (true) {
		size_t nread = fread(buf, 1, sizeof(buf), fp);
		if (nread <= 0)
			break;
		m_globalRules.append(buf, buf+nread);
	}
	fclose(fp);
	Log::getInstance().appendOK();
	return true;
}

bool
Application::readLocalRules()
{
	Log::getInstance().appendMessage("Reading local rules ... ");
	std::string fname = Glib::build_filename(m_confDir, "rules.local");
	FILE *fp = fopen(fname.c_str(), "r");
	if (!fp) {
		if (errno != ENOENT) {
			Log::getInstance().appendFailed();
			Log::getInstance().appendError(fname+": "+strerror(errno));
			return false;
		} else {
			Log::getInstance().appendMessage("[FILE NOT FOUND]", false);
			Log::getInstance().appendMessage("Copying the system-wide local rules file ... ");
			std::string gconf_fname = Glib::build_filename(CONFDIR, "rules.local");
			fp = fopen(gconf_fname.c_str(), "r");
			if (!fp) {
				Log::getInstance().appendFailed();
				Log::getInstance().appendError(gconf_fname+": "+strerror(errno));
				return false;
			}
			m_localRules.erase();
			char buf[4096];
			while (true) {
				size_t nread = fread(buf, 1, sizeof(buf), fp);
				if (nread <= 0)
					break;
				m_localRules.append(buf, buf+nread);
			}
			fclose(fp);
			fp = fopen(fname.c_str(), "w");
			if (!fp) {
				Log::getInstance().appendFailed();
				Log::getInstance().appendError(fname+": "+strerror(errno));
				return false;
			}
			fwrite(m_localRules.data(), 1, m_localRules.length(), fp);
			fclose(fp);
			Log::getInstance().appendOK();
		}
	} else {
		m_localRules.erase();
		char buf[4096];
		while (true) {
			size_t nread = fread(buf, 1, sizeof(buf), fp);
			if (nread <= 0)
				break;
			m_localRules.append(buf, buf+nread);
		}
		fclose(fp);
		Log::getInstance().appendOK();
	}
	return true;
}

bool
Application::processConfig(const std::string& config, ConfigIO& cio)
{
	Log::getInstance().appendMessage("Processing config ... ");
	std::istringstream strm(config);
	ConfigIO::Status status = cio.readConfig(strm);
	if (status.status == ConfigIO::Status::OK) {
		Log::getInstance().appendOK();
	} else {
		Log::getInstance().appendFailed();
	}
	for (ConfigIO::ErrorList::const_iterator iter = status.errors.begin();
	     iter != status.errors.end(); ++iter) {
		std::ostringstream msg;
		msg << "[config";
		if (iter->getLine()) {
			msg << ':' << iter->getLine();
			if (iter->getCol())
				msg << ':' << iter->getCol();
		}
		msg << "] " << iter->getMessage();
		if (iter->getType() == ConfIO::Error::T_ERROR) {
			Log::getInstance().appendError(msg.str());
		} else {
			Log::getInstance().appendWarning(msg.str());
		}
	}
	return (status.status != ConfigIO::Status::ERRORS);
}

bool
Application::processRules(const std::string& global_rules, const std::string& local_rules, RulesReader& rr)
{
	Log::getInstance().appendMessage("Processing rules ... ");
	std::istringstream strm_global(global_rules);
	std::istringstream strm_local(local_rules);
	RulesReader::Status status = rr.readRules(strm_global, strm_local);
	if (status.status == RulesReader::Status::OK) {
		Log::getInstance().appendOK();
	} else {
		Log::getInstance().appendFailed();
	}
	for (ConfIO::ErrorList::const_iterator iter = status.errors_global.begin();
	     iter != status.errors_global.end(); ++iter) {
		std::ostringstream msg;
		msg << "[rules";
		if (iter->getLine()) {
			msg << ':' << iter->getLine();
			if (iter->getCol())
				msg << ':' << iter->getCol();
		}
		msg << "] " << iter->getMessage();
		if (iter->getType() == ConfIO::Error::T_ERROR) {
			Log::getInstance().appendError(msg.str());
		} else {
			Log::getInstance().appendWarning(msg.str());
		}
	}
	for (ConfIO::ErrorList::const_iterator iter = status.errors_local.begin();
	     iter != status.errors_local.end(); ++iter) {
		std::ostringstream msg;
		msg << "[rules.local";
		if (iter->getLine()) {
			msg << ':' << iter->getLine();
			if (iter->getCol())
				msg << ':' << iter->getCol();
		}
		msg << "] " << iter->getMessage();
		if (iter->getType() == ConfIO::Error::T_ERROR) {
			Log::getInstance().appendError(msg.str());
		} else {
			Log::getInstance().appendWarning(msg.str());
		}
	}
	return (status.status != RulesReader::Status::ERRORS);
}

static bool resolveAddress(const char* host, in_addr* ia)
{
	if (inet_aton(host, ia)) {
		return true;
	}
	struct hostent *hp = gethostbyname(host);
	if (!hp) {
		return false;
	}
	memcpy(ia, hp->h_addr, sizeof(in_addr));
	return true;
}

int
Application::bindTo(const std::string& addr, int port)
{
	bool warnings = false;
	std::ostringstream strm;
	strm << port;
	std::string port_str = strm.str();
	Log::getInstance().appendMessage(Glib::ustring("Binding to ")+addr+":"+port_str+" ... ");
	struct sockaddr_in local;
	int on = 1;
	memset(&local, 0, sizeof(local));
	if (addr.empty()) {
		local.sin_addr.s_addr = htonl(INADDR_ANY);
	} else {
		if (!resolveAddress(addr.c_str(), &local.sin_addr)) {
			Log::getInstance().appendFailed();
			Log::getInstance().appendError("Could not resolve hostname "+addr);
			return -1;
		}
	}
	local.sin_port = htons(port);
	local.sin_family = AF_INET;
	
	int sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		Log::getInstance().appendFailed();
		Log::getInstance().appendError("socket() failed");
		return -1;
	}
	
	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on))) {
		warnings = true;
		Log::getInstance().appendWarning(Glib::ustring("setsockopt(SO_REUSEADDR) failed: ")+strerror(errno));
	}
	
	if (bind(sock, (struct sockaddr *)&local, sizeof(local))) {
		if (!warnings) {
			Log::getInstance().appendFailed();
		}
		Log::getInstance().appendError(Glib::ustring("bind() failed: ")+strerror(errno));
		return -1;
	}
	
	if (listen(sock, NLISTEN)) {
		if (!warnings) {
			Log::getInstance().appendFailed();
		}
		Log::getInstance().appendError(Glib::ustring("listen() failed: ")+strerror(errno));
		return -1;
	}
	
	if (!warnings) {
		Log::getInstance().appendOK();
	}
	
	return sock;
}

Application::BackendLink::BackendLink(Application* app, pid_t pid, int rfd, int wfd)
:	m_pApp(app), m_pid(pid), m_portal(rfd, wfd), m_waitingForWrite(false)
{
	Glib::signal_io().connect(SigC::slot(*this, &BackendLink::onIncomingData), rfd, Glib::IO_IN);
	Glib::signal_io().connect(SigC::slot(*this, &BackendLink::onErrorOrHUP), rfd,
		Glib::IO_ERR|Glib::IO_HUP|Glib::IO_NVAL);
	Glib::signal_io().connect(SigC::slot(*this, &BackendLink::onErrorOrHUP), wfd,
		Glib::IO_ERR|Glib::IO_HUP|Glib::IO_NVAL);
}

Application::BackendLink::~BackendLink()
{
	close(m_portal.getReadFD());
	close(m_portal.getWriteFD());
}

bool
Application::BackendLink::onIncomingData(Glib::IOCondition evt)
{
	if (!m_portal.readChunk()) {
		terminate();
		return true;
	}
	handleBackendMessages();
	updateWriteExpectation();
	return true;
}

void
Application::BackendLink::updateFilteringState()
{
	IPCMessage msg("updateFilteringState");
	msg.appendParameter(m_pApp->getState().isFilteringEnabled() ? "1" : "0");
	msg.appendParameter(m_pApp->getState().isScriptFilteringEnabled() ? "1" : "0");
	m_portal.sendMessage(msg);
	updateWriteExpectation();
}

void
Application::BackendLink::updateConfig()
{
	IPCMessage msg("updateConfig");
	msg.appendParameter(m_pApp->getState().getConfig().serialize());
	m_portal.sendMessage(msg);
	updateWriteExpectation();
}

void
Application::BackendLink::updateRules()
{
	IPCMessage msg("updateRules");
	msg.appendParameter(m_pApp->getState().getRules().serialize());
	m_portal.sendMessage(msg);
	updateWriteExpectation();
}

bool
Application::BackendLink::onOutgoingData(Glib::IOCondition evt)
{
	if (!m_portal.writeChunk()) {
		terminate();
		return true;
	}
	m_waitingForWrite = m_portal.hasDataToSend();
	return m_waitingForWrite;
}

bool
Application::BackendLink::onErrorOrHUP(Glib::IOCondition evt)
{
	terminate();
	return true;
}

void
Application::BackendLink::terminate()
{
	m_pApp->removeBackend(this);
}

void
Application::BackendLink::handleBackendMessages()
{
	for (const IPCMessage* req = 0; (req = m_portal.firstPendingRequest());) {
		if (req->getMethod() == "getState") {
			IPCMessage resp("returnState");
			resp.appendParameter(m_pApp->getState().getConfig().serialize());
			resp.appendParameter(m_pApp->getState().getRules().serialize());
			resp.appendParameter(m_pApp->getState().isFilteringEnabled() ? "1" : "0");
			resp.appendParameter(m_pApp->getState().isScriptFilteringEnabled() ? "1" : "0");
			m_portal.processFirstPendingRequest(resp);
#ifdef DEBUG
		} else if (req->getMethod() == "newConnection") {
			int conn_id = m_pApp->m_debugWindow.newConnection();
			IPCMessage resp("returnConnectionId");
			resp.appendParameter(conn_id);
			m_portal.processFirstPendingRequest(resp);
#endif
		} else {
			Log::getInstance().appendWarning("Unhandled request from backend: "+req->getMethod());
			m_portal.processFirstPendingRequest(IPCMessage("failure"));
		}
	}
	for (const IPCMessage* msg = 0; (msg = m_portal.firstPendingMessage()); m_portal.dropFirstPendingMessage()) {
#ifndef DEBUG
		if (false) {
#else
		if (msg->getMethod() == "connectionClosed") {
			m_pApp->m_debugWindow.connectionClosed(msg->getParameterAsInt(0));
		} else if (msg->getMethod() == "logMessageStarts") {
			int conn_id = msg->getParameterAsInt(0);
			std::string content = msg->getParameter(1);
			std::string type = msg->getParameter(2);
			std::string subtype = msg->getParameter(3);
			
			if (type == "INCOMING_REQUEST") {
				m_pApp->m_debugWindow.messageStarts(conn_id,
					content, DebugMessage::INCOMING_REQUEST);
			} else if (type == "OUTGOING_REQUEST") {
				m_pApp->m_debugWindow.messageStarts(conn_id,
					content, DebugMessage::OUTGOING_REQUEST);
			} else if (type == "INCOMING_RESPONSE") {
				m_pApp->m_debugWindow.messageStarts(conn_id,
					content, DebugMessage::INCOMING_RESPONSE);
			} else if (type == "OUTGOING_RESPONSE") {
				DebugMessage::Type type = DebugMessage::OUTGOING_RESPONSE;
				if (subtype == "ORES_NORMAL") {
					m_pApp->m_debugWindow.messageStarts(conn_id,
						content, type, DebugMessage::ORES_NORMAL);
				} else if (subtype == "ORES_FILTERED") {
					m_pApp->m_debugWindow.messageStarts(conn_id,
						content, type, DebugMessage::ORES_FILTERED);
				} else if (subtype == "ORES_GENERATED") {
					m_pApp->m_debugWindow.messageStarts(conn_id,
						content, type, DebugMessage::ORES_GENERATED);
				} else if (subtype == "ORES_ERROR") {
					m_pApp->m_debugWindow.messageStarts(conn_id,
						content, type, DebugMessage::ORES_ERROR);
				} else {
					Log::getInstance().appendWarning("Unhandled message subtype: "+subtype);
				}
			} else {
				Log::getInstance().appendWarning("Unhandled message type: "+type);
			}
		} else if (msg->getMethod() == "logMessageFinishes") {
			int conn_id = msg->getParameterAsInt(0);
			std::string type = msg->getParameter(1);
			if (type == "INCOMING_REQUEST") {
				m_pApp->m_debugWindow.messageFinishes(conn_id,
					DebugMessage::INCOMING_REQUEST);
			} else if (type == "OUTGOING_REQUEST") {
				m_pApp->m_debugWindow.messageFinishes(conn_id,
					DebugMessage::OUTGOING_REQUEST);
			} else if (type == "INCOMING_RESPONSE") {
				m_pApp->m_debugWindow.messageFinishes(conn_id,
					DebugMessage::INCOMING_RESPONSE);
			} else if (type == "OUTGOING_RESPONSE") {
				m_pApp->m_debugWindow.messageFinishes(conn_id,
					DebugMessage::OUTGOING_RESPONSE);
			} else {
				Log::getInstance().appendWarning("Unhandled message type: "+type);
			}
		} else if (msg->getMethod() == "logDebugMessage") {
			int conn_id = msg->getParameterAsInt(0);
			m_pApp->m_debugWindow.debugMessage(conn_id, msg->getParameter(1));
		} else if (msg->getMethod() == "logTraffic") {
			int conn_id = msg->getParameterAsInt(0);
			std::string type = msg->getParameter(1);
			std::string data = msg->getParameter(2);
			if (type == "FROM_CLIENT") {
				m_pApp->m_debugWindow.logTraffic(conn_id, DebugConnection::TRAF_FROM_CLIENT, data);
			} else if (type == "TO_CLIENT") {
				m_pApp->m_debugWindow.logTraffic(conn_id, DebugConnection::TRAF_TO_CLIENT, data);
			} else if (type == "FROM_SERVER") {
				m_pApp->m_debugWindow.logTraffic(conn_id, DebugConnection::TRAF_FROM_SERVER, data);
			} else if (type == "TO_SERVER") {
				m_pApp->m_debugWindow.logTraffic(conn_id, DebugConnection::TRAF_TO_SERVER, data);
			}
#endif
		} else {
			Log::getInstance().appendWarning("Unhandled message from backend: "+msg->getMethod());
		}
	}
}

void
Application::BackendLink::updateWriteExpectation()
{
	if (!m_waitingForWrite && m_portal.hasDataToSend()) {
		m_waitingForWrite = true;
		Glib::signal_io().connect(SigC::slot(*this, &BackendLink::onOutgoingData),
			m_portal.getWriteFD(), Glib::IO_OUT);
	}
}

