/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "debugconnection.h"
#include "debuglayout.h"
#include "debugmessage.h"
#include <sstream>
#include <fstream>
#include <list>
#include <gtkmm/drawingarea.h>
#include <gtkmm/notebook.h>
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/textview.h>
#include <gtkmm/frame.h>
#include <gtkmm/box.h>
#include <gtkmm/buttonbox.h>
#include <gtkmm/alignment.h>
#include <gtkmm/messagedialog.h>
#include <glibmm.h>

DebugConnection::DebugConnection(int id, DebugLayout* layout, int top)
:	m_id(id),
	m_pLayout(layout),
	m_top(top),
	m_height(DebugLayout::CONN_BORDER+DebugLayout::CONN_PADDING),
	m_closed(false),
	m_pLastIncomingRequest(0),
	m_pLastOutgoingRequest(0),
	m_pLastIncomingResponse(0),
	m_pLastOutgoingResponse(0),
	m_pContainer(manage(new Gtk::Fixed)),
	m_pViewWindow(0)
{
	for (int i = 0; i < 4; i++) {
		m_totalTrafficSize[i] = 0;
		m_storedTrafficSize[i] = 0;
	}
	modify_bg_pixmap(Gtk::STATE_NORMAL, "<none>");
	m_pContainer->show();
	add(*m_pContainer);
	m_pContainer->set_size_request(DebugLayout::CONN_WIDTH, m_height);
	Gdk::Color border_color;
	border_color.set("#ac40ef");
	for (int i = 0; i < 4; i++) {
		m_pBorders[i] = manage(new Gtk::DrawingArea);
		m_pBorders[i]->modify_bg_pixmap(Gtk::STATE_NORMAL, "<none>");
		m_pBorders[i]->modify_bg(Gtk::STATE_NORMAL, border_color);
		if (i != 3) {
			m_pBorders[i]->show();
		}
	}
	m_pBorders[0]->set_size_request(DebugLayout::CONN_WIDTH, DebugLayout::CONN_BORDER);
	m_pBorders[1]->set_size_request(DebugLayout::CONN_BORDER, m_height-DebugLayout::CONN_BORDER);
	m_pBorders[2]->set_size_request(DebugLayout::CONN_BORDER, m_height-DebugLayout::CONN_BORDER);
	m_pBorders[3]->set_size_request(DebugLayout::CONN_WIDTH, DebugLayout::CONN_BORDER);
	m_pContainer->put(*m_pBorders[0], 0, 0);
	m_pContainer->put(*m_pBorders[1], 0, DebugLayout::CONN_BORDER);
	m_pContainer->put(*m_pBorders[2], DebugLayout::CONN_WIDTH-DebugLayout::CONN_BORDER, DebugLayout::CONN_BORDER);
	m_pContainer->put(*m_pBorders[3], 0, m_height);
	
	signal_button_press_event().connect(SigC::slot(*this, &DebugConnection::onButtonPress));
	
	show();
}

DebugConnection::~DebugConnection()
{
	if (m_pViewWindow) {
		delete m_pViewWindow;
	}
	destroySignal();
}

void
DebugConnection::growBy(int increment)
{
	m_height += increment;
	m_pContainer->set_size_request(DebugLayout::CONN_WIDTH, m_height);
	m_pBorders[1]->set_size_request(DebugLayout::CONN_BORDER, m_height-DebugLayout::CONN_BORDER);
	m_pBorders[2]->set_size_request(DebugLayout::CONN_BORDER, m_height-DebugLayout::CONN_BORDER);
	m_pContainer->move(*m_pBorders[3], 0, m_height);
	if (m_pLastIncomingRequest && m_pLastIncomingRequest->isActive()) {
		m_pLastIncomingRequest->growBy(increment);
	}
	if (m_pLastOutgoingRequest && m_pLastOutgoingRequest->isActive()) {
		m_pLastOutgoingRequest->growBy(increment);
	}
	if (m_pLastIncomingResponse && m_pLastIncomingResponse->isActive()) {
		m_pLastIncomingResponse->growBy(increment);
	}
	if (m_pLastOutgoingResponse && m_pLastOutgoingResponse->isActive()) {
		m_pLastOutgoingResponse->growBy(increment);
	}
}

void
DebugConnection::markAsClosed()
{
	if (m_pLastIncomingRequest) {
		m_pLastIncomingRequest->finish();
	}
	if (m_pLastOutgoingRequest) {
		m_pLastOutgoingRequest->finish();
	}
	if (m_pLastIncomingResponse) {
		m_pLastIncomingResponse->finish();
	}
	if (m_pLastOutgoingResponse) {
		m_pLastOutgoingResponse->finish();
	}
	m_pLayout->growBy(DebugLayout::CONN_PADDING+DebugLayout::CONN_BORDER);
	m_pBorders[3]->set_size_request(DebugLayout::CONN_WIDTH-DebugLayout::CONN_BORDER*2, DebugLayout::CONN_BORDER);
	m_pContainer->move(*m_pBorders[3], DebugLayout::CONN_BORDER, m_height-DebugLayout::CONN_BORDER);
	m_pBorders[3]->show();
	m_closed = true;
}

void
DebugConnection::messageStarts(const std::string& content, DebugMessage::Type type, DebugMessage::Subtype subtype)
{
	DebugMessage** last_msg;
	switch (type) {
		case DebugMessage::INCOMING_REQUEST: {
			last_msg = &m_pLastIncomingRequest;
			break;
		}
		case DebugMessage::OUTGOING_REQUEST: {
			last_msg = &m_pLastOutgoingRequest;
			break;
		}
		case DebugMessage::INCOMING_RESPONSE: {
			last_msg = &m_pLastIncomingResponse;
			break;
		}
		default: { // OUTGOING_RESPONSE
			last_msg = &m_pLastOutgoingResponse;
		}
	}
	int left = 0;
	switch (type) {
		case DebugMessage::OUTGOING_RESPONSE:
			left += DebugLayout::MESSAGE_WIDTH+DebugLayout::MESSAGE_HOR_SPACING;
		case DebugMessage::INCOMING_RESPONSE:
			left += DebugLayout::MESSAGE_WIDTH+DebugLayout::MESSAGE_HOR_SPACING;
		case DebugMessage::OUTGOING_REQUEST:
			left += DebugLayout::MESSAGE_WIDTH+DebugLayout::MESSAGE_HOR_SPACING;
		case DebugMessage::INCOMING_REQUEST:
			left += DebugLayout::CONN_BORDER+DebugLayout::CONN_PADDING;
	};
	int top = getHeight();
	int grow_by = 0;
	if (*last_msg) {
		(*last_msg)->finish();
		int extra = (*last_msg)->getBottom()+DebugLayout::MESSAGE_VER_SPACING-getHeight();
		if (extra > 0) {
			grow_by = extra;
			top += extra;
		}
	}
	DebugMessage* msg = manage(new DebugMessage(top, content, type, subtype));
	grow_by += msg->getHeight();
	m_pLayout->growBy(grow_by);
	m_pContainer->put(*msg, left, top);
	*last_msg = msg;
}

void
DebugConnection::messageFinishes(DebugMessage::Type type)
{
	DebugMessage* last_msg;
	if (type == DebugMessage::INCOMING_REQUEST) {
		last_msg = m_pLastIncomingRequest;
	} else if (type == DebugMessage::OUTGOING_REQUEST) {
		last_msg = m_pLastOutgoingRequest;
	} else if (type == DebugMessage::INCOMING_RESPONSE) {
		last_msg = m_pLastIncomingResponse;
	} else {
		last_msg = m_pLastOutgoingResponse;
	}
	if (!last_msg) {
		return;
	}
	
	DebugMessage* messages[] = {
		m_pLastIncomingRequest, m_pLastOutgoingRequest,
		m_pLastIncomingResponse, m_pLastOutgoingResponse
	};
	for (int i = 0; i < 4; i++) {
		DebugMessage* msg = messages[i];
		if (msg && msg->isActive() && msg->getBottom() == getHeight()) {
			m_pLayout->growBy(DebugLayout::MESSAGE_VER_SPACING);
			break;
		}
	}
	
	last_msg->markAsClosed();
}

void
DebugConnection::debugMessage(const std::string& content)
{
	m_debug += content;
	m_debug += "\n";
	if (m_pViewWindow) {
		m_pViewWindow->appendDebugMessage(content);
	}
}

void
DebugConnection::logTraffic(TrafficType type, const std::string& content)
{
	if (content.empty())
		return;
	
	
	std::deque<std::string>& traffic = m_traffic[type];
	std::size_t traf_limit = m_pLayout->getTrafficLimit();
	if (content.length() > traf_limit) {
		traffic.clear();
		traffic.push_back(content.substr(content.length() - traf_limit));
		m_storedTrafficSize[type] = traf_limit;
	} else {
		if (m_storedTrafficSize[type] + content.length() <= traf_limit) {
			traffic.push_back(content);
			m_storedTrafficSize[type] += content.length();
		} else {
			std::size_t todo = m_storedTrafficSize[type] + content.length() - traf_limit;
			while (todo > 0 && !traffic.empty()) {
				if (traffic.front().length() > todo) {
					traffic.front().erase(0, todo);
					todo = 0;
					break;
				} else {
					todo -= traffic.front().length();
					traffic.pop_front();
				}
			}
			traffic.push_back(content);
			m_storedTrafficSize[type] = traf_limit;
		}
	}
	m_totalTrafficSize[type] += content.length();
	if (m_pViewWindow) {
		m_pViewWindow->trafficUpdateAvailable(type);
	}
}

bool
DebugConnection::onButtonPress(GdkEventButton* event)
{
	if (event->type == GDK_2BUTTON_PRESS) {
		if (m_pViewWindow) {
			m_pViewWindow->present();
		} else {
			m_pViewWindow = new ViewWindow(this);
			m_pViewWindow->signal_hide().connect(SigC::slot(*this, &DebugConnection::destroyViewWindow));
		}
	}
	return true;
}

void
DebugConnection::destroyViewWindow()
{
	delete m_pViewWindow;
	m_pViewWindow = 0;
}

void
DebugConnection::destroySaveTrafficDialog(SaveTrafficDialog* dlg)
{
	delete dlg;
}

DebugConnection::ViewWindow::ViewWindow(DebugConnection* debug_conn)
:	m_pDebugConnection(debug_conn),
	m_pDebugBuffer(Gtk::TextBuffer::create()),
	m_encoding("ISO-8859-1")
{
	set_title("HTTP Connection");
	set_size_request(550, 350);
	
	std::list<Glib::ustring> encodings;
	encodings.push_back(m_encoding);
	encodings.push_back("KOI8-R");
	encodings.push_back("CP1251");
	// Can't include UTF-8 here because if the traffic contains
	// other encodings or binary data, the conversion will fail.
	
	Gtk::Notebook* notebook = manage(new Gtk::Notebook);
	add(*notebook);
	
	Gtk::ScrolledWindow* debug_scrolled_window = manage(new Gtk::ScrolledWindow);
	debug_scrolled_window->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	debug_scrolled_window->set_shadow_type(Gtk::SHADOW_IN);
	Gtk::TextView* debug_text_view = manage(new Gtk::TextView(m_pDebugBuffer));
	debug_text_view->set_editable(false);
	debug_text_view->set_cursor_visible(false);
	debug_scrolled_window->add(*debug_text_view);
	
	notebook->append_page(*debug_scrolled_window, "Debug");
	m_pDebugBuffer->insert(m_pDebugBuffer->end(), Glib::convert_with_fallback(m_pDebugConnection->m_debug, "utf-8", "iso-8859-1"));
	
	Glib::ustring titles[4];
	titles[TRAF_FROM_CLIENT] = "From Client";
	titles[TRAF_TO_CLIENT] = "To Client";
	titles[TRAF_FROM_SERVER] = "From Server";
	titles[TRAF_TO_SERVER] = "To Server";
	
	for (int i = 0; i < 4; i++) {
		Gtk::Frame* frame = manage(new Gtk::Frame);
		frame->set_shadow_type(Gtk::SHADOW_NONE);
		notebook->append_page(*frame, titles[i]);
		
		Gtk::VBox* vbox = manage(new Gtk::VBox);
		frame->add(*vbox);
		
		Gtk::HBox* hbox1 = manage(new Gtk::HBox(true));
		hbox1->set_border_width(3);
		vbox->pack_start(*hbox1, Gtk::PACK_SHRINK);
		
		Gtk::HBox* hbox2 = manage(new Gtk::HBox);
		hbox1->pack_start(*hbox2);
		
		hbox2->pack_start(*manage(new Gtk::Label("Traffic size: ")), Gtk::PACK_SHRINK);
		std::ostringstream strm;
		strm << m_pDebugConnection->m_totalTrafficSize[i];
		m_pTrafficSizeLabels[i] = manage(new Gtk::Label(strm.str()));
		hbox2->pack_start(*m_pTrafficSizeLabels[i], Gtk::PACK_SHRINK);
		
		m_pUpdateButtons[i] = manage(new Gtk::Button("Update"));
		m_pUpdateButtons[i]->set_sensitive(false);
		m_pUpdateButtons[i]->signal_clicked().connect(
			SigC::bind(SigC::slot(*this, &DebugConnection::ViewWindow::onTrafUpdate), (TrafficType)i));
		Gtk::Alignment* align1 = manage(new Gtk::Alignment(0.5, 0.5, 0.0, 0.0));
		hbox1->pack_start(*align1);
		Gtk::HButtonBox* bbox = manage(new Gtk::HButtonBox(Gtk::BUTTONBOX_DEFAULT_STYLE, 4));
		align1->add(*bbox);
		bbox->pack_start(*m_pUpdateButtons[i], Gtk::PACK_SHRINK);
		Gtk::Button* save_btn = manage(new Gtk::Button("Save As"));
		save_btn->signal_clicked().connect(
			SigC::bind(SigC::slot(*this, &DebugConnection::ViewWindow::onSaveAs), (TrafficType)i));
		bbox->pack_start(*save_btn, Gtk::PACK_SHRINK);
		
		Gtk::Alignment* align2 = manage(new Gtk::Alignment(1.0, 0.5, 0.0, 0.0));
		hbox1->pack_start(*align2);
		Gtk::Combo* encoding_combo = manage(new Gtk::Combo);
		encoding_combo->disable_activate();
		encoding_combo->set_value_in_list(false, false);
		encoding_combo->set_case_sensitive(false);
		encoding_combo->set_popdown_strings(encodings);
		encoding_combo->get_entry()->signal_activate().connect(
			SigC::bind(SigC::bind(SigC::slot(*this, &DebugConnection::ViewWindow::onChangeEncoding),
			encoding_combo), (TrafficType)i));
		encoding_combo->get_list()->signal_select_child().connect(
			SigC::hide<Gtk::Widget&>(
			SigC::bind(SigC::bind(SigC::slot(*this, &DebugConnection::ViewWindow::onChangeEncoding),
			encoding_combo), (TrafficType)i)));
		align2->add(*encoding_combo);
		
		m_pTrafficBuffers[i] = Gtk::TextBuffer::create();
		Glib::RefPtr<Gtk::TextBuffer::Tag> tag = m_pTrafficBuffers[i]->create_tag("skipped");
		tag->property_foreground() = "#0000ff";
		Gtk::TextView* text_view = manage(new Gtk::TextView(m_pTrafficBuffers[i]));
		text_view->set_editable(false);
		text_view->set_cursor_visible(false);
		Gtk::ScrolledWindow* scrolled_window = manage(new Gtk::ScrolledWindow);
		scrolled_window->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
		scrolled_window->set_shadow_type(Gtk::SHADOW_IN);
		scrolled_window->add(*text_view);
		vbox->pack_start(*scrolled_window);
		onTrafUpdate((TrafficType)i);
	}
	
	show_all();
}

void
DebugConnection::ViewWindow::appendDebugMessage(const std::string& msg)
{
	m_pDebugBuffer->insert(m_pDebugBuffer->end(), Glib::convert_with_fallback(msg, "utf-8", "iso-8859-1"));
	m_pDebugBuffer->insert(m_pDebugBuffer->end(), "\n");
}

void
DebugConnection::ViewWindow::trafficUpdateAvailable(TrafficType type)
{
	m_pUpdateButtons[type]->set_sensitive(true);
}

void
DebugConnection::ViewWindow::onTrafUpdate(TrafficType type)
{
	std::size_t total_traf_size = m_pDebugConnection->m_totalTrafficSize[type];
	std::size_t traf_limit = m_pDebugConnection->m_pLayout->getTrafficLimit();
	
	std::ostringstream traf_size_strm;
	traf_size_strm << total_traf_size;
	m_pTrafficSizeLabels[type]->set_text(traf_size_strm.str());
	
	if (total_traf_size <= traf_limit) {
		m_pTrafficBuffers[type]->set_text(trafficToUnicode(type));
	} else {
		m_pTrafficBuffers[type]->set_text("");
		std::ostringstream skipped_strm;
		skipped_strm << (total_traf_size-traf_limit);
		Glib::ustring skipped = Glib::ustring("<")+skipped_strm.str()+" bytes skipped>\n";
		m_pTrafficBuffers[type]->insert_with_tag(
			m_pTrafficBuffers[type]->end(), skipped, "skipped");
		m_pTrafficBuffers[type]->insert(
			m_pTrafficBuffers[type]->end(), trafficToUnicode(type));
	}
	
	m_pUpdateButtons[type]->set_sensitive(false);
}

void
DebugConnection::ViewWindow::onSaveAs(TrafficType type)
{
	SaveTrafficDialog* dlg = new SaveTrafficDialog(m_pDebugConnection, type);
	m_pDebugConnection->destroySignal.connect(
		SigC::bind(SigC::slot(&DebugConnection::destroySaveTrafficDialog), dlg));
	dlg->show();
}

void
DebugConnection::ViewWindow::onChangeEncoding(TrafficType type, Gtk::Combo* combo)
{
	std::string new_encoding = combo->get_entry()->get_text().c_str();
	if (new_encoding == m_encoding)
		return;
	m_encoding = new_encoding;
	onTrafUpdate(type);
}

Glib::ustring
DebugConnection::ViewWindow::trafficToUnicode(TrafficType type)
{
	static const struct CharConvertor
	{
		char ch[256];
		CharConvertor() {
			for (int i = 0; i < 32; i++) {
				ch[i] = '?';
			}
			for (int i = 32; i < 256; i++) {
				ch[i] = i;
			}
			ch[(unsigned char)'\n'] = '\n';
			ch[(unsigned char)'\r'] = '\r';
			ch[(unsigned char)'\t'] = '\t';
		}
	} convertor;
	
	std::string data;
	data.reserve(m_pDebugConnection->m_storedTrafficSize[type]);
	std::deque<std::string>::iterator iter = m_pDebugConnection->m_traffic[type].begin();
	for (; iter != m_pDebugConnection->m_traffic[type].end(); ++iter) {
		data += *iter;
	}
	for (int i = data.length() - 1; i >= 0; --i) {
		data[i] = convertor.ch[static_cast<unsigned char>(data[i])];
	}
	try {
		Glib::ustring res(Glib::convert_with_fallback(data, "utf-8", m_encoding));
		return res;
	} catch (Glib::ConvertError& e) {}
	return "";
}

DebugConnection::SaveTrafficDialog::SaveTrafficDialog(DebugConnection* debug_conn, TrafficType type)
:	m_pDebugConnection(debug_conn),
	m_trafType(type)
{
	set_title("Save Traffic");
	Gtk::Alignment* align = manage(new Gtk::Alignment(0.5, 0.5, 0.0, 0.0));
	align->set_border_width(2);
	get_vbox()->pack_start(*align, Gtk::PACK_SHRINK);
	m_pBlockBoundriesCB = manage(new Gtk::CheckButton("Include block boundries"));
	align->add(*m_pBlockBoundriesCB);
	
	signal_response().connect(SigC::slot(*this, &DebugConnection::SaveTrafficDialog::onResponse));
	
	show_all_children();
}

void
DebugConnection::SaveTrafficDialog::onResponse(int response)
{
	if (response == Gtk::RESPONSE_OK) {
		std::string fname = get_filename();
		std::ofstream strm(fname.c_str(), std::ios::binary);
		if (!strm.good()) {
			Gtk::MessageDialog dlg(*this,
				"Couldn't open file for writing",
				Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
			dlg.run();
			return;
		}
		bool eob_marks = m_pBlockBoundriesCB->get_active();
		std::deque<std::string>::iterator iter = m_pDebugConnection->m_traffic[m_trafType].begin();
		for (; iter != m_pDebugConnection->m_traffic[m_trafType].end(); ++iter) {
			strm << *iter;
			if (eob_marks) {
				strm << "<<EOB>>";
			}
		}
		strm << std::flush;
		if (strm.bad()) {
			Gtk::MessageDialog dlg(*this,
				"Error writing to file",
				Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
			dlg.run();
			return;
		}
	}
	delete this;
}

