/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "debuglayout.h"
#include "debugconnection.h"
#include <vector>
#include <algorithm>

DebugLayout::DebugLayout()
:	m_height(0), m_width(0),
	m_connectionsToTrack(10),
	m_trafficLimit(200000)
{
	modify_bg_pixmap(Gtk::STATE_NORMAL, "<none>");
	set_size(LAYOUT_PADDING*2, LAYOUT_PADDING*2);
}

DebugLayout::~DebugLayout()
{
	for (std::list<Column*>::iterator iter = m_columns.begin(); iter != m_columns.end(); ++iter) {
		delete *iter;
	}
}

void
DebugLayout::growBy(int increment)
{
	if (increment == 0)
		return;
	
	m_height += increment;
	set_size(m_width+LAYOUT_PADDING*2, m_height+LAYOUT_PADDING*2);
	for (std::list<Column*>::iterator iter = m_columns.begin(); iter != m_columns.end(); ++iter) {
		(*iter)->growBy(increment);
	}
}

int
DebugLayout::newConnection()
{
	removeOldConnections(getConnectionsToTrack()-1);
	Column* col = 0;
	for (std::list<Column*>::iterator iter = m_columns.begin(); iter != m_columns.end(); ++iter) {
		if (!(*iter)->isInUse()) {
			col = *iter;
			break;
		}
	}
	if (!col) {
		int left = 0;
		if (!m_columns.empty()) {
			left = m_columns.back()->getRight()+CONN_HOR_SPACING;
		}
		col = new Column(this, left);
		m_width = col->getRight();
		set_size(m_width+LAYOUT_PADDING*2, m_height+LAYOUT_PADDING*2);
		m_columns.push_back(col);
	}
	int id = 1;
	if (!m_connections.empty()) {
		id = m_connections.rbegin()->first+1;
	}
	DebugConnection* conn = col->createConnection(id);
	std::pair<std::map<int, DebugConnection*>::iterator, bool> pos =
		m_connections.insert(std::map<int, DebugConnection*>::value_type(id, conn));
	conn->destroySignal.connect(SigC::bind(
		SigC::slot(*this, &DebugLayout::removeConnection), pos.first));
	return id;
}

void
DebugLayout::connectionClosed(int conn_id)
{
	std::map<int, DebugConnection*>::iterator fit = m_connections.find(conn_id);
	if (fit == m_connections.end()) {
		return;
	}
	fit->second->markAsClosed();
}

void
DebugLayout::messageStarts(int conn_id, const std::string& content, DebugMessage::Type type, DebugMessage::Subtype subtype)
{
	std::map<int, DebugConnection*>::iterator fit = m_connections.find(conn_id);
	if (fit == m_connections.end()) {
		return;
	}
	fit->second->messageStarts(content, type, subtype);
}

void
DebugLayout::messageFinishes(int conn_id, DebugMessage::Type type)
{
	std::map<int, DebugConnection*>::iterator fit = m_connections.find(conn_id);
	if (fit == m_connections.end()) {
		return;
	}
	fit->second->messageFinishes(type);
}

void
DebugLayout::debugMessage(int conn_id, const std::string& content)
{
	std::map<int, DebugConnection*>::iterator fit = m_connections.find(conn_id);
	if (fit == m_connections.end()) {
		return;
	}
	fit->second->debugMessage(content);
}

void
DebugLayout::logTraffic(int conn_id, DebugConnection::TrafficType type, const std::string& content)
{
	std::map<int, DebugConnection*>::iterator fit = m_connections.find(conn_id);
	if (fit == m_connections.end()) {
		return;
	}
	fit->second->logTraffic(type, content);
}

struct ConnectionIsOlder
{
	bool operator()(DebugConnection* conn1, DebugConnection* conn2)
	{
		return (conn1->getBottom() < conn2->getBottom());
	}
};

void
DebugLayout::removeOldConnections(int num2keep)
{
	int num2remove = (int)m_connections.size()-num2keep;
	if (num2remove <= 0) {
		return;
	}
	std::vector<DebugConnection*> cons;
	cons.reserve(m_connections.size());
	for (std::map<int, DebugConnection*>::iterator iter = m_connections.begin();
	     iter != m_connections.end(); ++iter) {
		cons.push_back(iter->second);
	}
	std::stable_sort(cons.begin(), cons.end(), ConnectionIsOlder());
	for (int i = 0; i < num2remove; i++) {
		delete cons[i];
	}
	
	bool size_changed = false;
	while (!m_columns.empty() && m_columns.back()->isEmpty()) {
		delete m_columns.back();
		m_columns.pop_back();
		size_changed = true;
	}
	if (size_changed) {
		if (m_columns.empty()) {
			m_width = 0;
		} else {
			m_width = m_columns.back()->getRight();
		}
	}
	
	int lowest_top = -1;
	for (std::list<Column*>::iterator iter = m_columns.begin(); iter != m_columns.end(); ++iter) {
		int top = (*iter)->getTop();
		if (lowest_top == -1 || top < lowest_top) {
			lowest_top = top;
		}
	}
	if (lowest_top > 0) {
		for (std::list<Column*>::iterator iter = m_columns.begin();
		     iter != m_columns.end(); ++iter) {
			(*iter)->trimTopBy(lowest_top);
		}
		m_height -= lowest_top;
		size_changed = true;
	} else if (lowest_top == -1) {
		// zero columns
		m_height = 0;
		size_changed = true;
	}
	
	if (size_changed) {
		set_size(m_width+LAYOUT_PADDING*2, m_height+LAYOUT_PADDING*2);
	}
}

void
DebugLayout::removeConnection(std::map<int, DebugConnection*>::iterator iter)
{
	m_connections.erase(iter);
}

DebugLayout::Column::Column(DebugLayout* layout, int left)
:	m_pLayout(layout), m_left(left)
{
}

int
DebugLayout::Column::getTop() const
{
	if (m_connections.empty())
		return 0;
	return (*m_connections.begin())->getTop();
}

bool
DebugLayout::Column::isInUse() const
{
	if (m_connections.empty())
		return false;
	return !m_connections.back()->isClosed();
}

DebugConnection*
DebugLayout::Column::createConnection(int id)
{
	int top = m_pLayout->getHeight();
	int grow_by = 0;
	if (!m_connections.empty()) {
		DebugConnection* last_conn = m_connections.back();
		int extra = last_conn->getBottom()+CONN_VER_SPACING-m_pLayout->getHeight();
		if (extra > 0) {
			grow_by = extra;
			top += extra;
		}
	}
	DebugConnection* conn = manage(new DebugConnection(id, m_pLayout, top));
	grow_by += conn->getHeight();
	m_pLayout->growBy(grow_by);
	m_connections.push_back(conn);
	conn->destroySignal.connect(SigC::bind(
		SigC::slot(*this, &DebugLayout::Column::removeConnection), --m_connections.end()));
	m_pLayout->put(*conn, getLeft()+LAYOUT_PADDING, top+LAYOUT_PADDING);
	return conn;
}

void
DebugLayout::Column::growBy(int increment)
{
	if (m_connections.empty())
		return;
	DebugConnection* last_conn = m_connections.back();
	if (!last_conn->isClosed()) {
		last_conn->growBy(increment);
	}
}

void
DebugLayout::Column::trimTopBy(int value)
{
	for (std::list<DebugConnection*>::iterator iter = m_connections.begin();
	     iter != m_connections.end(); ++iter) {
		int new_top = (*iter)->getTop()-value;
		m_pLayout->move(**iter, getLeft()+LAYOUT_PADDING, new_top+LAYOUT_PADDING);
		(*iter)->setTop(new_top);
	}
}

void
DebugLayout::Column::removeConnection(std::list<DebugConnection*>::iterator iter)
{
	m_connections.erase(iter);
}

