/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "debugmessage.h"
#include "debuglayout.h"
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/textbuffer.h>
#include <gtkmm/textview.h>
#include <glibmm.h>

DebugMessage::DebugMessage(int top, const std::string& content, Type type, Subtype subtype)
:	m_top(top),
	m_content(content),
	m_height(DebugLayout::MESSAGE_INITIAL_HEIGHT),
	m_active(true),
	m_pContainer(manage(new Gtk::Fixed)),
	m_pContentWindow(0)
{
	modify_bg_pixmap(Gtk::STATE_NORMAL, "<none>");
	m_pContainer->show();
	add(*m_pContainer);
	m_pContainer->set_size_request(DebugLayout::MESSAGE_WIDTH, m_height);
	for (int i = 0; i < 3; i++) {
		m_pBorders[i] = manage(new Gtk::DrawingArea);
		m_pBorders[i]->modify_bg_pixmap(Gtk::STATE_NORMAL, "<none>");
		m_pBorders[i]->modify_bg(Gtk::STATE_NORMAL, getColorFor(type, subtype));
		m_pBorders[i]->show();
	}
	m_pBorders[0]->set_size_request(DebugLayout::MESSAGE_WIDTH-DebugLayout::MESSAGE_BORDER*2, DebugLayout::MESSAGE_BORDER);
	m_pBorders[1]->set_size_request(DebugLayout::MESSAGE_BORDER, m_height);
	m_pBorders[2]->set_size_request(DebugLayout::MESSAGE_BORDER, m_height);
	m_pContainer->put(*m_pBorders[0], DebugLayout::MESSAGE_BORDER, 0);
	m_pContainer->put(*m_pBorders[1], 0, 0);
	m_pContainer->put(*m_pBorders[2], DebugLayout::MESSAGE_WIDTH-DebugLayout::MESSAGE_BORDER, 0);
	
	signal_button_press_event().connect(SigC::slot(*this, &DebugMessage::onButtonPress));
	
	show();
}

DebugMessage::~DebugMessage()
{
	if (m_pContentWindow)
		delete m_pContentWindow;
}

void
DebugMessage::growBy(int increment)
{
	m_height += increment;
	m_pContainer->set_size_request(DebugLayout::MESSAGE_WIDTH, m_height);
	m_pBorders[1]->set_size_request(DebugLayout::MESSAGE_BORDER, m_height);
	m_pBorders[2]->set_size_request(DebugLayout::MESSAGE_BORDER, m_height);
}

void
DebugMessage::markAsClosed()
{
	finish();
	m_pBorders[0]->set_size_request(DebugLayout::MESSAGE_WIDTH-DebugLayout::MESSAGE_BORDER*2, m_height);
}

const Gdk::Color&
DebugMessage::getColorFor(Type type, Subtype subtype)
{
	static const Gdk::Color by_type[] = {
		Gdk::Color("#4fa81f"), // INCOMING_REQUEST
		Gdk::Color("#63e24d"), // OUTGOING_REQUEST
		Gdk::Color("#4a52e8")  // INCOMING_RESPONSE
	};
	static const Gdk::Color by_subtype[] = {
		Gdk::Color("#000000"), // ORES_NA
		Gdk::Color("#4a52e8"), // ORES_NOTMAL
		Gdk::Color("#5abddd"), // ORES_FILTERED
		Gdk::Color("#f7e11b"), // ORES_GENERATED
		Gdk::Color("#e84c4c")  // ORES_ERROR
	};
	if (type == OUTGOING_RESPONSE) {
		return by_subtype[subtype];
	}
	return by_type[type];
}

bool
DebugMessage::onButtonPress(GdkEventButton* event)
{
	if (event->type == GDK_2BUTTON_PRESS) {
		showContentWindow();
	}
	return true;
}

void
DebugMessage::showContentWindow()
{
	if (m_pContentWindow) {
		m_pContentWindow->present();
		return;
	}
	
	m_pContentWindow = new Gtk::Window;
	m_pContentWindow->set_title("HTTP Message");
	m_pContentWindow->set_size_request(500, 250);
	
	Gtk::ScrolledWindow* scrolled_window = manage(new Gtk::ScrolledWindow);
	scrolled_window->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	scrolled_window->set_shadow_type(Gtk::SHADOW_IN);
	m_pContentWindow->add(*scrolled_window);
	
	Gtk::TextView* text_view = manage(new Gtk::TextView);
	scrolled_window->add(*text_view);
	text_view->set_editable(false);
	text_view->set_cursor_visible(false);
	try {
		text_view->get_buffer()->set_text(Glib::convert(m_content, "utf-8", "iso-8859-1"));
	} catch (Glib::ConvertError&) {}
	
	m_pContentWindow->signal_hide().connect(SigC::slot(*this, &DebugMessage::destroyContentWindow));
	
	m_pContentWindow->show_all();
}

void
DebugMessage::destroyContentWindow()
{
	if (m_pContentWindow) {
		delete m_pContentWindow;
		m_pContentWindow = 0;
	}
}

