/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "debugwindow.h"
#include "debuglayout.h"
#include "debugmessage.h"
#include <list>
#include <utility> // for std::pair
#include <gtkmm/box.h>
#include <gtkmm/label.h>
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/alignment.h>
#include <gtkmm/frame.h>
#include <gtkmm/table.h>
#include <gtkmm/fixed.h>
#include <gtkmm/drawingarea.h>

DebugWindow::DebugWindow()
:	m_pLayout(manage(new DebugLayout)),
	m_connectionsAdjustment(10, 1, 99, 1, 10),
	m_trafficLimitAdjustment(200, 100, 9999, 100, 1000),
	m_pConnections(manage(new Gtk::SpinButton(m_connectionsAdjustment))),
	m_pTrafficLimit(manage(new Gtk::SpinButton(m_trafficLimitAdjustment))),
	m_scrollAtBottom(true)
{
	set_title("Debug");
	
	Gtk::HBox* hbox = manage(new Gtk::HBox);
	add(*hbox);
	
	Gtk::VBox* vbox1 = manage(new Gtk::VBox);
	hbox->pack_start(*vbox1, Gtk::PACK_SHRINK);
	
	Gtk::VBox* vbox2 = manage(new Gtk::VBox);
	vbox2->set_border_width(3);
	vbox2->set_spacing(2);
	vbox1->pack_start(*vbox2, Gtk::PACK_SHRINK);
	
	Gtk::Label* connections_label = manage(new Gtk::Label("Connections to track:"));
	connections_label->set_alignment(Gtk::ALIGN_LEFT);
	vbox2->pack_start(*connections_label, Gtk::PACK_SHRINK);
	Gtk::Alignment* connections_align = manage(new Gtk::Alignment(0.0, 0.5, 0.0, 0.0));
	connections_align->add(*m_pConnections);
	vbox2->pack_start(*connections_align, Gtk::PACK_SHRINK);
	
	Gtk::Label* traf_limit_label = manage(new Gtk::Label("Traffic limit, KB:"));
	traf_limit_label->set_alignment(Gtk::ALIGN_LEFT);
	vbox2->pack_start(*traf_limit_label, Gtk::PACK_SHRINK);
	Gtk::Alignment* traf_limit_align = manage(new Gtk::Alignment(0.0, 0.5, 0.0, 0.0));
	traf_limit_align->add(*m_pTrafficLimit);
	vbox2->pack_start(*traf_limit_align, Gtk::PACK_SHRINK);
	
	Gtk::Alignment* legend_alignment = manage(new Gtk::Alignment(0.5, 1.0, 1.0, 0.0));
	legend_alignment->set_border_width(3);
	vbox1->pack_start(*legend_alignment, true, true);
	
	Gtk::Frame* legend_frame = manage(new Gtk::Frame("Legend"));
	legend_alignment->add(*legend_frame);
	legend_frame->set_label_align(Gtk::ALIGN_CENTER);
	legend_frame->add(*makeLegend());
	
	Gtk::ScrolledWindow* scrolled_window = manage(new Gtk::ScrolledWindow);
	scrolled_window->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	scrolled_window->set_size_request(400, 400);
	scrolled_window->add(*m_pLayout);
	hbox->pack_start(*scrolled_window);
	
	scrolled_window->get_vadjustment()->signal_changed().connect(SigC::bind(
		SigC::slot(*this, &DebugWindow::onVAdjustmentChanged),
		scrolled_window->get_vadjustment()));
	scrolled_window->get_vadjustment()->signal_value_changed().connect(SigC::bind(
		SigC::slot(*this, &DebugWindow::onVAdjustmentValueChanged),
		scrolled_window->get_vadjustment()));
	
	m_connectionsAdjustment.signal_value_changed().connect(SigC::slot(*this, &DebugWindow::onTrackConnectionsChange));
	m_trafficLimitAdjustment.signal_value_changed().connect(SigC::slot(*this, &DebugWindow::onTrafficLimitChange));
	onTrackConnectionsChange();
	onTrafficLimitChange();
	
	show_all_children();
}

int
DebugWindow::newConnection()
{
	return m_pLayout->newConnection();
}

void
DebugWindow::connectionClosed(int conn_id)
{
	m_pLayout->connectionClosed(conn_id);
}

void
DebugWindow::messageStarts(int conn_id, const std::string& content, DebugMessage::Type type, DebugMessage::Subtype subtype)
{
	m_pLayout->messageStarts(conn_id, content, type, subtype);
}

void
DebugWindow::messageFinishes(int conn_id, DebugMessage::Type type)
{
	m_pLayout->messageFinishes(conn_id, type);
}

void
DebugWindow::debugMessage(int conn_id, const std::string& content)
{
	m_pLayout->debugMessage(conn_id, content);
}

void
DebugWindow::logTraffic(int conn_id, DebugConnection::TrafficType type, const std::string& content)
{
	m_pLayout->logTraffic(conn_id, type, content);
}

void
DebugWindow::onVAdjustmentChanged(Gtk::Adjustment* a)
{
	if (m_scrollAtBottom) {
		a->set_value(a->get_upper() - a->get_page_size());
		a->value_changed();
	}
}

void
DebugWindow::onVAdjustmentValueChanged(Gtk::Adjustment* a)
{
	m_scrollAtBottom = (a->get_value() >= a->get_upper() - a->get_page_size());
}

void
DebugWindow::onTrackConnectionsChange()
{
	m_pLayout->setConnectionsToTrack(m_pConnections->get_value_as_int());
}

void
DebugWindow::onTrafficLimitChange()
{
	m_pLayout->setTrafficLimit(m_pTrafficLimit->get_value_as_int()*1000);
}

Gtk::Widget*
DebugWindow::makeLegend()
{
	static const int TEXT_WIDTH = 140;
	struct MessageLegends {
		typedef std::pair<Gtk::Widget*, Gtk::Label*> PairType;
		typedef std::list<PairType> ListType;
		ListType m_list;
		void add(const Glib::ustring& text, DebugMessage::Type type,
			DebugMessage::Subtype subtype=DebugMessage::ORES_NA)
		{
			Gtk::Label* label = manage(new Gtk::Label(text));
			label->set_line_wrap(true);
			label->set_size_request(TEXT_WIDTH);
			m_list.push_back(PairType(makeMessageSymbol(type, subtype), label));
		}
	} mls;
	
	mls.add("Incoming request", DebugMessage::INCOMING_REQUEST);
	mls.add("Outgoing request", DebugMessage::OUTGOING_REQUEST);
	mls.add("Incoming response", DebugMessage::INCOMING_RESPONSE);
	mls.add("Unmodified outgoing response", DebugMessage::OUTGOING_RESPONSE, DebugMessage::ORES_NORMAL);
	mls.add("Filtered outgoing response", DebugMessage::OUTGOING_RESPONSE, DebugMessage::ORES_FILTERED);
	mls.add("Generated response (ad replacement)", DebugMessage::OUTGOING_RESPONSE, DebugMessage::ORES_GENERATED);
	mls.add("Error response", DebugMessage::OUTGOING_RESPONSE, DebugMessage::ORES_ERROR);
	
	Gtk::Table* table = manage(new Gtk::Table(mls.m_list.size()+2, 2));
	table->set_border_width(3);
	table->set_col_spacings(3);
	table->set_row_spacings(4);
	
	int row = 0;
	Gtk::Widget* conn_symbol = makeConnectionSymbol(12, 16);
	table->attach(*conn_symbol, 0, 1, row, row+1, Gtk::SHRINK);
	Gtk::Label* label= manage(new Gtk::Label("Client connection"));
	label->set_line_wrap(true);
	label->set_size_request(TEXT_WIDTH);
	table->attach(*label, 1, 2, row, row+1);
	++row;
	
	MessageLegends::ListType::iterator iter = mls.m_list.begin();
	for (; iter != mls.m_list.end(); ++iter, ++row) {
		table->attach(*iter->first, 0, 1, row, row+1, Gtk::FILL);
		table->attach(*iter->second, 1, 2, row, row+1);
	}
	
	Gtk::Label* footer_label = manage(new Gtk::Label("Doubleclick any object for more info."));
	footer_label->set_size_request(TEXT_WIDTH+15);
	footer_label->set_line_wrap(true);
	table->attach(*footer_label, 0, 2, row, row+1);
	
	return table;
}

Gtk::Widget*
DebugWindow::makeMessageSymbol(DebugMessage::Type type, DebugMessage::Subtype subtype)
{
	Gtk::DrawingArea* box = manage(new Gtk::DrawingArea);
	box->set_size_request(DebugLayout::MESSAGE_WIDTH, DebugLayout::MESSAGE_INITIAL_HEIGHT);
	box->modify_bg_pixmap(Gtk::STATE_NORMAL, "<none>");
	box->modify_bg(Gtk::STATE_NORMAL, DebugMessage::getColorFor(type, subtype));
	return box;
}

Gtk::Widget*
DebugWindow::makeConnectionSymbol(int width, int height)
{
	Gtk::Fixed* box = manage(new Gtk::Fixed);
	box->set_size_request(width, height);
	Gtk::DrawingArea* borders[4];
	for (int i = 0; i < 4; i++) {
		borders[i] = manage(new Gtk::DrawingArea);
		borders[i]->modify_bg_pixmap(Gtk::STATE_NORMAL, "<none>");
		borders[i]->modify_bg(Gtk::STATE_NORMAL, Gdk::Color("#ac40ef"));
	}
	borders[0]->set_size_request(width, DebugLayout::CONN_BORDER);
	borders[1]->set_size_request(DebugLayout::CONN_BORDER, height-DebugLayout::CONN_BORDER*2);
	borders[2]->set_size_request(DebugLayout::CONN_BORDER, height-DebugLayout::CONN_BORDER*2);
	borders[3]->set_size_request(width, DebugLayout::CONN_BORDER);
	box->put(*borders[0], 0, 0);
	box->put(*borders[1], 0, DebugLayout::CONN_BORDER);
	box->put(*borders[2], width-DebugLayout::CONN_BORDER, DebugLayout::CONN_BORDER);
	box->put(*borders[3], 0, height-DebugLayout::CONN_BORDER);
	
	return box;
}

