/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "logdialog.h"
#include "application.h"
#include <gtkmm/box.h>
#include <gtkmm/buttonbox.h>
#include <gtkmm/button.h>
#include <gtkmm/alignment.h>
#include <gtkmm/stock.h>
#include <gtkmm/scrolledwindow.h>

LogDialog::LogDialog()
:	m_scrollAtBottom(true),
	m_pTextBuffer(Gtk::TextBuffer::create()),
	m_pWarningTag(Gtk::TextBuffer::Tag::create("warning")),
	m_pErrorTag(Gtk::TextBuffer::Tag::create("error")),
	m_pTextView(manage(new Gtk::TextView(m_pTextBuffer)))
{
	m_pWarningTag->property_foreground() = "#fd687c";
	m_pErrorTag->property_foreground() = "#f7005d";
	m_pTextBuffer->get_tag_table()->add(m_pWarningTag);
	m_pTextBuffer->get_tag_table()->add(m_pErrorTag);
	
	set_title("BFilter Log");
	set_border_width(5);
	
	Gtk::VBox* vbox = manage(new Gtk::VBox);
	add(*vbox);
	
	Gtk::ScrolledWindow* scrolled_window = manage(new Gtk::ScrolledWindow);
	scrolled_window->set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	scrolled_window->set_shadow_type(Gtk::SHADOW_IN);
	scrolled_window->set_size_request(400, 150);
	
	m_pTextView->set_editable(false);
	m_pTextView->set_cursor_visible(false);
	m_pTextView->set_wrap_mode(Gtk::WRAP_WORD);
	scrolled_window->add(*m_pTextView);
	vbox->pack_start(*scrolled_window);
	
	Gtk::Alignment* spacer1 = manage(new Gtk::Alignment());
	spacer1->set_size_request(8, 8);
	vbox->pack_start(*spacer1, Gtk::PACK_SHRINK);
	
	Gtk::Alignment* bbox_alignment = manage(new Gtk::Alignment(0.5, 0.5, 0.0, 0.0));
	vbox->pack_start(*bbox_alignment, Gtk::PACK_SHRINK);
	
	Gtk::Alignment* spacer2 = manage(new Gtk::Alignment());
	spacer2->set_size_request(3, 3);
	vbox->pack_start(*spacer2, Gtk::PACK_SHRINK);
	
	Gtk::HButtonBox* button_box = manage(new Gtk::HButtonBox);
	button_box->set_spacing(10);
	bbox_alignment->add(*button_box);
	
	Gtk::Button *close_btn = manage(new Gtk::Button(Gtk::Stock::CLOSE));
	close_btn->signal_clicked().connect(SigC::slot(*this, &LogDialog::hide));
	button_box->pack_start(*close_btn, Gtk::PACK_SHRINK);
	
	Gtk::Button *clear_btn = manage(new Gtk::Button(Gtk::Stock::CLEAR));
	clear_btn->signal_clicked().connect(SigC::slot(*this, &LogDialog::onClear));
	button_box->pack_start(*clear_btn, Gtk::PACK_SHRINK);
	
	close_btn->grab_focus();
	
	scrolled_window->get_vadjustment()->signal_changed().connect(SigC::bind(
		SigC::slot(*this, &LogDialog::onVAdjustmentChanged),
		scrolled_window->get_vadjustment()));
	scrolled_window->get_vadjustment()->signal_value_changed().connect(SigC::bind(
		SigC::slot(*this, &LogDialog::onVAdjustmentValueChanged),
		scrolled_window->get_vadjustment()));
	
	show_all_children();
	
	Log::getInstance().attachListener(this);
	std::list<Log::Entry>::const_iterator iter = Log::getInstance().getEntries().begin();
	for (; iter != Log::getInstance().getEntries().end(); ++iter) {
		append(*iter);
	}
}

LogDialog::~LogDialog()
{
	Log::getInstance().detachListener(this);
}

void
LogDialog::append(const Log::Entry& entry)
{
	if (m_pTextBuffer->get_char_count() > 0 && entry.isNewLine()) {
		m_pTextBuffer->insert(m_pTextBuffer->end(), "\n");
	}
	switch (entry.getType()) {
		case Log::Entry::T_ERROR:
			m_pTextBuffer->insert_with_tag(m_pTextBuffer->end(), entry.getMessage(), m_pErrorTag);
			break;
		case Log::Entry::T_WARNING:
			m_pTextBuffer->insert_with_tag(m_pTextBuffer->end(), entry.getMessage(), m_pWarningTag);
			break;
		case Log::Entry::T_MESSAGE:
			m_pTextBuffer->insert(m_pTextBuffer->end(), entry.getMessage());
			break;
	}
}

void
LogDialog::clear()
{
	m_pTextBuffer->set_text("");
}

void
LogDialog::onClear()
{
	Log::getInstance().clear();
}

void
LogDialog::onVAdjustmentChanged(Gtk::Adjustment* a)
{
	if (m_scrollAtBottom) {
		a->set_value(a->get_upper() - a->get_page_size());
		a->value_changed();
	}
}

void
LogDialog::onVAdjustmentValueChanged(Gtk::Adjustment* a)
{
	m_scrollAtBottom = (a->get_value() >= a->get_upper() - a->get_page_size());
}

