/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "application.h"
#include <sstream>
#include <fstream>
#include <string>
#include <list>
#include <algorithm>
#include <windows.h>
#include <winsock2.h>
#include <wx/wx.h>
#include <wx/dir.h>
#include <wx/log.h>
#include "logdialog.h"
#include "types.h"

IMPLEMENT_APP(Application);

Application::Application()
:	m_pSocketServer(0), m_pBasicConfig(0)
{
}

Application::~Application()
{
	if (m_pSocketServer)
		delete m_pSocketServer;
}

bool
Application::OnInit()
{
	SetTopWindow(new wxFrame);
	m_appDir.Assign(argv[0]);
	m_appDir.AssignDir(m_appDir.GetPath(wxPATH_GET_VOLUME|wxPATH_GET_SEPARATOR));
	m_confDir.Assign(m_appDir);
	m_confDir.AppendDir("conf");
	
	if (argc == 2 && std::string(argv[1]) == "--fix-config") {
		readConfig();
		processConfig(m_config, m_configIO);
		m_configIO.setConfig(m_configIO.getConfig());
		std::ostringstream strm;
		m_configIO.writeConfig(strm);
		m_config = strm.str();
		writeConfig();
		return false;
	}
	
	if (!locateBackend())
		return false;
	//Log::getInstance().appendMessage("Using backend: "+m_backendExecutable);
	
	m_pBasicConfig = new BasicConfigDialog(GetTopWindow());
	m_pAdvancedConfig = new AdvancedConfigWindow(GetTopWindow());
	
	bool ok = true;
	if (!readConfig())
		ok = false;
	bool gr_ok = readGlobalRules();
	bool lr_ok = readLocalRules();
	readConfigDefault();
	if (ok) {
		if (!processConfig(m_config, m_configIO)) {
			ok = false;
		} else {
			if (gr_ok && lr_ok) {
				processRules(m_globalRules, m_localRules, m_rulesReader);
			}
		}
	}
	
	if (!ok) {
		m_trayIcon.setUnopIcon();
		showLogDialog();
		return true;
	}
	
	m_masterState.setConfig(m_configIO.getConfig());
	m_masterState.setRules(m_rulesReader.getRules());
	
	std::ostringstream strm;
	strm << "Binding to " << m_configIO.getConfig().getListenAddr()
		<< ':' << m_configIO.getConfig().getListenPort();
	m_pSocketServer = new SocketServer(this,
		m_configIO.getConfig().getListenAddr(),
		m_configIO.getConfig().getListenPort());
	if (!m_pSocketServer->isOK()) {
		delete m_pSocketServer;
		m_pSocketServer = 0;
		strm << " [FAILED]";
		Log::getInstance().appendError(strm.str());
		m_trayIcon.setUnopIcon();
		showLogDialog();
		return true;
	}
	strm << " [OK]";
	Log::getInstance().appendMessage(strm.str());
	
	return true;
}

void
Application::showTrayMenu()
{
	m_trayIcon.PopupMenu(&m_trayMenu);
}

void
Application::showBasicConfigDialog()
{
	if (m_pBasicConfig->IsShown()) {
		m_pBasicConfig->Raise();
	} else {
		m_pBasicConfig->Show();
		m_pBasicConfig->setConfig(m_masterState.getConfig());
	}
}

void
Application::showAdvancedConfigWindow()
{
	if (m_pAdvancedConfig->IsShown()) {
		m_pAdvancedConfig->Raise();
	} else {
		m_pAdvancedConfig->loadConfig(m_config);
		m_pAdvancedConfig->loadConfigDefault(m_configDefault);
		m_pAdvancedConfig->loadRules(m_globalRules);
		m_pAdvancedConfig->loadRulesLocal(m_localRules);
		m_pAdvancedConfig->Show();
	}
}

void
Application::showLogDialog()
{
	LogDialog::show();
}

void
Application::setFiltering(bool enabled)
{
	m_masterState.setFilteringEnabled(enabled);
}

void
Application::setScriptFiltering(bool enabled)
{
	m_masterState.setScriptFilteringEnabled(enabled);
}

bool
Application::applyNewConfig(const Config& config)
{
	Log::getInstance().appendMessage("Applying the new config ...");
	if (config.getListenPort() < 1 || config.getListenPort() > 65535) {
		Log::getInstance().appendError("Listen port is out of range [1, 65535]");
		return false;
	}
	if (config.getUseProxy() && (config.getProxyPort() < 1 || config.getProxyPort() > 65535)) {
		Log::getInstance().appendError("Proxy port is out of range [1, 65535]");
		return false;
	}
	bool was_bound = m_pSocketServer;
	if (was_bound && m_configIO.getConfig().getListenAddr() == config.getListenAddr()
	    && m_configIO.getConfig().getListenPort() == config.getListenPort()) {
		Log::getInstance().appendMessage("Rebinding is not necessary");
	} else {
		if (was_bound) {
			delete m_pSocketServer;
			m_pSocketServer = 0;
			std::ostringstream strm;
			strm << "Unbinding from " << m_configIO.getConfig().getListenAddr()
				<< ':' << m_configIO.getConfig().getListenPort() << " [OK]";
			Log::getInstance().appendMessage(strm.str());
		}
		std::ostringstream strm;
		strm << "Binding to " << config.getListenAddr()
			<< ':' << config.getListenPort();
		m_pSocketServer = new SocketServer(this, config.getListenAddr(), config.getListenPort());
		if (m_pSocketServer->isOK()) {
			strm << " [OK]";
			Log::getInstance().appendMessage(strm.str());
			if (!was_bound) {
				m_trayIcon.setNormalIcon();
			}
		} else {
			strm << " [FAILED]";
			Log::getInstance().appendError(strm.str());
			if (was_bound) {
				Log::getInstance().appendMessage("Applying the new config failed, rollbacking changes ...");
				std::ostringstream strm2;
				strm2 << "Binding to " << m_configIO.getConfig().getListenAddr()
					<< ':' << m_configIO.getConfig().getListenPort();
				delete m_pSocketServer;
				m_pSocketServer = new SocketServer(this, m_configIO.getConfig().getListenAddr(),
					m_configIO.getConfig().getListenPort());
				if (m_pSocketServer->isOK()) {
					strm2 << " [OK]";
					Log::getInstance().appendMessage(strm2.str());
				} else {
					strm2 << " [FAILED]";
					Log::getInstance().appendError(strm2.str());
					delete m_pSocketServer;
					m_pSocketServer = 0;
					m_trayIcon.setUnopIcon();
				}
			}
			return false;
		}
	}
	m_configIO.setConfig(config);
	m_masterState.setConfig(config);
	std::ostringstream strm;
	m_configIO.writeConfig(strm);
	m_config = strm.str();
	return true;
}

bool
Application::applyNewConfigAndRules(const ConfigIO& cio, const Rules& rules,
	const std::string& config_source, const std::string& rules_source,
	const std::string& rules_local_source)
{
	if (!applyNewConfig(cio.getConfig()))
		return false;
	m_configIO = cio;
	m_rulesReader.setRules(rules);
	m_masterState.setRules(rules);
	Log::getInstance().appendMessage("Applying the new rules [OK]");
	m_config = config_source;
	m_globalRules = rules_source;
	m_localRules = rules_local_source;
	return true;
}

bool
Application::writeConfig()
{
	static std::string msg1 = "Writing config ...";
	static std::string msg2 = "Writing config [OK]";
	wxFileName fname(m_confDir);
	fname.SetName("config");
	std::string fn = fname.GetFullPath().c_str();
	std::ofstream fstrm(fn.c_str());
	if (!fstrm.good()) {
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendError("Could not open "+fn+" for writing");
		return false;
	}
	fstrm.write(m_config.data(), m_config.length());
	fstrm << std::flush;
	if (fstrm.bad()) {
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendError("Error writing to "+fn);
		return false;
	}
	Log::getInstance().appendMessage(msg2);
	return true;
}

bool
Application::writeGlobalRules()
{
	static std::string msg1 = "Writing global rules ...";
	static std::string msg2 = "Writing global rules [OK]";
	wxFileName fname(m_confDir);
	fname.SetName("rules");
	std::string fn = fname.GetFullPath().c_str();
	std::ofstream fstrm(fn.c_str());
	if (!fstrm.good()) {
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendError("Could not open "+fn+" for writing");
		return false;
	}
	fstrm.write(m_globalRules.data(), m_globalRules.length());
	fstrm << std::flush;
	if (fstrm.bad()) {
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendError("Error writing to "+fn);
		return false;
	}
	Log::getInstance().appendMessage(msg2);
	return true;
}

bool
Application::writeLocalRules()
{
	static std::string msg1 = "Writing local rules ...";
	static std::string msg2 = "Writing local rules [OK]";
	wxFileName fname(m_confDir);
	fname.SetName("rules.local");
	std::string fn = fname.GetFullPath().c_str();
	std::ofstream fstrm(fn.c_str());
	if (!fstrm.good()) {
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendError("Could not open "+fn+" for writing");
		return false;
	}
	fstrm.write(m_localRules.data(), m_localRules.length());
	fstrm << std::flush;
	if (fstrm.bad()) {
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendError("Error writing to "+fn);
		return false;
	}
	Log::getInstance().appendMessage(msg2);
	return true;
}

void
Application::quit()
{
	ExitMainLoop();
}

void
Application::incomingConnection(SOCKET socket)
{
	HANDLE hProcess = NULL;
	SOCKET isock = INVALID_SOCKET;
	DuplicateHandle(GetCurrentProcess(), GetCurrentProcess(), GetCurrentProcess(), &hProcess, 0, TRUE, DUPLICATE_SAME_ACCESS);
	DuplicateHandle(hProcess, (HANDLE)socket, hProcess, (LPHANDLE)&isock, 0, TRUE, DUPLICATE_SAME_ACCESS);
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	ZeroMemory(&si, sizeof(si));
	si.cb = sizeof(si);
	ZeroMemory(&pi, sizeof(pi));
	std::ostringstream fullpath;
	fullpath << '"' << m_backendExecutable << '"' << ' ' << (uintptr_t)hProcess << ' '
		<< (uintptr_t)m_masterState.getShmHandle() << ' ' << (uintptr_t)m_masterState.getShmMutex() << ' ' << isock;
	CreateProcess(m_backendExecutable.c_str(), (char *)fullpath.str().c_str(),
		NULL, NULL, TRUE, 0, NULL, NULL, &si, &pi);
	CloseHandle(pi.hProcess);
	CloseHandle(pi.hThread);
	closesocket(isock);
	closesocket(socket);
	CloseHandle(hProcess);
}

/*
 * On Windows, a running executable can't be overwriten, deleted, or renamed.
 * With bfilter.exe this is not a problem: the installer checks if it is
 * writable, and if not, asks the user to quit the program.
 * With backends, it's not that simple: closing bfilter.exe won't kill the
 * running backends. To solve this problem, the installer puts bfilter-backend.exe
 * to $INSTDIR\backend, but the frontend moves it to another location before running it.
 */
bool
Application::locateBackend()
{
	wxLogNull noLog;
	wxFileName fname = m_appDir;
	fname.AppendDir("backend");
	wxString backend_dir = fname.GetPath(wxPATH_GET_VOLUME);
	wxDir dir(backend_dir);
	if (!dir.IsOpened()) {
		wxMessageBox("Could not open the directory "+backend_dir+"\r\nThe installation is corrupt.",
			"BFilter Error", wxOK|wxICON_ERROR, GetTopWindow());
		return false;
	}
	wxArrayString subdirs;
	wxString subdir;
	if (dir.GetFirst(&subdir, wxEmptyString, wxDIR_DIRS)) {
		subdirs.Add(subdir);
		while (dir.GetNext(&subdir)) {
			subdirs.Add(subdir);
		}
	}
	long max_numeric = 0, tmp = 0;
	for (size_t i = 0; i < subdirs.Count(); ++i) {
		if (subdirs.Item(i).ToLong(&tmp)) {
			if (tmp > max_numeric)
				max_numeric = tmp;
		}
	}
	
	fname.AssignDir(backend_dir);
	fname.SetName("bfilter-backend.exe");
	wxString new_backend = fname.GetFullPath();
	if (!wxFileExists(new_backend)) {
		bool found = false;
		subdirs.Sort(true);
		for (size_t i = 0; i < subdirs.Count(); ++i) {
			if (subdirs.Item(i).ToLong(&tmp)) {
				fname.AssignDir(backend_dir);
				fname.AppendDir(subdirs.Item(i));
				fname.SetName("bfilter-backend.exe");
				wxString backend = fname.GetFullPath();
				if (wxFileExists(backend)) {
					m_backendExecutable = backend.c_str();
					found = true;
				}
			}
		}
		if (!found) {
			wxMessageBox("Could not locate the backend executable.\r\nThe installation is corrupt.",
				"BFilter Error", wxOK|wxICON_ERROR, GetTopWindow());
			return false;
		}
	} else {
		fname.AssignDir(backend_dir);
		wxString newdir;
		newdir.Printf("%d", (int)max_numeric+1);
		fname.AppendDir(newdir);
		newdir = fname.GetPath(wxPATH_GET_VOLUME);
		if (!fname.Mkdir()) {
			wxMessageBox("Could not create directory "+newdir,
				"BFilter Error", wxOK|wxICON_ERROR, GetTopWindow());
			return false;
		}
		fname.SetName("bfilter-backend.exe");
		wxString newfile = fname.GetFullPath();
		if (!wxRenameFile(new_backend, newfile)) {
			wxMessageBox("Could not move "+new_backend+" to "+newfile,
				"BFilter Error", wxOK|wxICON_ERROR, GetTopWindow());
			return false;
		}
		for (size_t i = 0; i < subdirs.Count(); ++i) {
			if (subdirs.Item(i).ToLong(&tmp)) {
				fname.AssignDir(backend_dir);
				fname.AppendDir(subdirs.Item(i));
				fname.SetName("bfilter-backend.exe");
				wxRemoveFile(fname.GetFullPath());
				fname.AssignDir(backend_dir);
				fname.AppendDir(subdirs.Item(i));
				if (fname.Rmdir()) {
					subdirs.Item(i) = wxEmptyString;
				}
			}
		}
		max_numeric = 0;
		for (size_t i = 0; i < subdirs.Count(); ++i) {
			if (subdirs.Item(i).ToLong(&tmp)) {
				if (tmp > max_numeric)
					max_numeric = tmp;
			}
		}
		m_backendExecutable = newfile.c_str();
		fname.AssignDir(backend_dir);
		wxString nd;
		nd.Printf("%d", (int)max_numeric+1);
		fname.AppendDir(nd);
		wxString movedir = fname.GetPath(wxPATH_GET_VOLUME);
		if (wxRenameFile(newdir, movedir)) {
			fname.SetName("bfilter-backend.exe");
			m_backendExecutable = fname.GetFullPath().c_str();
		}
	}
	return true;
}

bool
Application::readConfig()
{
	static const std::string msg1 = "Reading config ...";
	static const std::string msg2 = "Reading config [OK]";
	wxFileName fname(m_confDir);
	fname.SetName("config");
	std::string fn = fname.GetFullPath().c_str();
	std::ifstream strm(fn.c_str());
	if (!strm.good()) {
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendError("Could not open "+fn);
		return false;
	}
	m_config.erase();
	char buf[4096];
	while (strm.read(buf, sizeof(buf)).gcount()) {
		m_config.append(buf, strm.gcount());
	}
	if (strm.bad()) {
		m_config.erase();
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendError("Error reading "+fn);
		return false;
	}
	Log::getInstance().appendMessage(msg2);
	return true;
}

bool
Application::readConfigDefault()
{
	static const std::string msg1 = "Reading default config ...";
	static const std::string msg2 = "Reading default config [OK]";
	wxFileName fname(m_confDir);
	fname.SetName("config.default");
	std::string fn = fname.GetFullPath().c_str();
	std::ifstream strm(fn.c_str());
	if (!strm.good()) {
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendWarning("Could not open "+fn);
		return false;
	}
	m_configDefault.erase();
	char buf[4096];
	while (strm.read(buf, sizeof(buf)).gcount()) {
		m_configDefault.append(buf, strm.gcount());
	}
	if (strm.bad()) {
		m_configDefault.erase();
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendWarning("Error reading "+fn);
		return false;
	}
	Log::getInstance().appendMessage(msg2);
	return true;
}

bool
Application::readGlobalRules()
{
	static const std::string msg1 = "Reading global rules ...";
	static const std::string msg2 = "Reading global rules [OK]";
	wxFileName fname(m_confDir);
	fname.SetName("rules");
	std::string fn = fname.GetFullPath().c_str();
	std::ifstream strm(fn.c_str());
	if (!strm.good()) {
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendError("Could not open "+fn);
		return false;
	}
	m_globalRules.erase();
	char buf[4096];
	while (strm.read(buf, sizeof(buf)).gcount()) {
		m_globalRules.append(buf, strm.gcount());
	}
	if (strm.bad()) {
		m_globalRules.erase();
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendError("Error reading "+fn);
		return false;
	}
	Log::getInstance().appendMessage(msg2);
	return true;
}

bool
Application::readLocalRules()
{
	static const std::string msg1 = "Reading local rules ...";
	static const std::string msg2 = "Reading local rules [OK]";
	wxFileName fname(m_confDir);
	fname.SetName("rules.local");
	std::string fn = fname.GetFullPath().c_str();
	std::ifstream strm(fn.c_str());
	if (!strm.good()) {
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendError("Could not open "+fn);
		return false;
	}
	m_localRules.erase();
	char buf[4096];
	while (strm.read(buf, sizeof(buf)).gcount()) {
		m_localRules.append(buf, strm.gcount());
	}
	if (strm.bad()) {
		m_localRules.erase();
		Log::getInstance().appendMessage(msg1);
		Log::getInstance().appendError("Error reading "+fn);
		return false;
	}
	Log::getInstance().appendMessage(msg2);
	return true;
}

bool
Application::processConfig(const std::string& config, ConfigIO& cio)
{
	static const std::string msg1 = "Processing config ...";
	static const std::string msg2 = "Processing config [OK]";
	std::istringstream strm(config);
	ConfigIO::Status status = cio.readConfig(strm);
	if (status.status == ConfigIO::Status::OK) {
		Log::getInstance().appendMessage(msg2);
	} else {
		Log::getInstance().appendMessage(msg1);
	}
	for (ConfigIO::ErrorList::const_iterator iter = status.errors.begin();
	     iter != status.errors.end(); ++iter) {
		std::ostringstream msg;
		msg << "[config";
		if (iter->getLine()) {
			msg << ':' << iter->getLine();
			if (iter->getCol())
				msg << ':' << iter->getCol();
		}
		msg << "] " << iter->getMessage();
		if (iter->getType() == ConfIO::Error::T_ERROR) {
			Log::getInstance().appendError(msg.str());
		} else {
			Log::getInstance().appendWarning(msg.str());
		}
	}
	return (status.status != ConfigIO::Status::ERRORS);
}

bool
Application::processRules(const std::string& global_rules, const std::string& local_rules, RulesReader& rr)
{
	static const std::string msg1 = "Processing rules ...";
	static const std::string msg2 = "Processing rules [OK]";
	std::istringstream strm_global(global_rules);
	std::istringstream strm_local(local_rules);
	RulesReader::Status status = rr.readRules(strm_global, strm_local);
	if (status.status == RulesReader::Status::OK) {
		Log::getInstance().appendMessage(msg2);
	} else {
		Log::getInstance().appendMessage(msg1);
	}
	for (ConfIO::ErrorList::const_iterator iter = status.errors_global.begin();
	     iter != status.errors_global.end(); ++iter) {
		std::ostringstream msg;
		msg << "[rules";
		if (iter->getLine()) {
			msg << ':' << iter->getLine();
			if (iter->getCol())
				msg << ':' << iter->getCol();
		}
		msg << "] " << iter->getMessage();
		if (iter->getType() == ConfIO::Error::T_ERROR) {
			Log::getInstance().appendError(msg.str());
		} else {
			Log::getInstance().appendWarning(msg.str());
		}
	}
	for (ConfIO::ErrorList::const_iterator iter = status.errors_local.begin();
	     iter != status.errors_local.end(); ++iter) {
		std::ostringstream msg;
		msg << "[rules.local";
		if (iter->getLine()) {
			msg << ':' << iter->getLine();
			if (iter->getCol())
				msg << ':' << iter->getCol();
		}
		msg << "] " << iter->getMessage();
		if (iter->getType() == ConfIO::Error::T_ERROR) {
			Log::getInstance().appendError(msg.str());
		} else {
			Log::getInstance().appendWarning(msg.str());
		}
	}
	return (status.status != RulesReader::Status::ERRORS);
}

