/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "socketserver.h"
#include <sstream>
#include <cstring>

extern HINSTANCE wxhInstance;

static const char* CLASSNAME = "_SocketServerClass";
static const char* WINDOWNAME = "_SocketServerWindow";

const unsigned int SocketServer::UWM_SOCK;
const int SocketServer::NLISTEN;
std::map<SOCKET, SocketServer*> SocketServer::m_sSocketMap;

SocketServer::StaticInitializer::StaticInitializer()
:	err(0), hWin(NULL)
{
	if ((err = WSAStartup(MAKEWORD(2, 0), &wsadata))) {
		return;
	}
	
	WNDCLASS winClass;
	winClass.style         = 0;
	winClass.lpfnWndProc   = &SocketServer::wndProc;
	winClass.cbClsExtra    = 0;
	winClass.cbWndExtra    = 0;
	winClass.hInstance     = wxhInstance;
	winClass.hIcon         = (HICON)NULL;
	winClass.hCursor       = (HCURSOR)NULL;
	winClass.hbrBackground = (HBRUSH)NULL;
	winClass.lpszMenuName  = (LPCTSTR)NULL;
	winClass.lpszClassName = CLASSNAME;
	RegisterClass(&winClass);
	hWin = CreateWindow(CLASSNAME, WINDOWNAME,
	                    0, 0, 0, 0, 0,
	                    (HWND)NULL, (HMENU)NULL, wxhInstance, (LPVOID)NULL);
	if (!hWin) {
		err = GetLastError();
	}
}

SocketServer::StaticInitializer::~StaticInitializer()
{
	WSACleanup();
	if (hWin)
		DestroyWindow(hWin);
}

static int inet_aton(const char *cp, struct in_addr *pin)
{
	int rc = inet_addr(cp);
	if (rc == -1 && strcmp(cp, "255.255.255.255"))
		return 0;
	pin->s_addr = rc;
	return 1;
}

SocketServer::SocketServer(Listener* listener, const std::string& host, int port)
:	m_pListener(listener), m_socket(INVALID_SOCKET), m_lastError(0)
{
	const int on = 1;
	struct sockaddr_in local;
	static StaticInitializer si;
	if (si.err) {
		m_lastError = si.err;
		return;
	}
	
	memset(&local, 0, sizeof(local));
	if (host.empty()) {
		local.sin_addr.s_addr = htonl(INADDR_ANY);
	} else if (!inet_aton(host.c_str(), &local.sin_addr)) {
		struct hostent *hp = gethostbyname(host.c_str());
		if (!hp) {
			m_lastError = WSAGetLastError();
			return;
		}
		memcpy(&local.sin_addr, hp->h_addr, sizeof(local.sin_addr));
	}
	local.sin_port = htons(port);
	local.sin_family = AF_INET;
	
	m_socket = socket(AF_INET, SOCK_STREAM, 0);
	if (m_socket == INVALID_SOCKET) {
		m_lastError = WSAGetLastError();
		return;
	}
	
	setsockopt(m_socket, SOL_SOCKET, SO_REUSEADDR, (const char *)&on, sizeof(on));
	
	if (bind(m_socket, (struct sockaddr *)&local, sizeof(local))) {
		m_lastError = WSAGetLastError();
		closesocket(m_socket);
		m_socket = INVALID_SOCKET;
		return;
	}
	
	if (listen(m_socket, NLISTEN)) {
		m_lastError = WSAGetLastError();
		closesocket(m_socket);
		m_socket = INVALID_SOCKET;
		return;
	}
	
	if (WSAAsyncSelect(m_socket, si.hWin, UWM_SOCK, FD_ACCEPT)) {
		m_lastError = WSAGetLastError();
		closesocket(m_socket);
		m_socket = INVALID_SOCKET;
		return;
	}
	
	m_sSocketMap[m_socket] = this;
}

SocketServer::~SocketServer()
{
	if (m_socket != INVALID_SOCKET) {
		closesocket(m_socket);
		m_sSocketMap.erase(m_socket);
	}
}

namespace {
	struct WinSockError
	{
		int m_code;
		const char* m_message;
		
		WinSockError(int code, const char* msg)
		: m_code(code), m_message(msg) {}
		
		bool operator<(const WinSockError& rhs) const {
			return m_code < rhs.m_code;
		}
	};
	
	// The entries are sorted by error code, so we can use binary search on them
	const WinSockError wsockErrorList[] = {
		WinSockError(0,                  "No error"),
		WinSockError(WSAEINTR,           "Interrupted system call"),
		WinSockError(WSAEBADF,           "Bad file number"),
		WinSockError(WSAEACCES,          "Permission denied"),
		WinSockError(WSAEFAULT,          "Bad address"),
		WinSockError(WSAEINVAL,          "Invalid argument"),
		WinSockError(WSAEMFILE,          "Too many open sockets"),
		WinSockError(WSAEWOULDBLOCK,     "Operation would block"),
		WinSockError(WSAEINPROGRESS,     "Operation now in progress"),
		WinSockError(WSAEALREADY,        "Operation already in progress"),
		WinSockError(WSAENOTSOCK,        "Socket operation on non-socket"),
		WinSockError(WSAEDESTADDRREQ,    "Destination address required"),
		WinSockError(WSAEMSGSIZE,        "Message too long"),
		WinSockError(WSAEPROTOTYPE,      "Protocol wrong type for socket"),
		WinSockError(WSAENOPROTOOPT,     "Bad protocol option"),
		WinSockError(WSAEPROTONOSUPPORT, "Protocol not supported"),
		WinSockError(WSAESOCKTNOSUPPORT, "Socket type not supported"),
		WinSockError(WSAEOPNOTSUPP,      "Operation not supported on socket"),
		WinSockError(WSAEPFNOSUPPORT,    "Protocol family not supported"),
		WinSockError(WSAEAFNOSUPPORT,    "Address family not supported"),
		WinSockError(WSAEADDRINUSE,      "Address already in use"),
		WinSockError(WSAEADDRNOTAVAIL,   "Can't assign requested address"),
		WinSockError(WSAENETDOWN,        "Network is down"),
		WinSockError(WSAENETUNREACH,     "Network is unreachable"),
		WinSockError(WSAENETRESET,       "Net connection reset"),
		WinSockError(WSAECONNABORTED,    "Software caused connection abort"),
		WinSockError(WSAECONNRESET,      "Connection reset by peer"),
		WinSockError(WSAENOBUFS,         "No buffer space available"),
		WinSockError(WSAEISCONN,         "Socket is already connected"),
		WinSockError(WSAENOTCONN,        "Socket is not connected"),
		WinSockError(WSAESHUTDOWN,       "Can't send after socket shutdown"),
		WinSockError(WSAETOOMANYREFS,    "Too many references, can't splice"),
		WinSockError(WSAETIMEDOUT,       "Connection timed out"),
		WinSockError(WSAECONNREFUSED,    "Connection refused"),
		WinSockError(WSAELOOP,           "Too many levels of symbolic links"),
		WinSockError(WSAENAMETOOLONG,    "File name too long"),
		WinSockError(WSAEHOSTDOWN,       "Host is down"),
		WinSockError(WSAEHOSTUNREACH,    "No route to host"),
		WinSockError(WSAENOTEMPTY,       "Directory not empty"),
		WinSockError(WSAEPROCLIM,        "Too many processes"),
		WinSockError(WSAEUSERS,          "Too many users"),
		WinSockError(WSAEDQUOT,          "Disc quota exceeded"),
		WinSockError(WSAESTALE,          "Stale NFS file handle"),
		WinSockError(WSAEREMOTE,         "Too many levels of remote in path"),
		WinSockError(WSASYSNOTREADY,     "Network system is unavailable"),
		WinSockError(WSAVERNOTSUPPORTED, "Winsock version out of range"),
		WinSockError(WSANOTINITIALISED,  "WSAStartup not yet called"),
		WinSockError(WSAEDISCON,         "Graceful shutdown in progress"),
		WinSockError(WSAHOST_NOT_FOUND,  "Host not found"),
		WinSockError(WSANO_DATA,         "No host data of that type was found")
	};
}

std::string
SocketServer::getLastError() const
{
	WinSockError target(m_lastError, 0);
	const WinSockError* end = wsockErrorList+sizeof(wsockErrorList)/sizeof(*wsockErrorList);
	const WinSockError* it = std::lower_bound(wsockErrorList, end, target);
	if (it != end && it->m_code == target.m_code) {
		return it->m_message;
	} else {
		std::ostringstream strm;
		strm << "Unknown socket error (" << target.m_code << ")";
		return strm.str();
	}
}

LRESULT CALLBACK
SocketServer::wndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	struct sockaddr_in peer;
	if (uMsg == UWM_SOCK) {
		if (WSAGETSELECTERROR(lParam))
			return 0;
		std::map<SOCKET, SocketServer*>::iterator fit = m_sSocketMap.find(wParam);
		if (fit == m_sSocketMap.end())
			return 0;
		int addrlen = sizeof(peer);
		SOCKET newsock = accept(fit->second->m_socket, (struct sockaddr*)&peer, &addrlen);
		if (newsock != INVALID_SOCKET) {
			fit->second->m_pListener->incomingConnection(newsock);
		}
		return 0;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

