/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gzipcompressor.h"
#include <cstddef>
#include <cstring>

GzipCompressor::GzipCompressor()
:	m_isError(false),
	m_isEOF(false),
	m_isFirstBlock(true),
	m_dataLength(0),
	m_dataCRC32(0),
	m_unsyncedSize(0),
	m_outBufPos(m_outBuf+10) // 10 is the header size
{
	static const uint8_t header[10] = {
		0x1f, 0x8b,             // format id
		0x08,                   // compression: deflate
		0x00,                   // flags
		0x00, 0x00, 0x00, 0x00, // mtime: unknown
		0x00,                   // extra flags
		0xff                    // OS: unknown
	};
	std::memcpy(m_outBuf, header, sizeof(header));
	initStream();
}

void
GzipCompressor::reset()
{
	m_isError = false;
	m_isEOF = false;
	m_isFirstBlock = true;
	m_dataLength = 0;
	m_dataCRC32 = 0;
	m_unsyncedSize = 0;
	m_outBufPos = m_outBufPos+10;
	deflateReset(&m_strm);
}

void
GzipCompressor::consumeDataChunk(const uint8_t* begin, const uint8_t* end)
{
	if (begin == end) {
		return;
	}
	if (m_isError || m_isEOF) {
		m_isError = true;
		m_onErrorSignal();
		return;
	}
	m_strm.next_in = const_cast<Bytef*>(begin);
	m_strm.avail_in = end-begin;
	m_dataLength += m_strm.avail_in;
	m_dataCRC32 = crc32(m_dataCRC32, begin, m_strm.avail_in);
	int flush = 0;
	if ((m_unsyncedSize += m_strm.avail_in) >= SYNC_INTERVAL) {
		flush = Z_SYNC_FLUSH;
		m_unsyncedSize = 0;
	}
	int res;
	do {
		m_strm.next_out = m_outBufPos;
		m_strm.avail_out = m_outBuf+sizeof(m_outBuf)-8-m_outBufPos;
		res = deflate(&m_strm, flush);
		if (m_strm.next_out != m_outBufPos) { // if new output is available
			if (m_isFirstBlock) {
				m_onDataOutputSignal(m_outBuf, m_strm.next_out);
				m_isFirstBlock = false;
			} else {
				m_onDataOutputSignal(m_outBufPos, m_strm.next_out);
			}
		}
		m_outBufPos = m_strm.avail_out==0 ? m_outBuf+10 : m_strm.next_out;
	} while (res == Z_OK && m_strm.avail_out == 0);
	if (res != Z_OK && res != Z_BUF_ERROR) {
		m_isError = true;
		m_onErrorSignal();
	}
}

void
GzipCompressor::consumeEOF()
{
	if (m_isEOF) {
		return;
	} else {
		m_isEOF = true;
	}
	if (m_isError) {
		m_onErrorSignal();
		return;
	}
	m_strm.next_in = 0;
	m_strm.avail_in = 0;
	int res;
	do {
		m_strm.next_out = m_outBufPos;
		m_strm.avail_out = m_outBuf+sizeof(m_outBuf)-8-m_outBufPos;
		res = deflate(&m_strm, Z_FINISH);
		if (res == Z_STREAM_END) {
			writeFooter(m_strm.next_out);
			m_onDataOutputSignal(m_isFirstBlock ? m_outBuf : m_outBufPos, m_strm.next_out+8);
			return;
		}
		if (m_strm.next_out != m_outBufPos) { // if new output is available
			if (m_isFirstBlock) {
				m_onDataOutputSignal(m_outBuf, m_strm.next_out);
				m_isFirstBlock = false;
			} else {
				m_onDataOutputSignal(m_outBufPos, m_strm.next_out);
			}
			m_outBufPos = m_strm.avail_out==0 ? m_outBuf+10 : m_strm.next_out;
		}
	} while (res == Z_OK && m_strm.avail_out == 0);
	m_isError = true;
	m_onErrorSignal();
}

void
GzipCompressor::writeFooter(uint8_t* where)
{
	where[0] = m_dataCRC32;
	where[1] = m_dataCRC32>>8;
	where[2] = m_dataCRC32>>16;
	where[3] = m_dataCRC32>>24;
	where[4] = m_dataLength;
	where[5] = m_dataLength>>8;
	where[6] = m_dataLength>>16;
	where[7] = m_dataLength>>24;
}

void
GzipCompressor::initStream()
{
	m_strm.zalloc = Z_NULL;
	m_strm.zfree = Z_NULL;
	m_strm.opaque = Z_NULL;
	m_strm.next_in = Z_NULL;
	deflateInit2(&m_strm, Z_DEFAULT_COMPRESSION, Z_DEFLATED, -MAX_WBITS, 8, Z_DEFAULT_STRATEGY);
}

