/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gzipdecompressor.h"
#include <cstddef>
#include <cstring>
#include <algorithm>

GzipDecompressor::GzipDecompressor()
:	m_state(STATE_HEAD),
	m_isError(false),
	m_isEOF(false),
	m_stateOffset(0)
{
	initStream();
}

GzipDecompressor::~GzipDecompressor()
{
	inflateEnd(&m_strm);
}

void
GzipDecompressor::reset()
{
	inflateReset(&m_strm);
	m_state = STATE_HEAD;
	m_isError = false;
	m_isEOF = false;
	m_stateOffset = 0;
}

#define HANDLE_ERROR() { m_isError = true; m_onErrorSignal(); return; }

void
GzipDecompressor::consumeDataChunk(const uint8_t* begin, const uint8_t* end)
{
	if (begin == end) {
		return;
	}
	if (m_isError || m_isEOF) {
		HANDLE_ERROR();
	}
	const uint8_t* cur = begin;
	while (true) {
		switch (m_state) {
			case STATE_DATA: {
				m_strm.next_in = const_cast<Bytef*>(cur);
				m_strm.avail_in = end - cur;
				int res;
				do {
					m_strm.next_out = m_outBuf;
					m_strm.avail_out = sizeof(m_outBuf);
					res = inflate(&m_strm, Z_SYNC_FLUSH);
					if (m_strm.next_out != m_outBuf) {
						std::size_t len = m_strm.next_out - m_outBuf;
						m_dataLength += len;
						m_dataCRC32 = crc32(m_dataCRC32, m_outBuf, len);
						m_onDataOutputSignal(m_outBuf, m_strm.next_out);
					}
				} while (res == Z_OK && m_strm.avail_out == 0);
				/*
				If inflate() returns Z_OK, it always consumes all of the input,
				unless there is not enough space in the output buffer.
				Now if there is indeed not enough space in the output buffer,
				it may still consume all of the input data. In this case
				(avail_in == 0 && avail_out == 0) we can't say for sure if all the
				consumed input has been processed or not. In this case we have to
				call inflate once more (with avail_in == 0 and avail_out != 0) and be
				prepared to handle Z_BUF_ERROR that will be returned if all of the input
				has been processed on the previous call.
				*/
				if (res == Z_OK || res == Z_BUF_ERROR) {
					return;
				} else if (res == Z_STREAM_END) {
					cur = m_strm.next_in;
					m_state = STATE_CRC32_ISIZE;
					inflateReset(&m_strm);
				} else {
					HANDLE_ERROR();
				}
			}
			case STATE_CRC32_ISIZE: {
				std::size_t len;
				std::memcpy(m_footer+m_stateOffset, cur,
					len=std::min(std::size_t(8-m_stateOffset), std::size_t(end-cur)));
				m_stateOffset += len;
				if (m_stateOffset < 8) {
					return;
				}
				uint32_t crc32 = uint32_t(m_footer[0])|(uint32_t(m_footer[1])<<8)
					|(uint32_t(m_footer[2])<<16)|(uint32_t(m_footer[3])<<24);
				uint32_t length = uint32_t(m_footer[4])|(uint32_t(m_footer[5])<<8)
					|(uint32_t(m_footer[6])<<16)|(uint32_t(m_footer[7])<<24);
				if (crc32 != m_dataCRC32) {
					HANDLE_ERROR();
				}
				if (length != m_dataLength) {
					HANDLE_ERROR();
				}
				cur += len;
				m_state = STATE_HEAD;
				m_stateOffset = 0;
			}
			case STATE_HEAD: {
				std::size_t len;
				std::memcpy(m_header+m_stateOffset, cur,
					len=std::min(std::size_t(10-m_stateOffset), std::size_t(end-cur)));
				m_stateOffset += len;
				if (m_stateOffset < 10) {
					return;
				}
				if (m_header[0] != 0x1f || m_header[1] != 0x8b) {
					// signature doesn't match
					HANDLE_ERROR();
				}
				if (m_header[2] != 8) {
					// wrong compression method
					HANDLE_ERROR();
				}
				m_flags = m_header[3];
				if (m_flags & 0xe0) {
					// reserved flags are not clear
					HANDLE_ERROR();
				}
				m_headCRC32 = crc32(0, m_header, 10);
				cur += len;
				m_state = STATE_FEXTRA;
				m_stateOffset = 0;
			}
			case STATE_FEXTRA: {
				if (m_flags & (1<<2)) {
					if (cur == end)
						return;
					if (m_stateOffset == 0) {
						m_headCRC32 = crc32(m_headCRC32, cur, 1);
						m_fExtraLeft = *cur++;
						++m_stateOffset;
						if (cur == end)
							return;
					}
					if (m_stateOffset == 1) {
						m_headCRC32 = crc32(m_headCRC32, cur, 1);
						m_fExtraLeft |= std::size_t(*cur++)<<8;
						++m_stateOffset;
					}
					if (std::size_t(end-cur) < m_fExtraLeft) {
						m_headCRC32 = crc32(m_headCRC32, cur, end-cur);
						m_fExtraLeft -= end-cur;
						return;
					}
					m_headCRC32 = crc32(m_headCRC32, cur, m_fExtraLeft);
					cur += m_fExtraLeft;
					m_stateOffset = 0;
				}
				m_state = STATE_FNAME;
			}
			case STATE_FNAME: {
				if (m_flags & (1<<3)) {
					const uint8_t* endpos = reinterpret_cast<const uint8_t*>(std::memchr(cur, 0, end-cur));
					if (!endpos) {
						m_headCRC32 = crc32(m_headCRC32, cur, end-cur);
						return;
					}
					++endpos;
					m_headCRC32 = crc32(m_headCRC32, cur, endpos-cur);
					cur = endpos;
				}
				m_state = STATE_FCOMMENT;
			}
			case STATE_FCOMMENT: {
				if (m_flags & (1<<4)) {
					const uint8_t* endpos = reinterpret_cast<const uint8_t*>(std::memchr(cur, 0, end-cur));
					if (!endpos) {
						m_headCRC32 = crc32(m_headCRC32, cur, end-cur);
						return;
					}
					++endpos;
					m_headCRC32 = crc32(m_headCRC32, cur, endpos-cur);
					cur = endpos;
				}
				m_state = STATE_FHCRC;
			}
			case STATE_FHCRC: {
				if (m_flags & (1<<1)) {
					if (cur == end)
						return;
					if (m_stateOffset == 0) {
						m_fhCRC = *cur++;
						++m_stateOffset;
						if (cur == end)
							return;
					}
					if (m_stateOffset == 1) {
						m_fhCRC |= uint16_t(*cur++)<<8;
					}
					if (m_fhCRC != m_headCRC32 & 0xFFFF) {
						HANDLE_ERROR();
					}
					m_stateOffset = 0;
				}
				m_state = STATE_DATA;
				m_dataLength = 0;
				m_dataCRC32 = 0;
			}
		}
	}
}

void
GzipDecompressor::consumeEOF()
{
	if (m_isEOF) {
		return;
	} else {
		m_isEOF = true;
	}
	if (m_isError || m_state != STATE_HEAD || m_stateOffset != 0) {
		m_isError = true;
		m_onErrorSignal();
	}
}

void
GzipDecompressor::initStream()
{
	m_strm.zalloc = Z_NULL;
	m_strm.zfree = Z_NULL;
	m_strm.opaque = Z_NULL;
	m_strm.next_in = Z_NULL;
	inflateInit2(&m_strm, -MAX_WBITS);
}

