/*
    BFilter - a web proxy which removes banner ads
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <fstream>
#include <sstream>
#include "lexgen.h"
#include "nfachar.h"
#include "nfacharclass.h"
#include "nfanegatedcharclass.h"
#include "nfaconcatenation.h"
#include "nfaclosure.h"
#include "nfaoption.h"
#include "nfastring.h"
#include "nfaistring.h"
#include "nfaunion.h"
#include "nfaeof.h"

#include "htmllexerdefinitions.h"

int main(int argc, char** argv)
{
	if (argc < 8) {
		std::cerr << "Usage: makelexer <OutputClass> <out.h> <out.cpp> "
			"<Definitions> <definitions.h> <Subclass> <subclass.h>" << std::endl;
		return 1;
	}
	const char* out_class = argv[1];
	const char* out_header = argv[2];
	const char* out_impl = argv[3];
	const char* def_class = argv[4];
	const char* def_header = argv[5];
	const char* subclass = argv[6];
	const char* subclass_header = argv[7];
	
	std::ofstream header_strm(out_header);
	if (header_strm.fail()) {
		std::cerr << "Failed opening " << out_header << " for writing" << std::endl;
	}
	
	std::ofstream impl_strm(out_impl);
	if (impl_strm.fail()) {
		std::cerr << "Failed opening " << out_impl << " for writing" << std::endl;
	}
	
	LexGen lexgen(out_class, subclass);
	
	NfaCharClass space(" \t\r\n");
	NfaCharClass alpha('A', 'Z');
	alpha.addChars('a', 'z');
	NfaCharClass ident = alpha;
	ident.addChars('0', '9');
	ident.addChars("-:");
	NfaNegatedCharClass not_ident('0', '9');
	not_ident.addChars('a', 'z');
	not_ident.addChars('A', 'Z');
	not_ident.addChars("-:");
	NfaNegatedCharClass any;
	NfaEof eof;
	NfaChar dquote('"');
	NfaChar squote('\'');
	NfaChar lbracket('<');
	NfaChar rbracket('>');
	NfaNegatedCharClass not_lbracket("<");
	NfaChar dash('-');
	NfaNegatedCharClass not_dash("-");
	
	NfaString comment_begin("<!--");
	NfaString comment_end("-->");
	
	NfaConcatenation cdata_end(NfaString("</"), alpha);
	cdata_end.addComponent(NfaClosure(ident));
	cdata_end.addComponent(rbracket);
	
	typedef HtmlLexerDefinitions Defs;
	
	/*
	When we enter the INITIAL state, the current token is always empty.
	Actions that do transition to the INITIAL state can't use MORE() or
	START_GROWING() or CONTINUE_GROWING().
	*/
	lexgen.addRule(Defs::INITIAL, NfaString("<!DOCTYPE"), space,
		"BEGIN(INSIDE_DOCTYPE); MORE();"
	).setLazy();
	lexgen.addRule(Defs::INITIAL, lbracket, alpha,
		"BEGIN(TAG_NAME); MORE();"
	).setLazy();
	lexgen.addRule(Defs::INITIAL, NfaString("</"), alpha,
		"BEGIN(CLOSING_TAG_NAME); MORE();"
	).setLazy();
	lexgen.addRule(Defs::INITIAL, comment_begin,
		"BEGIN(INSIDE_COMMENT); MORE();"
	).setLazy();
	lexgen.addRule(Defs::INITIAL, any, "BEGIN(TEXT); IGNORE_MATCH();");
	
	lexgen.addRule(Defs::TEXT, NfaConcatenation(NfaOption(lbracket), NfaClosure(not_lbracket)),
		"BEGIN(INITIAL); obj->processText(tok_begin, tok_end);"
	).setLazyOnBlockEnd();
	
	// When we enter the INSIDE_COMMENT state, the current token already contains some part of the comment
	lexgen.addRule(Defs::INSIDE_COMMENT, NfaConcatenation(NfaClosure(dash), comment_end),
		"BEGIN(AFTER_COMMENT); MORE();"
	).setLazy();
	lexgen.addRule(Defs::INSIDE_COMMENT, any, "BEGIN(INSIDE_COMMENT_NOEND); MORE(); IGNORE_MATCH();");
	lexgen.addRule(Defs::INSIDE_COMMENT, eof, "obj->processText(tok_begin, tok_end);");
	
	lexgen.addRule(Defs::INSIDE_COMMENT_NOEND, NfaConcatenation(NfaClosure(dash), NfaClosure(not_dash)),
		"BEGIN(INSIDE_COMMENT); if (tok_end.isAtEnd())\n\t"
		"obj->processText(tok_begin, tok_end);\n\t"
		"else MORE();"
	).setLazyOnBlockEnd();
	
	// When we enter the AFTER_COMMENT state, the current token still contains some part of the comment
	lexgen.addRule(Defs::AFTER_COMMENT, NfaClosure(not_lbracket),
		"BEGIN(INITIAL); obj->processText(tok_begin, tok_end);"
	).setLazyOnBlockEnd();
	lexgen.addRule(Defs::AFTER_COMMENT, eof, "obj->processText(tok_begin, tok_end);");
	
	// When we enter the CDATA state, the current token may already contain some part of it
	lexgen.addRule(Defs::CDATA, cdata_end,
		"ConstIterator begin(match_begin); begin += 2;\n\t"
		"ConstIterator end(match_end); --end;\n\t"
		"if (obj->isCDATAEnding(begin, end))\n\t"
		"{ BEGIN(INITIAL); IGNORE_MATCH(); obj->processText(tok_begin, tok_end); }\n\t"
		"else MORE();"
	).setLazy();
	lexgen.addRule(Defs::CDATA, comment_begin,
		"BEGIN(CDATA_COMM); MORE(); obj->m_cdataCommentLevel = 1;"
	).setLazy();
	lexgen.addRule(Defs::CDATA, any, "BEGIN(CDATA_NOEND); MORE(); IGNORE_MATCH();");
	lexgen.addRule(Defs::CDATA, eof, "obj->processText(tok_begin, tok_end);");
	
	lexgen.addRule(Defs::CDATA_NOEND, NfaConcatenation(NfaOption(lbracket), NfaClosure(not_lbracket)),
		"BEGIN(CDATA); if (tok_end.isAtEnd())\n\t"
		"obj->processText(tok_begin, tok_end);\n\t"
		"else MORE();"
	).setLazyOnBlockEnd();
	
	lexgen.addRule(Defs::CDATA_COMM, NfaConcatenation(NfaClosure(any), comment_end),
		"MORE(); if (--obj->m_cdataCommentLevel == 0) BEGIN(CDATA);"
	).setLazy();
	lexgen.addRule(Defs::CDATA_COMM, NfaConcatenation(NfaClosure(any), comment_begin),
		"MORE(); ++obj->m_cdataCommentLevel;"
	).setLazy();
	lexgen.addRule(Defs::CDATA_COMM, NfaUnion(any, NfaEof()),
		"BEGIN(CDATA); MORE(); IGNORE_MATCH();"
	);
	
	lexgen.addRule(Defs::INSIDE_DOCTYPE, NfaConcatenation(NfaClosure(any), rbracket),
		"BEGIN(INITIAL); obj->processDocType(tok_begin, tok_end);"
	).setLazy();
	lexgen.addRule(Defs::INSIDE_DOCTYPE, NfaConcatenation(NfaClosure(any), NfaChar('[')),
		"BEGIN(INSIDE_DOCTYPE_INTERNAL_SUBSET); MORE();"
	).setLazy();
	
	lexgen.addRule(Defs::INSIDE_DOCTYPE_INTERNAL_SUBSET, NfaConcatenation(NfaClosure(any), NfaChar(']')),
		"BEGIN(INSIDE_DOCTYPE); MORE();"
	).setLazy();
	
	lexgen.addRule(Defs::TAG_NAME, NfaClosure(ident),
		"BEGIN(INSIDE_TAG); MORE(); obj->processOpeningTagName(match_begin, tok_end);"
	);
	
	lexgen.addRule(Defs::CLOSING_TAG_NAME, NfaClosure(ident),
		"BEGIN(INSIDE_CLOSING_TAG); MORE(); obj->processClosingTagName(match_begin, tok_end);"
	);
	
	lexgen.addRule(Defs::INSIDE_TAG, NfaConcatenation(alpha, NfaClosure(ident)),
		"BEGIN(ATTR_EQUAL);\n\t"
		"if (obj->processAttrName(match_begin, tok_end))\n\t"
		"MORE();\n\t"
		"else { BEGIN(INITIAL); CONSUME_NOTHING(); }"
	);
	lexgen.addRule(Defs::INSIDE_TAG, NfaString("/>"),
		"BEGIN(INITIAL); obj->processOpeningTag(tok_begin, tok_end, true);"
	);
	lexgen.addRule(Defs::INSIDE_TAG, rbracket,
		"BEGIN(obj->isCDATAStarting() ? CDATA : INITIAL);\n\t"
		"obj->processOpeningTag(tok_begin, tok_end, false);"
	);
	lexgen.addRule(Defs::INSIDE_TAG, any, "MORE();");
	
	lexgen.addRule(Defs::ATTR_EQUAL, NfaConcatenation(NfaClosure(space), NfaChar('=')).addComponent(NfaClosure(space)),
		"BEGIN(ATTR_VALUE); MORE();"
	);
	
	lexgen.addRule(Defs::ATTR_EQUAL, any,
		"BEGIN(INSIDE_TAG); MORE(); IGNORE_MATCH();"
	);
	
	lexgen.addRule(Defs::ATTR_VALUE, dquote, "BEGIN(ATTR_VALUE_DQ); MORE();");
	lexgen.addRule(Defs::ATTR_VALUE, squote, "BEGIN(ATTR_VALUE_SQ); MORE();");
	lexgen.addRule(Defs::ATTR_VALUE, any, "BEGIN(ATTR_VALUE_NQ); MORE(); IGNORE_MATCH();");
	
	lexgen.addRule(Defs::ATTR_VALUE_NQ, NfaClosure(NfaNegatedCharClass(" \t\n\r>")),
		"BEGIN(INSIDE_TAG); MORE(); obj->processAttrValue(match_begin, tok_end);"
	);
	
	lexgen.addRule(Defs::ATTR_VALUE_DQ, NfaClosure(NfaNegatedCharClass("\"")), dquote,
		"BEGIN(INSIDE_TAG); MORE(); obj->processAttrValue(match_begin, tok_end);"
	);
	lexgen.addRule(Defs::ATTR_VALUE_DQ, any, "BEGIN(ATTR_VALUE_NQ); MORE(); IGNORE_MATCH();");
	
	lexgen.addRule(Defs::ATTR_VALUE_SQ, NfaClosure(NfaNegatedCharClass("'")), squote,
		"BEGIN(INSIDE_TAG); MORE(); obj->processAttrValue(match_begin, tok_end);"
	);
	lexgen.addRule(Defs::ATTR_VALUE_SQ, any, "BEGIN(ATTR_VALUE_NQ); MORE(); IGNORE_MATCH();");
	
	lexgen.addRule(Defs::INSIDE_CLOSING_TAG, rbracket,
		"BEGIN(INITIAL); obj->processClosingTag(tok_begin, tok_end);"
	);
	lexgen.addRule(Defs::INSIDE_CLOSING_TAG, any, "MORE();");
	
	lexgen.writeHeader(header_strm, def_class, def_header);
	lexgen.writeCode(impl_strm, subclass_header);
	return 0;
}

