/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef HTMLNODE_H_
#define HTMLNODE_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "types.h"
#include "intrusiveptr.h"
#include "buffer.h"
#include <cstddef>
#include <string>
#include <map>

class HtmlNode;

typedef IntrusivePtr<HtmlNode> HtmlNodePtr;

class HtmlNode
{
	friend class IntrusivePtr<HtmlNode>;
public:
	enum NodeType {
		DOCUMENT_NODE,
		FRAGMENT_NODE,
		ELEMENT_NODE,
		TEXT_NODE
	};
	typedef std::map<std::string, std::string> AttributeMap;
private:
	HtmlNode(const HtmlNode&);            // forbidden
	
	HtmlNode& operator=(const HtmlNode&); // forbidden
protected:
	HtmlNode();
	
	HtmlNode(HtmlNode* parent); // to be used only by HtmlNodeWithChildren::Sentinel
	
	static void* operator new(std::size_t size) { return ::operator new(size); }
public:
	virtual ~HtmlNode();
	
	virtual HtmlNodePtr clone() const = 0;
	
	virtual NodeType getNodeType() const = 0;
	
	virtual const Buffer& getContent() const = 0;
	
	virtual void setContent(const Buffer& content) = 0;
	
	virtual void appendContent(const Buffer::ConstIterator& begin, const Buffer::ConstIterator& end) = 0;
	
	virtual void appendContent(const uint8_t* begin, const uint8_t* end) = 0;
	
	virtual void appendContent(const std::string& text) = 0;
	
	virtual const std::string& getTagName() const = 0;
	
	virtual void setTagName(const std::string& name) = 0;
	
	virtual int getTagType() const = 0;
	
	virtual void setTagType(int type) = 0;
	
	virtual bool isClosed() const = 0;
	
	virtual void setClosed(bool closed) = 0;
	
	virtual bool hasAdMark() const = 0;
	
	virtual void setAdMark(bool mark) = 0;
	
	virtual const AttributeMap& getRawAttributes() const = 0;
	
	virtual bool hasAttribute(const std::string& name) const = 0;
	
	virtual std::string getRawAttribute(const std::string& name) const = 0;
	
	virtual std::string getAttribute(const std::string& name) const = 0;
	
	virtual void setRawAttribute(const std::string& name, const std::string& value) = 0;
	
	virtual void setAttribute(const std::string& name, const std::string& value) = 0;
	
	virtual void removeAttribute(const std::string& name) = 0;
	
	virtual bool hasChildren() const = 0;
	
	virtual HtmlNode* getFirstChild() = 0;
	
	virtual const HtmlNode* getFirstChild() const = 0;
	
	virtual HtmlNode* getLastChild() = 0;
	
	virtual const HtmlNode* getLastChild() const = 0;
	
	virtual void appendChild(HtmlNode* node) = 0;
	
	virtual void prependChild(HtmlNode* node) = 0;
	
	virtual void removeAllChildren() = 0;
	
	void insertAfter(HtmlNode* node);
	
	void insertBefore(HtmlNode* node);
	
	void detachFromParent();
	
	HtmlNode* getParent() { return m_pParent; }
	
	const HtmlNode* getParent() const { return m_pParent; }
	
	HtmlNode* getNextSibling();
	
	const HtmlNode* getNextSibling() const;
	
	HtmlNode* getPrevSibling();
	
	const HtmlNode* getPrevSibling() const;
protected:
	static void makeSentinelPair(HtmlNode* left, HtmlNode* right);
private:
	void intrusivePtrRef() { ++m_refCounter; }
	
	void intrusivePtrUnref();
private:
	HtmlNode* m_pParent;
	HtmlNode* m_pPrev;
	HtmlNode* m_pNext;
	unsigned m_refCounter;
};

#endif

