/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "htmlprocessor.h"
#include "jsenvironment.h"
#include "htmlelementnode.h"
#include "htmltextnode.h"
#include "httpfetcher.h"
#include "state.h"
#include "stringutils.h"
#include "debug.h"
#include <algorithm>
#include <sstream>

HtmlProcessor::HtmlProcessor(BFilter *owner)
:	m_pOwner(owner),
	m_pResult(0),
	m_nestLevel(0),
	m_contentType(CONTENT_HTML),
	m_foundMovementHooks(false),
	m_pContainingNode(&m_docFragment),
	m_baseURL(""),
	m_scriptsRule(0),
	m_jsengineRule(0),
	m_targetBlankRule(0),
	m_isInsideHead(false),
	m_webbugOpenComment(false)
{
}

HtmlProcessor::HtmlProcessor(HtmlProcessor& parent, ScriptResult* pres)
:	m_pOwner(parent.m_pOwner),
	m_pResult(pres),
	m_nestLevel(parent.m_nestLevel+1),
	m_jsEnv(parent.m_jsEnv),
	m_contentType(parent.m_contentType),
	m_foundMovementHooks(false),
	m_pContainingNode(&m_docFragment),
	m_ptrContainingAnchor(parent.m_ptrContainingAnchor ? parent.m_ptrContainingAnchor->clone() : HtmlNodePtr()),
	m_ptrContainingMap(parent.m_ptrContainingMap ? parent.m_ptrContainingMap->clone() : HtmlNodePtr()),
	m_mapTags(parent.m_mapTags),
	m_imgTagsUsingMap(parent.m_imgTagsUsingMap),
	m_baseURL(parent.m_baseURL),
	m_userAgent(parent.m_userAgent),
	m_scriptsRule(parent.m_scriptsRule),
	m_jsengineRule(parent.m_jsengineRule),
	m_targetBlankRule(parent.m_targetBlankRule),
	m_isInsideHead(parent.m_isInsideHead),
	m_webbugOpenComment(m_webbugOpenComment)
{
}

HtmlProcessor::~HtmlProcessor()
{
}

void
HtmlProcessor::setBaseURL(const URI& url)
{
	m_baseURL = url;
	std::string url_str = url.toString();
	m_scriptsRule = m_pOwner->getState().getRules().getScriptsParam(url_str);
	m_jsengineRule = m_pOwner->getState().getRules().getJSengineParam(url_str);
	m_targetBlankRule = m_pOwner->getState().getRules().getTargetBlankParam(url_str);
}

void
HtmlProcessor::setUserAgent(const std::string& ua)
{
	m_userAgent = ua;	
}

void
HtmlProcessor::consumeEOF()
{
	HtmlLexer::consumeEOF();
	outputPendingData();
	if (!m_onload.empty() || !m_onunload.empty()) {
		processOnloadHooks();
	}
}

void
HtmlProcessor::reset()
{
	HtmlLexerCore::reset();
	m_nestLevel = 0;
	m_jsEnv.reset();
	m_contentType = CONTENT_HTML;
	m_ptrCurTag.reset(0);
	m_foundMovementHooks = false;
	m_docFragment.removeAllChildren();
	m_pContainingNode = &m_docFragment;
	m_ptrContainingAnchor.reset(0);
	m_ptrContainingMap.reset(0);
	m_mapTags.clear();
	m_imgTagsUsingMap.clear();
	m_isInsideHead = false;
	m_webbugOpenComment = false;
	std::string().swap(m_onload);
	std::string().swap(m_onunload);
	//m_baseURL = URI("");
	//m_userAgent = "";
	//m_scriptsRule = 0;
	//m_jsengineRule = 0;
	//m_targetBlankRule = 0;
}

void
HtmlProcessor::processDocType(const ConstIterator& begin, const ConstIterator& end)
{
	m_contentType = (tokenAsString(begin, end).find("XHTML") != std::string::npos)
		? CONTENT_XHTML : CONTENT_HTML;
	processText(begin, end);
}

void
HtmlProcessor::processText(const ConstIterator& begin, const ConstIterator& end)
{
	if (m_docFragment.hasChildren()) {
		tagTextAppend(m_pContainingNode, begin, end);
	} else {
		outputData(begin, end);
	}
}

void
HtmlProcessor::processText(const std::string& text)
{
	if (m_docFragment.hasChildren()) {
		tagTextAppend(m_pContainingNode, text);
	} else {
		outputData(text);
	}
}

void
HtmlProcessor::processOpeningTagName(const ConstIterator& begin, const ConstIterator& end)
{
	m_curTagName = StringUtils::toLowerStr(begin, end);
	
	if (m_ptrContainingAnchor && tagTerminatesAnchor(m_curTagName)) {
		m_ptrContainingAnchor.reset(0);
	}
	
	TagType type = getTagType(m_curTagName);
	if (type != TAG_OTHER || m_foundMovementHooks) {
		m_ptrCurTag = HtmlElementNode::create(m_curTagName, false);
		m_ptrCurTag->setTagType(type);
	} else {
		m_ptrCurTag.reset(0);
	}
}

void
HtmlProcessor::processOpeningTag(const ConstIterator& begin, const ConstIterator& end, bool explicit_empty)
{
	if (!m_ptrCurTag) {
		processText(begin, end);
		return;
	}
	
	bool modified = false;
	if (m_foundMovementHooks) {
		modified = processMovementHooks(m_ptrCurTag.get());
		m_foundMovementHooks = false;
	}
	
	typedef bool (HtmlProcessor::*FuncPtr)(HtmlNode*, bool);
	static const FuncPtr func_table[] = {
		&HtmlProcessor::processOpeningOther,    // TAG_OTHER
		&HtmlProcessor::processOpeningA,        // TAG_A
		&HtmlProcessor::processOpeningArea,     // TAG_AREA
		&HtmlProcessor::processOpeningBase,     // TAG_BASE
		&HtmlProcessor::processOpeningBody,     // TAG_BODY
		&HtmlProcessor::processOpeningEmbed,    // TAG_EMBED
		&HtmlProcessor::processOpeningForm,     // TAG_FORM
		&HtmlProcessor::processOpeningFrameset, // TAG_FRAMESET
		&HtmlProcessor::processOpeningHead,     // TAG_HEAD
		&HtmlProcessor::processOpeningIframe,   // TAG_IFRAME
		&HtmlProcessor::processOpeningImg,      // TAG_IMG
		&HtmlProcessor::processOpeningMap,      // TAG_MAP
		&HtmlProcessor::processOpeningObject,   // TAG_OBJECT
		&HtmlProcessor::processOpeningParam,    // TAG_PARAM
		&HtmlProcessor::processOpeningScript    // TAG_SCRIPT
	};
	bool taken_care = (this->*func_table[m_ptrCurTag->getTagType()])(m_ptrCurTag.get(), explicit_empty);
	
	if (!taken_care) {
		if (modified) {
			processOpeningTagAsText(m_ptrCurTag.get(), explicit_empty);
		} else {
			processText(begin, end);
		}
	}
}

void
HtmlProcessor::processClosingTagName(const ConstIterator& begin, const ConstIterator& end)
{
	m_closingTagName = StringUtils::toLowerStr(begin, end);
}

void
HtmlProcessor::processClosingTag(const ConstIterator& begin, const ConstIterator& end)
{
	TagType type = getTagType(m_closingTagName);
	
	typedef void (HtmlProcessor::*FuncPtr)(const ConstIterator&, const ConstIterator&);
	static const FuncPtr func_table[] = {
		&HtmlProcessor::processClosingOther,  // TAG_OTHER
		&HtmlProcessor::processClosingA,      // TAG_A
		&HtmlProcessor::processClosingOther,  // TAG_AREA
		&HtmlProcessor::processClosingOther,  // TAG_BASE
		&HtmlProcessor::processClosingBody,   // TAG_BODY
		&HtmlProcessor::processClosingOther,  // TAG_EMBED
		&HtmlProcessor::processClosingOther,  // TAG_FORM
		&HtmlProcessor::processClosingOther,  // TAG_FRAMESET
		&HtmlProcessor::processClosingHead,   // TAG_HEAD
		&HtmlProcessor::processClosingIframe, // TAG_IFRAME
		&HtmlProcessor::processClosingOther,  // TAG_IMG
		&HtmlProcessor::processClosingMap,    // TAG_MAP
		&HtmlProcessor::processClosingObject, // TAG_OBJECT
		&HtmlProcessor::processClosingOther,  // TAG_PARAM
		&HtmlProcessor::processClosingScript  // TAG_SCRIPT
	};
	(this->*func_table[type])(begin, end);
}

bool
HtmlProcessor::processAttrName(const ConstIterator& begin, const ConstIterator& end)
{
	static const char onmouseover[] = { 'o','n','m','o','u','s','e','o','v','e','r' };
	static const char onmouseout[] = { 'o','n','m','o','u','s','e','o','u','t' };
	
	if (m_ptrCurTag) {
		m_curAttrName = StringUtils::toLowerStr(begin, end);
	} else {
		if (StringUtils::ciEquals(begin, end, onmouseover, onmouseover+sizeof(onmouseover)) ||
		    StringUtils::ciEquals(begin, end, onmouseout, onmouseout+sizeof(onmouseout))) {
			m_foundMovementHooks = true;
			return false;
		}
	}
	return true;
}

void
HtmlProcessor::processAttrValue(const ConstIterator& begin, const ConstIterator& end)
{
	if (m_ptrCurTag) {
		std::string value(tokenAsString(begin, end));
		if (value.find('"') != std::string::npos) {
			value = StringUtils::replace("\"", "&quot;", value);
		}
		m_ptrCurTag->setRawAttribute(m_curAttrName, value);
	}
}

bool
HtmlProcessor::isCDATAStarting()
{
	return m_curTagName == "script" || m_curTagName == "style";
}

bool
HtmlProcessor::isCDATAEnding(const ConstIterator& begin, const ConstIterator& end) {
	return StringUtils::ciEquals(m_curTagName, begin, end);
}

bool
HtmlProcessor::processOpeningOther(HtmlNode* node, bool explicit_empty)
{
	return false;
}

bool
HtmlProcessor::processOpeningA(HtmlNode* anchor, bool explicit_empty)
{
	if (!explicit_empty) {
		m_ptrContainingAnchor.reset(anchor);
	}
	if (m_ptrContainingMap) {
		m_ptrContainingMap->appendChild(anchor);
	}
	
	if (m_targetBlankRule == 1 && anchor->getAttribute("target") == "_blank") {
		HtmlNodePtr replacement = anchor->clone();
		replacement->removeAttribute("target");
		processOpeningTagAsText(replacement.get(), explicit_empty);
		return true;
	}
	
	return false;
}

bool
HtmlProcessor::processOpeningArea(HtmlNode* area, bool explicit_empty)
{
	if (m_ptrContainingMap) {
		m_ptrContainingMap->appendChild(area);
	}
	return false;
}

bool
HtmlProcessor::processOpeningBase(HtmlNode* base, bool explicit_empty)
{
	setBaseURL(URI(m_baseURL, URI(base->getAttribute("href"))));
	return false;
}

bool
HtmlProcessor::processOpeningBody(HtmlNode* body, bool explicit_empty)
{
	m_isInsideHead = false;
	if (!m_onload.empty() || !m_onunload.empty()) {
		// second <body>?
		return false;
	}
	bool modified = false;
	m_onload = body->getAttribute("onload");
	m_onunload = body->getAttribute("onunload");
	if (!m_onload.empty()) {
		body->setAttribute("onload", "if (typeof bfilter_onload_ok != 'undefined') { "+m_onload+" }");
		modified = true;
	}
	if (!m_onunload.empty()) {
		body->setAttribute("onunload", "if (typeof bfilter_onunload_ok != 'undefined') { "+m_onunload+" }");
		modified = true;
	}
	if (modified) {
		processOpeningTagAsText(body, explicit_empty);
		return true;
	}
	return false;
}

bool
HtmlProcessor::processOpeningEmbed(HtmlNode* embed, bool explicit_empty)
{
	static const char flash_mimetype[] = { 'a','p','p','l','i','c','a','t','i','o','n',
		'/','x','-','s','h','o','c','k','w','a','v','e','-','f','l','a','s','h' };
	static const char false_str[] = { 't','r','u','e' };
	
	if (!embed->hasAttribute("type") ||
	    StringUtils::ciEquals(embed->getAttribute("type"),
	    flash_mimetype, flash_mimetype+sizeof(flash_mimetype))) {
		std::string url_str = StringUtils::trimStr(embed->getAttribute("src"));
		URI url(m_baseURL, URI(url_str));
		std::string width_str = embed->getAttribute("width");
		std::string height_str = embed->getAttribute("height");
		int width = StringUtils::toNumber<int>(width_str);
		int height = StringUtils::toNumber<int>(height_str);
		if (m_pOwner->getState().getConfig().getNoFlash()) {
			if (m_pResult) {
				m_pResult->setType(ScriptResult::T_FLASH_AS_IMG);
				m_pResult->setObjectURL(url_str);
				m_pResult->setDimentions(width, height);
				stop();
				return true;
			}
			HtmlNodePtr anchor(HtmlElementNode::create("a", true));
			anchor->setAttribute("href", url_str);
			anchor->setRawAttribute("alt", "A replacement for a flash file");
			HtmlNodePtr img(HtmlElementNode::create("img", true));
			anchor->appendChild(img.get());
			URI subst_url(url);
			subst_url.setAbsoluteRawPath("bf.intercept-request/image_substitute");
			subst_url.setRawQuery("w="+width_str+"&h="+height_str);
			img->setRawAttribute("src", subst_url.toString());
			img->setRawAttribute("border", "0");
			if (width > 0 && height > 0) {
				img->setRawAttribute("width", width_str);
				img->setRawAttribute("height", height_str);
			}
			processIndependentNode(anchor.get());
			return true;
		}
		int ad_rule = m_pOwner->getState().getRules().getAdParam(url.toString());
		bool external_url = isExternalURL(url);
		bool suspicious_url = isSuspiciousURL(url);
		bool is_loop = embed->hasAttribute("loop")
			&& !StringUtils::ciEquals(embed->getAttribute("loop"),
			false_str, false_str+sizeof(false_str));
		bool is_ad = is_loop || isCommonAdSize(width, height);
		bool could_be_ad_size = couldBeAdSize(width, height);
		if (!is_ad && m_pResult && (external_url || suspicious_url || could_be_ad_size)) {
			is_ad = true;
		}
		if ((is_ad || ad_rule==2) && ad_rule!=1) {
			if (m_pResult) {
				m_pResult->setType(ScriptResult::T_FLASH);
				m_pResult->setObjectURL(url_str);
				m_pResult->setDimentions(width, height);
				stop();
				return true;
			}
			URI subst_url(url);
			subst_url.setAbsoluteRawPath("bf.intercept-request/flash_substitute");
			std::ostringstream strm;
			strm << "w=" << width << "&h=" << height << "&url=" << URI::escape(url.toString());
			subst_url.setRawQuery(strm.str());
			embed->setAttribute("src", subst_url.toString());
			m_pContainingNode->setAdMark(true);
			processOpeningTagAsText(embed, explicit_empty);
			return true;
		} else if (!m_pResult && (external_url || suspicious_url || could_be_ad_size)) {
			URI subst_url(url);
			subst_url.setAbsoluteRawPath("bf.intercept-request/flash_analyze");
			std::ostringstream strm;
			strm << "w=" << width << "&h=" << height << "&url=" << URI::escape(url.toString());
			subst_url.setRawQuery(strm.str());
			embed->setAttribute("src", subst_url.toString());
			processOpeningTagAsText(embed, explicit_empty);
			return true;
		}
	}
	return false;
}

bool
HtmlProcessor::processOpeningForm(HtmlNode* form, bool explicit_empty)
{
	std::string name = form->getAttribute("name");
	if (name.empty()) {
		name = form->getAttribute("id");
	}
	m_jsEnv->addForm(name, form->getAttribute("action"), form->getAttribute("target"));
	return false;
}

bool
HtmlProcessor::processOpeningFrameset(HtmlNode* frameset, bool explicit_empty)
{
	m_isInsideHead = false;
	bool modified = false;
	std::string onload = frameset->getAttribute("onload");
	std::string onunload = frameset->getAttribute("onunload");
	if (!onload.empty()) {
		if (checkCodeForPopups(onload)) {
			frameset->setRawAttribute("onload", "/* popup detected by BFilter */");
			modified = true;
		}
	}
	if (!onunload.empty()) {
		if (checkCodeForPopups(onunload)) {
			frameset->setRawAttribute("onunload", "/* popup detected by BFilter */");
			modified = true;
		}
	}
	if (modified) {
		processOpeningTagAsText(frameset, explicit_empty);
		return true;
	}
	return false;
}

bool
HtmlProcessor::processOpeningHead(HtmlNode* head, bool explicit_empty)
{
	m_isInsideHead = true;
	return false;
}

bool
HtmlProcessor::processOpeningIframe(HtmlNode* iframe, bool explicit_empty)
{
	docFragmentPutContainer(iframe);
	if (explicit_empty) {
		processClosingIframe(m_emptyBuffer.begin(), m_emptyBuffer.end());
	}
	return true;
}

bool
HtmlProcessor::processOpeningImg(HtmlNode* img, bool explicit_empty)
{
	img->setClosed(true);
	return processImg(img, explicit_empty, m_ptrContainingAnchor.get());
}

bool
HtmlProcessor::processOpeningMap(HtmlNode* map, bool explicit_empty)
{
	m_ptrContainingMap.reset(map);
	m_mapTags.insert(TagsByName::value_type(map->getAttribute("name"), m_ptrContainingMap));
	if (explicit_empty) {
		processClosingMap(m_emptyBuffer.begin(), m_emptyBuffer.end());
	}
	return false;
}

bool
HtmlProcessor::processOpeningObject(HtmlNode* object, bool explicit_empty)
{
	if (explicit_empty) {
		return false;
	}
	docFragmentPutContainer(object);
	return true;
}

bool
HtmlProcessor::processOpeningScript(HtmlNode* script, bool explicit_empty)
{
	docFragmentPutContainer(script);
	if (explicit_empty) {
		processClosingScript(m_emptyBuffer.begin(), m_emptyBuffer.end());
	}
	return true;
}

bool
HtmlProcessor::processOpeningParam(HtmlNode* param, bool explicit_empty)
{
	if (explicit_empty) {
		param->setClosed(true);
	}
	docFragmentPutElement(param);
	return true;
}

void
HtmlProcessor::processClosingOther(const ConstIterator& begin, const ConstIterator& end)
{
	processText(begin, end);
}

void
HtmlProcessor::processClosingA(const ConstIterator& begin, const ConstIterator& end)
{
	m_ptrContainingAnchor.reset(0);
	processText(begin, end);
}

void
HtmlProcessor::processClosingBody(const ConstIterator& begin, const ConstIterator& end)
{
	if (!m_onload.empty() || !m_onunload.empty()) {
		processOnloadHooks();
	}
	processText(begin, end);
}

void
HtmlProcessor::processClosingHead(const ConstIterator& begin, const ConstIterator& end)
{
	m_isInsideHead = false;
	processText(begin, end);
}

void
HtmlProcessor::processClosingIframe(const ConstIterator& begin, const ConstIterator& end)
{
	HtmlNode* iframe = findAncestorOrSelf(m_pContainingNode, "iframe");
	if (!iframe) {
		processText(begin, end);
		return;
	}
	iframe->setClosed(true);
	processUnclosedTags(iframe);
	
	std::string url_str = StringUtils::trimStr(iframe->getAttribute("src"));
	URI url(m_baseURL, URI(url_str));
	int width = StringUtils::toNumber<int>(iframe->getAttribute("width"));
	int height = StringUtils::toNumber<int>(iframe->getAttribute("height"));
	int ad_rule = m_pOwner->getState().getRules().getAdParam(url.toString());
	bool is_ad = isExternalURL(url) || iframe->hasAdMark()
		|| (isSuspiciousURL(url) && isCommonAdSize(width, height));
	if ((is_ad || ad_rule==2) && ad_rule!=1) {
		if (m_pResult) {
			m_pResult->setType(ScriptResult::T_IFRAME);
			m_pResult->setObjectURL(url_str);
			m_pResult->setDimentions(width, height);
			stop();
			return;
		}
		std::string width_str = StringUtils::fromNumber(width);
		std::string height_str = StringUtils::fromNumber(height);
		HtmlNodePtr anchor(HtmlElementNode::create("a", true));
		anchor->setAttribute("href", url_str);
		anchor->setRawAttribute("alt", "A replacement for an iframe ad");
		HtmlNodePtr img(HtmlElementNode::create("img", true));
		anchor->appendChild(img.get());
		URI subst_url(url);
		subst_url.setAbsoluteRawPath("bf.intercept-request/image_substitute");
		subst_url.setRawQuery("w="+width_str+"&h="+height_str);
		img->setRawAttribute("src", subst_url.toString());
		img->setRawAttribute("border", "0");
		if (width > 0 && height > 0) {
			img->setRawAttribute("width", width_str);
			img->setRawAttribute("height", height_str);
		}
		iframe->getParent()->setAdMark(true);
		processDocFragmentNodeReplacement(iframe, anchor.get());
		return;
	}
	processDocFragmentNode(iframe);
}

void
HtmlProcessor::processClosingMap(const ConstIterator& begin, const ConstIterator& end)
{
	if (m_ptrContainingMap) {
		HtmlNode* anchor = findMostEvilMapLink(m_ptrContainingMap.get());
		typedef std::pair<MultiTagsByName::iterator, MultiTagsByName::iterator> IterPair;
		IterPair p = m_imgTagsUsingMap.equal_range(m_ptrContainingMap->getAttribute("name"));
		for (MultiTagsByName::iterator iter = p.first; iter != p.second; ++iter) {
			if (!iter->second->getParent()) {
				// this means it has been removed from the tree (together with its parent)
				continue;
			}
			processImg(iter->second.get(), m_contentType==CONTENT_XHTML, anchor);
		}
		m_imgTagsUsingMap.erase(p.first, p.second);
		
		m_ptrContainingMap.reset(0);
	}
	processText(begin, end);
}

void
HtmlProcessor::processClosingObject(const ConstIterator& begin, const ConstIterator& end)
{
	HtmlNode* object = findAncestorOrSelf(m_pContainingNode, "object");
	if (!object) {
		processText(begin, end);
		return;
	}
	object->setClosed(true);
	processUnclosedTags(object);
	
	static const char class_id_str[] = {
		'c','l','s','i','d',':','d','2','7','c','d','b','6','e',
		'-','a','e','6','d','-','1','1','c','f','-','9','6','b','8',
		'-','4','4','4','5','5','3','5','4','0','0','0','0'
	};
	static const char loop_str[] = { 'l', 'o', 'o', 'p' };
	static const char true_str[] = { 't', 'r', 'u', 'e' };
	static const char movie_str[] = { 'm','o','v','i','e' };
	
	if (StringUtils::ciEquals(object->getAttribute("classid"),
	    class_id_str, class_id_str+sizeof(class_id_str))) {
		std::string width_str = object->getAttribute("width");
		std::string height_str = object->getAttribute("height");
		int width = StringUtils::toNumber<int>(width_str);
		int height = StringUtils::toNumber<int>(height_str);
		bool is_loop = false;
		std::string url_str;
		HtmlNode* movie_node = 0;
		for (HtmlNode* child = object->getFirstChild(); child; child = child->getNextSibling()) {
			if (child->getTagName() != "param") {
				continue;
			}
			if (StringUtils::ciEquals(child->getAttribute("name"),
			    loop_str, loop_str+sizeof(loop_str))) {
				is_loop = StringUtils::ciEquals(child->getAttribute("value"),
					true_str, true_str+sizeof(true_str));
			} else if (StringUtils::ciEquals(child->getAttribute("name"),
			           movie_str, movie_str+sizeof(movie_str))) {
				url_str = StringUtils::trimStr(child->getAttribute("value"));
				movie_node = child;
			}
		}
		if (movie_node) {
			URI url(m_baseURL, URI(url_str));
			if (m_pOwner->getState().getConfig().getNoFlash()) {
				if (m_pResult) {
					m_pResult->setType(ScriptResult::T_FLASH_AS_IMG);
					m_pResult->setObjectURL(url_str);
					m_pResult->setDimentions(width, height);
					stop();
					return;
				}
				HtmlNodePtr anchor(HtmlElementNode::create("a", true));
				anchor->setAttribute("href", url_str);
				anchor->setRawAttribute("alt", "A replacement for a flash file");
				HtmlNodePtr img(HtmlElementNode::create("img", true));
				anchor->appendChild(img.get());
				URI subst_url(url);
				subst_url.setAbsoluteRawPath("bf.intercept-request/image_substitute");
				subst_url.setRawQuery("w="+width_str+"&h="+height_str);
				img->setRawAttribute("src", subst_url.toString());
				img->setRawAttribute("border", "0");
				if (width > 0 && height > 0) {
					img->setRawAttribute("width", width_str);
					img->setRawAttribute("height", height_str);
				}
				processDocFragmentNodeReplacement(object, anchor.get());
				return;
			}
			int ad_rule = m_pOwner->getState().getRules().getAdParam(url.toString());
			bool external_url = isExternalURL(url);
			bool suspicious_url = isSuspiciousURL(url);
			bool is_ad = is_loop || isCommonAdSize(width, height);
			bool could_be_ad_size = couldBeAdSize(width, height);
			if (!is_ad && m_pResult && (external_url || suspicious_url || could_be_ad_size)) {
				is_ad = true;
			}
			if ((is_ad || ad_rule==2) && ad_rule!=1) {
				if (m_pResult) {
					m_pResult->setType(ScriptResult::T_FLASH);
					m_pResult->setObjectURL(url_str);
					m_pResult->setDimentions(width, height);
					stop();
					return;
				}
				URI subst_url(url);
				subst_url.setAbsoluteRawPath("bf.intercept-request/flash_substitute");
				std::ostringstream strm;
				strm << "w=" << width << "&h=" << height << "&url=" << URI::escape(url.toString());
				subst_url.setRawQuery(strm.str());
				movie_node->setAttribute("value", subst_url.toString());
				object->getParent()->setAdMark(true);
			} else if (!m_pResult && (external_url || suspicious_url || could_be_ad_size)) {
				URI subst_url(url);
				subst_url.setAbsoluteRawPath("bf.intercept-request/flash_analyze");
				std::ostringstream strm;
				strm << "w=" << width << "&h=" << height << "&url=" << URI::escape(url.toString());
				subst_url.setRawQuery(strm.str());
				movie_node->setAttribute("value", subst_url.toString());
			}
		}
	}
	processDocFragmentNode(object);
}

namespace {
	class JSPopupCounter : public SigC::Object
	{
	public:
		JSPopupCounter(JSEnvironment& je) : m_popupCounter(0)
		{
			je.onPopup().connect(SigC::slot(*this, &JSPopupCounter::processPopup));
		}
		int getPopupCount() const { return m_popupCounter; }
	private:
		void processPopup() { ++m_popupCounter; }
		int m_popupCounter;
	};
	
	class JSOutputCollector : public JSPopupCounter
	{
	public:
		JSOutputCollector(JSEnvironment& je) : JSPopupCounter(je)
		{
			je.onDataOutput().connect(SigC::slot(*this, &JSOutputCollector::processDataChunk));
		}
		Buffer& getBuffer() { return m_buffer; }
	private:
		void processDataChunk(const uint8_t* begin, const uint8_t* end) { m_buffer.appendData(begin, end); }
		Buffer m_buffer;
	};
}

void
HtmlProcessor::processClosingScript(const ConstIterator& begin, const ConstIterator& end)
{
	HtmlNode* script = m_pContainingNode;
	if (!m_pOwner->getState().isScriptFilteringEnabled() || script->getTagName() != "script") {
		processText(begin, end);
		return;
	}
	
	script->setClosed(true);
	
	static const char js_str[] = { 'j','a','v','a','s','c','r','i','p','t' };
	static const char js_mimetype[] = { 't','e','x','t','/','j','a','v','a','s','c','r','i','p','t' };
	
	std::string script_code;
	std::string script_src;
	int ad_rule = 0;
	bool script_is_js = (!script->hasAttribute("type") ||
		StringUtils::ciEquals(script->getAttribute("type"),
			js_mimetype, js_mimetype+sizeof(js_mimetype))) &&
		(!script->hasAttribute("language") ||
		StringUtils::ciStartsWith(script->getAttribute("language"),
			js_str, js_str+sizeof(js_str)));
	bool script_external = script->hasAttribute("src");
	if (!script_external) {
		DEBUGLOG("processing inline script");
		if (script->hasChildren()) {
			// A script node has either zero or one child. Can't be more because
			// script content is CDATA (no child tags allowed) and because of
			// the way tagTextAppend() works.
			const Buffer& buf = script->getFirstChild()->getContent();
			Buffer::appendToString(script_code, buf.begin(), buf.end());
		}
	} else {
		script_src = StringUtils::trimStr(script->getAttribute("src"));
		URI url(m_baseURL, URI(script_src));
		std::string url_str = url.toString();
		ad_rule = m_pOwner->getState().getRules().getAdParam(url_str);
		// We might have checked if ad_rule == 2 here, and if it is,
		// remove the script without fetching or processing it. That would be
		// a bad idea, because other scripts that generate ads may depend
		// on this one, so if we don't process it, those ads won't be detected.
		int script_rule = m_pOwner->getState().getRules().getScriptsParam(url_str);
		if (m_pOwner->getState().getRules().getJSengineParam(url_str) == 1
		    || script_rule == 3 || script_rule == 4) {
			// rules 3 and 4 check script contents
			DEBUGLOG("Fetching external script from " << script_src);
			try {
				HttpFetcher fetcher(m_pOwner);
				for (int i = 0;; ++i) {
					if (i >= MAX_SCRIPT_FETCH_REDIRECTS) {
						throw HttpFetcher::Exception(
							"too many redirects ("+StringUtils::fromNumber(i)+")");
					}
					HttpRequest request("GET", url);
					request.headers().setHeader(HttpHeader("User-Agent", m_userAgent));
					request.headers().setHeader(HttpHeader("Referrer", m_baseURL.toString()));
					request.headers().setHeader(HttpHeader("Accept", "*" "/" "*"));
					HttpResponse response = fetcher.request(request);
					int rcode = response.getStatusCode();
					if ((rcode >= 301 && rcode <= 303) || rcode == 307) {
						url = URI(m_baseURL, URI(response.headers().getHeader("Location").getValue()));
						continue;
					} else if (rcode != 200) {
						throw HttpFetcher::Exception(StringUtils::fromNumber(rcode)
							+" "+response.getStatusMessage());
					} else {
						script_code = response.body();
						break;
					}
				}
			} catch (HttpFetcher::Exception& e) {
				DEBUGLOG("Fetching the script failed: " << e.getMessage());
				std::ostringstream strm;
				strm << "<!--\r\nBFilter: Error fetching the script from "
					<< url.toString() << "\r\nReason: " << e.getMessage() << "\r\n-->";
				if (m_pResult) {
					m_pResult->setType(ScriptResult::T_SCRIPT_FETCHING_FAILED);
					stop();
					return;
				}
				processDocFragmentNodeReplacement(script, strm.str());
				return;
			}
			DEBUGLOG("External script fetched successfully");
		}
	}
	
	if (script_is_js && m_jsengineRule == 1 && m_nestLevel < MAX_NEST_LEVEL && ad_rule != 1) {
		std::string language = script->getAttribute("language");
		std::string js_version;
		if (StringUtils::ciStartsWith(language, js_str, js_str+sizeof(js_str))) {
			js_version.append(language, sizeof(js_str), std::string::npos);
		}
		
		JSOutputCollector js_collector(*m_jsEnv);
		m_jsEnv->executeScript(script_code, js_version);
		
		if (m_webbugOpenComment && StringUtils::trimStr(
		    js_collector.getBuffer().begin(), js_collector.getBuffer().end()) == "-->") {
			m_webbugOpenComment = false;
			processDocFragmentNodeReplacement(script,
				"<!--\r\nScript removed by BFilter (jsengine -> webbug tail)\r\n-->");
			return;
		}
		
		if (js_collector.getPopupCount() > 0) {
			DEBUGLOG("The script opens a popup");
			if (StringUtils::endsWith(js_collector.getBuffer().begin(),
			    StringUtils::rtrim(js_collector.getBuffer().begin(),
			    js_collector.getBuffer().end()), "<!--")) {
				m_webbugOpenComment = true;
			}
			if (m_pResult) {
				m_pResult->setType(ScriptResult::T_POPUP);
				stop();
				return;
			} else {
				processDocFragmentNodeReplacement(script,
					"<!--\r\nScript removed by BFilter (jsengine -> popup)\r\n-->");
				return;
			}
		}
		
		if (ad_rule == 2) {
			if (m_pResult) {
				m_pResult->setType(ScriptResult::T_SCRIPT_RULES);
				stop();
				return;
			}
			processDocFragmentNodeReplacement(script,
				"<!--\r\nScript removed by BFilter (rule)\r\n-->");
			return;
		}
		
		DEBUGLOG("Launching subprocessor to examine the output of the script (nest_level=" << m_nestLevel+1 << ")");
		ScriptResult res;
		ScriptResult* pres = m_pResult ? m_pResult : &res;
		HtmlProcessor child_proc(*this, pres);
		Buffer::ConstRangeIterator range_iter(js_collector.getBuffer().begin(), js_collector.getBuffer().end());
		for (const Buffer::ConstRange* range; (range = range_iter.next()); ) {
			child_proc.consumeDataChunk(range->begin(), range->end());
		}
		child_proc.consumeEOF();
		if (pres->getType() == ScriptResult::T_NOTHING) {
			DEBUGLOG("Subprocessor finishes - no ads found");
			if (script_external) {
				tryInliningExternalScript(script, script_src, script_code, script_is_js);
			}
			processDocFragmentNode(script);
			return;
		} else if (ad_rule == 1) {
			DEBUGLOG("Subprocessor has caught something, but a rule is forcing the negative decision");
			if (script_external) {
				tryInliningExternalScript(script, script_src, script_code, script_is_js);
			}
			processDocFragmentNode(script);
			return;
		}
		
		DEBUGLOG("Subprocessor has caught something: " << pres->typeToString());
		if (StringUtils::endsWith(js_collector.getBuffer().begin(),
		    StringUtils::rtrim(js_collector.getBuffer().begin(),
		    js_collector.getBuffer().end()), "<!--")) {
			m_webbugOpenComment = true;
		}
		if (m_pResult) {
			pres->incNestLevel();
			stop();
			return;
		}
		
		int width = pres->getWidth();
		int height = pres->getHeight();
		bool do_replacement = (!pres->getObjectURL().empty() &&
			(width > 2 || width == 0) && (height > 2 || height == 0));
		std::ostringstream repl_strm;
		repl_strm << "<!--\r\nScript removed by BFilter (jsengine -> ";
		for (int i = pres->getNestLevel(); i > 0; --i) {
			repl_strm << "jsengine -> ";
		}
		repl_strm << pres->typeToString() << ')';
		if (do_replacement) {
			repl_strm << " [replacement follows]";
		}
		repl_strm << "\r\n-->";
		if (do_replacement) {
			if (width == 0)
				width = 10;
			if (height == 0)
				height = 10;
			std::string width_str = StringUtils::fromNumber(width);
			std::string height_str = StringUtils::fromNumber(height);
			std::ostringstream rhtml;
			if (pres->getType() == ScriptResult::T_FLASH) {
				URI subst_url(pres->getObjectURL());
				subst_url.setAbsoluteRawPath("bf.intercept-request/flash_substitute");
				std::ostringstream strm;
				strm << "w=" << width_str << "&h=" << height_str << "&url="
					<< URI::escape(pres->getObjectURL());
				subst_url.setRawQuery(strm.str());
				std::string subst_url_str = subst_url.toString();
				HtmlElementNode object("object", true);
				object.setRawAttribute("classid", "clsid:d27cdb6e-ae6d-11cf-96b8-444553540000");
				object.setRawAttribute("width", width_str);
				object.setRawAttribute("height", height_str);
				HtmlElementNode param_movie("param", true);
				param_movie.setRawAttribute("name", "movie");
				param_movie.setAttribute("value", subst_url_str);
				HtmlElementNode embed("embed", true);
				embed.setRawAttribute("width", width_str);
				embed.setRawAttribute("height", height_str);
				embed.setRawAttribute("type", "application/x-shockwave-flash");
				embed.setAttribute("src", subst_url_str);
				rhtml << openingTagToText(&object, false)
					<< openingTagToText(&param_movie, isXHTML())
					<< openingTagToText(&embed, isXHTML()) << "</object>";
			} else {
				URI subst_url(pres->getObjectURL());
				subst_url.setAbsoluteRawPath("bf.intercept-request/image_substitute");
				std::ostringstream strm;
				strm << "w=" << width_str << "&h=" << height_str;
				subst_url.setRawQuery(strm.str());
				HtmlElementNode anchor("a", true);
				if (pres->getLinkURL().empty()) {
					anchor.setAttribute("href", pres->getObjectURL());
				} else {
					anchor.setAttribute("href", pres->getLinkURL());
				}
				anchor.setRawAttribute("alt", "Ad replacement");
				HtmlElementNode img("img", true);
				img.setAttribute("src", subst_url.toString());
				img.setRawAttribute("border", "0");
				img.setRawAttribute("width", width_str);
				img.setRawAttribute("height", height_str);
				rhtml << openingTagToText(&anchor, false)
					<< openingTagToText(&img, isXHTML()) << "</a>";
			}
			repl_strm << "<script type=\"text/javascript\">\r\n<!--\r\n"
				"document.writeln(unescape('" << URI::escape(rhtml.str())
				<< "'));\r\n//--></script>";
		}
		processDocFragmentNodeReplacement(script, repl_strm.str());
		return;
	} else {
		if (ad_rule != 1 && !allowScript(script_code, script_external && isExternalURL(URI(script_src)))) {
			if (m_pResult) {
				m_pResult->setType(ScriptResult::T_SCRIPT_RULES);
				stop();
				return;
			}
			processDocFragmentNodeReplacement(script,
				"<!--\r\nScript removed by BFilter (rule)\r\n-->");
			return;
		}
	}
	m_webbugOpenComment = false;
	processDocFragmentNode(script);
}

bool
HtmlProcessor::processImg(HtmlNode* img, bool explicit_empty, const HtmlNode* anchor)
{
	std::string imgurl_str = StringUtils::trimStr(img->getAttribute("src"));
	URI imgurl(m_baseURL, URI(imgurl_str));
	int img_ad_rule = m_pOwner->getState().getRules().getAdParam(imgurl.toString());
	bool external_img = isExternalURL(imgurl);
	bool suspicious_img = isSuspiciousURL(imgurl);
	int width = StringUtils::toNumber<int>(img->getAttribute("width"));
	int height = StringUtils::toNumber<int>(img->getAttribute("height"));
	bool size_defined = (width > 0 && height > 0);
	if (width == 1 && height == 1 && img_ad_rule != 1 && (external_img || suspicious_img)) {
		if (m_pResult) {
			m_pResult->setType(ScriptResult::T_WEBBUG);
			m_pResult->setObjectURL(imgurl_str);
			m_pResult->setDimentions(width, height);
			stop();
			return true;
		}
		URI subst_url(imgurl);
		subst_url.setScheme("http");
		subst_url.setAbsoluteRawPath("bf.intercept-request/image_substitute");
		subst_url.setRawQuery("w=1&h=1");
		img->setAttribute("src", subst_url.toString());
		if (HtmlNode* parent = img->getParent()) {
			parent->setAdMark(true);
			processDocFragmentNode(img);
		} else {
			m_pContainingNode->setAdMark(true);
			processOpeningTagAsText(img, explicit_empty);
		}
		return true;
	}
	if (!anchor && !img->getParent()) {
		if (img->hasAttribute("usemap")) {
			std::string mapname = img->getAttribute("usemap");
			if ('#' == mapname[0]) {
				mapname.erase(0, 1);
			}
			TagsByName::iterator it = m_mapTags.find(mapname);
			if (it != m_mapTags.end()) {
				anchor = findMostEvilMapLink(it->second.get());
			} else {
				docFragmentPutElement(img);
				m_imgTagsUsingMap.insert(MultiTagsByName::value_type(mapname, HtmlNodePtr(img)));
				return true;
			}
		}
	}
	if (anchor) {
		std::string aurl_str = StringUtils::trimStr(anchor->getAttribute("href"));
		URI aurl(m_baseURL, URI(aurl_str));
		bool external_link = isExternalURL(aurl);
		bool suspicious_link = isSuspiciousURL(aurl);
		std::string target = anchor->getAttribute("target");
		bool is_ad = false;
		bool need_more_info = false;
		if (size_defined) {
			int res = imageIsAd(img_ad_rule, width, height, target, external_link,
				external_img, suspicious_link, suspicious_img);
			if (res == 1) {
				is_ad = true;
			} else if (res == 0) {
				need_more_info = true;
			}
		} else { // size not defined
			need_more_info = true;
		}
		if (need_more_info) {
			if (m_pResult) {
				// TODO: a better solution is needed here
				if (size_defined) {
					m_pResult->setType(ScriptResult::T_IMG);
					m_pResult->setLinkURL(aurl_str);
					m_pResult->setObjectURL(imgurl_str);
					m_pResult->setDimentions(width, height);
					stop();
				} else {
					int w = size_defined ? width : 311;
					int h = size_defined ? height : 41;
					// 311x41 is not a common ad size, but couldBeAdSize() returns true for it
					if (imageIsAd(img_ad_rule, w, h, target, external_link,
					    external_img, suspicious_link, suspicious_img) == 1) {
						m_pResult->setType(ScriptResult::T_IMG);
						m_pResult->setLinkURL(aurl_str);
						m_pResult->setObjectURL(imgurl_str);
						stop();
					}
				}
				return true;
			}
			if (imgurl.getScheme() == "http") {
				URI subst_url(imgurl);
				subst_url.setAbsoluteRawPath("bf.intercept-request/image_analyze");
				std::ostringstream strm;
				strm << "url=" << URI::escape(imgurl.toString()) << "&w=" << width
				<< "&h=" << height << "&iar=" << img_ad_rule << "&ea="
				<< (external_link ? '1' : '0') << "&ei=" << (external_img ? '1' : '0')
				<< "&sa=" << (suspicious_link ? '1' : '0') << "&si=" << (suspicious_img ? '1' : '0')
				<< "&target=" << URI::escape(target);
				subst_url.setRawQuery(strm.str());
				img->setAttribute("src", subst_url.toString());
				if (img->getParent()) {
					processDocFragmentNode(img);
				} else {
					processOpeningTagAsText(img, explicit_empty);
				}
				return true;
			}
		} else if (is_ad) {
			if (m_pResult) {
				m_pResult->setType(ScriptResult::T_IMG);
				m_pResult->setLinkURL(aurl_str);
				m_pResult->setObjectURL(imgurl_str);
				m_pResult->setDimentions(width, height);
				stop();
				return true;
			}
			URI subst_url(imgurl);
			subst_url.setScheme("http");
			subst_url.setAbsoluteRawPath("bf.intercept-request/image_substitute");
			subst_url.setRawQuery("w="+StringUtils::fromNumber(width)+"&h="+StringUtils::fromNumber(height));
			img->setAttribute("src", subst_url.toString());
			if (HtmlNode* parent = img->getParent()) {
				parent->setAdMark(true);
				processDocFragmentNode(img);
			} else {
				m_pContainingNode->setAdMark(true);
				processOpeningTagAsText(img, explicit_empty);
			}
			return true;
		}
	}
	if (img->getParent()) {
		processDocFragmentNode(img);
		return true;
	}
	return false;
}

bool
HtmlProcessor::processMovementHooks(HtmlNode* node)
{
	bool modified = false;
	if (node->hasAttribute("onmouseover")) {
		if (checkCodeForPopups(node->getAttribute("onmouseover"))) {
			node->setRawAttribute("onmouseover", "/* popup detected by BFilter */");
			modified = true;
		}
	}
	if (node->hasAttribute("onmouseout")) {
		if (checkCodeForPopups(node->getAttribute("onmouseout"))) {
			node->setRawAttribute("onmouseout", "/* popup detected by BFilter */");
			modified = true;
		}
	}
	return modified;
}

void
HtmlProcessor::processOnloadHooks()
{
	std::ostringstream strm;
	strm << "<script type=\"text/javascript\"> ";
	if (!m_onload.empty()) {
		if (!checkCodeForPopups(m_onload)) {
			strm << "bfilter_onload_ok = true; ";
		} else {
			strm << "/* onload produces a popup */ ";
		}
	}
	if (!m_onunload.empty()) {
		if (!checkCodeForPopups(m_onunload)) {
			strm << "bfilter_onunload_ok = true; ";
		} else {
			strm << "/* onunload produces a popup */ ";
		}
	}
	strm << "</script>\r\n";
	processText(strm.str());
	m_onload.resize(0);
	m_onunload.resize(0);
}

void
HtmlProcessor::processUnclosedTags(HtmlNode* to)
{
	while (m_pContainingNode != to) {
		m_closingTagName = m_pContainingNode->getTagName();
		processClosingTag(m_emptyBuffer.begin(), m_emptyBuffer.end());
	}
}

void
HtmlProcessor::processOpeningTagAsText(HtmlNode* node, bool explicit_empty)
{
	DBG_ASSERT(!node->getParent());
	
	std::string text = openingTagToText(node, explicit_empty);
	if (m_docFragment.hasChildren()) {
		tagTextAppend(m_pContainingNode, text);
	} else {
		outputData(text);
	}
}

void
HtmlProcessor::processIndependentNode(HtmlNode* node)
{
	DBG_ASSERT(!node->getParent());
	
	if (m_docFragment.hasChildren()) {
		docFragmentPutElement(node);
		processNode(node, false, false);
	} else {
		HtmlFragmentNode parent;
		parent.appendChild(node);
		processNode(node, false, true);
	}
}

void
HtmlProcessor::processDocFragmentNode(HtmlNode* node)
{
	DBG_ASSERT(node->getParent());
	
	if (isAncestorOrSelf(node, m_pContainingNode)) {
		m_pContainingNode = node->getParent();
	}
	processNode(node, true, m_docFragment.getFirstChild() == node);
	outputDocFragmentLeadingText();
}

void
HtmlProcessor::processDocFragmentNodeReplacement(HtmlNode* original, HtmlNode* replacement)
{
	DBG_ASSERT(original->getParent());
	
	if (isAncestorOrSelf(original, m_pContainingNode)) {
		m_pContainingNode = original->getParent();
	}
	replacement->insertBefore(original);
	original->detachFromParent();
	processNode(replacement, true, m_docFragment.getFirstChild() == replacement);
	outputDocFragmentLeadingText();
}

void
HtmlProcessor::processDocFragmentNodeReplacement(HtmlNode* node, const std::string& replacement)
{
	HtmlNodePtr text_node(HtmlTextNode::create());
	text_node->appendContent(replacement);
	processDocFragmentNodeReplacement(node, text_node.get());
}

// this method destroys the node (removes its children and detaches the node from its parent)
bool
HtmlProcessor::processNode(HtmlNode* node, bool skip_stoppers, bool can_output_directly, bool toplevel)
{
	DBG_ASSERT(node->getParent());
	
	if (node->getNodeType() == HtmlNode::TEXT_NODE) {
		if (can_output_directly) {
			outputData(node->getContent());
			node->detachFromParent();
		}
		return can_output_directly;
	}
	if (!toplevel && skip_stoppers && isStopperTag(node->getTagType())) {
		return false;
	}
	
	bool explicit_empty = true; // true will prevent outputing a closing tag for a fragment node
	if (node->getNodeType() == HtmlNode::ELEMENT_NODE) {
		explicit_empty = isXHTML() && !node->hasChildren() && isClosingTagForbidden(node->getTagName());
		std::string opening(openingTagToText(node, explicit_empty));
		if (can_output_directly) {
			outputData(opening);
		} else {
			HtmlNodePtr text_node(HtmlTextNode::create());
			text_node->appendContent(opening);
			text_node->insertBefore(node);
		}
	}
	
	for (HtmlNodePtr child(node->getFirstChild()); child; ) {
		HtmlNode* next_child = child->getNextSibling();
		child->detachFromParent();
		child->insertBefore(node);
		can_output_directly = processNode(child.get(), skip_stoppers, can_output_directly, false);
		child.reset(next_child);
	}
	
	if (!explicit_empty && node->isClosed()) {
		std::string ctag;
		ctag.reserve(node->getTagName().length()+3);
		ctag += "</";
		ctag += node->getTagName();
		ctag += '>';
		if (can_output_directly) {
			outputData(ctag);
		} else {
			HtmlNodePtr text_node(HtmlTextNode::create());
			text_node->appendContent(ctag);
			text_node->insertBefore(node);
		}
	}
	
	node->detachFromParent();
	return can_output_directly;
}

void
HtmlProcessor::tryInliningExternalScript(HtmlNode* script_tag, const std::string& script_src,
	const std::string& script_code, bool script_is_js)
{
	bool safe = (StringUtils::toLowerStr(script_code).find("</script") == std::string::npos);
	bool leave_as_is = false;
	std::string script_body;
	if (safe) {
		script_body = script_code;
	} else if (script_is_js) {
		script_body = "\r\neval(unescape('"+URI::escape(script_code)+"'))\r\n;";
	} else {
		leave_as_is = true;
	}
	if (!leave_as_is) {
		script_tag->removeAttribute("src");
		script_tag->removeAllChildren();
		tagTextAppend(script_tag, script_body);
		HtmlNodePtr repl_node(HtmlTextNode::create());
		repl_node->appendContent("<!-- The following script has been fetched from ");
		repl_node->appendContent(StringUtils::replace(">", "&gt;", script_src));
		if (!safe) {
			repl_node->appendContent(" and escaped");
		}
		repl_node->appendContent(" by BFilter -->");
		repl_node->insertBefore(script_tag);
		processDocFragmentNode(repl_node.get());
	}
}

void
HtmlProcessor::docFragmentPutContainer(HtmlNode* node)
{
	m_pContainingNode->appendChild(node);
	m_pContainingNode = node;
}

void
HtmlProcessor::docFragmentPutElement(HtmlNode* node)
{
	m_pContainingNode->appendChild(node);
}

bool
HtmlProcessor::isAncestorOrSelf(const HtmlNode* ancestor, const HtmlNode* descendant)
{
	for (; descendant; descendant = descendant->getParent()) {
		if (descendant == ancestor) {
			return true;
		}
	}
	return false;
}

HtmlNode*
HtmlProcessor::findAncestorOrSelf(HtmlNode* node, const std::string& name)
{
	for (; node; node = node->getParent()) {
		if (node->getTagName() == name) {
			break;
		}
	}
	return node;
}

HtmlNode*
HtmlProcessor::findMostEvilMapLink(HtmlNode* map)
{
	HtmlNode* most_evil_link = 0;
	float max_factor = 0;
	
	for (HtmlNode* node = map->getFirstChild(); node; node = node->getNextSibling()) {
		float factor = 0;
		URI url(node->getAttribute("href"));
		if (isExternalURL(url))
			factor += 1.0;
		if (isSuspiciousURL(url))
			factor += 1.0;
		std::string target = node->getAttribute("target");
		if (target == "_blank" || target == "_new") {
			factor += 1.2;
		} else if (!target.empty() && target[0] != '_') {
			factor += 1.0;
		}
		if (factor > max_factor) {
			most_evil_link = node;
			max_factor = factor;
		}
	}
	
	return most_evil_link;
}

bool
HtmlProcessor::checkCodeForPopups(const std::string& code)
{
	JSPopupCounter counter(*m_jsEnv);
	m_jsEnv->executeScriptAsFunction(code);
	return (counter.getPopupCount() != 0);
}

void
HtmlProcessor::outputData(const std::string& data)
{
	m_onDataOutputSignal(reinterpret_cast<const uint8_t*>(data.data()),
		reinterpret_cast<const uint8_t*>(data.data())+data.length());
}

void
HtmlProcessor::outputData(const Buffer& data)
{
	outputData(data.begin(), data.end());
}

void
HtmlProcessor::outputData(const ConstIterator& begin, const ConstIterator& end)
{
	Buffer::ConstRangeIterator iter(begin, end);
	const Buffer::ConstRange* range;
	while ((range = iter.next())) {
		m_onDataOutputSignal(range->begin(), range->end());
	}
}

void
HtmlProcessor::outputDocFragmentLeadingText()
{
	HtmlNode* node;
	while ((node = m_docFragment.getFirstChild()) && node->getNodeType() == HtmlNode::TEXT_NODE) {
		outputData(node->getContent());
		node->detachFromParent();
	}
}

#ifdef DEBUG
namespace {
	class TrafficMeter : public SigC::Object
	{
	public:
		TrafficMeter() : m_trafSize(0) {
			m_trafStart.reserve(STORE_DATA);
		}
		
		std::size_t getTrafficSize() const { return m_trafSize; }
		
		const std::string& getTrafficStart() const { return m_trafStart; }
		
		void processData(const uint8_t* begin, const uint8_t* end) {
			std::size_t len = std::min<std::size_t>(STORE_DATA - m_trafStart.length(), end - begin);
			m_trafStart.append(reinterpret_cast<const char*>(begin), len);
			m_trafSize += end - begin;
		}
	private:
		enum { STORE_DATA = 60 };
		std::size_t m_trafSize;
		std::string m_trafStart;
	};
}
#endif

void
HtmlProcessor::outputPendingData()
{
	if (m_pResult) {
		return;
	}
#ifdef DEBUG
	TrafficMeter tmeter;
	onDataOutput().connect(SigC::slot(tmeter, &TrafficMeter::processData));
#endif
	for (HtmlNode* child; (child = m_docFragment.getFirstChild()); ) {
		processNode(child, false, true);
	}
	m_pContainingNode = &m_docFragment;
#ifdef DEBUG
	DEBUGLOG("HtmlProcessor::outputPendingData(): " << tmeter.getTrafficSize() << " bytes");
	if (tmeter.getTrafficSize() != 0) {
		DEBUGLOG("Pending data start: " << tmeter.getTrafficStart());
	}
#endif
}

void
HtmlProcessor::tagTextAppend(HtmlNode* el, const Buffer::ConstIterator& begin, const Buffer::ConstIterator& end)
{
	HtmlNode* last_child = el->getLastChild();
	if (last_child && last_child->getNodeType() == HtmlNode::TEXT_NODE) {
		last_child->appendContent(begin, end);
	} else {
		HtmlNodePtr new_child(HtmlTextNode::create());
		new_child->appendContent(begin, end);
		el->appendChild(new_child.get());
	}
}

void
HtmlProcessor::tagTextAppend(HtmlNode* el, const std::string& text)
{
	HtmlNode* last_child = el->getLastChild();
	if (last_child && last_child->getNodeType() == HtmlNode::TEXT_NODE) {
		last_child->appendContent(text);
	} else {
		HtmlNodePtr new_child(HtmlTextNode::create());
		new_child->appendContent(text);
		el->appendChild(new_child.get());
	}
}

std::string
HtmlProcessor::openingTagToText(const HtmlNode* node, bool explicit_empty)
{
	std::ostringstream strm;
	strm << '<' << node->getTagName();
	HtmlNode::AttributeMap::const_iterator iter = node->getRawAttributes().begin();
	for (; iter != node->getRawAttributes().end(); ++iter) {
		strm << ' ' << iter->first << '=' << '"' << iter->second << '"';
	}
	if (explicit_empty) {
		strm << " />";
	} else {
		strm << '>';
	}
	return strm.str();
}

bool
HtmlProcessor::isExternalURL(const URI& url)
{
	std::string rhost = StringUtils::toLowerStr(url.getHost());
	if (rhost.find("www.") == 0)
		rhost.erase(0, 4);
	std::string lhost = StringUtils::toLowerStr(m_baseURL.getHost());
	if (lhost.find("www.") == 0)
		lhost.erase(0, 4);
	if (rhost.empty()) {
		return false;
	}
	if (rhost == lhost) {
		return false;
	}
	// if one is the subdomain of the other, return false
	if (StringUtils::endsWith(rhost, "."+lhost) || StringUtils::endsWith(lhost, "."+rhost)) {
		return false;
	}
	return true;
}

bool
HtmlProcessor::isSuspiciousURL(const URI& url)
{
	//if (url.getQueryEncoded().length() > 35)
	//	return true;
	
	if (url.getDecodedPath().find("http://") != std::string::npos ||
	    url.getDecodedQuery().find("http://") != std::string::npos) {
		return true;
	}
	
	std::string path_plus_query = url.getRawPath()+"?"+url.getRawQuery();
	static std::string digits = "0123456789";
	std::string::size_type pos1 = 0;
	int ndigits = 0;
	while (true) {
		std::string::size_type pos2 = path_plus_query.find_first_of(digits, pos1);
		if (pos2 == std::string::npos)
			break;
		pos1 = pos2+1;
		++ndigits;
	}
	if (ndigits > 6)
		return true;
	
	pos1 = 0;
	int nequals = 0;
	while (true) {
		std::string::size_type pos2 = path_plus_query.find('=', pos1);
		if (pos2 == std::string::npos)
			break;
		pos1 = pos2+1;
		++nequals;
	}
	return (nequals > 3);
}

HtmlProcessor::TagType
HtmlProcessor::getTagType(const std::string& tname)
{
	static const TagTypePair type_map[] = {
		// sorted by tag name
		{ "a",        TAG_A },
		{ "area",     TAG_AREA },
		{ "base",     TAG_BASE },
		{ "body",     TAG_BODY },
		{ "embed",    TAG_EMBED },
		{ "form",     TAG_FORM },
		{ "frameset", TAG_FRAMESET },
		{ "head",     TAG_HEAD },
		{ "iframe",   TAG_IFRAME },
		{ "img",      TAG_IMG },
		{ "map",      TAG_MAP },
		{ "object",   TAG_OBJECT },
		{ "param",    TAG_PARAM },
		{ "script",   TAG_SCRIPT }
	};
	const TagTypePair* end = &type_map[sizeof(type_map)/sizeof(*type_map)];
	const TagTypePair* found = std::lower_bound(&type_map[0], end, tname, TagTypeComparator());
	if (found != end && found->tname == tname) {
		return found->type;
	}
	return TAG_OTHER;
}

bool
HtmlProcessor::isStopperTag(int type)
{
	static const bool table[] = {
		false, // TAG_OTHER
		false, // TAG_A
		false, // TAG_AREA
		false, // TAG_BASE
		false, // TAG_BODY
		false, // TAG_EMBED
		false, // TAG_FORM
		false, // TAG_FRAMESET
		false, // TAG_HEAD
		true,  // TAG_IFRAME
		true,  // TAG_IMG
		false, // TAG_MAP
		true,  // TAG_OBJECT
		false, // TAG_PARAM
		true,  // TAG_SCRIPT
	};
	return table[type];
}

bool
HtmlProcessor::isClosingTagForbidden(const std::string& tname)
{
	static const char* tags[] = { // sorted
		"img", "param", "area"
	};
	return std::binary_search(&tags[0], &tags[sizeof(tags)/sizeof(*tags)], tname);
}

bool
HtmlProcessor::tagTerminatesAnchor(const std::string& tname)
{
	static const char* tags[] = { // sorted
		"a", "div", "table", "td", "tr"
	};
	return std::binary_search(&tags[0], &tags[sizeof(tags)/sizeof(*tags)], tname);
}

bool
HtmlProcessor::isCommonAdSize(int width, int height)
{
	if (width == 120 && height == 60) {
		return true;
	} else if (width == 120 && height == 600) {
		return true;
	} else if (width == 120 && height == 60) {
		return true;
	} else if (width == 125 && height == 125) {
		return true;
	} else if (width == 234 && height == 60) {
		return true;
	} else if (width == 468 && height == 60) {
		return true;
	} else if (width == 468 && height == 80) {
		return true;
	} else if (width == 80 && height == 468) {
		return true;
	} else if (width == 88 && height == 31) {
		return true;
	} else if (width == 728 && height == 90) {
		return true;
	} else if (width == 160 && height == 600) {
		return true;
	} else if (width == 300 && height == 250) {
		return true;
	} else if (width == 180 && height == 150) {
		return true;
	}
	return false;
}

bool
HtmlProcessor::couldBeAdSize(int width, int height)
{
	if (width == 0 || height == 0) {
		return false;
	}
	return (width * height >= 10000 && width * height < 120000 &&
		(float)std::max(width, height)/std::min(width, height) >= 2.5);
}

int
HtmlProcessor::imageIsAd(int img_ad_rule, int width, int height, const std::string& target,
                         bool external_link, bool external_img, bool suspicious_link, bool suspicious_img,
                         int animated, int redir, int cookie) {
	enum {YES=1, NO=-1, NOT_SURE=0};
	if (img_ad_rule == 2) {
		return YES;
	}
	if (img_ad_rule == 1) {
		return NO;
	}
	if (width * height < 600) {
		// too small
		return NO;
	}
	if (width * height > 260000 && !external_img && !suspicious_img) {
		// too big
		return NO;
	}
	bool first_pass = (animated == 0);
	double factor = 0.0;
	if (isCommonAdSize(width, height)) {
		factor += first_pass ? 2.1 : 1.6;
	} else if (couldBeAdSize(width, height)) {
		factor += 1.4;
	}
	if (external_link) {
		factor += 1.0;
	} else {
		factor -= 0.4;
	}
	if (external_img) {
		factor += 1.0;
	} else {
		factor -= 0.3;
	}
	if (suspicious_link) factor += 1.0;
	if (suspicious_img) factor += 1.0;
	factor += redir > 0 ? 2.0 : (double)redir;
	if (cookie > 0) factor += 1.5;
	if (animated == YES) {
		factor += 2.0;
	}
	if (target == "_blank" || target == "_new") {
		factor += 1.2;
	} else if (!target.empty() && target[0] != '_') {
		factor += 1.0;
	}
	if (factor > 4.0) {
		return YES;
	} else if (factor <= 2.0) {
		return NO;
	} else if (factor > 3.0 && !first_pass) {
		return YES;
	}
	return NOT_SURE;
}

bool
HtmlProcessor::allowScript(const std::string& script, bool is_3rd_party)
{
	switch (m_scriptsRule) {
	case 0:
		return true;
	case 1:
		return (m_isInsideHead || !is_3rd_party);
	case 2:
		return !is_3rd_party;
	case 3:
		return (m_isInsideHead || (!is_3rd_party && script.find(".write")==std::string::npos));
	case 4:
		return (m_isInsideHead || (!is_3rd_party && script.find("function ")!=std::string::npos));
	case 5:
		return m_isInsideHead;
	case 6:
		return (m_isInsideHead && !is_3rd_party);
	case 7:
		return false;
	default:
		return true;
	}
}


HtmlProcessor::ScriptResult::ScriptResult()
:	m_type(T_NOTHING),
	m_width(0),
	m_height(0),
	m_nestLevel(0)
{
}

std::string
HtmlProcessor::ScriptResult::typeToString() const
{
	static const char* table[] = {
		"nothing",
		"img ad",
		"img/map ad",
		"iframe ad",
		"flash ad",
		"flash (as img)",
		"webbug",
		"popup",
		"rule",
		"script fetching failed"
	};
	return table[m_type];
}

