/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef HTMLPROCESSOR_H_
#define HTMLPROCESSOR_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "types.h"
#include "buffer.h"
#include "htmllexer.h"
#include "htmlnode.h"
#include "htmlfragmentnode.h"
#include "uri.h"
#include "jsenvproxy.h"
#include <string>
#include <map>
#include <sigc++/sigc++.h>

class BFilter;

class HtmlProcessor : private HtmlLexer, public SigC::Object
{
public:
	HtmlProcessor(BFilter *owner);
private:
	HtmlProcessor(const HtmlProcessor&);            //
	HtmlProcessor& operator=(const HtmlProcessor&); // forbidden
public:
	virtual ~HtmlProcessor();
	
	void setBaseURL(const URI& url);
	
	void setUserAgent(const std::string& ua);
	
	using HtmlLexer::consumeDataChunk;
	
	void consumeEOF();
	
	void reset();
	
	static bool isCommonAdSize(int width, int height);
	
	static bool couldBeAdSize(int width, int height);
	
	static int imageIsAd(int ad_rule, int width, int height, const std::string& target,
		bool external_link, bool external_img, bool suspicious_link, bool suspicious_img,
		int animated=0, int redir=0, int cookie=0);
	
	SigC::Signal2<void, const uint8_t*, const uint8_t*>& onDataOutput() {
		return m_onDataOutputSignal;
	}
private:
	enum ContentType { CONTENT_HTML, CONTENT_XHTML };
	
	enum TagType {
		TAG_OTHER,
		TAG_A,
		TAG_AREA,
		TAG_BASE,
		TAG_BODY,
		TAG_EMBED,
		TAG_FORM,
		TAG_FRAMESET,
		TAG_HEAD,
		TAG_IFRAME,
		TAG_IMG,
		TAG_MAP,
		TAG_OBJECT,
		TAG_PARAM,
		TAG_SCRIPT
	};
	
	enum { MAX_SCRIPT_FETCH_REDIRECTS = 2 };
	enum { MAX_NEST_LEVEL = 2 };
	
	typedef SigC::Slot2<void, const uint8_t*, const uint8_t*> DataOutputFunctor;
	
	struct TagTypePair
	{
		const char* tname;
		TagType type;
	};
	
	struct TagTypeComparator
	{
		bool operator()(const TagTypePair& pair, const std::string& tname) {
			return pair.tname < tname;
		}
	};
	
	class ScriptResult
	{
	public:
		enum Type {
			T_NOTHING, T_IMG, T_MAP, T_IFRAME, T_FLASH, T_FLASH_AS_IMG,
			T_WEBBUG, T_POPUP, T_SCRIPT_RULES, T_SCRIPT_FETCHING_FAILED
		};
		
		ScriptResult();
		
		Type getType() const { return m_type; }
		
		void setType(Type t) { m_type = t; }
		
		const std::string& getLinkURL() const { return m_linkURL; }
		
		void setLinkURL(const std::string& url) { m_linkURL = url; }
		
		const std::string& getObjectURL() const { return m_objectURL; }
		
		void setObjectURL(const std::string& url) { m_objectURL = url; }
		
		int getWidth() const { return m_width; }
		
		int getHeight() const { return m_height; }
		
		void setDimentions(int w, int h) { m_width = w; m_height = h; }
		
		int getNestLevel() const { return m_nestLevel; }
		
		void incNestLevel() { ++m_nestLevel; }
		
		std::string typeToString() const;
	private:
		Type m_type;
		std::string m_linkURL;
		std::string m_objectURL;
		int m_width;
		int m_height;
		int m_nestLevel;
	};
	
	HtmlProcessor(HtmlProcessor& parent, ScriptResult* result);
	
	virtual void processDocType(const ConstIterator& begin, const ConstIterator& end);
	
	virtual void processText(const ConstIterator& begin, const ConstIterator& end);
	
	void processText(const std::string& text);
	
	virtual void processOpeningTagName(const ConstIterator& begin, const ConstIterator& end);
	
	virtual void processOpeningTag(const ConstIterator& begin, const ConstIterator& end, bool explicit_empty);
	
	virtual void processClosingTagName(const ConstIterator& begin, const ConstIterator& end);
	
	virtual void processClosingTag(const ConstIterator& begin, const ConstIterator& end);
	
	virtual bool processAttrName(const ConstIterator& begin, const ConstIterator& end);
	
	virtual void processAttrValue(const ConstIterator& begin, const ConstIterator& end);
	
	virtual bool isCDATAStarting();
	
	virtual bool isCDATAEnding(const ConstIterator& begin, const ConstIterator& end);
	
	bool processOpeningOther(HtmlNode* node, bool explicit_empty);
	
	bool processOpeningA(HtmlNode* anchor, bool explicit_empty);
	
	bool processOpeningArea(HtmlNode* area, bool explicit_empty);
	
	bool processOpeningBase(HtmlNode* base, bool explicit_empty);
	
	bool processOpeningBody(HtmlNode* body, bool explicit_empty);
	
	bool processOpeningEmbed(HtmlNode* embed, bool explicit_empty);
	
	bool processOpeningForm(HtmlNode* form, bool explicit_empty);
	
	bool processOpeningFrameset(HtmlNode* frameset, bool explicit_empty);
	
	bool processOpeningHead(HtmlNode* head, bool explicit_empty);
	
	bool processOpeningIframe(HtmlNode* iframe, bool explicit_empty);
	
	bool processOpeningImg(HtmlNode* img, bool explicit_empty);
	
	bool processOpeningMap(HtmlNode* map, bool explicit_empty);
	
	bool processOpeningObject(HtmlNode* object, bool explicit_empty);
	
	bool processOpeningParam(HtmlNode* param, bool explicit_empty);
	
	bool processOpeningScript(HtmlNode* script, bool explicit_empty);
	
	void processClosingOther(const ConstIterator& begin, const ConstIterator& end);
	
	void processClosingA(const ConstIterator& begin, const ConstIterator& end);
	
	void processClosingBody(const ConstIterator& begin, const ConstIterator& end);
	
	void processClosingHead(const ConstIterator& begin, const ConstIterator& end);
	
	void processClosingIframe(const ConstIterator& begin, const ConstIterator& end);
	
	void processClosingMap(const ConstIterator& begin, const ConstIterator& end);
	
	void processClosingObject(const ConstIterator& begin, const ConstIterator& end);
	
	void processClosingScript(const ConstIterator& begin, const ConstIterator& end);
	
	bool processImg(HtmlNode* img, bool explicit_empty, const HtmlNode* anchor);
	
	bool processMovementHooks(HtmlNode* node);
	
	void processOnloadHooks();
	
	void processUnclosedTags(HtmlNode* to);
	
	void processOpeningTagAsText(HtmlNode* node, bool explicit_empty);
	
	void processIndependentNode(HtmlNode* node);
	
	void processDocFragmentNode(HtmlNode* node);
	
	void processDocFragmentNodeReplacement(HtmlNode* original, HtmlNode* replacement);
	
	void processDocFragmentNodeReplacement(HtmlNode* node, const std::string& replacement);
	
	bool processNode(HtmlNode* node, bool skip_stoppers,
		bool can_output_directly, bool toplevel=true);
	
	void tryInliningExternalScript(HtmlNode* script_tag, const std::string& script_src,
		const std::string& script_code, bool script_is_js);
	
	void docFragmentPutContainer(HtmlNode* node);
	
	void docFragmentPutElement(HtmlNode* node);
	
	static bool isAncestorOrSelf(const HtmlNode* ancestor, const HtmlNode* descendant);
	
	static HtmlNode* findAncestorOrSelf(HtmlNode* node, const std::string& name);
	
	HtmlNode* findMostEvilMapLink(HtmlNode* map);
	
	bool checkCodeForPopups(const std::string& code);
	
	void outputData(const std::string& data);
	
	void outputData(const Buffer& data);
	
	void outputData(const ConstIterator& begin, const ConstIterator& end);
	
	void outputDocFragmentLeadingText();
	
	void outputPendingData();
	
	static void tagTextAppend(HtmlNode* el, const Buffer::ConstIterator& begin, const Buffer::ConstIterator& end);
	
	static void tagTextAppend(HtmlNode* el, const std::string& text);
	
	static std::string openingTagToText(const HtmlNode* node, bool explicit_empty);
	
	bool isExternalURL(const URI& url);
	
	static bool isSuspiciousURL(const URI& url);
	
	static TagType getTagType(const std::string& tname);
	
	static bool isStopperTag(int type);
	
	static bool isClosingTagForbidden(const std::string& tname);
	
	static bool tagTerminatesAnchor(const std::string& tname);
	
	bool allowScript(const std::string& script, bool is_3rd_party);
	
	bool isXHTML() const { return m_contentType == CONTENT_XHTML; }
	
	BFilter *m_pOwner;
	ScriptResult* m_pResult;
	int m_nestLevel;
	JSEnvProxy m_jsEnv;
	SigC::Signal2<void, const uint8_t*, const uint8_t*> m_onDataOutputSignal;
	ContentType m_contentType;
	HtmlNodePtr m_ptrCurTag;
	bool m_foundMovementHooks;
	HtmlFragmentNode m_docFragment;
	HtmlNode* m_pContainingNode; // points to either m_docFragment or one of its descendants
	HtmlNodePtr m_ptrContainingAnchor;
	HtmlNodePtr m_ptrContainingMap;
	typedef std::map<std::string, HtmlNodePtr> TagsByName;
	typedef std::multimap<std::string, HtmlNodePtr> MultiTagsByName;
	TagsByName m_mapTags;
	MultiTagsByName m_imgTagsUsingMap;
	std::string m_curTagName;
	std::string m_curAttrName;
	std::string m_closingTagName;
	URI m_baseURL;
	std::string m_userAgent;
	int m_scriptsRule;
	int m_jsengineRule;
	int m_targetBlankRule;
	bool m_isInsideHead;
	bool m_webbugOpenComment; // true if the last script was rejected, and its output ends with <!--
	std::string m_onload;
	std::string m_onunload;
	const Buffer m_emptyBuffer;
};

#endif
