/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "htmlresponsefilter.h"
#include "responsefilterchain.h"
#include "decompressorresponsefilter.h"
#include "stringutils.h"
#include <string>

AbstractResponseFilter*
HtmlResponseFilter::create(NextFilterProvider* follower,
	HttpResponseMetadata* metadata, HtmlProcessor* delegate)
{
	int status_code = metadata->statusLine().getCode();
	if (status_code < 400 && status_code != 200) {
		return 0;
	}
	std::string content_type = metadata->headers().getHeader("Content-Type").getValue();
	if (!StringUtils::startsWith(content_type, "text/html") &&
	    !StringUtils::startsWith(content_type, "application/xhtml+xml")) {
		return 0;
	}
	if (metadata->getBodyStatus() == HttpResponseMetadata::NO_BODY_BUT_PRESERVE_HEADERS ||
	    (metadata->getBodyStatus() == HttpResponseMetadata::BODY_SIZE_KNOWN && metadata->getBodySize() == 0)) {
		// no body in response
		return 0;
	}
	
	if (ResponseFilterChain::isBodyEncoded(*metadata)) {
		NextFilterProvider provider =
			SigC::bind(SigC::slot(&HtmlResponseFilter::construct), follower, delegate);
		return DecompressorResponseFilter::createStack(&provider, metadata, 2);
	} else {
		return construct(metadata, follower, delegate);
	}
}

HtmlResponseFilter::HtmlResponseFilter(AbstractResponseFilter* follower, HtmlProcessor* delegate)
:	m_pFollower(follower), m_pDelegate(delegate)
{
	delegate->onDataOutput().connect(SigC::slot(*this, &HtmlResponseFilter::processOutputData));
}

HtmlResponseFilter::~HtmlResponseFilter()
{
	m_pDelegate->reset();
	m_pFollower->destroy();
}

void
HtmlResponseFilter::consumeDataChunk(const uint8_t* begin, const uint8_t* end)
{
	m_pDelegate->consumeDataChunk(begin, end);
}

void
HtmlResponseFilter::consumeEOF()
{
	m_pDelegate->consumeEOF();
	m_pFollower->consumeEOF();
}

void
HtmlResponseFilter::processOutputData(const uint8_t* begin, const uint8_t* end)
{
	m_pFollower->consumeDataChunk(begin, end);
}

AbstractResponseFilter*
HtmlResponseFilter::construct(HttpResponseMetadata* metadata,
	NextFilterProvider* follower, HtmlProcessor* delegate)
{
	metadata->setBodyStatus(HttpMessageMetadata::BODY_SIZE_UNKNOWN);
#ifdef DEBUG
	metadata->headers().addHeader("X-filtered-by", "HtmlResponseFilter");
#endif
	return new HtmlResponseFilter((*follower)(metadata), delegate);
}

