/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httperrormessages.h"
#include "date.h"
#include "htmlescaper.h"
#include "stringutils.h"
#include <sstream>

const char *HttpErrorMessages::m_html_error_header = "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\r\n"
"<html>\r\n"
"<head>\r\n"
"  <title>BFilter Error</title>\r\n"
"  <meta http-equiv=\"content-type\" content=\"text/html; charset=iso-8859-1\">\r\n"
"</head>\r\n"
"<body bgcolor=\"#ffffff\">\r\n"
"<div align=\"center\"><font size=\"+1\" color=\"#ff6600\"><b>BFilter Error</b></font></div><br>\r\n"
"<hr width=\"100%\" size=\"2\"><br>";

std::string
HttpErrorMessages::errMalformedRequestLine(const HttpRequestLine& request_line)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_BAD_REQUEST),
		"Malformed request line"
	);
}

std::string
HttpErrorMessages::errBadRequestURI(const HttpRequestLine& request_line, const std::string& details)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_BAD_REQUEST),
		"Bad request URI: "+HtmlEscaper::escape(details)+"<br><br>[Request: "
		+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::errUnsupportedRequestMethod(const HttpRequestLine& request_line)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_METHOD_NOT_ALLOWED),
		"Unsupported method: "+HtmlEscaper::escape(request_line.getMethod())
		+"<br><br>[Request: "+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::errUnsupportedExpectation(const HttpRequestLine& request_line, const std::string& expectation)
{
	return formatMessage(request_line,
		HttpStatusLine(HttpVersion::HTTP_1_1, HttpStatusLine::SC_EXPECTATION_FAILED),
		"Unrecognised expectation: "+HtmlEscaper::escape(expectation)
		+"<br><br>[Request: "+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::errMalformedRequestHeader(const HttpRequestLine& request_line)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_BAD_REQUEST),
		"Malformed request header"
	);
}

std::string
HttpErrorMessages::errMalformedResponseHeader(const HttpRequestLine& request_line)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_SERVICE_UNAVAILABLE),
		"Malformed response header"
	);
}

std::string
HttpErrorMessages::errRequestBodyLengthRequired(const HttpRequestLine& request_line)
{
	int status_code = (request_line.getHttpVersion() >= HttpVersion::HTTP_1_1)
		? HttpStatusLine::SC_LENGTH_REQUIRED : HttpStatusLine::SC_BAD_REQUEST;
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), status_code),
		"Bad request: body length required<br><br>[Request: "
		+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::errMalformedChunkInRequest(const HttpRequestLine& request_line)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_BAD_REQUEST),
		"Malformed chunk in request body<br><br>[Request: "
		+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::errMalformedChunkInResponse(const HttpRequestLine& request_line)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_SERVICE_UNAVAILABLE),
		"Malformed chunk in response body<br><br>[Request: "
		+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::errUnexpectedServerEOF(const HttpRequestLine& request_line)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_SERVICE_UNAVAILABLE),
		"Server has unexpectedly closed connection<br><br>[Request: "
		+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::errDNS(const HttpRequestLine& request_line, const std::string& host)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_SERVICE_UNAVAILABLE),
		"Could not resolve host name <i>"+HtmlEscaper::escape(host)
		+"</i><br><br>[Request: "+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::errConnRefused(const HttpRequestLine& request_line, const std::string& host, int port)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_SERVICE_UNAVAILABLE),
		"Connection refused to <i>"+HtmlEscaper::escape(host)+":"+StringUtils::fromNumber(port)
		+"</i><br><br>[Request: "+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::errConnTimeout(const HttpRequestLine& request_line, const std::string& host, int port)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_GATEWAY_TIMEOUT),
		"Connection timeout while connecting to <i>"+HtmlEscaper::escape(host)
		+":"+StringUtils::fromNumber(port)
		+"</i><br><br>[Request: "+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::errNetworkUnreachable(const HttpRequestLine& request_line, const std::string& host, int port)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_SERVICE_UNAVAILABLE),
		"Connection to <i>"+HtmlEscaper::escape(host)+":"+StringUtils::fromNumber(port)
		+"</i> failed:<br>Network is unreachable<br><br>[Request: "
		+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::errHostUnreachable(const HttpRequestLine& request_line, const std::string& host, int port)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_SERVICE_UNAVAILABLE),
		"Connection to <i>"+HtmlEscaper::escape(host)+":"+StringUtils::fromNumber(port)
		+"</i> failed:<br>No route to host<br><br>[Request: "
		+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::errInternal(const HttpRequestLine& request_line, const std::string& details)
{
	return formatMessage(request_line,
		HttpStatusLine(request_line.getHttpVersion(), HttpStatusLine::SC_INTERNAL_SERVER_ERROR),
		"Internal error:<br>"+HtmlEscaper::escape(details)+"<br><br>[Request: "
		+HtmlEscaper::escape(request_line.toString())+"]"
	);
}

std::string
HttpErrorMessages::formatMessage(const HttpRequestLine& request_line,
	const HttpStatusLine& status, const std::string& error_text)
{
	bool no_body = (request_line.getMethod() == "HEAD");
	std::string body;
	if (!no_body) {
		body = m_html_error_header+error_text+"</body></html>";
	}
	std::ostringstream msg;
	msg << status.toString() << "\r\n";
	msg << "Content-Type: text/html\r\nContent-Length: " << body.length()
	<< "\r\nDate: " << Date::formatCurrentTime() << "\r\nConnection: close\r\n\r\n";
	msg << body;
	return msg.str();
}

