/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httpfetcher.h"
#include <sstream>
#include "conf.h"
#include "network.h"
#include "syscall.h"
#include "tselect.h"
#include "stringutils.h"
#include "debug.h"

#ifdef WIN32

#include <winsock2.h>

#else

#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>

#endif

HttpFetcher::HttpFetcher(BFilter* owner)
:	m_pOwner(owner),
	m_timeout(10),
	m_is_head_request(false),
	m_status_line(HttpVersion::HTTP_1_0, 200, "OK")
{
}

HttpResponse
HttpFetcher::request(const HttpRequest& request)
{
	m_body.erase();
	m_headers.clear();
	if (request.getURI().getScheme() != "http") {
		throw Exception("unhandled protocol: "+request.getURI().getScheme());
	}
	m_is_head_request = (request.getMethod() == "HEAD");
	bool through_proxy = m_pOwner->getState().getConfig().willGoThroughProxy(request.getURI().getHost());
	std::string host;
	unsigned port;
	if (through_proxy) {
		host = m_pOwner->getState().getConfig().getProxyHost();
		port = m_pOwner->getState().getConfig().getProxyPort();
	} else {
		host = request.getURI().getHost();
		port = request.getURI().guessPort();
	}
	NetworkError err;
	Network::Socket sock = Network::tcpConnect(host.c_str(), port, &err);
	if (!err.isOK()) {
		throw Exception("couldn't connect: "+err.getErrorMessage());
	}
	try {
		// TODO: make this write asynchronous
		if (!Network::tcpSendData(sock, request.toString(through_proxy), &err)) {
			throw Exception("error sending request: "+err.getErrorMessage());
		}
		TSelect tselect;
		struct timeval tv;
		tv.tv_sec = m_timeout;
		tv.tv_usec = 0;
		int tid = tselect.addTimer(tv);
		fd_set rfds;
		unsigned char buf[8192];
		while (true) {
			FD_ZERO(&rfds);
			FD_SET(sock, &rfds);
			int res = tselect.select(sock+1, &rfds, 0, 0);
			if (!tselect.getTriggeredTimers().empty()) {
				throw Exception("read timeout");
			}
			tselect.resetTimer(tid);
			if (res < 0) {
				throw Exception("error in select: "+Network::getLastError());
			} else if (res == 0) {
				throw Exception("unknown error (probably a bug in TSelect)");
			}
			int nread = -1;
			SYSCALL(nread = recv(sock, (char*)buf, sizeof(buf), 0));
			if (nread < 0) {

				throw Exception("error reading from server: "+Network::getLastError());
			}
			if (nread == 0) {
				consumeEOF();
				throw Exception("unknown error (probably a bug in HttpFetcher or HttpStream)");
			}
			consumeDataChunk(buf, buf+nread);
		}
	} catch (Exception& e) {
		Network::closeSocket(sock);
		throw;
	} catch (ResponseReceivedException& e) {
		Network::closeSocket(sock);
		return HttpResponse(m_status_line, m_headers, m_body);
	}
}

void
HttpFetcher::eventError(int ecode, const char *info)
{
	switch (ecode) {
	case ERR_PARSING_HEADER:
		throw Exception("error parsing response header");
	case ERR_PARSING_CHUNK_HEADER:
		throw Exception("error parsing response chunk");
	case ERR_UNEXPECTED_EOF:
		throw Exception("unexpected EOF from server");
	default:
		throw Exception("unexpected error code ("+StringUtils::fromNumber(ecode)+") in HttpFetcher");
	}
}

void
HttpFetcher::eventGotMetadata(HttpResponseMetadata& metadata)
{
	m_status_line = metadata.statusLine();
	m_headers = metadata.headers();
}

void
HttpFetcher::eventGotContentPart(const uint8_t* begin, const uint8_t* end)
{
	m_body.append(reinterpret_cast<const char*>(begin), reinterpret_cast<const char*>(end));
}

void
HttpFetcher::eventMessageEnds(bool connection_needs_closing)
{
	throw ResponseReceivedException();
}

