/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httpheader.h"
#include "stringutils.h"
#include "debug.h"
#include <string>
#include <list>
#include <cstring>
#include <sstream>

HttpHeader::HttpHeader(const std::string& name, const std::string& value)
:	m_elements(0)
{
	m_name = name;
	m_lcname = StringUtils::toLowerStr(name);
	m_values.push_back(value);
}

HttpHeader::HttpHeader(const HttpHeader& header)
:	m_name(header.m_name),
	m_lcname(header.m_lcname),
	m_values(header.m_values)
{
	if (header.m_elements) {
		m_elements = new Elements(*header.m_elements);
	} else {
		m_elements = 0;
	}
}

HttpHeader&
HttpHeader::operator=(const HttpHeader& header)
{
	if (this == &header)
		return *this;
	
	m_name = header.m_name;
	m_lcname = header.m_lcname;
	m_values = header.m_values;
	if (m_elements)
		delete m_elements;
	if (header.m_elements) {
		m_elements = new Elements(*header.m_elements);
	} else {
		m_elements = 0;
	}
	return *this;
}

HttpHeader::~HttpHeader()
{
	if (m_elements)
		delete m_elements;
}

const std::list<std::string>&
HttpHeader::getValues() const
{
	if (m_elements && m_elements->isModified()) {
		updateValuesFromElements();
	}
	return m_values;
}

void
HttpHeader::setValues(const std::list<std::string>& values)
{
	m_values = values;
	if (m_elements) {
		m_elements->valueToParse(getValue());
	}
}

void
HttpHeader::appendValue(const std::string& value)
{
	m_values.push_back(value);
	if (m_elements) {
		m_elements->valueToParse(getValue());
	}
}

std::string
HttpHeader::getValue() const
{
	const std::list<std::string>& values = getValues();
	return StringUtils::join(", ", values.begin(), values.end());
}

void
HttpHeader::setValue(const std::string& value)
{
	m_values.clear();
	m_values.push_back(value);
	if (m_elements) {
		m_elements->valueToParse(value);
	}
}

HttpHeader::Elements&
HttpHeader::elements()
{
	if (!m_elements) {
		m_elements = new Elements(getValue());
	}
	return *m_elements;
}

const HttpHeader::Elements&
HttpHeader::elements() const
{
	if (!m_elements) {
		m_elements = new Elements(getValue());
	}
	return *m_elements;
}

std::string
HttpHeader::toString() const
{
	std::ostringstream strm;
	toStream(strm);
	return strm.str();
}

void
HttpHeader::toStream(std::ostream& strm) const
{
	if (m_elements && m_elements->isModified()) {
		updateValuesFromElements();
	}
	std::list<std::string>::const_iterator iter = m_values.begin();
	for (; iter != m_values.end(); ++iter) {
		strm << getName() << ": " << *iter << "\r\n";
	}
}

void
HttpHeader::updateValuesFromElements() const
{
	if (!m_elements)
		return;
	m_values.clear();
	m_values.push_back(m_elements->toString());
	m_elements->setModified(false);
}

/* ======================================= HttpHeader::Elements ====================================== */

HttpHeader::Elements::Elements(const std::string& value)
:	m_value(value), m_is_modified(false)
{
	parse();
}

bool
HttpHeader::Elements::isParsingFailed() const
{
	if (m_needs_parsing) {
		parse();
	}
	return m_parsing_failed;
}

bool
HttpHeader::Elements::empty() const
{
	return getElements().empty();
}

bool
HttpHeader::Elements::hasElement(const std::string& name) const
{
	std::string lcname = StringUtils::toLowerStr(name);
	getElements(); // to make sure they get parsed
	std::list<HttpHeaderElement>::const_iterator iter = m_elements.begin();
	for (; iter != m_elements.end(); ++iter) {
		if (iter->getLowercaseName() == lcname) {
			return true;
		}
	}
	return false;
}

const std::list<HttpHeaderElement>&
HttpHeader::Elements::getElements() const
{
	if (m_needs_parsing) {
		parse();
	}
	return m_elements;
}

const HttpHeaderElement*
HttpHeader::Elements::getFirstElement() const
{
	return empty() ? 0 : &m_elements.front();
}

const HttpHeaderElement*
HttpHeader::Elements::getLastElement() const
{
	return empty() ? 0 : &m_elements.back();
}

void
HttpHeader::Elements::removeFirstElement()
{
	if (empty()) {
		return;
	}
	m_elements.pop_front();
	m_is_modified = true;
}

void
HttpHeader::Elements::removeLastElement()
{
	if (empty()) {
		return;
	}
	m_elements.pop_back();
	m_is_modified = true;
}

void
HttpHeader::Elements::removeElementsByName(const std::string& name)
{
	std::string lcname = StringUtils::toLowerStr(name);
	getElements(); // to make sure they get parsed
	std::list<HttpHeaderElement>::iterator iter = m_elements.begin();
	int cnt = 0;
	while (iter != m_elements.end()) {
		if (iter->getLowercaseName() == lcname) {
			++cnt;
			m_elements.erase(iter++);
		} else {
			++iter;
		}
	}
	if (cnt) {
		m_is_modified = true;
	}
}

void
HttpHeader::Elements::clear()
{
	m_elements.clear();
	m_is_modified = true;
	m_needs_parsing = false;
}

void
HttpHeader::Elements::prependElement(const HttpHeaderElement& element)
{
	getElements(); // to make sure they get parsed
	m_elements.push_front(element);
	m_is_modified = true;
}

void
HttpHeader::Elements::appendElement(const HttpHeaderElement& element)
{
	getElements(); // to make sure they get parsed
	m_elements.push_back(element);
	m_is_modified = true;
}

void
HttpHeader::Elements::appendElements(const std::list<HttpHeaderElement>& elements)
{
	getElements(); // to make sure they get parsed
	m_elements.insert(m_elements.end(), elements.begin(), elements.end());
	m_is_modified = true;
}

void
HttpHeader::Elements::appendElements(const Elements& elements)
{
	appendElements(elements.getElements());
}

namespace {
	struct ElementToString
	{
		std::string operator()(const HttpHeaderElement& el) {
			return el.toString();
		}
	};
}

std::string
HttpHeader::Elements::toString() const
{
	std::string res;
	getElements(); // to make sure they get parsed
	/*
	std::list<HttpHeaderElement>::const_iterator iter = m_elements.begin();
	for (; iter != m_elements.end(); ++iter) {
		if (iter != m_elements.begin()) {
			res += ", ";
		}
		res += iter->toString();
	}
	return res;
	*/
	return StringUtils::join(", ", m_elements.begin(), m_elements.end(), ElementToString());
}

void
HttpHeader::Elements::valueToParse(const std::string& value)
{
	m_value = value;
	m_needs_parsing = true;
}

void
HttpHeader::Elements::parse() const
{
	HttpHeader::Parser parser(m_value);
	m_elements = parser.getElements();
	m_is_modified = false;
	m_needs_parsing = false;
	m_parsing_failed = parser.isError();
}

/* ======================================= HttpHeader::Parser ====================================== */

/*
Many headers can be parsed into the following structure:

header        = [ element ] *( ",", [ element ] )
element       = name [ "=" [ value ] ] *( ";" [ param ] )
param         = name [ "=" [ value ] ]
name          = token
value         = ( token | quoted_string )
token         = 1*<any char except "=", ",", ";", <"> and white space>
quoted_string = <"> *( text | quoted_char ) <">
text          = any char except <">
quoted_char   = "\" char


The grammar suitable for predicative parsing:

header                      -> more_elements_after_quote
more_elements_after_qoute   -> EOF | WSCHAR spaces_if_any EOF | element more_elements
more_elements               -> EOF | WSCHAR spaces_if_any EOF | "," spaces_if_any more_elements_after_qoute
element                     -> name spaces_if_any equal_plus_value spaces_if_any more_params
more_params                 -> ";" spaces_if_any more_params_after_semicolon | E
more_params_after_semicolon -> EOF | WSCHAR spaces_if_any EOF | param more_params
param                       -> name spaces_if_any equal_plus_value
equal_plus_value            -> "=" spaces_if_any value spaces_if_any | E
value                       -> '"' quoted_content '"' | TCHAR more_tchars | E
name                        -> TCHAR more_tchars
more_tchars                 -> TCHAR more_tchars | E
quoted_content              -> "\" ANY_CHAR quoted_content | ANY_CHAR_BUT_QUOTE quoted_content | E
spaces_if_any               -> WSCHAR spaces_if_any | E
*/

const char *HttpHeader::Parser::m_unescape_table = HttpHeader::Parser::prepareUnescapeTable();

HttpHeader::Parser::Parser(const std::string& header_value)
:	m_data_ptr(header_value.data()), m_chars_left(header_value.length()), m_is_error(false)
{
	if (m_chars_left == 0) {
		return;
	}
	try {
		next();
		consumeMoreElements();
	} catch (ErrorException& e) {
		DEBUGLOG("Error parsing header: " << header_value << " [" << __FILE__ << ":" << e.getLine() << "]");
		m_is_error = true;
	}
}

bool
HttpHeader::Parser::nextIsSpace()
{
	const char spaces[] = "\f\n\r\t ";
	if (nextIsEOF()) {
		return false;
	}
	return (strchr(spaces, nextChar()) != 0);
}

bool
HttpHeader::Parser::nextIsTChar()
{
	const char forbidden[] = "=,;\"\f\n\r\t ";
	// technically, the non-ascii characters are also forbidden, but we don't enforce that
	if (nextIsEOF()) {
		return false;
	}
	return (strchr(forbidden, nextChar()) == 0);
}

/* consume_more_elements_after_quote -> EOF | WSCHAR spaces_if_any EOF | element more_elements */
/* consume_more_elements             -> EOF | WSCHAR spaces_if_any EOF | "," spaces_if_any more_elements_after_qoute */
// merged into one and got rid of recursion
void
HttpHeader::Parser::consumeMoreElements()
{
	while (true) {
		consumeSpacesIfAny();
		if (nextIsEOF()) {
			return;
		}
		if (nextIs(',')) {
			next();
			continue;
		}
		consumeElement();
	}
}

/* element -> name spaces_if_any equal_plus_value spaces_if_any more_params */
void
HttpHeader::Parser::consumeElement()
{
	std::string name, value;
	std::list<NameValuePair> params;
	consumeName(name);
	consumeSpacesIfAny();
	consumeEqualPlusValue(value);
	consumeSpacesIfAny();
	consumeMoreParams(params);
	m_elements.push_back(HttpHeaderElement(name, value, params));
}

/* consume_more_params                 -> ";" spaces_if_any more_params_after_semicolon | E */
/* consume_more_params_after_semicolon -> EOF | WSCHAR spaces_if_any EOF | param more_params */
// merged into one and got rid of recursion
void
HttpHeader::Parser::consumeMoreParams(std::list<NameValuePair>& params)
{
	while (true) {
		consumeSpacesIfAny();
		if (nextIsEOF()) {
			return;
		}
		if (nextIs(';')) {
			next();
			consumeSpacesIfAny();
			if (nextIsEOF()) {
				return;
			}
			std::string name, value;
			consumeParam(name, value);
			params.push_back(NameValuePair(name, value));
			continue;
		}
		return;
	}
}

/* param -> name spaces_if_any equal_plus_value */
void
HttpHeader::Parser::consumeParam(std::string& name, std::string& value)
{
	consumeName(name);
	consumeSpacesIfAny();
	consumeEqualPlusValue(value);
}

/* equal_plus_value -> "=" spaces_if_any value spaces_if_any | E */
void
HttpHeader::Parser::consumeEqualPlusValue(std::string& value)
{
	if (nextIs('=')) {
		next();
		consumeSpacesIfAny();
		consumeValue(value);
		consumeSpacesIfAny();
		return;
	}
}

/* value -> '"' quoted_content '"' | TCHAR more_tchars | E */
void
HttpHeader::Parser::consumeValue(std::string& value)
{
	if (nextIs('"')) {
		next();
		consumeQuotedContent(value);
		if (!nextIs('"')) {
			throw ErrorException(__LINE__);
		}
		next();
		return;
	} else if (nextIsTChar()) {
		consumeMoreTChars(value);
		if (nextIs(',')) {
			// dates require special attention
			if (value == "Mon" || value == "Monday" ||
			    value == "Tue" || value == "Tuesday" ||
			    value == "Wed" || value == "Wednesday" ||
			    value == "Thu" || value == "Thursday" ||
			    value == "Fri" || value == "Friday" ||
			    value == "Sat" || value == "Saturday" ||
			    value == "Sun" || value == "Sunday") {
				value += ",";
				next();
				while (nextIsSpace() || nextIsTChar()) {
					value += nextChar();
					next();
				}
			}
		}
		return;
	}
}

/* name -> TCHAR more_tchars */
void
HttpHeader::Parser::consumeName(std::string& name)
{
	if (!nextIsTChar()) {
		throw ErrorException(__LINE__);
	}
	consumeMoreTChars(name);
}

/* more_tchars -> TCHAR more_tchars | E */
// optimized to avoid recursion
void
HttpHeader::Parser::consumeMoreTChars(std::string& append_to)
{
	while (nextIsTChar()) {
		append_to.append(1, nextChar());
		next();
	}
}

/* quoted_content -> "\" ANY_CHAR quoted_content | ANY_CHAR_BUT_QUOTE quoted_content | E */
// optimized to avoid recursion
void
HttpHeader::Parser::consumeQuotedContent(std::string& append_to)
{
	while (true) {
		if (nextIsEOF() || nextIs('"')) {
			break;
		}
		if (nextIs('\\')) {
			next();
			append_to.append(1, m_unescape_table[(unsigned char)nextChar()]);
			next();
		} else {
			append_to.append(1, nextChar());
			next();
		}
	}
}

/* spaces_if_any -> WSCHAR spaces_if_any | E */
// optimized to avoid recursion
void
HttpHeader::Parser::consumeSpacesIfAny()
{
	while (nextIsSpace()) {
		next();
	}
}

#define MAKE_ESCAPE(from,to) table[(unsigned char)(from)] = to

const char *
HttpHeader::Parser::prepareUnescapeTable()
{
	static char table[256];
	for (unsigned int i = 0; i < 256; i++) {
		table[i] = (unsigned char)i;
	}
	MAKE_ESCAPE('a', '\a');
	MAKE_ESCAPE('b', '\b');
	MAKE_ESCAPE('f', '\f');
	MAKE_ESCAPE('n', '\n');
	MAKE_ESCAPE('r', '\r');
	MAKE_ESCAPE('t', '\t');
	MAKE_ESCAPE('v', '\v');
	return table;
}

