/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef HTTPHEADER_H_
#define HTTPHEADER_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httpheaderelement.h"
#include <string>
#include <list>
#include <iosfwd>

class HttpHeader
{
public:
	class Elements
	{
		friend class HttpHeader;
	public:
		bool isParsingFailed() const;
		bool empty() const;
		bool hasElement(const std::string& name) const;
		const std::list<HttpHeaderElement>& getElements() const;
		const HttpHeaderElement* getFirstElement() const;
		const HttpHeaderElement* getLastElement() const;
		void removeFirstElement();
		void removeLastElement();
		void removeElementsByName(const std::string& name);
		void clear();
		void prependElement(const HttpHeaderElement& element);
		void appendElement(const HttpHeaderElement& element);
		void appendElements(const std::list<HttpHeaderElement>& elements);
		void appendElements(const Elements& elements);
		std::string toString() const;
	private:
		Elements(const std::string& value);
		void valueToParse(const std::string& value);
		void parse() const;
		bool isModified() const { return m_is_modified; }
		void setModified(bool value) { m_is_modified = value; }
		std::string m_value;
		mutable std::list<HttpHeaderElement> m_elements;
		mutable bool m_is_modified;
		mutable bool m_needs_parsing;
		mutable bool m_parsing_failed;
	};
	HttpHeader(const std::string& name, const std::string& value);
	HttpHeader(const HttpHeader& header);
	HttpHeader& operator=(const HttpHeader& header);
	~HttpHeader();
	const std::string& getName() const { return m_name; }
	const std::string& getLowercaseName() const { return m_lcname; }
	const std::list<std::string>& getValues() const;
	void setValues(const std::list<std::string>& values);
	void appendValue(const std::string& value);
	std::string getValue() const;
	void setValue(const std::string& value);
	Elements& elements();
	const Elements& elements() const;
	std::string toString() const;
	void toStream(std::ostream& strm) const;
private:
	class Parser
	{
	public:
		Parser(const std::string& header_value);
		bool isError() const { return m_is_error; }
		const std::list<HttpHeaderElement>& getElements() const { return m_elements; }
	private:
		class ErrorException
		{
		public:
			ErrorException(int line) : m_line(line) {}
			int getLine() const { return m_line; }
		private:
			const int m_line;
		};
		void next();
		char nextChar();
		bool nextIs(char ch);
		bool nextIsEOF();
		bool nextIsSpace();
		bool nextIsTChar();
		void consumeMoreElements();
		void consumeElement();
		void consumeMoreParams(std::list<NameValuePair>& params);
		void consumeParam(std::string& name, std::string& value);
		void consumeEqualPlusValue(std::string& value);
		void consumeValue(std::string& value);
		void consumeName(std::string& name);
		void consumeMoreTChars(std::string& append_to);
		void consumeQuotedContent(std::string& append_to);
		void consumeSpacesIfAny();
		static const char *prepareUnescapeTable();
		static const char *m_unescape_table;
		const char* m_data_ptr;
		unsigned int m_chars_left;
		unsigned int m_lookahead;
		std::list<HttpHeaderElement> m_elements;
		bool m_is_error;
	};
private:
	bool operator==(const HttpHeader& other) const; //
	bool operator!=(const HttpHeader& other) const; // not implemented
	void updateValuesFromElements() const;
	std::string m_name;
	std::string m_lcname;
	mutable std::list<std::string> m_values;
	mutable Elements *m_elements;
};

inline std::ostream&
operator<<(std::ostream& strm, const HttpHeader& header)
{
	header.toStream(strm);
	return strm;
}

inline void
HttpHeader::Parser::next()
{
	if (m_chars_left == 0) {
		m_lookahead = std::string::npos;
	} else {
		m_lookahead = (unsigned char)*m_data_ptr;
		--m_chars_left;
		++m_data_ptr;
	}
}

inline char
HttpHeader::Parser::nextChar()
{
	return (unsigned char)m_lookahead;
}

inline bool
HttpHeader::Parser::nextIs(char ch)
{
	return (m_lookahead == (unsigned char)ch);
}

inline bool
HttpHeader::Parser::nextIsEOF()
{
	return (m_lookahead == std::string::npos);
}

#endif

