/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httpstatefooters.h"
#include "httpstream.h"
#include "httpheader.h"
#include "httpheaderscollection.h"
#include <cstring>
#include <cstddef>

HttpStateFooters::HttpStateFooters(HttpStream* owner)
:	m_pOwner(owner), m_state(ST_CHECKING_FOR_END)
{
}

void
HttpStateFooters::consumeDataChunk(const uint8_t*& begin, const uint8_t* end)
{
	const uint8_t* unbuffered_begin = begin;
	switch (m_state) {
		case ST_CHECKING_FOR_END:
		case_ST_CHECKING_FOR_END: {
			if (*begin == '\n') {
				++begin;
				m_pOwner->eventGotFooters(m_headers);
				m_pOwner->eventMessageEnds(m_pOwner->m_isConnectionToBeClosed);
				m_pOwner->switchState(HttpStream::STATE_INITIAL);
				return;
			}
			if (*begin == '\r') {
				if (begin+1 == end) {
					++begin;
					return;
				} else if (begin[1] == '\n') {
					begin += 2;
					m_pOwner->eventGotFooters(m_headers);
					m_pOwner->eventMessageEnds(m_pOwner->m_isConnectionToBeClosed);
					m_pOwner->switchState(HttpStream::STATE_INITIAL);
					return;
				}
			}
			m_state = ST_LOOKING_FOR_COLON;
		}
		case ST_LOOKING_FOR_COLON: {
			const uint8_t* cpos = reinterpret_cast<const uint8_t*>(std::memchr(begin, ':', end-begin));
			if (!cpos) {
				if (std::memchr(begin, '\n', end-begin)) {
					m_pOwner->eventError(HttpStream::ERR_PARSING_HEADER);
					m_pOwner->switchState(HttpStream::STATE_ERROR);
					return;
				}
				bufAppend(unbuffered_begin, end);
				begin = end;
				return;
			}
			if (std::memchr(begin, '\n', cpos-begin)) {
				m_pOwner->eventError(HttpStream::ERR_PARSING_HEADER);
				m_pOwner->switchState(HttpStream::STATE_ERROR);
				return;
			}
			m_colonPos = cpos-begin+m_buf.size();
			begin = cpos+1;
			m_state = ST_LOOKING_FOR_NEWLINE;
		}
		case ST_LOOKING_FOR_NEWLINE:
		case_ST_LOOKING_FOR_NEWLINE: {
			const uint8_t* nlpos = reinterpret_cast<const uint8_t*>(std::memchr(begin, '\n', end-begin));
			if (!nlpos) {
				bufAppend(unbuffered_begin, end);
				begin = end;
				return;
			}
			begin = nlpos+1;
			m_state = ST_CHECKING_FOR_LWS;
		}
		case ST_CHECKING_FOR_LWS: {
			if (begin == end) {
				bufAppend(unbuffered_begin, end);
				return;
			}
			if (*begin == ' ' || *begin == '\t') {
				bufAppend(unbuffered_begin, begin);
				m_buf.pop_back(); // remove '\n'
				// m_buf can't be empty at this point, at least it contains a colon
				if (m_buf.back() == '\r') {
					m_buf.pop_back();
				}
				m_state = ST_CONSUMING_LWS;
			} else {
				if (m_buf.empty()) {
					const uint8_t* header_end = begin-1;
					// begin-1 is inside the initial interval (otherwise
					// m_buf would not be empty) and points to '\n'
					if (*(header_end-1) == '\r') {
						// header_end-1 is also inside the initial interval
						// because otherwise either m_buf would not be empty
						// or we would never reach this state
						--header_end;
					}
					addHeader(unbuffered_begin, header_end);
				} else {
					bufAppend(unbuffered_begin, begin);
					m_buf.pop_back(); // remove '\n'
					// m_buf can't be empty at this point, at least it contains a colon
					if (m_buf.back() == '\r') {
						m_buf.pop_back();
					}
					addHeader(&m_buf[0], &m_buf[m_buf.size()]);
					m_buf.clear();
				}
				unbuffered_begin = begin;
				m_state = ST_CHECKING_FOR_END;
				goto case_ST_CHECKING_FOR_END;
			}
		}
		case ST_CONSUMING_LWS: {
			for (const uint8_t* cur = begin;; ++cur) {
				if (cur == end) {
					return;
				}
				if (*cur != ' ' && *cur != '\t') {
					m_buf.push_back(' ');
					begin = cur;
					unbuffered_begin = cur;
					break;
				}
			}
			m_state = ST_LOOKING_FOR_NEWLINE;
			goto case_ST_LOOKING_FOR_NEWLINE;
		}
	}
}

void
HttpStateFooters::consumeEOF()
{
	m_pOwner->eventError(HttpStream::ERR_UNEXPECTED_EOF);
	m_pOwner->switchState(HttpStream::STATE_ERROR);
}

void
HttpStateFooters::cleanup()
{
	m_state = ST_CHECKING_FOR_END;
	m_headers.clear();
	std::vector<uint8_t>().swap(m_buf);
}

void
HttpStateFooters::addHeader(const uint8_t* begin, const uint8_t* end)
{
	std::string name(reinterpret_cast<const char*>(begin), reinterpret_cast<const char*>(begin+m_colonPos));
	for (begin = begin+m_colonPos+1; begin != end; ++begin) {
		if (*begin != ' ' && *begin != '\t') {
			break;
		}
	}
	for (; end != begin; --end) {
		char ch = *(end-1);
		if (ch != ' ' && ch != '\t') {
			break;
		}
	}
	std::string value(reinterpret_cast<const char*>(begin), reinterpret_cast<const char*>(end));
	m_headers.addHeader(name, value);
}

