/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httpstatefooters.h"
#include "httpstream.h"
#include "httpheaderscollection.h"
#include "httprequestline.h"
#include "httpstatusline.h"
#include "stringutils.h"
#include <string>
#include <cstring>
#include <cstddef>

HttpStateHeaders::HttpStateHeaders(HttpStream* owner)
:	m_pOwner(owner), m_state(ST_CHECKING_FOR_END),
	m_requestMetadata(HttpRequestLine("", URI(""), HttpVersion::HTTP_1_1)),
	m_responseMetadata(HttpStatusLine(HttpVersion::HTTP_1_1, 200))
{
}

void
HttpStateHeaders::consumeDataChunk(const uint8_t*& begin, const uint8_t* end)
{
	const uint8_t* unbuffered_begin = begin;
	switch (m_state) {
		case ST_CHECKING_FOR_END:
		case_ST_CHECKING_FOR_END: {
			if (*begin == '\n') {
				++begin;
				headersReceived();
				return;
			}
			if (*begin == '\r') {
				if (begin+1 == end) {
					++begin;
					return;
				} else if (begin[1] == '\n') {
					begin += 2;
					headersReceived();
					return;
				}
			}
			m_state = ST_LOOKING_FOR_COLON;
		}
		case ST_LOOKING_FOR_COLON: {
			const uint8_t* cpos = reinterpret_cast<const uint8_t*>(std::memchr(begin, ':', end-begin));
			if (!cpos) {
				if (std::memchr(begin, '\n', end-begin)) {
					m_pOwner->eventError(HttpStream::ERR_PARSING_HEADER);
					m_pOwner->switchState(HttpStream::STATE_ERROR);
					return;
				}
				bufAppend(unbuffered_begin, end);
				begin = end;
				return;
			}
			if (std::memchr(begin, '\n', cpos-begin)) {
				m_pOwner->eventError(HttpStream::ERR_PARSING_HEADER);
				m_pOwner->switchState(HttpStream::STATE_ERROR);
				return;
			}
			m_colonPos = cpos-begin+m_buf.size();
			begin = cpos+1;
			m_state = ST_LOOKING_FOR_NEWLINE;
		}
		case ST_LOOKING_FOR_NEWLINE:
		case_ST_LOOKING_FOR_NEWLINE: {
			const uint8_t* nlpos = reinterpret_cast<const uint8_t*>(std::memchr(begin, '\n', end-begin));
			if (!nlpos) {
				bufAppend(unbuffered_begin, end);
				begin = end;
				return;
			}
			begin = nlpos+1;
			m_state = ST_CHECKING_FOR_LWS;
		}
		case ST_CHECKING_FOR_LWS: {
			if (begin == end) {
				bufAppend(unbuffered_begin, end);
				return;
			}
			if (*begin == ' ' || *begin == '\t') {
				bufAppend(unbuffered_begin, begin);
				m_buf.pop_back(); // remove '\n'
				// m_buf can't be empty at this point, at least it contains a colon
				if (m_buf.back() == '\r') {
					m_buf.pop_back();
				}
				m_state = ST_CONSUMING_LWS;
			} else {
				if (m_buf.empty()) {
					const uint8_t* header_end = begin-1;
					// begin-1 is inside the initial interval (otherwise
					// m_buf would not be empty) and points to '\n'
					if (*(header_end-1) == '\r') {
						// header_end-1 is also inside the initial interval
						// because otherwise either m_buf would not be empty
						// or we would never reach this state
						--header_end;
					}
					addHeader(unbuffered_begin, header_end);
				} else {
					bufAppend(unbuffered_begin, begin);
					m_buf.pop_back(); // remove '\n'
					// m_buf can't be empty at this point, at least it contains a colon
					if (m_buf.back() == '\r') {
						m_buf.pop_back();
					}
					addHeader(&m_buf[0], &m_buf[m_buf.size()]);
					m_buf.clear();
				}
				unbuffered_begin = begin;
				m_state = ST_CHECKING_FOR_END;
				goto case_ST_CHECKING_FOR_END;
			}
		}
		case ST_CONSUMING_LWS: {
			for (const uint8_t* cur = begin;; ++cur) {
				if (cur == end) {
					return;
				}
				if (*cur != ' ' && *cur != '\t') {
					m_buf.push_back(' ');
					begin = cur;
					unbuffered_begin = cur;
					break;
				}
			}
			m_state = ST_LOOKING_FOR_NEWLINE;
			goto case_ST_LOOKING_FOR_NEWLINE;
		}
	}
}

void
HttpStateHeaders::consumeEOF()
{
	m_pOwner->eventError(HttpStream::ERR_UNEXPECTED_EOF);
	m_pOwner->switchState(HttpStream::STATE_ERROR);
}

void
HttpStateHeaders::cleanup()
{
	m_state = ST_CHECKING_FOR_END;
	m_requestMetadata.headers().clear();
	m_responseMetadata.headers().clear();
	std::vector<uint8_t>().swap(m_buf);
}

void
HttpStateHeaders::setRequestLine(const std::string& method, const URI& uri, const HttpVersion& http_version)
{
	HttpRequestLine& line = m_requestMetadata.requestLine();
	line.setMethod(method);
	line.setURI(uri);
	line.setHttpVersion(http_version);
}

void
HttpStateHeaders::setStatusLine(const HttpVersion& http_version, int status_code, const std::string message)
{
	HttpStatusLine& line = m_responseMetadata.statusLine();
	line.setHttpVersion(http_version);
	line.setCodeAndMessage(status_code, message);
}

void
HttpStateHeaders::addHeader(const uint8_t* begin, const uint8_t* end)
{
	HttpHeadersCollection* headers = m_pOwner->isRequestStream()
		? &m_requestMetadata.headers() : &m_responseMetadata.headers();
	
	std::string name(reinterpret_cast<const char*>(begin), reinterpret_cast<const char*>(begin+m_colonPos));
	for (begin = begin+m_colonPos+1; begin != end; ++begin) {
		if (*begin != ' ' && *begin != '\t') {
			break;
		}
	}
	for (; end != begin; --end) {
		char ch = *(end-1);
		if (ch != ' ' && ch != '\t') {
			break;
		}
	}
	std::string value(reinterpret_cast<const char*>(begin), reinterpret_cast<const char*>(end));
	
	headers->addHeader(name, value);
}

void
HttpStateHeaders::headersReceived()
{
	HttpHeadersCollection* headers = m_pOwner->isRequestStream()
		? &m_requestMetadata.headers() : &m_responseMetadata.headers();
	
	bool conn_close = (headers->getHeader("Connection").elements().hasElement("close")
		|| headers->getHeader("Proxy-Connection").elements().hasElement("close"));
	bool keep_alive = !conn_close && ((headers->getHeader("Connection").elements().hasElement("keep-alive")
		|| headers->getHeader("Proxy-Connection").elements().hasElement("keep-alive")));
	bool no_body = false;
	
	if (m_pOwner->isRequestStream()) {
		if (m_requestMetadata.requestLine().getHttpVersion() < HttpVersion::HTTP_1_1 && !keep_alive) {
			conn_close = true;
		}
		std::string method = m_requestMetadata.requestLine().getMethod();
		if (method == "GET" || method == "HEAD") {
			no_body = true;
		}
	} else {
		if (m_responseMetadata.statusLine().getHttpVersion() < HttpVersion::HTTP_1_1 && !keep_alive) {
			conn_close = true;
		}
		int status = m_responseMetadata.statusLine().getCode();
		if (status == 304 || (status >= 100 && status < 200)) {
			no_body = true;
		} else if (m_pOwner->isResponseOnHeadRequest()) {
			no_body = true;
		}
	}
	
	if (no_body) {
		if (m_pOwner->isRequestStream()) {
			m_requestMetadata.setBodyStatus(HttpMessageMetadata::NO_BODY_BUT_PRESERVE_HEADERS);
			m_pOwner->eventGotMetadata(m_requestMetadata);
		} else {
			m_responseMetadata.setBodyStatus(HttpMessageMetadata::NO_BODY_BUT_PRESERVE_HEADERS);
			m_pOwner->eventGotMetadata(m_responseMetadata);
		}
		m_pOwner->switchState(HttpStream::STATE_INITIAL);
		m_pOwner->eventMessageEnds(conn_close);
		return;
	}
		
	HttpMessageMetadata::BodyStatus body_status = HttpMessageMetadata::BODY_SIZE_UNKNOWN;
	std::size_t body_size = 0;
	
	if (headers->hasHeader("Content-Length")) {
		body_status = HttpMessageMetadata::BODY_SIZE_KNOWN;
		body_size = StringUtils::parseUnsigned(headers->getHeader("Content-Length").getValue());
	}
	
	bool chunked_body = false;
	HttpHeader *tenc_header = headers->getHeaderPtr("Transfer-Encoding");
	if (tenc_header) {
		if (tenc_header->elements().empty()) {
			headers->removeHeader("Transfer-Encoding");
		} else {
			static const char chunked[] = { 'c','h','u','n','k','e','d' };
			static const char identity[] = { 'i','d','e','n','t','i','t','y' };
			std::string last_encoding = tenc_header->elements().getLastElement()->getName();
			chunked_body = StringUtils::ciEquals(last_encoding, chunked, chunked+sizeof(chunked));
			if (chunked_body || StringUtils::ciEquals(last_encoding,  identity, identity+sizeof(identity))) {
				tenc_header->elements().removeLastElement();
				if (tenc_header->elements().empty()) {
					headers->removeHeader("Transfer-Encoding");
				}
			}
		}
	}
	
	if (body_status == HttpMessageMetadata::BODY_SIZE_KNOWN &&
	    (chunked_body || (conn_close && !m_pOwner->isRequestStream() && headers->hasHeader("Content-Encoding")))) {
		/*
		We can't trust the Content-Length header of encoded content because
		some servers set it to the length of original content, while others
		set it to the length of encoded content.
		However, if connection is not going to be closed after the current message,
		we have no choise but to trust it, as we have no other way to determine the
		length of the message.
		*/
		body_status = HttpMessageMetadata::BODY_SIZE_UNKNOWN;
	}
	
	if (!chunked_body && body_status == HttpMessageMetadata::BODY_SIZE_UNKNOWN) {
		if (m_pOwner->isRequestStream()) {
			std::string method = m_requestMetadata.requestLine().getMethod();
			if (method == "POST" || method == "PUT") {
				m_pOwner->eventError(HttpStream::ERR_LENGTH_REQUIRED);
				m_pOwner->switchState(HttpStream::STATE_ERROR);
				return;
			} else {
				body_status = HttpMessageMetadata::BODY_SIZE_KNOWN;
				body_size = 0;
			}
		}
	}
	
	m_pOwner->m_isConnectionToBeClosed = (!m_pOwner->isRequestStream() && (conn_close
		|| (body_status == HttpMessageMetadata::BODY_SIZE_UNKNOWN && !chunked_body)));
	
	if (m_pOwner->isRequestStream()) {
		m_requestMetadata.setBodyStatus(body_status);
		m_requestMetadata.setBodySize(body_size);
		m_pOwner->eventGotMetadata(m_requestMetadata);
	} else {
		m_responseMetadata.setBodyStatus(body_status);
		m_responseMetadata.setBodySize(body_size);
		m_pOwner->eventGotMetadata(m_responseMetadata);
	}
	
	if (chunked_body) {
		m_pOwner->switchState(HttpStream::STATE_CHUNK_HEADER);
	} else {
		if (body_status == HttpMessageMetadata::BODY_SIZE_KNOWN) {
			if (body_size == 0) {
				m_pOwner->switchState(HttpStream::STATE_INITIAL);
				m_pOwner->eventMessageEnds(m_pOwner->isRequestStream() ? false : conn_close);
				return;
			} else {
				m_pOwner->m_stateFlatBody.setBodySize(body_size);
				// this line must only be executed if we are going to switch to STATE_FLAT_BODY,
				// otherwise HttpStateFlatBody::cleanup() won't be called, and this body
				// size will affect the next response (in case it has undefined body size).
			}
		}
		m_pOwner->switchState(HttpStream::STATE_FLAT_BODY);
	}
}

