/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httpstatestartline.h"
#include "httpstream.h"
#include "httpresponsemetadata.h"
#include "httpversion.h"
#include "stringutils.h"
#include "uri.h"
#include <cstddef>
#include <cstring>
#include <string>

HttpStateStartLine::HttpStateStartLine(HttpStream* owner)
:	m_pOwner(owner), m_lineStartPosInBuffer(0)
{
}

void
HttpStateStartLine::consumeDataChunk(const uint8_t*& begin, const uint8_t* end)
{
	const uint8_t* spos = begin;
	while (true) {
		const uint8_t* epos = reinterpret_cast<const uint8_t*>(std::memchr(begin, '\n', end-begin));
		if (!epos) {
			std::size_t size = m_buf.size();
			m_buf.resize(size+end-begin);
			std::memcpy(&m_buf[size], begin, end-begin);
			begin = end;
			return;
		} else {
			++epos;
			if (m_buf.empty()) {
				if (processStartLine(begin, epos, spos)) {
					begin = epos;
					return;
				} else {
					if ((begin = epos) == end) {
						m_pOwner->switchState(HttpStream::STATE_INITIAL);
						return;
					}
					continue;
				}
			} else {
				std::size_t bufpos = m_buf.size();
				m_buf.resize(epos-begin+bufpos);
				std::memcpy(&m_buf[bufpos], begin, epos-begin);
				if (processStartLine(&m_buf[m_lineStartPosInBuffer], &m_buf[m_buf.size()], &m_buf[0])) {
					begin = epos;
					return;
				} else {
					m_lineStartPosInBuffer = m_buf.size();
					if ((begin = epos) == end) {
						m_pOwner->switchState(HttpStream::STATE_INITIAL);
						return;
					}
					continue;
				}
			}
		}
	}
}

void
HttpStateStartLine::consumeEOF()
{
	if (m_pOwner->isRequestStream()) {
		m_pOwner->eventError(HttpStream::ERR_UNEXPECTED_EOF);
		m_pOwner->switchState(HttpStream::STATE_ERROR);
	} else {
		HttpResponseMetadata metadata(HttpStatusLine(HttpVersion::HTTP_0_9, 200));
		metadata.setBodyStatus(HttpMessageMetadata::BODY_SIZE_UNKNOWN);
		m_pOwner->eventGotMetadata(metadata);
		m_pOwner->eventGotContentPart(&m_buf[0], &m_buf[m_buf.size()]);
		m_pOwner->switchState(HttpStream::STATE_INITIAL);
		m_pOwner->eventMessageEnds(true);
	}
}

void
HttpStateStartLine::cleanup()
{
	std::vector<uint8_t>().swap(m_buf);
	m_lineStartPosInBuffer = 0;
}

// returns false if the start line is empty and thus should be ignored
bool
HttpStateStartLine::processStartLine(const uint8_t* begin, const uint8_t* end, const uint8_t* data_start)
{
	const uint8_t* data_end = end;
	--end; // now it points to '\n'
	if (begin != end-1 && *(end-1) == '\r') {
		--end; // now it points to '\r'
	}
	if (begin == end) {
		return false;
	}
	if (m_pOwner->isRequestStream()) {
		const uint8_t* method_end = reinterpret_cast<const uint8_t*>(std::memchr(begin, ' ', end-begin));
		if (!method_end) {
			m_pOwner->eventError(HttpStream::ERR_PARSING_REQUEST_LINE);
			m_pOwner->switchState(HttpStream::STATE_ERROR);
			return true;
		}
		std::string method(reinterpret_cast<const char*>(begin), reinterpret_cast<const char*>(method_end));
		const uint8_t* uri_end = end;
		while (*--uri_end != ' ');
		std::string uri_str(reinterpret_cast<const char*>(method_end+1),
			reinterpret_cast<const char*>(uri_end == method_end ? end : uri_end));
		HttpVersion http_version = HttpVersion::HTTP_0_9;
		if (uri_end != method_end) {
			if (!http_version.parse(uri_end+1, end)) {
				m_pOwner->eventError(HttpStream::ERR_PARSING_REQUEST_LINE);
				m_pOwner->switchState(HttpStream::STATE_ERROR);
				return true;
			}
		}
		m_pOwner->m_stateHeaders.setRequestLine(method, URI(uri_str), http_version);
		m_pOwner->switchState(HttpStream::STATE_HEADERS);
	} else {
		HttpVersion http_version = HttpVersion::HTTP_0_9;
		const uint8_t* pos = reinterpret_cast<const uint8_t*>(std::memchr(begin, ' ', end-begin));
		if (!pos) {
			switchToSimpleResponse(data_start, data_end);
			return true;
		}
		if (http_version.parse(begin, pos)) {
			begin = pos+1;
		}
		std::string message;
		pos = reinterpret_cast<const uint8_t*>(std::memchr(begin, ' ', end-begin));
		if (pos) {
			message.assign(reinterpret_cast<const char*>(pos+1), reinterpret_cast<const char*>(end));
		} else {
			pos = end;
		}
		const uint8_t* epos = pos;
		int code = StringUtils::parseUnsigned(begin, epos);
		if (epos != pos) {
			switchToSimpleResponse(data_start, data_end);
			return true;
		}
		m_pOwner->m_stateHeaders.setStatusLine(http_version, code, message);
		m_pOwner->switchState(HttpStream::STATE_HEADERS);
	}
	return true;
}

void
HttpStateStartLine::switchToSimpleResponse(const uint8_t* begin, const uint8_t* end)
{
	m_pOwner->m_isConnectionToBeClosed = true;
	HttpResponseMetadata metadata(HttpStatusLine(HttpVersion::HTTP_0_9, 200));
	metadata.setBodyStatus(HttpMessageMetadata::BODY_SIZE_UNKNOWN);
	m_pOwner->eventGotMetadata(metadata);
	m_pOwner->eventGotContentPart(begin, end);
	m_pOwner->switchState(HttpStream::STATE_SIMPLE_RESPONSE);
}

