/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httpstreamwriter.h"
#include "httpheader.h"
#include "httpheaderelement.h"
#include "date.h"
#include "stringutils.h"
#include <string>
#include <cstring>

HttpStreamWriter::HttpStreamWriter(Receiver* receiver,
	HttpMessageMetadata& metadata, bool chunked_ok, bool force_conn_close)
:	m_pReceiver(receiver), m_isChunkedOutput(false), m_offset(DATA_BEGIN)
{
	m_noBody = (metadata.getBodyStatus() == HttpMessageMetadata::NO_BODY_BUT_PRESERVE_HEADERS);
	if (m_noBody) {
		m_isConnectionToBeClosed = force_conn_close;
	} else {
		bool body_size_known = (metadata.getBodyStatus() == HttpMessageMetadata::BODY_SIZE_KNOWN);
		if (!body_size_known) {
			metadata.headers().removeHeader("Content-Length");
			if (chunked_ok) {
				m_isChunkedOutput = true;
				HttpHeader* teh = metadata.headers().getHeaderPtr("Transfer-Encoding");
				if (!teh || teh->elements().getElements().empty()) {
					metadata.headers().setHeader(HttpHeader("Transfer-Encoding", "chunked"));
				} else {
					teh->elements().appendElement(HttpHeaderElement("chunked"));
				}
			}
		} else {
			metadata.headers().setHeader(HttpHeader("Content-Length",
				StringUtils::fromNumber(metadata.getBodySize())));
		}
		m_isConnectionToBeClosed = (!m_isChunkedOutput && !body_size_known) || force_conn_close;
		HttpHeader* ch = metadata.headers().getHeaderPtr("Connection");
		if (m_isConnectionToBeClosed) {
			if (ch) {
				if (!ch->elements().hasElement("close")) {
					ch->elements().appendElement(HttpHeaderElement("close"));
				}
			} else {
				metadata.headers().setHeader(HttpHeader("Connection", "close"));
			}
		} else {
			if (ch) {
				ch->elements().removeElementsByName("close");
				if (ch->elements().empty()) {
					metadata.headers().removeHeader("Connection");
				}
			}
		}
	}
	if (!metadata.headers().hasHeader("Date")) {
		metadata.headers().setHeader(HttpHeader("Date", Date::formatCurrentTime()));
	}
	std::ostringstream strm;
	strm << metadata.getStartLine() << "\r\n" << metadata.headers() << "\r\n";
	outputString(strm.str());
}

void
HttpStreamWriter::write(const std::string& data)
{
	write(
		reinterpret_cast<const uint8_t*>(data.data()),
		reinterpret_cast<const uint8_t*>(data.data()+data.length())
	);
}

void
HttpStreamWriter::write(const uint8_t* begin, const uint8_t* end)
{
	if (begin == end || m_noBody) {
		return;
	}
	if (!m_isChunkedOutput) {
		m_pReceiver->moreData(begin, end);
	} else {
		for (std::size_t todo; (todo = end - begin); ) {
			if (m_offset + todo < DATA_END) {
				std::memcpy(&m_buffer[m_offset], begin, todo);
				m_offset += todo;
				return;
			}
			std::size_t len = DATA_END - m_offset;
			std::memcpy(&m_buffer[m_offset], begin, len);
			m_offset += len;
			begin += len;
			writeChunkHeader(m_offset - DATA_BEGIN);
			m_buffer[DATA_END] = '\r';
			m_buffer[DATA_END+1] = '\n';
			m_offset = DATA_BEGIN;
			m_pReceiver->moreData(m_buffer, m_buffer+BUF_SIZE);
		}
	}
}

void
HttpStreamWriter::end()
{
	if (m_isChunkedOutput) {
		static const uint8_t trailer[] = { '0', '\r', '\n', '\r', '\n' };
		if (m_offset > DATA_BEGIN) {
			writeChunkHeader(m_offset - DATA_BEGIN);
			m_buffer[m_offset++] = '\r';
			m_buffer[m_offset++] = '\n';
			if (m_offset + sizeof(trailer) <= BUF_SIZE) {
				std::memcpy(&m_buffer[m_offset], trailer, sizeof(trailer));
				m_pReceiver->moreData(m_buffer, m_buffer + m_offset + sizeof(trailer));
			} else {
				m_pReceiver->moreData(m_buffer, m_buffer + m_offset);
				m_pReceiver->moreData(trailer, trailer + sizeof(trailer));
			}
		} else {
			m_pReceiver->moreData(trailer, trailer + sizeof(trailer));
		}
	}
	m_pReceiver->done(m_isConnectionToBeClosed);
}

void
HttpStreamWriter::outputString(const std::string& str)
{
	m_pReceiver->moreData(
		reinterpret_cast<const uint8_t*>(str.data()),
		reinterpret_cast<const uint8_t*>(str.data()+str.length())
	);
}

void
HttpStreamWriter::writeChunkHeader(std::size_t size)
{
	static const char symbols[] = {
		'0', '1', '2', '3', '4', '5', '6', '7',
		'8', '9', 'a', 'b', 'c', 'd', 'e', 'f'
	};
	for (int i = CHUNK_SIZE_DIGITS-1; i >= 0; --i) {
		m_buffer[i] = symbols[size & 0xf];
		size >>= 4;
	}
	m_buffer[CHUNK_SIZE_DIGITS] = '\r';
	m_buffer[CHUNK_SIZE_DIGITS+1] = '\n';
}

