/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "imageinspector.h"
#include "gifimageinspector.h"
#include "pngimageinspector.h"
#include "jpegimageinspector.h"
#include <cstring>
#include <cstddef>
#include <algorithm>

#define SigJPEG JPEGImageInspector::SIGNATURE
#define SigGIF  GIFImageInspector::SIGNATURE
#define SigPNG  PNGImageInspector::SIGNATURE

ImageInspector::ImageInspector()
:	m_state(ST_SIGNATURE), m_type(T_UNKNOWN),
	m_typesBeingConsidered((1<<T_UNKNOWN)-1)
{
}

void
ImageInspector::consumeDataChunk(const uint8_t* begin, const uint8_t* end)
{
	if (m_state == ST_DELEGATION) {
		m_ptrDelegate->consumeDataChunk(begin, end);
		return;
	}
	
	if (m_state == ST_SIGNATURE) {
		if (considerType(T_JPEG, begin, end, SigJPEG, sizeof(SigJPEG))) {
			m_ptrDelegate.reset(new JPEGImageInspector);
		} else if (considerType(T_GIF, begin, end, SigGIF, sizeof(SigGIF))) {
			m_ptrDelegate.reset(new GIFImageInspector);
		} else if (considerType(T_PNG, begin, end, SigPNG, sizeof(SigPNG))) {
			m_ptrDelegate.reset(new PNGImageInspector);
		}
		if (m_ptrDelegate.get()) {
			if (!m_data.empty()) {
				m_ptrDelegate->consumeDataChunk(&m_data[0], &m_data[m_data.size()]);
			}
			m_ptrDelegate->consumeDataChunk(begin, end);
			m_state = ST_DELEGATION;
		} else {
			std::size_t size = m_data.size();
			m_data.resize(end - begin + size);
			std::memcpy(&m_data[size], begin, end - begin);
			if (m_typesBeingConsidered == 0) {
				m_state = ST_UNKNOWN_TYPE;
			}
		}
	}
}

void
ImageInspector::consumeEOF()
{
	if (m_state == ST_DELEGATION) {
		m_ptrDelegate->consumeEOF();
	} else {
		m_state = ST_UNKNOWN_TYPE;
	}
}

void
ImageInspector::reset()
{
	m_ptrDelegate.reset(0);
	m_state = ST_SIGNATURE;
	m_type = T_UNKNOWN;
	std::vector<uint8_t>().swap(m_data);
}

std::pair<const uint8_t*, const uint8_t*>
ImageInspector::getAccumulatedData() const
{
	if (m_state == ST_DELEGATION) {
		return m_ptrDelegate->getAccumulatedData();
	} else {
		return std::pair<const uint8_t*, const uint8_t*>(&m_data[0], &m_data[m_data.size()]);
	}
}

bool
ImageInspector::isInspectionInProgress() const
{
	if (m_state == ST_DELEGATION) {
		return m_ptrDelegate->isInspectionInProgress();
	} else {
		return (m_state == ST_SIGNATURE);
	}
}

bool
ImageInspector::isInspectionFailed() const
{
	if (m_state == ST_DELEGATION) {
		return m_ptrDelegate->isInspectionFailed();
	} else {
		return (m_state == ST_UNKNOWN_TYPE);
	}
}

int
ImageInspector::getWidth() const
{
	if (m_state == ST_DELEGATION) {
		return m_ptrDelegate->getWidth();
	} else {
		return 0;
	}
}

int
ImageInspector::getHeight() const
{
	if (m_state == ST_DELEGATION) {
		return m_ptrDelegate->getHeight();
	} else {
		return 0;
	}
}

bool
ImageInspector::isAnimated() const
{
	if (m_state == ST_DELEGATION) {
		return m_ptrDelegate->isAnimated();
	} else {
		return false;
	}
}

bool
ImageInspector::considerType(ImageType type, const uint8_t* begin,
	const uint8_t* end, const uint8_t* sig, std::size_t sig_size)
{
	if (!(m_typesBeingConsidered & (1<<type))) {
		return false;
	}
	std::size_t len = std::min(std::size_t(end-begin), sig_size-m_data.size());
	if (std::memcmp(sig+m_data.size(), begin, len)) {
		m_typesBeingConsidered &= ~(1<<type);
		return false;
	}
	return (m_data.size()+len == sig_size);
}

