/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "imageresponsefilter.h"
#include "responsefilterchain.h"
#include "decompressorresponsefilter.h"
#include "htmlprocessor.h"
#include "httpheader.h"
#include "replacementimage.h"
#include "conf.h"

AbstractResponseFilter*
ImageResponseFilter::create(NextFilterProvider* follower,
	HttpResponseMetadata* metadata, const Params& params)
{
	int status_code = metadata->statusLine().getCode();
	if (status_code != 200) {
		return 0;
	}
	if (metadata->getBodyStatus() == HttpResponseMetadata::NO_BODY_BUT_PRESERVE_HEADERS ||
	    (metadata->getBodyStatus() == HttpResponseMetadata::BODY_SIZE_KNOWN && metadata->getBodySize() == 0)) {
		// no body in response
		return 0;
	}
	
	if (ResponseFilterChain::isBodyEncoded(*metadata)) {
		NextFilterProvider provider =
			SigC::bind(SigC::slot(&ImageResponseFilter::construct), follower, &params);
		return DecompressorResponseFilter::createStack(&provider, metadata, 2);
	} else {
		return construct(metadata, follower, &params);
	}
}

ImageResponseFilter::ImageResponseFilter(NextFilterProvider* follower,
	HttpResponseMetadata* metadata, const Params* params)
:	m_followerProvider(*follower), m_metadata(*metadata), m_params(*params),
	m_state(ST_INSPECTING), m_pFollower(0)
{
}

ImageResponseFilter::~ImageResponseFilter()
{
	if (m_pFollower) {
		m_pFollower->destroy();
	}
}

void
ImageResponseFilter::consumeDataChunk(const uint8_t* begin, const uint8_t* end)
{
	if (m_state == ST_INSPECTING) {
		m_inspector.consumeDataChunk(begin, end);
		if (!m_inspector.isInspectionInProgress()) {
			if (isAd()) {
				m_state = ST_SUBSTITUTING;
				performSubstitution();
			} else {
				m_state = ST_FORWARDING;
				m_pFollower = m_followerProvider(&m_metadata);
				m_pFollower->consumeDataChunk(
					m_inspector.getAccumulatedData().first,
					m_inspector.getAccumulatedData().second
				);
			}
		}
	} else if (m_state == ST_FORWARDING) {
		m_pFollower->consumeDataChunk(begin, end);
	}
	// do nothing if m_state == ST_SUBSTITUTING
}

void
ImageResponseFilter::consumeEOF()
{
	if (m_state == ST_SUBSTITUTING) {
		return;
	}
	if (m_state == ST_INSPECTING) {
		m_inspector.consumeEOF();
		if (isAd()) {
			m_state = ST_SUBSTITUTING;
			performSubstitution();
			return;
		} else {
			m_state = ST_FORWARDING;
			m_pFollower = m_followerProvider(&m_metadata);
			m_pFollower->consumeDataChunk(
				m_inspector.getAccumulatedData().first,
				m_inspector.getAccumulatedData().second
			);
		}
	}
	m_pFollower->consumeEOF();
}

bool
ImageResponseFilter::isAd() const
{
	int res = HtmlProcessor::imageIsAd(m_params.ad_rule, m_inspector.getWidth(), m_inspector.getHeight(),
		m_params.target, m_params.external_link, m_params.external_img, m_params.suspicious_link,
		m_params.suspicious_img, (m_inspector.isAnimated() ? 1 : -1), (m_params.redir ? 1 : -1),
		(m_params.cookies ? 1 : -1)
	);
	if (res == 1) {
		return true;
	}
	return false;
}

void
ImageResponseFilter::performSubstitution()
{
	ReplacementImage image(m_params.width==0 ? m_inspector.getWidth() : m_params.width,
		m_params.height==0 ? m_inspector.getHeight() : m_params.height,
		m_params.config->getDrawBorders(), m_params.config->getBorderColor()
	);
	m_metadata.headers().setHeader(HttpHeader("Content-Type", image.getContentType()));
	m_metadata.setBodyStatus(HttpMessageMetadata::BODY_SIZE_KNOWN);
	m_metadata.setBodySize(image.getBinaryData().second-image.getBinaryData().first);
	m_pFollower = m_followerProvider(&m_metadata);
	m_pFollower->consumeDataChunk(image.getBinaryData().first, image.getBinaryData().second);
	m_pFollower->consumeEOF();
}

AbstractResponseFilter*
ImageResponseFilter::construct(HttpResponseMetadata* metadata,
	NextFilterProvider* follower, const Params* params)
{
#ifdef DEBUG
	metadata->headers().addHeader("X-filtered-by", "ImageResponseFilter");
#endif
	return new ImageResponseFilter(follower, metadata, params);
}

