/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef INTRUSIVEPTR_H_
#define INTRUSIVEPTR_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

class IntrusiveRefCounter
{
	template<typename T> friend class IntrusivePtr;
public:
	IntrusiveRefCounter() : m_refCounter(0) {}
	
	virtual ~IntrusiveRefCounter() {}
private:
	void intrusivePtrRef() { ++m_refCounter; }
	
	void intrusivePtrUnref() {
		if (--m_refCounter == 0) {
			delete this;
		}
	}
private:
	unsigned m_refCounter;
};


template<typename T>
class IntrusivePtr
{
public:
	IntrusivePtr() : m_pObj(0) {}
	
	explicit
	IntrusivePtr(T* obj);
	
	IntrusivePtr(const IntrusivePtr& other);
	
	template<typename OT>
	IntrusivePtr(const IntrusivePtr<OT>& other);
	
	~IntrusivePtr();
public:
	IntrusivePtr& operator=(const IntrusivePtr& rhs);
	
	template<typename OT>
	IntrusivePtr& operator=(const IntrusivePtr<OT>& rhs);
	
	T& operator*() const { return *m_pObj; }
	
	T* operator->() const { return m_pObj; }
	
	bool operator==(const IntrusivePtr& rhs) {
		return get() == rhs.get();
	}
	
	bool operator!=(const IntrusivePtr& rhs) {
		return get() != rhs.get();
	}
	
private:
	typedef void (IntrusivePtr::*MemberPtr)(T*);
public:
	operator MemberPtr() const {
		return m_pObj ? &IntrusivePtr::reset : 0;
	}
	
	T* get() const { return m_pObj; }
	
	void reset(T* obj);
	
	void swap(IntrusivePtr& other);
private:
	T* m_pObj;
};


template<typename T>
IntrusivePtr<T>::IntrusivePtr(T* obj)
:	m_pObj(obj)
{
	if (obj)
		obj->intrusivePtrRef();
}

template<typename T>
IntrusivePtr<T>::IntrusivePtr(const IntrusivePtr& other)
:	m_pObj(other.m_pObj)
{
	if (other.m_pObj)
		other.m_pObj->intrusivePtrRef();
}

template<typename T>
template<typename OT>
IntrusivePtr<T>::IntrusivePtr(const IntrusivePtr<OT>& other)
:	m_pObj(other.m_pObj)
{
	if (other.m_pObj)
		other.m_pObj->intrusivePtrRef();
}

template<typename T>
IntrusivePtr<T>::~IntrusivePtr()
{
	if (m_pObj)
		m_pObj->intrusivePtrUnref();
}

template<typename T>
IntrusivePtr<T>&
IntrusivePtr<T>::operator=(const IntrusivePtr& rhs)
{
	IntrusivePtr(rhs).swap(*this);
	return *this;
}

template<typename T>
template<typename OT>
IntrusivePtr<T>&
IntrusivePtr<T>::operator=(const IntrusivePtr<OT>& rhs)
{
	IntrusivePtr(rhs).swap(*this);
	return *this;
}

template<typename T>
void
IntrusivePtr<T>::reset(T* obj)
{
	IntrusivePtr(obj).swap(*this);
}

template<typename T>
void
IntrusivePtr<T>::swap(IntrusivePtr& other)
{
	T* obj = other.m_pObj;
	other.m_pObj = m_pObj;
	m_pObj = obj;
}

#endif

