/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "ipcportal.h"
#include "stringutils.h"
#include <sstream>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <syscall.h>

using namespace std;

IPCPortal::IPCPortal()
:	m_rfd(-1), m_wfd(-1), m_lastMessageId(0),
	m_pResponse(0), m_receivingHeader(true)
{
}

IPCPortal::IPCPortal(int rfd, int wfd)
:	m_rfd(rfd), m_wfd(wfd), m_lastMessageId(0),
	m_pResponse(0), m_receivingHeader(true)
{
}

const IPCMessage *
IPCPortal::firstPendingMessage() const
{
	if (m_receivedMessages.empty())
		return 0;
	return &m_receivedMessages.front().message();
}

const IPCMessage *
IPCPortal::firstPendingRequest() const
{
	if (m_receivedRequests.empty())
		return 0;
	return &m_receivedRequests.front().message();
}

void
IPCPortal::processFirstPendingRequest(const IPCMessage& response)
{
	if (m_receivedRequests.empty())
		return;
	Message m(Message::T_RESPONSE, m_receivedRequests.front().getId(), response);
	string serialized = m.serialize();
	m_dataToSend += StringUtils::fromNumber(serialized.length());
	m_dataToSend += ":";
	m_dataToSend += serialized;
	m_receivedRequests.pop_front();
}

void
IPCPortal::sendMessage(const IPCMessage& msg)
{
	Message m(Message::T_ASYNC, ++m_lastMessageId, msg);
	string serialized = m.serialize();
	m_dataToSend += StringUtils::fromNumber(serialized.length());
	m_dataToSend += ":";
	m_dataToSend += serialized;
}

bool
IPCPortal::sendRequest(const IPCMessage& msg, IPCMessage& response)
{
	Message m(Message::T_REQUEST, ++m_lastMessageId, msg);
	string serialized = m.serialize();
	m_dataToSend += StringUtils::fromNumber(serialized.length());
	m_dataToSend += ":";
	m_dataToSend += serialized;
	if (!flushPendingData())
		return false;
	while (true) {
		if (!readChunk())
			return false;
		if (m_pResponse.get()) {
			if (m_pResponse->getId() == m.getId()) {
				break;
			} else {
				m_pResponse.reset(0);
			}
		}
	}
	response = m_pResponse->message();
	return true;
}

bool
IPCPortal::flushPendingData()
{
	while (hasDataToSend()) {
		if (!writeChunk())
			return false;
	}
	return true;
}

void
IPCPortal::readChunk(const string& data)
{
	string::size_type pos = 0;
	string::size_type len = data.length();
	while (pos < len) {
		if (m_receivingHeader) {
			string::size_type fpos = data.find(':', pos);
			if (fpos == string::npos) {
				m_bufferedData.append(data, pos, len-pos);;
				return;
			} else {
				m_bufferedData.append(data, pos, fpos-pos);
				istringstream strm(m_bufferedData);
				m_bodySize = 0;
				strm >> m_bodySize;
				m_receivingHeader = false;
				m_bufferedData.erase();
				pos = fpos+1;
				continue;
			}
		} else {
			string::size_type rest = m_bodySize-m_bufferedData.length();
			if (len-pos < rest) {
				m_bufferedData.append(data, pos, len-pos);
				return;
			} else {
				m_bufferedData.append(data, pos, rest);
				Message msg;
				msg.unserialize(m_bufferedData);
				if (msg.getType() == Message::T_RESPONSE) {
					m_pResponse.reset(new Message(msg));
				} else if (msg.getType() == Message::T_REQUEST) {
					m_receivedRequests.push_back(msg);
				} else {
					m_receivedMessages.push_back(msg);
				}
				m_receivingHeader = true;
				m_bufferedData.erase();
				pos += rest;
				continue;
			}
		}
	}
}

bool
IPCPortal::readChunk()
{
	if (m_rfd == -1 || m_wfd == -1) {
		return false;
	}
	char buf[8192];
	int res = 0;
	SYSCALL(res = read(m_rfd, buf, sizeof(buf)));
	if (res <= 0)
		return false;
	readChunk(string(buf, buf+res));
	return true;
}

bool
IPCPortal::writeChunk()
{
	if (m_rfd == -1 || m_wfd == -1) {
		m_dataToSend.erase();
		return false;
	}
	if (!m_dataToSend.empty()) {
		int res = 0;
		SYSCALL(res = write(m_wfd, m_dataToSend.data(), m_dataToSend.length()));
		if (res < 0)
			return false;
		m_dataToSend.erase(0, res);
	}
	return true;
}

bool
IPCPortal::readWontBlock()
{
	if (m_rfd == -1) {
		return true;
	}
	fd_set rfds;
	FD_ZERO(&rfds);
	FD_SET(m_rfd, &rfds);
	struct timeval tv;
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	int res = 0;
	SYSCALL(res = select(m_rfd+1, &rfds, 0, 0, &tv));
	return (res > 0);
}

bool
IPCPortal::writeWontBlock()
{
	if (m_wfd == -1) {
		return true;
	}
	fd_set wfds;
	FD_ZERO(&wfds);
	FD_SET(m_wfd, &wfds);
	struct timeval tv;
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	int res = 0;
	SYSCALL(res = select(m_wfd+1, 0, &wfds, 0, &tv));
	return (res > 0);
}

string
IPCPortal::Message::serialize() const
{
	string serialized = message().serialize();
	ostringstream strm;
	strm << getType() << ':' << getId() << ':' << serialized.length() << ':' << serialized;
	return strm.str();
}

bool
IPCPortal::Message::unserialize(const string& data)
{
	istringstream strm(data);
	int type = Message::T_ASYNC;
	strm >> type;
	if (type < 0 || type >= Message::T_END)
		return false;
	if (strm.get() != ':')
		return false;
	unsigned int id = 0;
	strm >> id;
	if (strm.get() != ':')
		return false;
	string::size_type len = 0;
	strm >> len;
	if (strm.get() != ':')
		return false;
	string msg_str;
	msg_str.resize(len);
	strm.read(const_cast<char *>(msg_str.data()), len);
	msg_str.resize(strm.gcount());
	m_type = static_cast<Type>(type);
	m_id = id;
	return m_msg.unserialize(msg_str);
}

