/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "jpegimageinspector.h"
#include <cstring>

const uint8_t JPEGImageInspector::SIGNATURE[3] = { 0xff, 0xd8, 0xff };

JPEGImageInspector::JPEGImageInspector()
:	m_state(ST_SIGNATURE), m_isInProgress(true),
	m_isFailed(false), m_width(0), m_height(0),
	m_curPos(0)
{
}

void
JPEGImageInspector::consumeDataChunk(const uint8_t* begin, const uint8_t* end)
{
	std::size_t size = m_data.size();
	m_data.resize(end-begin+size);
	std::memcpy(&m_data[size], begin, end-begin);
	size = m_data.size();
	if (!m_isInProgress) {
		return;
	}
	switch (m_state) {
		case ST_SIGNATURE: {
			if (size < sizeof(SIGNATURE)) {
				return;
			}
			if (std::memcmp(&m_data[0], SIGNATURE, sizeof(SIGNATURE))) {
				m_isInProgress = false;
				m_isFailed = true;
				return;
			}
			m_curPos = sizeof(SIGNATURE);
			m_state = ST_MARKER_OR_PADDING;
		}
		case ST_MARKER_OR_PADDING:
		case_ST_MARKER_OR_PADDING: {
			uint8_t byte;
			do {
				if (size <= m_curPos) {
					return;
				}
				byte = m_data[m_curPos++];
				if (byte != 0xff) {
					break;
				}
			} while (true);
			if (byte >= 0xc0 && byte <= 0xcf && byte != 0xc4 && byte != 0xc8 && byte != 0xcc) {
				m_state = ST_SOF;
				goto case_ST_SOF;
			} else if (byte == 0xd9 || byte == 0xda) {
				m_isInProgress = false;
				m_isFailed = true;
				return;
			}
			m_state = ST_SKIP_VAR;
		}
		case ST_SKIP_VAR: {
			if (size < m_curPos+2) {
				return;
			}
			m_curBlockSize = (uint32_t(m_data[m_curPos])<<8)+uint32_t(m_data[m_curPos+1]);
			if (m_curBlockSize < 2) {
				m_isInProgress = false;
				m_isFailed = true;
				return;
			}
			m_state = ST_SKIP_VAR_DATA;
		}
		case ST_SKIP_VAR_DATA: {
			if (size < m_curPos+m_curBlockSize) {
				return;
			}
			m_curPos += m_curBlockSize;
			m_state = ST_MARKER_OR_PADDING;
			goto case_ST_MARKER_OR_PADDING;
		}
		case ST_SOF:
		case_ST_SOF: {
			if (size < m_curPos+2) {
				return;
			}
			m_curBlockSize = (uint32_t(m_data[m_curPos])<<8)+uint32_t(m_data[m_curPos+1]);
			if (m_curBlockSize < 8) {
				m_isInProgress = false;
				m_isFailed = true;
				return;
			}
			m_state = ST_SOF_DATA;
		}
		case ST_SOF_DATA: {
			if (size < m_curPos+m_curBlockSize) {
				return;
			}
			const uint8_t *bytes = &m_data[m_curPos];
			m_height = ((unsigned)bytes[3]<<8)+((unsigned)bytes[4]);
			m_width = ((unsigned)bytes[5]<<8)+((unsigned)bytes[6]);
			m_isInProgress = false;
			return;
		}
	}
}

void
JPEGImageInspector::consumeEOF()
{
	if (m_isInProgress) {
		m_isInProgress = false;
		m_isFailed = true;
	}
}

void
JPEGImageInspector::reset()
{
	m_state = ST_SIGNATURE;
	std::vector<uint8_t>().swap(m_data);
	m_isInProgress = true;
	m_isFailed = false;
	m_width = 0;
	m_height = 0;
	m_curPos = 0;
}

std::pair<const uint8_t*, const uint8_t*>
JPEGImageInspector::getAccumulatedData() const
{
	return std::pair<const uint8_t*, const uint8_t*>(&m_data[0], &m_data[m_data.size()]);
}

