/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "jsenvironment.h"
#include "debug.h"

#include <string>
#include <cstring>
#include <cctype>
#include <sstream>
#include <list>
#include <map>
#include <algorithm>

JSClass JSEnvironment::m_generic_class = {
	"generic", 0,
	JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub,
	JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSEnvironment::JSEnvironment()
:	m_global_class(m_generic_class), m_document_class(m_generic_class),
	m_navigator_class(m_generic_class), m_location_class(m_generic_class),
	m_screen_class(m_generic_class), m_image_class(m_generic_class),
	m_form_class(m_generic_class), m_mimetype_class(m_generic_class),
	m_plugin_class(m_generic_class), m_body_class(m_generic_class),
	m_branch_cnt(0), m_document_cookie("somecookie")
{
	m_global_class.name = "Window";
	m_document_class.name = "HTMLDocument";
	m_navigator_class.name = "Navigator";
	m_location_class.name = "Location";
	m_screen_class.name = "Screen";
	m_image_class.name = "HTMLImageElement";
	m_form_class.name = "HTMLFormElement";
	m_mimetype_class.name = "MimeType";
	m_plugin_class.name = "Plugin";
	m_body_class.name = "HTMLBodyElement";
	
	m_runtime = JS_Init(40000);
	m_ctx = JS_NewContext(m_runtime, 0x1000);
	JS_SetErrorReporter(m_ctx, &errorReporter);
	JS_SetBranchCallback(m_ctx, &branchCallback);
	
	m_global_obj = JS_NewObject(m_ctx, &m_global_class, 0, 0);
	JS_InitStandardClasses(m_ctx, m_global_obj);
	JS_SetPrivate(m_ctx, m_global_obj, this);
	
	JS_DefineProperty(m_ctx, m_global_obj, "self", OBJECT_TO_JSVAL(m_global_obj), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, m_global_obj, "window", OBJECT_TO_JSVAL(m_global_obj), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, m_global_obj, "top", OBJECT_TO_JSVAL(m_global_obj), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, m_global_obj, "parent", OBJECT_TO_JSVAL(m_global_obj), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, m_global_obj, "onload", JSVAL_NULL, 0, &onloadSetter, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, m_global_obj, "onunload", JSVAL_NULL, 0, &onunloadSetter, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JSObject *history_array = JS_NewArrayObject(m_ctx, 0, 0);
	JS_DefineProperty(m_ctx, m_global_obj, "history", OBJECT_TO_JSVAL(history_array), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_global_obj, "open", &windowOpen, 1, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_global_obj, "setTimeout", &setTimeout, 2, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_global_obj, "setInterval", &setTimeout, 2, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_global_obj, "clearTimeout", &doNothing, 1, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_global_obj, "clearInterval", &doNothing, 1, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_global_obj, "focus", &doNothing, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_global_obj, "blur", &doNothing, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_global_obj, "resizeTo", &doNothing, 2, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_global_obj, "moveTo", &doNothing, 2, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_global_obj, "Image", &imageConstructor, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
#ifdef DEBUG
	JS_DefineFunction(m_ctx, m_global_obj, "bfilterDebugLog", &bfilterDebugLog, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
#endif
	
	JSObject *location_obj = JS_DefineObject(m_ctx, m_global_obj, "location", &m_location_class, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, location_obj, "hostname", JS_GetEmptyStringValue(m_ctx), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, location_obj, "protocol", STRING_TO_JSVAL(JS_NewStringCopyZ(m_ctx, "http:")), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, location_obj, "host", JS_GetEmptyStringValue(m_ctx), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, location_obj, "port", INT_TO_JSVAL(80), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, location_obj, "pathname", STRING_TO_JSVAL(JS_NewStringCopyZ(m_ctx, "/")), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, location_obj, "hash", JS_GetEmptyStringValue(m_ctx), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, location_obj, "href", JS_GetEmptyStringValue(m_ctx), 0, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, location_obj, "search", JS_GetEmptyStringValue(m_ctx), 0, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	
	JSObject *nav_obj = JS_DefineObject(m_ctx, m_global_obj, "navigator", &m_navigator_class, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, nav_obj, "javaEnabled", &javaEnabled, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	std::map<std::string, std::string> nav_props;
	nav_props["appCodeName"] = "Mozilla";
	nav_props["appName"] = "Netscape";
	nav_props["appVersion"] = "5.0 (Windows; en-US)";
	nav_props["userAgent"] = "Mozilla/5.0 (Windows; U; Win98; en-US) Netscape6/6.0";
	std::map<std::string, std::string>::const_iterator iter = nav_props.begin();
	for (; iter != nav_props.end(); ++iter) {
		JS_DefineProperty(m_ctx, nav_obj, iter->first.c_str(),
			STRING_TO_JSVAL(JS_NewStringCopyN(m_ctx, iter->second.c_str(), iter->second.length())),
			0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	}
	
	JSObject *mimetypes_array = JS_NewArrayObject(m_ctx, 0, 0);
	JS_DefineProperty(m_ctx, nav_obj, "mimeTypes", OBJECT_TO_JSVAL(mimetypes_array), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JSObject *plugins_array = JS_NewArrayObject(m_ctx, 0, 0);
	JS_DefineProperty(m_ctx, nav_obj, "plugins", OBJECT_TO_JSVAL(plugins_array), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JSObject *flash_mimetype_obj = JS_DefineObject(m_ctx, mimetypes_array, "application/x-shockwave-flash", &m_mimetype_class, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	jsval flash_mimetype_jsval = OBJECT_TO_JSVAL(flash_mimetype_obj);
	JS_SetElement(m_ctx, mimetypes_array, 0, &flash_mimetype_jsval);
	JS_DefineProperty(m_ctx, flash_mimetype_obj, "description", STRING_TO_JSVAL(JS_NewStringCopyZ(m_ctx, "Shockwave Flash 6.0")), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, flash_mimetype_obj, "type", STRING_TO_JSVAL(JS_NewStringCopyZ(m_ctx, "application/x-shockwave-flash")), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, flash_mimetype_obj, "suffixes", STRING_TO_JSVAL(JS_NewStringCopyZ(m_ctx, "swf")), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JSObject *flash_plugin_obj = JS_DefineObject(m_ctx, flash_mimetype_obj, "enabledPlugin", &m_plugin_class, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	jsval flash_plugin_jsval = OBJECT_TO_JSVAL(flash_plugin_obj);
	JS_SetElement(m_ctx, plugins_array, 0, &flash_plugin_jsval);
	JS_DefineProperty(m_ctx, plugins_array, "Shockwave Flash", flash_plugin_jsval, 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, flash_plugin_obj, "name", STRING_TO_JSVAL(JS_NewStringCopyZ(m_ctx, "Shockwave Flash")), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, flash_plugin_obj, "description", STRING_TO_JSVAL(JS_NewStringCopyZ(m_ctx, "Shockwave Flash 5.0 r50")), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, flash_plugin_obj, "length", INT_TO_JSVAL(1), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, flash_plugin_obj, "refresh", &doNothing, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	
	JSObject *screen_obj = JS_DefineObject(m_ctx, m_global_obj, "screen", &m_screen_class, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, screen_obj, "width", INT_TO_JSVAL(1024), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, screen_obj, "height", INT_TO_JSVAL(768), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, screen_obj, "availWidth", INT_TO_JSVAL(1024), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, screen_obj, "availHeight", INT_TO_JSVAL(738), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, screen_obj, "colorDepth", INT_TO_JSVAL(16), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, screen_obj, "pixelDepth", INT_TO_JSVAL(16), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	
	JSObject *frames_obj = JS_NewArrayObject(m_ctx, 0, 0);
	JS_DefineProperty(m_ctx, m_global_obj, "frames", OBJECT_TO_JSVAL(frames_obj), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	
	m_doc_obj = JS_DefineObject(m_ctx, m_global_obj, "document", &m_document_class, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, m_doc_obj, "cookie", JS_GetEmptyStringValue(m_ctx), &cookieGetter, &cookieSetter, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, m_doc_obj, "location", OBJECT_TO_JSVAL(location_obj), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, m_doc_obj, "domain", JS_GetEmptyStringValue(m_ctx), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, m_doc_obj, "referrer", STRING_TO_JSVAL(JS_NewStringCopyZ(m_ctx, "http://whatever")), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JSObject *images_array = JS_NewArrayObject(m_ctx, 0, 0);
	JS_DefineProperty(m_ctx, m_doc_obj, "images", OBJECT_TO_JSVAL(images_array), 0, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT|JSPROP_READONLY);
	JS_DefineProperty(m_ctx, m_doc_obj, "URL", JS_GetEmptyStringValue(m_ctx), 0, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_doc_obj, "write", &documentWrite, 1, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_doc_obj, "writeln", &documentWriteln, 1, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_doc_obj, "open", &doNothing, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, m_doc_obj, "close", &doNothing, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	
	m_form_array = JS_NewArrayObject(m_ctx, 0, 0);
	JS_DefineProperty(m_ctx, m_doc_obj, "forms", OBJECT_TO_JSVAL(m_form_array), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);

	JSObject* body_obj = JS_DefineObject(m_ctx, m_doc_obj, "body", &m_body_class, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, body_obj, "clientWidth", INT_TO_JSVAL(1024), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, body_obj, "clientHeight", INT_TO_JSVAL(602), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
}

JSEnvironment::~JSEnvironment()
{	
	JS_DestroyContext(m_ctx);
	JS_DestroyRuntime(m_runtime);
}

bool
JSEnvironment::executeScript(const std::string& script, const std::string& version)
{
	setJSVersion(version);
	jsval rval;
	JSBool res = JS_EvaluateScript(m_ctx, m_global_obj, script.c_str(), script.length(), "[unknown]", 1, &rval);
	executeScheduledActions();
	return (res == JS_TRUE);
}

bool
JSEnvironment::executeScriptAsFunction(const std::string& script, const std::string& version)
{
	setJSVersion(version);
	JSFunction *func = JS_CompileFunction(m_ctx, m_global_obj, 0, 0, 0, script.c_str(), script.length(), "[unknown]", 1);
	if (!func) {
		return false;
	}
	jsval rval;
	JSBool res = JS_CallFunction(m_ctx, m_global_obj, func, 0, 0, &rval);
	return (res == JS_TRUE);
}

void
JSEnvironment::addForm(const std::string& name, const std::string& action, const std::string& target)
{
	JSObject *form_obj = JS_NewObject(m_ctx, &m_form_class, 0, 0);
	if (!form_obj) {
		return;
	}
	JS_DefineProperty(m_ctx, form_obj, "name", STRING_TO_JSVAL(JS_NewStringCopyN(m_ctx, name.c_str(), name.length())), 0, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, form_obj, "action", STRING_TO_JSVAL(JS_NewStringCopyN(m_ctx, action.c_str(), action.length())), 0, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineProperty(m_ctx, form_obj, "target", STRING_TO_JSVAL(JS_NewStringCopyN(m_ctx, target.c_str(), target.length())), 0, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, form_obj, "submit", &formSubmit, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	JS_DefineFunction(m_ctx, form_obj, "reset", &doNothing, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT);
	jsval form_jsval = OBJECT_TO_JSVAL(form_obj);
	jsuint nforms = 0;
	JS_GetArrayLength(m_ctx, m_form_array, &nforms);
	JS_SetElement(m_ctx, m_form_array, nforms, &form_jsval);
	if (!name.empty()) {
		JS_DefineProperty(m_ctx, m_form_array, name.c_str(), OBJECT_TO_JSVAL(form_obj), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
		JS_DefineProperty(m_ctx, m_doc_obj, name.c_str(), OBJECT_TO_JSVAL(form_obj), 0, 0, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT);
	}
}

void
JSEnvironment::setJSVersion(const std::string& version)
{
	float vers = 0.0;
	std::istringstream strm(version);
	strm >> vers;
	JSVersion jsv;
	if (vers == 1.0) {
		jsv = JSVERSION_1_0;
	} else if (vers == 1.1) {
		jsv = JSVERSION_1_1;
	} else if (vers == 1.2) {
		jsv = JSVERSION_1_2;
	} else if (vers == 1.3) {
		jsv = JSVERSION_1_3;
	} else if (vers == 1.4) {
		jsv = JSVERSION_1_4;
	} else if (vers == 1.5) {
		jsv = JSVERSION_1_5;
	} else {
		jsv = JSVERSION_DEFAULT;
	}
	JS_SetVersion(m_ctx, jsv);
}

void
JSEnvironment::executeScheduledActions()
{
	if (m_scheduled_actions.empty())
		return;
	std::multimap<int, std::string> sheduled_actions = m_scheduled_actions;
	// a copy is needed in case more actions are shedulled by these actions
	std::multimap<int, std::string>::const_iterator iter = sheduled_actions.begin();
	jsval rval;
	for (; iter != sheduled_actions.end(); ++iter) {
		JS_EvaluateScript(m_ctx, m_global_obj, iter->second.c_str(), iter->second.length(), "[unknown]", 1, &rval);
	}
	m_scheduled_actions.clear();
}

JSEnvironment *
JSEnvironment::getEnvironment(JSContext *cx)
{
	return reinterpret_cast<JSEnvironment *>(JS_GetPrivate(cx, JS_GetGlobalObject(cx)));
}

void
JSEnvironment::errorReporter(JSContext *cx, const char *msg, JSErrorReport *report)
{
	DEBUGLOG("JS_ERR: " << msg);
	DEBUGLOG("        " << report->linebuf);
	std::string err_marker;
	int skip_chars = report->tokenptr-report->linebuf;
	for (int i = 0; i < skip_chars; i++) {
		char ch = report->linebuf[i];
		err_marker += isspace(ch) ? ch : ' '; // this will take care of tabs and such
	}
	DEBUGLOG("        " << err_marker << '^');
}

JSBool
JSEnvironment::branchCallback(JSContext *cx, JSScript *script)
{
	JSEnvironment *env = getEnvironment(cx);
	if (++env->m_branch_cnt < BRANCH_LIMIT) {
		return JS_TRUE;
	} else {
		// infinite loop?
		env->m_branch_cnt = 0;
		return JS_FALSE; // terminate the script
	}
}

JSBool
JSEnvironment::cookieGetter(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
	JSEnvironment* env = getEnvironment(cx);
	JSString *cookie = JS_NewStringCopyN(cx, env->m_document_cookie.c_str(), env->m_document_cookie.length());
	if (!cookie) {
		*vp = JSVAL_NULL;
		return JS_FALSE;
	}
	*vp = STRING_TO_JSVAL(cookie);
	return JS_TRUE;
}

JSBool
JSEnvironment::cookieSetter(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
	JSEnvironment* env = getEnvironment(cx);
	std::string cookie(JS_GetStringBytes(JS_ValueToString(cx, *vp)));
	if (env->m_document_cookie.empty()) {
		env->m_document_cookie = cookie;
	} else {
		// ok, I know this is a wrong behavior, but it's enough to convince
		// some scripts that getting/setting a cookie works
		env->m_document_cookie += "; "+cookie;
	}
	return JS_TRUE;
}

JSBool
JSEnvironment::onloadSetter(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
	JSEnvironment* env = getEnvironment(cx);
	JSFunction *func = JS_ValueToFunction(cx, *vp);
	if (!func)
		return JS_FALSE;
	env->m_scheduled_actions.insert(
		std::multimap<int, std::string>::value_type(
		2000, std::string(JS_GetFunctionName(func))+"()")
	);
	return JS_TRUE;
}

JSBool
JSEnvironment::onunloadSetter(JSContext *cx, JSObject *obj, jsval id, jsval *vp)
{
	JSEnvironment* env = getEnvironment(cx);
	JSFunction *func = JS_ValueToFunction(cx, *vp);
	if (!func)
		return JS_FALSE;
	env->m_scheduled_actions.insert(
		std::multimap<int, std::string>::value_type(10000,
		std::string(JS_GetFunctionName(func))+"()")
	);
	return JS_TRUE;
}

JSBool
JSEnvironment::javaEnabled(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	*rval = JSVAL_FALSE;
	return JS_TRUE;
}

JSBool
JSEnvironment::windowOpen(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc < 2) {
		getEnvironment(cx)->onPopup().emit();
	} else {
		std::string target(JS_GetStringBytes(JS_ValueToString(cx, argv[1])));
		if (target.length() == 0 || (target[0] != '_' || target == "_blank" || target == "_new")) {
			getEnvironment(cx)->onPopup().emit();
		}
	}
	*rval = JSVAL_NULL;
	return JS_TRUE;
}

JSBool
JSEnvironment::setTimeout(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	static int idc = 1;
	*rval = INT_TO_JSVAL(idc++);
	if (argc < 2)
		return JS_TRUE;
	JSEnvironment* env = getEnvironment(cx);
	std::string code(JS_GetStringBytes(JS_ValueToString(cx, argv[0])));
	int delay = JSVAL_TO_INT(argv[1]);
	env->m_scheduled_actions.insert(std::multimap<int, std::string>::value_type(delay, code));
	return JS_TRUE;
}

JSBool
JSEnvironment::documentWrite(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	*rval = JSVAL_VOID;
	if (argc < 1)
		return JS_TRUE;
	const uint8_t* data = reinterpret_cast<const uint8_t*>(JS_GetStringBytes(JS_ValueToString(cx, argv[0])));
	getEnvironment(cx)->onDataOutput().emit(data, data + strlen(reinterpret_cast<const char*>(data)));
	return JS_TRUE;
}

JSBool
JSEnvironment::documentWriteln(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	*rval = JSVAL_VOID;
	if (argc < 1)
		return JS_TRUE;
	const uint8_t* data = reinterpret_cast<const uint8_t*>(JS_GetStringBytes(JS_ValueToString(cx, argv[0])));
	DataOutputSignal& signal = getEnvironment(cx)->onDataOutput();
	signal.emit(data, data + strlen(reinterpret_cast<const char*>(data)));
	static const uint8_t crlf[] = { '\r', '\n' };
	signal.emit(crlf, crlf + 2);
	return JS_TRUE;
}

JSBool
JSEnvironment::imageConstructor(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	JSObject *image_obj = JS_NewObject(cx, &getEnvironment(cx)->m_image_class, 0, 0);
	if (!image_obj)
		return JS_FALSE;
	JS_DefineProperty(cx, image_obj, "complete", JSVAL_TRUE, 0, 0, JSPROP_ENUMERATE|JSPROP_PERMANENT|JSPROP_READONLY);
	*rval = OBJECT_TO_JSVAL(image_obj);
	return JS_TRUE;
}

JSBool
JSEnvironment::formSubmit(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	jsval target_val;
	JS_GetProperty(cx, obj, "target", &target_val);
	if (JSVAL_IS_STRING(target_val)) {
		std::string target(JS_GetStringBytes(JS_ValueToString(cx, target_val)));
		if (target.length() > 0 && (target[0] != '_' || target == "_blank" || target == "_new")) {
			getEnvironment(cx)->onPopup().emit();
		}
	}
	*rval = JSVAL_VOID;
	return JS_TRUE;
}

JSBool
JSEnvironment::doNothing(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	*rval = JSVAL_VOID;
	return JS_TRUE;
}

JSBool
JSEnvironment::bfilterDebugLog(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	*rval = JSVAL_VOID;
	if (argc < 1)
		return JS_TRUE;
	DEBUGLOG("JS: bfilterDebugLog(" << JS_GetStringBytes(JS_ValueToString(cx, argv[0])) << ")");
	return JS_TRUE;
}

