#include <signal.h>
#include <conf.h>
#include <rules.h>
#include <ipcportal.h>
#include "bfilter.h"
#include "state.h"
#include "network.h"
#include "debug.h"

class BackendState : public State
{
public:
	BackendState(IPCPortal* portal);
	virtual const Config& getConfig() { processMasterMessages(); return m_config; }
	virtual const Rules& getRules() { processMasterMessages(); return m_rules; }
	virtual bool isFilteringEnabled() { processMasterMessages(); return m_filtering; }
	virtual bool isScriptFilteringEnabled() { processMasterMessages(); return m_scriptFiltering; }
	virtual bool isMasterDead() { processMasterMessages(); return m_sMasterDead || m_commError; }
	bool isCommError() const { return m_commError; }
	void processMasterMessages();
	static void signalMasterDeadHandler(int signum);
private:
	IPCPortal *m_pPortal;
	bool m_commError;
	bool m_filtering;
	bool m_scriptFiltering;
	static volatile bool m_sMasterDead;
	Config m_config;
	Rules m_rules;
};

volatile bool BackendState::m_sMasterDead = false;

BackendState::BackendState(IPCPortal* portal)
:	m_pPortal(portal), m_commError(false),
	m_filtering(true), m_scriptFiltering(true)
{
	IPCMessage response;
	if (!m_pPortal->sendRequest(IPCMessage("getState"), response)) {
		m_commError = true;
		return;
	}
	if (response.getMethod() != "returnState") {
		m_commError = true;
		return;
	}
	m_config.unserialize(response.getParameter(0));
	m_rules.unserialize(response.getParameter(1));
	m_filtering = (response.getParameter(2) == "1");
	m_scriptFiltering = (response.getParameter(3) == "1");
}

void
BackendState::processMasterMessages()
{
	if (m_commError) {
		return;
	}
	bool read = false;
	while (m_pPortal->readWontBlock()) {
		if (!m_pPortal->readChunk()) {
			m_commError = true;
			break;
		}
		read = true;
	}
	if (!read) {
		return;
	}
	for (const IPCMessage* msg = 0; (msg = m_pPortal->firstPendingMessage()); m_pPortal->dropFirstPendingMessage()) {
		if (msg->getMethod() == "updateConfig") {
			m_config.unserialize(msg->getParameter(0));
		} else if (msg->getMethod() == "updateRules") {
			m_rules.unserialize(msg->getParameter(0));
		} else if (msg->getMethod() == "updateFilteringState") {
			m_filtering = (msg->getParameter(0) == "1");
			m_scriptFiltering = (msg->getParameter(1) == "1");
		} else {
			DEBUGLOG("Unhandled message from master: "+msg->getMethod());
		}
	}
}

void
BackendState::signalMasterDeadHandler(int signum)
{
	m_sMasterDead = true;
}

static void setup_child_signals()
{
	struct sigaction sa;
	
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	
	sa.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &sa, NULL);
	
	sa.sa_handler = &BackendState::signalMasterDeadHandler;
	sigaction(SIGUSR1, &sa, NULL);
}

int main_unix_backend(Network::Socket csock, IPCPortal* portal)
{
	setup_child_signals();
	
	BackendState state(portal);
	if (state.isCommError()) {
		return 1;
	}
	
#ifdef DEBUG
	Debug::getInstance().setPortal(portal);
#endif
	
	Network::sockSetNodelay(csock, true);
	BFilter filter(csock, &state);
	filter.run();
	Network::disconnectAndCloseSocket(csock, 10);
	
#ifdef DEBUG
	Debug::getInstance().signalSessionFinished();
#endif
	
	return 0;
}

