#include "bfilter.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <list>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#ifdef HAVE_SYS_FCNTL_H
#include <sys/fcntl.h>
#endif
#include <conf.h>
#include <configio.h>
#include <rules.h>
#include <rulesreader.h>
#include "network.h"
#include "httperrormessages.h"
#include "state.h"
#include <ipcportal.h>
#include "syscall.h"

class StandaloneState : public State
{
public:
	StandaloneState(const Config& config, const Rules& rules)
	: m_config(config), m_rules(rules), m_filtering(true), m_scriptFiltering(true) {}
	virtual const Config& getConfig() { return m_config; }
	virtual const Rules& getRules() { return m_rules; }
	virtual bool isFilteringEnabled() { return m_filtering; }
	virtual bool isScriptFilteringEnabled() { return m_scriptFiltering; }
	virtual bool isMasterDead() { return m_sMasterDead; }
	static void sigMasterDeadHandler(int signum);
private:
	Config m_config;
	Rules m_rules;
	bool m_filtering;
	bool m_scriptFiltering;
	static volatile bool m_sMasterDead;
};

volatile bool StandaloneState::m_sMasterDead = false;

void
StandaloneState::sigMasterDeadHandler(int signum)
{
	m_sMasterDead = true;
}

static void sig_child(int signum)
{
	if (signum != SIGCHLD) {
		return;
	}
	int saved_errno = errno;
	while (waitpid(-1, NULL, WNOHANG) > 0);
	errno = saved_errno;
}

static void sig_term(int signum)
{
	kill(0, SIGUSR1);
	exit(0);
}

static void setup_parent_signals()
{
	struct sigaction sa;
	
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	
	sa.sa_handler = &sig_term;
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGINT, &sa, NULL);
	
	sa.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &sa, NULL);
	sigaction(SIGUSR1, &sa, NULL);
	
	sa.sa_handler = &sig_child;
	sa.sa_flags = SA_NOCLDSTOP;
	sigaction(SIGCHLD, &sa, NULL);
}

static void setup_child_signals()
{
	struct sigaction sa;
	
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	
	sa.sa_handler = SIG_DFL;
	sigaction(SIGCHLD, &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGINT, &sa, NULL);
	
	sa.sa_handler = &StandaloneState::sigMasterDeadHandler;
	sigaction(SIGUSR1, &sa, NULL);
}

#ifndef HAVE_DAEMON
static int daemon(int nocd, int noclose)
{
	struct rlimit rlim;
	pid_t pid;
	int i;
	
	umask(0);
	
	if (getrlimit(RLIMIT_NOFILE, &rlim) < 0)
		perror("getrlimit() failed");
	
	pid = fork();
	if (pid < 0)
		return -1;
	if (pid != 0)
		exit(0);
	
	setsid();
	
	signal(SIGHUP, SIG_IGN);
	pid = fork();
	if (pid < 0)
		return -1;
	if (pid != 0)
		exit(0);
	
	if (!nocd)
		chdir("/");
	
	if (!noclose) {
#if 0
		if (rlim.rlim_max == RLIM_INFINITY)
			rlim.rlim_max = 1024;
		for (i = 0; i < rlim.rlim_max; i++) {
			close(i);
		}
#endif
		i = open("/dev/null", O_RDWR);
		if (i < 0)
			return -1;
		dup2(i, 0);
		dup2(i, 1);
		dup2(i, 2);
		if (i > 2)
			close(i);
	}
	return 0;
}
#endif

static std::string format_config_error(const ConfIO::Error& err, const std::string& fname)
{
	std::ostringstream strm;
	strm << '[' << fname;
	if (err.getLine()) {
		strm << ':' << err.getLine();
		if (err.getCol()) {
			strm << ':' << err.getCol();
		}
	}
	strm << "] ";
	if (err.getType() == ConfIO::Error::T_WARNING) {
		strm << "Warning: ";
	}
	strm << err.getMessage();
	return strm.str();
}

int main_unix_standalone(const std::string& confdir, bool nodaemon)
{
	Network::Socket serv_sock = Network::INVALID_SOCK;
	Network::Socket clnt_sock = Network::INVALID_SOCK;
	struct sockaddr_in client;
	
	ConfigIO creader;
	std::ifstream cf_strm((confdir+"/config").c_str());
	if (cf_strm.fail()) {
		std::cerr << "Could not open " << confdir << "/config" << std::endl;
		return 1;
	}
	ConfigIO::Status cf_status = creader.readConfig(cf_strm);
	for (std::list<ConfigIO::Error>::const_iterator iter = cf_status.errors.begin();
	     iter != cf_status.errors.end(); ++iter) {
		std::cerr << format_config_error(*iter,  confdir+"/config") << std::endl;
		if (iter->getType() == ConfigIO::Error::T_ERROR) {
			break;
		}
	}
	cf_strm.close();
	if (cf_status.status == ConfigIO::Status::ERRORS) {
		return 1;
	}
	
	RulesReader rreader;
	std::ifstream gr_strm((confdir+"/rules").c_str());
	if (!gr_strm.good()) {
		std::cerr << "Could not open " << confdir << "/rules" << std::endl;
		return 1;
	}
	std::ifstream lr_strm((confdir+"/rules.local").c_str());
	if (!lr_strm.good()) {
		std::cerr << "Could not open " << confdir <<  "/rules.local" << std::endl;
		return 1;
	}
	RulesReader::Status r_status= rreader.readRules(gr_strm, lr_strm);
	bool error_found = false;
	for (std::list<RulesReader::Error>::const_iterator iter = r_status.errors_global.begin();
	     !error_found && iter != r_status.errors_global.end(); ++iter) {
		std::cerr << format_config_error(*iter, confdir+"/rules") << std::endl;
		if (iter->getType() == RulesReader::Error::T_ERROR) {
			error_found = true;
		}
	}
	for (std::list<RulesReader::Error>::const_iterator iter = r_status.errors_local.begin();
	     !error_found && iter != r_status.errors_local.end(); ++iter) {
		std::cerr << format_config_error(*iter, confdir+"/rules.local") << std::endl;
		if (iter->getType() == RulesReader::Error::T_ERROR) {
			error_found = true;
		}
	}
	gr_strm.close();
	lr_strm.close();
	if (r_status.status == RulesReader::Status::ERRORS) {
		return 1;
	}
	
	StandaloneState state(creader.getConfig(), rreader.getRules());
	
	NetworkError err;
	serv_sock = Network::tcpServer(state.getConfig().getListenAddr().c_str(),
		state.getConfig().getListenPort(), &err);
	if (!err.isOK()) {
		std::cerr << err.getErrorMessage() << std::endl;
		return 1;
	}
	
	if (nodaemon) {
		setsid();
	} else {
		daemon(1, 0);
	}
	setup_parent_signals();
	
	while (true) {
		clnt_sock = Network::tcpServerWaitConn(serv_sock, &client);
		if (!Network::isValidSocket(clnt_sock))
			continue;
		pid_t pid = fork();
		if (pid == -1) {
			return 1;
		} else if (pid == 0) { // child
			setup_child_signals();
			Network::closeSocket(serv_sock);
			Network::sockSetNodelay(clnt_sock, true);
			BFilter filter(clnt_sock, &state);
			filter.run();
			Network::disconnectAndCloseSocket(clnt_sock, 10);
			return 0;
		} else { // parent
			Network::closeSocket(clnt_sock);
		}
	}
	return 0;
}

