/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <winsock2.h>
#include <cstddef>
#include <string>
#include <sstream>
#include "bfilter.h"
#include "network.h"
#include "state.h"
#include "types.h"

class BackendState : public State
{
private:
	struct SharedMemory
	{
		int m_configMod;
		int m_rulesMod;
		HANDLE m_hConfigShm;
		HANDLE m_hRulesShm;
		bool m_filteringEnabled;
		bool m_scriptFilteringEnabled;
		bool m_masterDead;
	};
public:
	BackendState(HANDLE hParentProcess, HANDLE hShm, HANDLE hShmMutex);
	~BackendState();
	bool configNeedsUpdating() const;
	bool rulesNeedUpdating() const;
	void updateConfig();
	void updateRules();
	virtual const Config& getConfig();
	virtual const Rules& getRules();
	virtual bool isFilteringEnabled();
	virtual bool isScriptFilteringEnabled();
	virtual bool isMasterDead();
private:
	HANDLE m_hParentProcess;
	HANDLE m_hShm;
	HANDLE m_hShmMutex;
	const SharedMemory *m_pShm;
	Config m_config;
	Rules m_rules;
	int m_configMod;
	int m_rulesMod;
};

BackendState::BackendState(HANDLE hParentProcess, HANDLE hShm, HANDLE hShmMutex)
:	m_hParentProcess(hParentProcess), m_hShm(hShm), m_hShmMutex(hShmMutex)
{
	m_pShm = (const SharedMemory *)MapViewOfFile(hShm, FILE_MAP_READ, 0, 0, 0);
	updateConfig();
	updateRules();
}

BackendState::~BackendState()
{
	UnmapViewOfFile((void *)m_pShm);
}

bool
BackendState::configNeedsUpdating() const
{
	return (!m_pShm->m_masterDead && m_configMod != m_pShm->m_configMod);
}

bool
BackendState::rulesNeedUpdating() const
{
	return (!m_pShm->m_masterDead && m_rulesMod != m_pShm->m_rulesMod);
}

void
BackendState::updateConfig()
{
	WaitForSingleObject(m_hShmMutex, INFINITE);
	m_configMod = m_pShm->m_configMod;
	HANDLE hConfig = NULL;
	DuplicateHandle(m_hParentProcess, m_pShm->m_hConfigShm, GetCurrentProcess(), &hConfig, 0, FALSE, DUPLICATE_SAME_ACCESS);
	void *config = MapViewOfFile(hConfig, FILE_MAP_READ, 0, 0, 0);
	if (!config) {
		// master is dead?
		return;
	}
	const std::size_t *size = (std::size_t *)config;
	const char *data = (const char *)config+sizeof(std::size_t);
	m_config.unserialize(std::string(data, data+*size));
	UnmapViewOfFile(config);
	CloseHandle(hConfig);
	ReleaseMutex(m_hShmMutex);
}

void
BackendState::updateRules()
{
	WaitForSingleObject(m_hShmMutex, INFINITE);
	m_rulesMod = m_pShm->m_rulesMod;
	HANDLE hRules = NULL;
	DuplicateHandle(m_hParentProcess, m_pShm->m_hRulesShm, GetCurrentProcess(), &hRules, 0, FALSE, DUPLICATE_SAME_ACCESS);
	void *rules = MapViewOfFile(hRules, FILE_MAP_READ, 0, 0, 0);
	if (!rules) {
		// master is dead?
		return;
	}
	const std::size_t *size = (std::size_t *)rules;
	const char *data = (const char *)rules+sizeof(std::size_t);
	m_rules.unserialize(std::string(data, data+*size));
	UnmapViewOfFile(rules);
	CloseHandle(hRules);
	ReleaseMutex(m_hShmMutex);
}

const Config&
BackendState::getConfig()
{
	if (configNeedsUpdating())
		updateConfig();
	return m_config;
}

const Rules&
BackendState::getRules()
{
	if (rulesNeedUpdating())
		updateRules();
	return m_rules;
}

bool
BackendState::isFilteringEnabled()
{
	return m_pShm->m_filteringEnabled;
}

bool
BackendState::isScriptFilteringEnabled()
{
	return m_pShm->m_scriptFilteringEnabled;
}

bool
BackendState::isMasterDead()
{
	return m_pShm->m_masterDead;
}

int WINAPI
WinMain(HINSTANCE hThisInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	std::istringstream strm(lpCmdLine);
	
	uintptr_t iParentProcess = 0;
	uintptr_t iShm = 0;
	uintptr_t iShmMutex = 0;
	Network::Socket csock = Network::INVALID_SOCK;
	strm >> iParentProcess >> iShm >> iShmMutex >> csock;
	HANDLE hParentProcess = reinterpret_cast<HANDLE>(iParentProcess);
	HANDLE hShm = reinterpret_cast<HANDLE>(iShm);
	HANDLE hShmMutex = reinterpret_cast<HANDLE>(iShmMutex);
	if (hParentProcess == NULL || hShm == NULL || hShmMutex == NULL || csock == Network::INVALID_SOCK) {
		return 1;
	}
	WSAData wsadata;
	if (WSAStartup(MAKEWORD(2, 0), &wsadata)) {
		CloseHandle(hParentProcess);
		CloseHandle(hShm);
		CloseHandle(hShmMutex);
		return 1;
	}
	{ // this block is needed to make the BackendState destructor run before we close the handles
		BackendState state(hParentProcess, hShm, hShmMutex);
		BFilter filter(csock, &state);
		filter.run();
	}
	Network::disconnectAndCloseSocket(csock, 10);
	CloseHandle(hParentProcess);
	CloseHandle(hShm);
	CloseHandle(hShmMutex);
	return 0;
}

