/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "network.h"
#include "syscall.h"
#include "tselect.h"
#include "stringutils.h"
#include <string>
#include <cstring>

#ifdef WIN32

#include <windows.h>
#include <winsock2.h>

#else

#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <cstring>
#include <signal.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#endif

#define NLISTEN 5

#ifndef WIN32
std::map<std::string, Network::addr_list_type> Network::m_resolver_cache;
#endif

#ifdef WIN32
static int inet_aton(const char *cp, struct in_addr *pin)
{
	int rc = inet_addr(cp);
	if (rc == -1 && strcmp(cp, "255.255.255.255"))
		return 0;
	pin->s_addr = rc;
	return 1;
}
#endif

Network::addr_list_type
Network::resolve(const char *host, NetworkError *err)
{
#ifndef WIN32
	std::string lchost = StringUtils::toLowerStr(host);
	std::map<std::string, addr_list_type>::const_iterator fit = m_resolver_cache.find(lchost);
	if (fit != m_resolver_cache.end()) {
		return fit->second;
	}
#endif
	in_addr ia;
	if (inet_aton(host, &ia)) {
		return addr_list_type(1, ia);
	}
	struct hostent *hp = gethostbyname(host);
	if (!hp) {
		if (err) {
			err->setState(NetworkError::ERR_UNKNOWN_HOST, std::string("Unknown host: ")+host);
		}
		return addr_list_type();
	}
	in_addr *iap;
	addr_list_type res;
	for (int i = 0; (iap = (struct in_addr *)hp->h_addr_list[i]); i++) {
		res.push_back(*iap);
	}
#ifndef WIN32
	m_resolver_cache[lchost] = res;
#endif
	return res;
}

Network::Socket
Network::tcpServer(const char *host, unsigned port, NetworkError *err)
{
	struct sockaddr_in local;
	Socket sock;
	const int on = 1;
	
	memset(&local, 0, sizeof(local));
	if (!host || host[0] == '\0') {
		local.sin_addr.s_addr = htonl(INADDR_ANY);
	} else {
		addr_list_type addr_list = resolve(host, err);
		if (addr_list.empty())
			return INVALID_SOCK;
		local.sin_addr = *addr_list.begin();
	}
	local.sin_port = htons(port);
	local.sin_family = AF_INET;
	
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (!isValidSocket(sock)) {
		if (err) {
			std::string errmsg = "socket() failed: "+getLastError();
			err->setState(NetworkError::ERR_OTHER, errmsg);
		}
		return INVALID_SOCK;
	}
	
	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (const char *)&on, sizeof(on))) {
		if (err) {
			std::string errmsg = "setsockopt() failed: "+getLastError();
			err->setState(NetworkError::ERR_OTHER, errmsg);
		}
		return INVALID_SOCK;
	}
	
	if (bind(sock, (struct sockaddr *)&local, sizeof(local))) {
		if (err) {
			std::string errmsg = "bind() failed: "+getLastError();
			err->setState(NetworkError::ERR_OTHER, errmsg);
		}
		return INVALID_SOCK;
	}
	
	if (listen(sock, NLISTEN)) {
		if (err) {
			std::string errmsg = "listen() failed: "+getLastError();
			err->setState(NetworkError::ERR_OTHER, errmsg);
		}
		return INVALID_SOCK;
	}
	
	return sock;
}

Network::Socket
Network::tcpServerWaitConn(Socket sock, struct sockaddr_in *client, NetworkError *err)
{
	Socket newsock = INVALID_SOCK;
#ifdef WIN32
	int addrlen = sizeof(struct sockaddr_in);
#else
	socklen_t addrlen = sizeof(struct sockaddr_in);
#endif
	
	SYSCALL(newsock = accept(sock, (struct sockaddr *)client, &addrlen));
	if (!isValidSocket(newsock)) {
		if (err) {
			std::string errmsg = "accept() failed: "+getLastError();
			err->setState(NetworkError::ERR_OTHER, errmsg);
		}
		return INVALID_SOCK;
	}
	
	return newsock;
}

Network::Socket
Network::tcpConnect(const char *host, unsigned port, NetworkError *err)
{
	addr_list_type addr_list = resolve(host, err);
	if (addr_list.empty())
		return INVALID_SOCK;
	
	Socket sock = socket(AF_INET, SOCK_STREAM, 0);
	if (!isValidSocket(sock)) {
		if (err) {
			std::string errmsg = "socket() failed: "+getLastError();
			err->setState(NetworkError::ERR_OTHER, errmsg);
		}
		return INVALID_SOCK;
	}
	
	struct sockaddr_in peer;
	memset(&peer, 0, sizeof(peer));
	peer.sin_family = AF_INET;
	peer.sin_port = htons(port);
	int res = -1;
	addr_list_type::const_iterator iter = addr_list.begin();
	for (; iter != addr_list.end(); ++iter) {
		peer.sin_addr = *iter;
		SYSCALL(res = connect(sock, (struct sockaddr *)&peer, sizeof(peer)));
		if (res) {
			continue;
		} else {
			return sock;
		}
	}
	
	NetworkError::ErrorCode ecode = NetworkError::ERR_OTHER;
#ifdef WIN32
	switch (WSAGetLastError()) {
	case WSAECONNREFUSED:
		ecode = NetworkError::ERR_CONNECTION_REFUSED;
		break;
	case WSAETIMEDOUT:
		ecode = NetworkError::ERR_CONNECTION_TIMEOUT;
		break;
	case WSAENETUNREACH:
		ecode = NetworkError::ERR_NETWORK_UNREACHABLE;
		break;
	case WSAEHOSTUNREACH:
		ecode = NetworkError::ERR_HOST_UNREACHABLE;
		break;
	default: break;
	}
#else
	switch (errno) {
	case ECONNREFUSED:
		ecode = NetworkError::ERR_CONNECTION_REFUSED;
		break;
	case ETIMEDOUT:
		ecode = NetworkError::ERR_CONNECTION_TIMEOUT;
		break;
	case ENETUNREACH:
		ecode = NetworkError::ERR_NETWORK_UNREACHABLE;
		break;
	case EHOSTUNREACH:
		ecode = NetworkError::ERR_HOST_UNREACHABLE;
		break;
	default: break;
	}
#endif
	std::string errmsg = "connect() failed: "+getLastError();
	if (err) {
		err->setState(ecode, errmsg);
	}
	
	return INVALID_SOCK;
}

bool
Network::tcpSendData(Socket sock, const std::string& data, NetworkError *err)
{
	std::string tmpdata = data;
	while (!tmpdata.empty()) {
		int len = -1;
		SYSCALL(len = send(sock, tmpdata.c_str(), tmpdata.length(), 0));
		if (len < 0) {
			if (err) {
				std::string errmsg = "send() failed: "+getLastError();
				err->setState(NetworkError::ERR_OTHER, errmsg);
			}
			return false;
		}
		tmpdata.erase(0, len);
	}
	
	return true;
}

void
Network::sockSetNodelay(Socket sock, bool state)
{
	const int st = state ? 1 : 0;
	setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (const char *)&st, sizeof(st));
}

int
Network::writev(int sock, const IoVec* iov, int cnt)
{
#ifdef WIN32
	DWORD sent;
	if (WSASend(sock, (WSABUF*)iov, cnt, &sent, 0, NULL, NULL)) {
		if (WSAGetLastError() == WSAEWOULDBLOCK) {
			// a bug in Win9x
			return 0;
		}
		return -1;
	}
	return sent;
#else
	int sent;
	SYSCALL(sent = ::writev(sock, iov, cnt));
	return sent;
#endif
}

void
Network::closeSocket(Socket sock)
{
#ifdef WIN32
	closesocket(sock);
#else
	close(sock);
#endif
}

void
Network::disconnectAndCloseSocket(Socket sock, int timeout)
{
#ifdef WIN32
	shutdown(sock, SD_SEND);
#else
	shutdown(sock, SHUT_WR);
#endif
	TSelect tselect;
	fd_set rfds;
	struct timeval tv;
	tv.tv_sec = timeout;
	tv.tv_usec = 0;
	tselect.addTimer(tv);
	char buf[2048];
	while (true) {
		FD_ZERO(&rfds);
		FD_SET(sock, &rfds);
		int res = tselect.select(sock+1, &rfds, NULL, NULL);
		if (res <= 0) {
			// error or timeout
			break;
		}
		int len = -1;
		SYSCALL(len = recv(sock, buf, sizeof(buf), 0));
		if (len <= 0) {
			// error or EOF
			break;
		}
	}
	closeSocket(sock);
}

bool
Network::isValidSocket(Socket sock)
{
#ifdef WIN32
	return (sock != INVALID_SOCKET);
#else
	return (sock >= 0);
#endif
}

#ifdef WIN32
namespace {
	struct WinSockError
	{
		int m_code;
		const char* m_message;
		
		WinSockError(int code, const char* msg)
		: m_code(code), m_message(msg) {}
		
		bool operator<(const WinSockError& rhs) const {
			return m_code < rhs.m_code;
		}
	};
	
	// The entries are sorted by error code, so we can use binary search on them
	const WinSockError wsockErrorList[] = {
		WinSockError(0,                  "No error"),
		WinSockError(WSAEINTR,           "Interrupted system call"),
		WinSockError(WSAEBADF,           "Bad file number"),
		WinSockError(WSAEACCES,          "Permission denied"),
		WinSockError(WSAEFAULT,          "Bad address"),
		WinSockError(WSAEINVAL,          "Invalid argument"),
		WinSockError(WSAEMFILE,          "Too many open sockets"),
		WinSockError(WSAEWOULDBLOCK,     "Operation would block"),
		WinSockError(WSAEINPROGRESS,     "Operation now in progress"),
		WinSockError(WSAEALREADY,        "Operation already in progress"),
		WinSockError(WSAENOTSOCK,        "Socket operation on non-socket"),
		WinSockError(WSAEDESTADDRREQ,    "Destination address required"),
		WinSockError(WSAEMSGSIZE,        "Message too long"),
		WinSockError(WSAEPROTOTYPE,      "Protocol wrong type for socket"),
		WinSockError(WSAENOPROTOOPT,     "Bad protocol option"),
		WinSockError(WSAEPROTONOSUPPORT, "Protocol not supported"),
		WinSockError(WSAESOCKTNOSUPPORT, "Socket type not supported"),
		WinSockError(WSAEOPNOTSUPP,      "Operation not supported on socket"),
		WinSockError(WSAEPFNOSUPPORT,    "Protocol family not supported"),
		WinSockError(WSAEAFNOSUPPORT,    "Address family not supported"),
		WinSockError(WSAEADDRINUSE,      "Address already in use"),
		WinSockError(WSAEADDRNOTAVAIL,   "Can't assign requested address"),
		WinSockError(WSAENETDOWN,        "Network is down"),
		WinSockError(WSAENETUNREACH,     "Network is unreachable"),
		WinSockError(WSAENETRESET,       "Net connection reset"),
		WinSockError(WSAECONNABORTED,    "Software caused connection abort"),
		WinSockError(WSAECONNRESET,      "Connection reset by peer"),
		WinSockError(WSAENOBUFS,         "No buffer space available"),
		WinSockError(WSAEISCONN,         "Socket is already connected"),
		WinSockError(WSAENOTCONN,        "Socket is not connected"),
		WinSockError(WSAESHUTDOWN,       "Can't send after socket shutdown"),
		WinSockError(WSAETOOMANYREFS,    "Too many references, can't splice"),
		WinSockError(WSAETIMEDOUT,       "Connection timed out"),
		WinSockError(WSAECONNREFUSED,    "Connection refused"),
		WinSockError(WSAELOOP,           "Too many levels of symbolic links"),
		WinSockError(WSAENAMETOOLONG,    "File name too long"),
		WinSockError(WSAEHOSTDOWN,       "Host is down"),
		WinSockError(WSAEHOSTUNREACH,    "No route to host"),
		WinSockError(WSAENOTEMPTY,       "Directory not empty"),
		WinSockError(WSAEPROCLIM,        "Too many processes"),
		WinSockError(WSAEUSERS,          "Too many users"),
		WinSockError(WSAEDQUOT,          "Disc quota exceeded"),
		WinSockError(WSAESTALE,          "Stale NFS file handle"),
		WinSockError(WSAEREMOTE,         "Too many levels of remote in path"),
		WinSockError(WSASYSNOTREADY,     "Network system is unavailable"),
		WinSockError(WSAVERNOTSUPPORTED, "Winsock version out of range"),
		WinSockError(WSANOTINITIALISED,  "WSAStartup not yet called"),
		WinSockError(WSAEDISCON,         "Graceful shutdown in progress"),
		WinSockError(WSAHOST_NOT_FOUND,  "Host not found"),
		WinSockError(WSANO_DATA,         "No host data of that type was found")
	};
}
#endif

std::string
Network::getLastError()
{
#ifndef WIN32
	return strerror(errno);
#else
	WinSockError target(WSAGetLastError(), 0);
	const WinSockError* end = wsockErrorList+sizeof(wsockErrorList)/sizeof(*wsockErrorList);
	const WinSockError* it = std::lower_bound(wsockErrorList, end, target);
	if (it != end && it->m_code == target.m_code) {
		return it->m_message;
	} else {
		std::ostringstream strm;
		strm << "Unknown socket error (" << target.m_code << ")";
		return strm.str();
	}
#endif
}

