/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "replacementimage.h"

#include <algorithm>

using namespace std;

const std::string ReplacementImage::CONTENT_TYPE = "image/gif";
const uint8_t ReplacementImage::HEADER[6] = { 'G', 'I', 'F', '8', '9', 'a' };
const uint8_t ReplacementImage::GRAPHIC_CONTROL_EXTENSION[8] = {
	0x21, // extension introducer
	0xf9, // graphic control label
	0x04, // block size
	0x05, // flags (transparent + disposal_method==1)
	0x00, 0x00, // delay time
	0x00, // transparent color index
	0x00  // block terminator
};
const ReplacementImage::LZWCodeStream::size_type ReplacementImage::TABLE_SIZE = 4096;
const ReplacementImage::LZWCodeStream::size_type ReplacementImage::FIRST_AVAILABLE_CODE = 6;
const ReplacementImage::LZWCode ReplacementImage::CLEAR_CODE = 4;
const ReplacementImage::LZWCode ReplacementImage::EOI_CODE = 5;

static const uint8_t transgif_1x1[] = {
	0x47, 0x49, 0x46, 0x38, 0x39, 0x61, 0x01, 0x00,
	0x01, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00,
	0xff, 0xff, 0xff, 0x21, 0xf9, 0x04, 0x01, 0x00,
	0x00, 0x00, 0x00, 0x2c, 0x00, 0x00, 0x00, 0x00,
	0x01, 0x00, 0x01, 0x00, 0x40, 0x02, 0x01, 0x44,
	0x00, 0x3b
};

ReplacementImage::ReplacementImage(int width, int height, bool draw_border, const unsigned char *border_color)
{
	if (width < 1)
		width = 1;
	if (height < 1)
		height = 1;
	
	// just in case
	if (width > 2000)
		width = 20;
	if (height > 2000)
		height = 20;
	
	if (width == 1 && height == 1) {
		m_data.assign(transgif_1x1, transgif_1x1+sizeof(transgif_1x1));
		return;
	}
	
	m_data.append(HEADER, HEADER+sizeof(HEADER));
	m_data.append(makeScreenDescriptor(width, height));
	m_data.append(GRAPHIC_CONTROL_EXTENSION, GRAPHIC_CONTROL_EXTENSION+sizeof(GRAPHIC_CONTROL_EXTENSION));
	if (!draw_border || width < 3 || height < 3) {
		m_data.append(makeImageDescriptor(0, 0, width, height));
		m_data.append(makeLocalColorTable(border_color));
		m_data.append(formatLZWByteStream(packLZWCodeStream(makeLZWCodeStream(width*height, 0x00))));
	} else {
		/*
		m_data.append(makeImageDescriptor(1, 1, width-2, height-2));
		m_data.append(makeLocalColorTable(border_color));
		m_data.append(formatLZWByteStream(packLZWCodeStream(makeLZWCodeStream((width-2)*(height-2), 0x00))));
		*/
		m_data.append(makeImageDescriptor(0, 0, width, 1));
		m_data.append(makeLocalColorTable(border_color));
		string horiz_line = formatLZWByteStream(packLZWCodeStream(makeLZWCodeStream(width, 0x01)));
		m_data.append(horiz_line);
		m_data.append(makeImageDescriptor(0, height-1, width, 1));
		m_data.append(makeLocalColorTable(border_color));
		m_data.append(horiz_line);
		m_data.append(makeImageDescriptor(0, 1, 1, height-2));
		m_data.append(makeLocalColorTable(border_color));
		string vert_line = formatLZWByteStream(packLZWCodeStream(makeLZWCodeStream(height-2, 0x01)));
		m_data.append(vert_line);
		m_data.append(makeImageDescriptor(width-1, 1, 1, height-2));
		m_data.append(makeLocalColorTable(border_color));
		m_data.append(vert_line);
	}
	m_data.append(1, (unsigned char)0x3b);
}

std::pair<const uint8_t*, const uint8_t*>
ReplacementImage::getBinaryData() const
{
	return std::pair<const uint8_t*, const uint8_t*>(
		reinterpret_cast<const uint8_t*>(m_data.data()),
		reinterpret_cast<const uint8_t*>(m_data.data()+m_data.length())
	);
}

std::string
ReplacementImage::makeScreenDescriptor(int width, int height)
{
	unsigned char sd[] = {
		(unsigned)width, (unsigned)width >> 8,
		(unsigned)height, (unsigned)height >> 8,
		0x00, // flags
		0x00, // background color index
		0x00  // pixel aspect ratio
	};
	return string(sd, sd+sizeof(sd));
}

string
ReplacementImage::makeImageDescriptor(int xpos, int ypos, int width, int height)
{
	unsigned char id[] = {
		0x2c, // image separator
		(unsigned)xpos, ((unsigned)xpos>>8),
		(unsigned)ypos, ((unsigned)ypos>>8),
		(unsigned)width, ((unsigned)width >> 8),
		(unsigned)height, ((unsigned)height >> 8),
		0x80 // flags
	};
	return string(id, id+sizeof(id));
}

string
ReplacementImage::makeLocalColorTable(const unsigned char *border_color)
{
	unsigned char bytes[] = {
		0x00, 0x00, 0x00, // this one will be transparent
		border_color[0], border_color[1], border_color[2]
	};
	return string(bytes, bytes+sizeof(bytes));
}

/*
We only need to encode a sequence of data_size color_idx bytes.
The LZW code size is 2 (minimal possible value).
The imaginary code table would look like this:
0: 0
1: 1
10: 2 // not used
11: 3 // not used
100: clear code
101: end of information
110: col col
111: col col col
1000: col col col col
and so on
*/
ReplacementImage::LZWCodeStream
ReplacementImage::makeLZWCodeStream(LZWCodeStream::size_type data_size, unsigned char color_idx)
{
	LZWCodeStream res;
	LZWCodeStream::size_type tablepos = 0;
	LZWCodeStream::size_type buflen = 1;
	--data_size;
	res.push_back(CLEAR_CODE);
	while (data_size-- > 0) {
		if (isCodeInTable(buflen+1, tablepos)) {
			++buflen;
		} else {
			res.push_back(getCodeFor(buflen, color_idx));
			buflen = 1;
			if (++tablepos + FIRST_AVAILABLE_CODE >= TABLE_SIZE && data_size > 0) {
				res.push_back(getCodeFor(buflen, color_idx));
				res.push_back(CLEAR_CODE);
				tablepos = 0;
			}
		}
	}
	res.push_back(getCodeFor(buflen, color_idx));
	res.push_back(EOI_CODE);
	return res;
}

string
ReplacementImage::packLZWCodeStream(const LZWCodeStream& stream)
{
	string res;
	int code_bits = 3;
	unsigned int max_code = (1<<3)-1;
	unsigned char cur_byte = 0;
	int cur_byte_bits_occupied = 0;
	LZWCodeStream::const_iterator iter = stream.begin();
	for (; iter != stream.end(); ++iter) {
		LZWCode code = *iter;
		int bits_injected = 0;
		while (bits_injected < code_bits) {
			cur_byte |= ((code >> bits_injected) << cur_byte_bits_occupied);
			int injection = std::min(code_bits-bits_injected, 8-cur_byte_bits_occupied);
			bits_injected += injection;
			cur_byte_bits_occupied += injection;
			if (cur_byte_bits_occupied == 8) {
				res.append(1, cur_byte);
				cur_byte = 0;
				cur_byte_bits_occupied = 0;
			}
		}
		if (code == CLEAR_CODE) {
			code_bits = 3;
			max_code = (1<<3)-1;
		} else if (code >= max_code) {
			max_code = (1 << ++code_bits)-1;
		}
	}
	if (cur_byte_bits_occupied > 0) {
		res.append(1, cur_byte);
	}
	return res;
}

string
ReplacementImage::formatLZWByteStream(const string& stream)
{
	string res;
	string::size_type data_left = stream.length();
	string::size_type data_done = 0;
	res.append(1, (unsigned char)0x02); // LZW code size
	while (data_left > 0) {
		string::size_type block_size = std::min<string::size_type>(255, data_left);
		res.append(1, (unsigned char)block_size);
		res.append(stream, data_done, block_size);
		data_left -= block_size;
		data_done += block_size;
	}
	res.append(1, (unsigned char)0x00);
	return res;
}

bool
ReplacementImage::isCodeInTable(LZWCodeStream::size_type seqlen, LZWCodeStream::size_type tablepos)
{
	if (seqlen == 1) {
		return true;
	} else {
		return (tablepos > seqlen-2);
	}
}

ReplacementImage::LZWCode
ReplacementImage::getCodeFor(LZWCodeStream::size_type seqlen, unsigned char color_idx)
{
	if (seqlen == 1) {
		return color_idx;
	} else {
		return FIRST_AVAILABLE_CODE+seqlen-2;
	}
}

