/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "serverresponsestream.h"
#include "bfilter.h"
#include "httperrormessages.h"
#include "state.h"
#include "conf.h"
#include "rules.h"
#include "network.h"
#include "acceptencoding.h"
#include "htmlresponsefilter.h"
#include "imageresponsefilter.h"
#include "flashresponsefilter.h"
#include "compressorresponsefilter.h"
#include "decompressorresponsefilter.h"
#include "blackholeresponsefilter.h"
#include "httpresponsemetadata.h"
#include "httpstatusline.h"
#include "httpversion.h"
#include "replacementimage.h"
#include "replacementflash.h"
#include "sessionfinishexception.h"
#include "stringutils.h"
#include "debug.h"
#include <algorithm>
#include <sstream>

ServerResponseStream::ServerResponseStream(BFilter* owner)
:	m_pOwner(owner),
	m_nextFilterProvider(SigC::slot(*this, &ServerResponseStream::provideNextFilter)),
	m_is1xxResponse(false)
{
}

bool
ServerResponseStream::isResponseOnHeadRequest() const
{
	if (m_pOwner->getNumPendingRequests() == 0) {
		DEBUGLOG2("response to a non existent request");
		return false;
	}
	return (m_pOwner->requestsToBeServed().front()->metadata().requestLine().getMethod() == "HEAD");
}

void
ServerResponseStream::eventError(int ecode, const char *info)
{
	if (!info) {
		info = "";
	}
	DEBUGLOG("ServerResponseStream::eventError(ecode = " << ecode << ")");
	
	if (!m_ptrActiveRequest.get()) {
		throw SessionFinishException(false, "Unexpected data from server");
	}
	
	std::string response;
	std::string fail_reason;
	switch (ecode) {
	case ERR_PARSING_HEADER:
		response = HttpErrorMessages::errMalformedResponseHeader(
			m_ptrActiveRequest->getOrigMetadata().requestLine());
		fail_reason = "Malformed header in server response";
		break;
	case ERR_PARSING_CHUNK_HEADER:
		response = HttpErrorMessages::errMalformedChunkInResponse(
			m_ptrActiveRequest->getOrigMetadata().requestLine());
		fail_reason = "Malformed chunk header in server response";
		break;
	case ERR_UNEXPECTED_EOF:
		response = HttpErrorMessages::errUnexpectedServerEOF(
			m_ptrActiveRequest->getOrigMetadata().requestLine());
		fail_reason = "Unexpected server EOF";
		break;
	default:
		response = HttpErrorMessages::errInternal(
			m_ptrActiveRequest->getOrigMetadata().requestLine(),
			"Unhandled error in ServerResponseStream: "+StringUtils::fromNumber(ecode));
		fail_reason = "Unhandled error in ServerResponseStream::eventError";
	}
	if (!isForwardingInProgress()) {
		m_pOwner->moreDataForClient(response);
		LOG_ERROR_OUTGOING_RESPONSE(response);
	}
	throw SessionFinishException(false, fail_reason);
}

void
ServerResponseStream::eventGotMetadata(HttpResponseMetadata& metadata)
{
	LOG_INCOMING_RESPONSE_STARTS(metadata.statusLine().toString(), metadata.headers().toString());
	
	if (m_pOwner->getNumPendingRequests() == 0) {
		DEBUGLOG2("PANIC: response to a non existent request");
		throw SessionFinishException(false, "Unexpected message from server");
	}
	
	m_ptrActiveRequest = m_pOwner->requestsToBeServed().front();
	
	DEBUGLOG("Response begins (active request: " << m_ptrActiveRequest->metadata().requestLine().getURI().toString() << ")");
	
	if (metadata.statusLine().getHttpVersion() > m_pOwner->serverInfo().getProtocolCompability()) {
		m_pOwner->serverInfo().setProtocolCompability(metadata.statusLine().getHttpVersion());
	}
	
	prepareResponseForForwarding(metadata);
	
	// Some servers set Content-Encoding: gzip for .tar.gz files.
	// This is very wrong: it indicates to the user agent that it has to decompress the content.
	// As a result, you might get a tar file with the .tar.gz extension.
	if (metadata.headers().getHeader("Content-Encoding").getValue() == "gzip") {
		if (metadata.headers().getHeader("Content-Type").getValue() == "application/x-tar") {
			metadata.headers().removeHeader("Content-Encoding");
			metadata.headers().setHeader(HttpHeader("Content-Type", "application/x-tgz"));
		} else if (metadata.headers().getHeader("Content-Type").getValue() == "application/x-gzip") {
			metadata.headers().removeHeader("Content-Encoding");
			metadata.headers().setHeader(HttpHeader("Content-Type", "application/octet-stream"));
		} else {
			static const char gz[] = { '.','g','z' };
			static const char tgz[] = { '.','t','g','z' };
			static const char tar_gz[] = { '.','t','a','r','.','g','z' };
			std::string req_path = m_ptrActiveRequest->metadata().requestLine().getURI().getDecodedPath();
			if (StringUtils::ciEndsWith(req_path, tgz, tgz+sizeof(tgz)) ||
			    StringUtils::ciEndsWith(req_path, tar_gz, tar_gz+sizeof(tar_gz))) {
				metadata.headers().removeHeader("Content-Encoding");
				metadata.headers().setHeader(HttpHeader("Content-Type", "application/x-tgz"));
			} else {
				if (StringUtils::ciEndsWith(req_path, gz, gz+sizeof(gz))) {
					metadata.headers().removeHeader("Content-Encoding");
					metadata.headers().setHeader(HttpHeader("Content-Type", "application/octet-stream"));
				}
			}
		}
	}
	
	m_filterTryList.clear();
	m_ptrFilterChain.reset(new ResponseFilterChain(m_nextFilterProvider));
	m_ptrFilterChain->onMetadataOutput().connect(
		SigC::slot(*this, &ServerResponseStream::processFilteredMetadata)
	);
	m_ptrFilterChain->onDataOutput().connect(SigC::slot(*this, &ServerResponseStream::processFilteredData));
	m_ptrFilterChain->onEOF().connect(SigC::slot(*this, &ServerResponseStream::processFilterEOF));
	
	int status_code = metadata.statusLine().getCode();
	m_pOwner->serverInfo().registerResponse(status_code);
	
	if ((m_is1xxResponse = (status_code >= 100 && status_code < 200))) { // asynchronous message
		if (m_ptrActiveRequest->getOrigMetadata().requestLine().getHttpVersion() < HttpVersion::HTTP_1_1) {
			m_filterTryList.push_back(SigC::slot(*this, &ServerResponseStream::tryBlackHoleFilter));
		}
		m_ptrFilterChain->consumeMetadata(&metadata);
		return;
	}
	
	HttpClientRequest::InterceptHandler intercept_handler = m_ptrActiveRequest->getInterceptHandler();
	
	if ((status_code >= 301 && status_code <= 303) || status_code == 307) { // redirect
		if (intercept_handler == HttpClientRequest::IH_IMAGE_ANALYZE) {
			handleRedirectWhileAnalyzingImage(metadata);
		} else if (intercept_handler == HttpClientRequest::IH_FLASH_ANALYZE) {
			handleRedirectWhileAnalyzingFlash(metadata);
		}
		m_ptrFilterChain->consumeMetadata(&metadata);
		return;
	}
	
	if (metadata.getBodyStatus() == HttpResponseMetadata::NO_BODY_BUT_PRESERVE_HEADERS) {
		m_ptrFilterChain->consumeMetadata(&metadata);
		return;
	}
	
	if (m_pOwner->getState().isFilteringEnabled() &&
	    m_pOwner->getState().getRules().getFilterParam(
	      m_ptrActiveRequest->getOrigMetadata().requestLine().getURI().toString()) != 0) {
		if (intercept_handler == HttpClientRequest::IH_IMAGE_ANALYZE) {
			m_filterTryList.push_back(SigC::slot(*this, &ServerResponseStream::tryImageFilter));
		} else if (intercept_handler == HttpClientRequest::IH_FLASH_ANALYZE) {
			m_filterTryList.push_back(SigC::slot(*this, &ServerResponseStream::tryFlashFilter));
		} else {
			m_filterTryList.push_back(SigC::slot(*this, &ServerResponseStream::tryHtmlFilter));
		}
	}
	
	// the response may have been encoded in some way that the client can't handle but we can
	m_filterTryList.push_back(SigC::slot(*this, &ServerResponseStream::tryDecompressorFilter));
	
	if (m_pOwner->getState().getConfig().getClientCompression()) {
		m_filterTryList.push_back(SigC::slot(*this, &ServerResponseStream::tryCompressorFilter));
	}
	
	m_ptrFilterChain->consumeMetadata(&metadata);
}

void
ServerResponseStream::eventGotContentPart(const uint8_t* begin, const uint8_t* end)
{
	m_ptrFilterChain->consumeDataChunk(begin, end);
}

void
ServerResponseStream::eventMessageEnds(bool connection_needs_closing)
{
	LOG_INCOMING_RESPONSE_FINISHES();
	DEBUGLOG("ServerResponseStream::eventMessageEnds(conn_close = " << connection_needs_closing << ")");
	if (connection_needs_closing) {
		m_pOwner->serverInfo().setPipeliningCapability(ServerInfo::P_NOT_OK);
	}
	
	bool response_came_through = m_ptrFilterChain->isSealed();
	m_ptrFilterChain->consumeEOF();
	m_ptrFilterChain.reset(0);
	
	if (!m_is1xxResponse) {
		m_ptrActiveRequest->setAlreadyServed();
		DEBUGLOG("The request have been served");
		m_pOwner->shiftRequests();
		if (m_pOwner->getNumPendingRequests() == 0) {
			if (m_pOwner->serverInfo().getPipeliningCapability() == ServerInfo::P_TESTING) {
				m_pOwner->serverInfo().setPipeliningCapability(ServerInfo::P_OK);
				DEBUGLOG("The server has responded to a pipelined request - marking as P_OK");
			}
		}
	}
	
	m_ptrActiveRequest.reset(0);
	if (connection_needs_closing) {
		m_pOwner->closeServerConnection();
	}
	if (response_came_through) {
		m_pOwner->requestStream().dispatchMoreRequests();
	}
}

void
ServerResponseStream::moreData(const uint8_t* begin, const uint8_t* end)
{
	m_pOwner->moreDataForClient(begin, end);
}

void
ServerResponseStream::done(bool conn_close)
{
	DEBUGLOG("ServerResponseStream::done(conn_close = " << conn_close << ")");
	m_ptrStreamWriter.reset(0);
	if (conn_close) {
		throw SessionFinishException(true);
	}
}

AbstractResponseFilter*
ServerResponseStream::provideNextFilter(HttpResponseMetadata* metadata)
{
	AbstractResponseFilter* filter = 0;
	while (!filter && !m_filterTryList.empty()) {
		AbstractResponseFilter::NextFilterProvider provider(m_filterTryList.front());
		m_filterTryList.pop_front();
		// we must remove the provider from the list before calling it, because
		// it may recursively call this method again, and we'd get infinite recursion
		filter = provider(metadata);
	}
	return filter;
}

AbstractResponseFilter*
ServerResponseStream::tryBlackHoleFilter(HttpResponseMetadata* metadata)
{
	return BlackHoleResponseFilter::create(m_ptrFilterChain->getNextFilterProvider(), metadata);
}

AbstractResponseFilter*
ServerResponseStream::tryHtmlFilter(HttpResponseMetadata* metadata)
{
	m_pOwner->htmlProcessor()->setBaseURL(
		m_ptrActiveRequest->getOrigMetadata().requestLine().getURI()
	);
	m_pOwner->htmlProcessor()->setUserAgent(
		m_ptrActiveRequest->getOrigMetadata().headers().getHeader("User-Agent").getValue()
	);
	return HtmlResponseFilter::create(
		m_ptrFilterChain->getNextFilterProvider(), metadata, m_pOwner->htmlProcessor()
	);
}

AbstractResponseFilter*
ServerResponseStream::tryImageFilter(HttpResponseMetadata* metadata)
{
	ImageResponseFilter::Params params(&m_pOwner->getState().getConfig());
	params.ad_rule = StringUtils::toNumber<int>(m_ptrActiveRequest->getInterceptParam("iar"));
	params.width = StringUtils::toNumber<int>(m_ptrActiveRequest->getInterceptParam("w"));
	params.height = StringUtils::toNumber<int>(m_ptrActiveRequest->getInterceptParam("h"));
	params.target = m_ptrActiveRequest->getInterceptParam("target");
	params.external_link = (m_ptrActiveRequest->getInterceptParam("ea") == "1");
	params.external_img = (m_ptrActiveRequest->getInterceptParam("ei") == "1");
	params.suspicious_link = (m_ptrActiveRequest->getInterceptParam("sa") == "1");
	params.suspicious_img = (m_ptrActiveRequest->getInterceptParam("si") == "1");
	params.redir = (m_ptrActiveRequest->getInterceptParam("redir") == "1");
	params.cookies = (m_ptrActiveRequest->getInterceptParam("cookie") == "1");
	if (!params.cookies && metadata->headers().hasHeader("Set-Cookie")
		|| metadata->headers().hasHeader("Set-Cookie2")) {
		params.cookies = true;
	}
	return ImageResponseFilter::create(
		m_ptrFilterChain->getNextFilterProvider(), metadata, params);
}

AbstractResponseFilter*
ServerResponseStream::tryFlashFilter(HttpResponseMetadata* metadata)
{
	FlashResponseFilter::Params params(&m_pOwner->getState().getConfig());
	params.width = StringUtils::toNumber<int>(m_ptrActiveRequest->getInterceptParam("w"));
	params.height = StringUtils::toNumber<int>(m_ptrActiveRequest->getInterceptParam("h"));
	params.url = m_ptrActiveRequest->getInterceptParam("url");
	return FlashResponseFilter::create(
		m_ptrFilterChain->getNextFilterProvider(), metadata, params);
}

AbstractResponseFilter*
ServerResponseStream::tryDecompressorFilter(HttpResponseMetadata* metadata)
{
	if (!ResponseFilterChain::isBodyEncoded(*metadata)) {
		return 0;
	}
	AcceptEncoding aenc(m_ptrActiveRequest->getOrigMetadata().headers().getHeader("Accept-Encoding"));
	if (ResponseFilterChain::isEncodingAccepted(*metadata, aenc)) {
		return 0;
	}
	return DecompressorResponseFilter::createStack(m_ptrFilterChain->getNextFilterProvider(), metadata, 2);
}

AbstractResponseFilter*
ServerResponseStream::tryCompressorFilter(HttpResponseMetadata* metadata)
{
	if (!StringUtils::startsWith(metadata->headers().getHeader("Content-Type").getValue(), "text/")) {
		return 0;
	}
	if (ResponseFilterChain::isBodyEncoded(*metadata)) {
		return 0;
	}
	AcceptEncoding aenc(m_ptrActiveRequest->metadata().headers().getHeader("Accept-Encoding"));
	return CompressorResponseFilter::create(m_ptrFilterChain->getNextFilterProvider(), metadata, aenc);
}

void
ServerResponseStream::processFilteredMetadata(HttpResponseMetadata* metadata)
{
	bool chunked_ok = (m_ptrActiveRequest->getOrigMetadata().requestLine().getHttpVersion() >= HttpVersion::HTTP_1_1);
	bool force_conn_close = m_ptrActiveRequest->isClosingRequest();
	m_ptrStreamWriter.reset(new HttpStreamWriter(this, *metadata, chunked_ok, force_conn_close));
#ifdef DEBUG
	if (metadata->statusLine().getCode() >= 400) {
		LOG_ERROR_OUTGOING_RESPONSE_STARTS(metadata->statusLine().toString(), metadata->headers().toString());
	} else if (isFilteredResponse(*metadata)) {
		LOG_FILTERED_OUTGOING_RESPONSE_STARTS(metadata->statusLine().toString(), metadata->headers().toString());
	} else {
		LOG_NORMAL_OUTGOING_RESPONSE_STARTS(metadata->statusLine().toString(), metadata->headers().toString());
	}
#endif
}

void
ServerResponseStream::processFilteredData(const uint8_t* begin, const uint8_t* end)
{
	m_ptrStreamWriter->write(begin, end);
}

void
ServerResponseStream::processFilterEOF()
{
	LOG_OUTGOING_RESPONSE_FINISHES();
	m_ptrStreamWriter->end(); /* may throw a SessionFinishException,
	so it must not be above LOG_OUTGOING_RESPONSE_FINISHES() */
}

#ifdef DEBUG
bool
ServerResponseStream::isFilteredResponse(const HttpResponseMetadata& metadata)
{
	const HttpHeader* hp = metadata.headers().getHeaderPtr("X-filtered-by");
	if (!hp) {
		return false;
	}
	typedef std::list<std::string> VList;
	const VList& values = hp->getValues();
	for (VList::const_iterator iter = values.begin(); iter != values.end(); ++iter) {
		if (*iter == "HtmlResponseFilter" ||
		    *iter == "ImageResponseFilter" ||
		    *iter == "FlashResponseFilter") {
			return true;
		}
	}
	return false;
}
#endif

void
ServerResponseStream::handleRedirectWhileAnalyzingImage(HttpResponseMetadata& metadata)
{
	URI redirect(m_ptrActiveRequest->metadata().requestLine().getURI(),
		URI(metadata.headers().getHeader("Location").getValue()));
	std::string orig_redirect = redirect.toString();
	redirect.removeFragment();
	redirect.setAbsoluteRawPath("bf.intercept-request/image_analyze");
	std::ostringstream strm;
	strm << "url=" << URI::escape(orig_redirect)
		<< "&w=" << m_ptrActiveRequest->getInterceptParam("w")
		<< "&h=" << m_ptrActiveRequest->getInterceptParam("h")
		<< "&iar=" << m_ptrActiveRequest->getInterceptParam("iar")
		<< "&ea=" << m_ptrActiveRequest->getInterceptParam("ea")
		<< "&ei=" << m_ptrActiveRequest->getInterceptParam("ei")
		<< "&sa=" << m_ptrActiveRequest->getInterceptParam("sa")
		<< "&si=" << m_ptrActiveRequest->getInterceptParam("si")
		<< "&target=" << URI::escape(m_ptrActiveRequest->getInterceptParam("target"))
		<< "&redir=1&cookie="
		<< ((metadata.headers().hasHeader("Set-Cookie") ||
		     metadata.headers().hasHeader("Set-Cookie2")) ? "1" : "0");
	redirect.setRawQuery(strm.str());
	metadata.headers().setHeader(HttpHeader("Location", redirect.toString()));
}

void
ServerResponseStream::handleRedirectWhileAnalyzingFlash(HttpResponseMetadata& metadata)
{
	URI redirect(m_ptrActiveRequest->metadata().requestLine().getURI(),
		URI(metadata.headers().getHeader("Location").getValue()));
	std::string orig_redirect = redirect.toString();
	redirect.removeFragment();
	redirect.setAbsoluteRawPath("bf.intercept-request/flash_analyze");
	redirect.setRawQuery(
		"w="+m_ptrActiveRequest->getInterceptParam("w")+
		"&h="+m_ptrActiveRequest->getInterceptParam("h")+
		"&url="+URI::escape(orig_redirect)
	);
	metadata.headers().setHeader(HttpHeader("Location", redirect.toString()));
}


void
ServerResponseStream::prepareResponseForForwarding(HttpResponseMetadata& metadata)
{
	HttpHeader connection_hdr = metadata.headers().getHeader("Connection");
	connection_hdr.elements().appendElements(metadata.headers().getHeader("Proxy-Connection").elements());
	std::list<HttpHeaderElement> hop_by_hop_headers = connection_hdr.elements().getElements();
	std::list<HttpHeaderElement>::iterator iter = hop_by_hop_headers.begin();
	for (; iter != hop_by_hop_headers.end(); ++iter) {
		metadata.headers().removeHeader(iter->getName());
	}
	metadata.headers().removeHeader("Keep-Alive");
	metadata.headers().removeHeader("Connection");
	metadata.headers().removeHeader("Proxy-Connection");
	metadata.statusLine().setHttpVersion(
		std::min(m_ptrActiveRequest->getOrigMetadata().requestLine().getHttpVersion(), HttpVersion::HTTP_1_1)
	);
	if (metadata.statusLine().getHttpVersion() < HttpVersion::HTTP_1_1) {
		if (metadata.statusLine().getCode() == 303 || metadata.statusLine().getCode() == 307) {
			metadata.statusLine().setCodeAndMessage(302);
		}
	}
#ifdef DEBUG
	metadata.headers().setHeader(HttpHeader("X-in-response-to", m_ptrActiveRequest->getOrigMetadata().requestLine().getURI().toString()));
#endif
}

void
ServerResponseStream::processSubstitutionRequest(const HttpClientRequest *request)
{
	DEBUGLOG("processSubstitutionRequest()");
	HttpClientRequest::InterceptHandler ih = request->getInterceptHandler();
	if (ih == HttpClientRequest::IH_IMAGE_SUBSTITUTE) {
		performImageSubstitution(request);
	} else if (ih == HttpClientRequest::IH_FLASH_SUBSTITUTE) {
		performFlashSubstitution(request);
	} else {
		performErrorSubstitution(request);
	}
}

void
ServerResponseStream::performImageSubstitution(const HttpClientRequest *request)
{
	DEBUGLOG("performImageSubstitution()");
	bool closing_request = request->isClosingRequest();
	HttpVersion request_version = request->metadata().requestLine().getHttpVersion();
	HttpStatusLine status_line(std::min(request_version, HttpVersion::HTTP_1_1), 200);
	HttpResponseMetadata metadata(status_line, HttpHeadersCollection());
	if (request->metadata().requestLine().getMethod() == "HEAD") {
		metadata.setBodyStatus(HttpMessageMetadata::NO_BODY_BUT_PRESERVE_HEADERS);
		metadata.headers().setHeader(HttpHeader("Content-Type", "image/gif"));
#ifdef DEBUG
		metadata.headers().setHeader(HttpHeader("X-in-response-to", request->metadata().requestLine().getURI().toString()));
#endif
		LOG_GENERATED_OUTGOING_RESPONSE(metadata.statusLine().toString(), metadata.headers().toString());
		m_ptrStreamWriter.reset(new HttpStreamWriter(this, metadata, request_version>=HttpVersion::HTTP_1_1, closing_request));
		m_ptrStreamWriter->end();
	} else {
		int width = StringUtils::toNumber<int>(request->getInterceptParam("w"));
		int height = StringUtils::toNumber<int>(request->getInterceptParam("h"));
		ReplacementImage image(width, height, m_pOwner->getState().getConfig().getDrawBorders(),
			m_pOwner->getState().getConfig().getBorderColor());
		HttpHeadersCollection headers;
		metadata.setBodyStatus(HttpMessageMetadata::BODY_SIZE_KNOWN);
		metadata.setBodySize(image.getBinaryData().second-image.getBinaryData().first);
		metadata.headers().setHeader(HttpHeader("Content-Type", image.getContentType()));
#ifdef DEBUG
		metadata.headers().setHeader(HttpHeader("X-in-response-to", request->metadata().requestLine().getURI().toString()));
#endif
		LOG_GENERATED_OUTGOING_RESPONSE(metadata.statusLine().toString(), metadata.headers().toString());
		m_ptrStreamWriter.reset(new HttpStreamWriter(this, metadata, request_version>=HttpVersion::HTTP_1_1, closing_request));
		m_ptrStreamWriter->write(image.getBinaryData().first, image.getBinaryData().second);
		m_ptrStreamWriter->end();
	}
}

void
ServerResponseStream::performFlashSubstitution(const HttpClientRequest *request)
{
	DEBUGLOG("performFlashSubstitution()");
	
	bool closing_request = request->isClosingRequest();
	HttpVersion request_version = request->metadata().requestLine().getHttpVersion();
	HttpStatusLine status_line(std::min(request_version, HttpVersion::HTTP_1_1), 200);
	HttpResponseMetadata metadata(status_line, HttpHeadersCollection());
	if (request->metadata().requestLine().getMethod() == "HEAD") {
		metadata.setBodyStatus(HttpMessageMetadata::NO_BODY_BUT_PRESERVE_HEADERS);
		metadata.headers().setHeader(HttpHeader("Content-Type", ReplacementFlash::getContentType()));
#ifdef DEBUG
		metadata.headers().setHeader(HttpHeader("X-in-response-to", request->metadata().requestLine().getURI().toString()));
#endif
		LOG_GENERATED_OUTGOING_RESPONSE(metadata.statusLine().toString(), metadata.headers().toString());
		m_ptrStreamWriter.reset(new HttpStreamWriter(this, metadata, request_version>=HttpVersion::HTTP_1_1, closing_request));
		m_ptrStreamWriter->end();
	} else {
		int width = StringUtils::toNumber<int>(request->getInterceptParam("w"));
		int height = StringUtils::toNumber<int>(request->getInterceptParam("h"));
		std::string orig_url = request->getInterceptParam("url");
		ReplacementFlash flash(width, height, orig_url.c_str(),
			m_pOwner->getState().getConfig().getDrawBorders(),
			m_pOwner->getState().getConfig().getBorderColor()
		);
		HttpHeadersCollection headers;
		metadata.setBodyStatus(HttpMessageMetadata::BODY_SIZE_KNOWN);
		metadata.setBodySize(flash.getBinaryData().second-flash.getBinaryData().first);
		metadata.headers().setHeader(HttpHeader("Content-Type", ReplacementFlash::getContentType()));
#ifdef DEBUG
		metadata.headers().setHeader(HttpHeader("X-in-response-to", request->metadata().requestLine().getURI().toString()));
#endif
		LOG_GENERATED_OUTGOING_RESPONSE(metadata.statusLine().toString(), metadata.headers().toString());
		m_ptrStreamWriter.reset(new HttpStreamWriter(this, metadata, request_version>=HttpVersion::HTTP_1_1, closing_request));
		m_ptrStreamWriter->write(flash.getBinaryData().first, flash.getBinaryData().second);
		m_ptrStreamWriter->end();
	}
}

void
ServerResponseStream::performErrorSubstitution(const HttpClientRequest *request)
{
	LOG_ERROR_OUTGOING_RESPONSE(request->getErrorResponse());
	DEBUGLOG("performErrorSubstitution()");
	m_pOwner->moreDataForClient(request->getErrorResponse());
	throw SessionFinishException(false, "Finishing session after performErrorSubstitution()");
}

