/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef SHAREDPTR_H_
#define SHAREDPTR_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <memory>
#include <cstddef>

class SharedPtrRefCounter
{
public:
	SharedPtrRefCounter() : m_counter(1) {}

	void ref() { ++m_counter; }
	
	bool unref() { return --m_counter == 0; }
	
	static void* operator new(std::size_t size) {
		return allocator().allocate(1);
	}
	
	static void operator delete(void* addr) {
		allocator().deallocate(reinterpret_cast<SharedPtrRefCounter*>(addr), 1);
	}
private:
	static std::allocator<SharedPtrRefCounter>& allocator();
private:
	unsigned m_counter;
};


template<typename T>
class SharedPtr
{
public:
	SharedPtr() : m_pObj(0), m_pRefCounter(0) {}
	
	explicit
	SharedPtr(T* obj);
	
	SharedPtr(const SharedPtr& other);
	
	template<typename OT>
	SharedPtr(const SharedPtr<OT>& other);
	
	~SharedPtr();
public:
	SharedPtr& operator=(const SharedPtr& rhs);
	
	template<typename OT>
	SharedPtr& operator=(const SharedPtr<OT>& rhs);
	
	T& operator*() const { return *m_pObj; }
	
	T* operator->() const { return m_pObj; }
	
	bool operator==(const SharedPtr& rhs) {
		return get() == rhs.get();
	}
	
	bool operator!=(const SharedPtr& rhs) {
		return get() != rhs.get();
	}
private:
	typedef void (SharedPtr::*MemberPtr)(T*);
public:
	operator MemberPtr() const {
		return m_pObj ? &SharedPtr::reset : 0;
	}
	
	T* get() const { return m_pObj; }
	
	void reset(T* obj);
	
	void swap(SharedPtr& other);
private:
	T* m_pObj;
	SharedPtrRefCounter* m_pRefCounter;
};


template<typename T>
SharedPtr<T>::SharedPtr(T* obj)
:	m_pObj(obj), m_pRefCounter(new SharedPtrRefCounter)
{
}

template<typename T>
SharedPtr<T>::SharedPtr(const SharedPtr& other)
:	m_pObj(other.m_pObj), m_pRefCounter(other.m_pRefCounter)
{
	other.m_pRefCounter->ref();
}

template<typename T>
SharedPtr<T>::~SharedPtr()
{
	if (m_pRefCounter) {
		if (m_pRefCounter->unref()) {
			delete m_pObj;
			delete m_pRefCounter;
		}
	}
}

template<typename T>
SharedPtr<T>&
SharedPtr<T>::operator=(const SharedPtr& rhs)
{
	SharedPtr(rhs).swap(*this);
	return *this;
}

template<typename T>
template<typename OT>
SharedPtr<T>&
SharedPtr<T>::operator=(const SharedPtr<OT>& rhs)
{
	SharedPtr(rhs).swap(*this);
	return *this;
}

template<typename T>
void
SharedPtr<T>::reset(T* obj)
{
	SharedPtr(obj).swap(*this);
}

template<typename T>
void
SharedPtr<T>::swap(SharedPtr& other)
{
	T* obj = other.m_pObj;
	SharedPtrRefCounter* cnt = other.m_pRefCounter;
	other.m_pObj = m_pObj;
	other.m_pRefCounter = m_pRefCounter;
	m_pObj = obj;
	m_pRefCounter = cnt;
}

std::allocator<SharedPtrRefCounter>&
SharedPtrRefCounter::allocator()
{
	static std::allocator<SharedPtrRefCounter> allocator;
	return allocator;
}

#endif

