/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "tselect.h"
#include "syscall.h"
#include <algorithm>
#include <utility>

#ifdef WIN32

#include <sys/timeb.h>

struct timezone
{
	long tz_minuteswest;
	long tz_dsttime;
};

int gettimeofday(struct timeval *tvp, struct timezone *tzp)
{
	struct timeb tb;
	ftime(&tb);
	if (tvp) {
		tvp->tv_sec = tb.time;
		tvp->tv_usec = tb.millitm * 1000;
	}
	if (tzp) {
		tzp->tz_minuteswest = tb.timezone;
		tzp->tz_dsttime = tb.dstflag;
	}
	return 0;
}

#endif

int
TSelect::addTimer(struct timeval& tv)
{
	int tid;
	if (m_timers_by_id.empty()) {
		tid = 1;
	} else if ((tid = m_timers_by_id.begin()->first) > 1) {
		--tid;
	} else {
		tid = m_timers_by_id.rbegin()->first+1;
	}
	Timeval timeval(tv.tv_sec, tv.tv_usec);
	m_timers_by_id.insert(std::map<int, Timeval>::value_type(tid, timeval));
	m_timers.insert(TimerMap::value_type(timeval+m_time_passed, tid));
	return tid;
}

void
TSelect::removeTimer(int id)
{
	std::map<int, Timeval>::iterator fit = m_timers_by_id.find(id);
	if (fit != m_timers_by_id.end()) {
		m_timers_by_id.erase(fit);
	} else {
		return;
	}
	TimerMap::iterator iter = m_timers.begin();
	for (; iter != m_timers.end(); ++iter) {
		if (iter->second == id) {
			m_timers.erase(iter);
			return;
		}
	}
}

void
TSelect::resetTimer(int id)
{
	std::map<int, Timeval>::iterator fit = m_timers_by_id.find(id);
	if (fit == m_timers_by_id.end()) {
		return;
	}
	TimerMap::iterator iter = m_timers.begin();
	for (; iter != m_timers.end(); ++iter) {
		if (iter->second == id) {
			m_timers.erase(iter);
			m_timers.insert(TimerMap::value_type(fit->second+m_time_passed, id));
			return;
		}
	}
}

int
TSelect::selectNoRetry(int n, fd_set *rfds, fd_set *wfds, fd_set *efds)
{
	struct timeval before, after, timeout, *timeout_ptr;
	
	m_triggered_timers.clear();
	int res = -1;
	if (m_timers.empty()) {
		timeout_ptr = NULL;
	} else {
		Timeval tout = m_timers.begin()->first - m_time_passed;
		if (tout < Timeval(0, 0))
			tout = Timeval(0, 0);
		timeout.tv_sec = tout.getSec();
		timeout.tv_usec = tout.getUsec();
		timeout_ptr = &timeout;
	}
	gettimeofday(&before, NULL);
	res = ::select(n, rfds, wfds, efds, timeout_ptr);
	gettimeofday(&after, NULL);
	Timeval delta = Timeval(after.tv_sec, after.tv_usec)-Timeval(before.tv_sec, before.tv_usec);
	m_time_passed = m_time_passed+delta;
	
	if (res != 0)
		return res;
	
	if (!m_timers.empty()) {
		std::pair<TimerMap::iterator, TimerMap::iterator> eq_range =
			m_timers.equal_range(m_timers.begin()->first);
		TimerMap::iterator iter = eq_range.first;
		for (; iter != eq_range.second; ++iter) {
			m_triggered_timers.insert(iter->second);
			m_timers_by_id.erase(iter->second);
		}
		m_timers.erase(eq_range.first, eq_range.second);
	}
	TimerMap::iterator threshold = m_timers.upper_bound(m_time_passed);
	TimerMap::iterator iter = m_timers.begin();
	while (iter != threshold) {
		m_triggered_timers.insert(iter->second);
		m_timers_by_id.erase(iter->second);
		m_timers.erase(iter++);
	}
	
	return 0;
}

int
TSelect::select(int n, fd_set *rfds, fd_set *wfds, fd_set *efds)
{
	int res = -1;
	do {
		res = selectNoRetry(n, rfds, wfds, efds);
#ifdef WIN32
	} while (false);
#else
	} while (res == -1 && errno == EINTR);
#endif
	return res;
}

void
TSelect::Timeval::normalize()
{
	if (m_usec < 0) {
		m_sec += (m_usec+1)/1000000-1;
		m_usec = 1000000+m_usec%1000000;
	} else {
		m_sec += m_usec/1000000;
		m_usec = m_usec%1000000;
	}
}

bool
TSelect::Timeval::operator<(const TSelect::Timeval& tv) const
{
        if (m_sec < tv.m_sec) {
		return true;
	} else if (m_sec > tv.m_sec) {
		return false;
	} else if (m_usec < tv.m_usec) {
		return true;
	} else {
		return false;
	}
}

TSelect::Timeval
TSelect::Timeval::operator+(const TSelect::Timeval& tv) const
{
	return Timeval(m_sec+tv.m_sec, m_usec+tv.m_usec);
}

TSelect::Timeval
TSelect::Timeval::operator-(const TSelect::Timeval& tv) const
{
	return Timeval(m_sec-tv.m_sec, m_usec-tv.m_usec);
}

