/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef URI_H_
#define URI_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "namevaluepair.h"
#include <string>
#include <list>
#include <iosfwd>

class URI
{
public:
	explicit URI(const std::string& str);
	
	URI(const URI& base, const URI& relative);
	
	virtual ~URI() {}
	
	static std::string escape(const std::string& str, const char *allow_extra=0);
	
	static std::string unescape(const std::string& str, bool decode_plus=false);
	
	bool isGood() const { return m_is_good; }
	
	bool isAbsolute() const { return m_is_absolute; }
	
	bool isRelativePath() const { return m_is_relative_path; }
	
	bool hasAuthority() const { return m_has_authority; }
	
	bool hasUserInfo() const { return m_has_user_info; }
	
	bool hasQuery() const { return m_has_query; }
	
	bool hasFragment() const { return m_has_fragment; }
	
	const std::string& getScheme() const { return m_scheme; }
	
	void setScheme(const std::string& scheme);
	
	const std::string& getRawSchemeSpecificPart() const { return m_raw_scheme_specific_part; }
	
	void setRawSchemeSpecificPart(const std::string& str) { parseSchemeSpecificPart(str); }
	
	const std::string& getRawAuthority() const { return m_raw_authority; }
	
	std::string getDecodedAuthority() const { return unescape(m_raw_authority); }
	
	void setRawAuthority(const std::string& authority);
	
	void removeAuthority();
	
	const std::string& getRawUserInfo() const { return m_raw_user_info; }
	
	std::string getDecodedUserInfo() const { return unescape(m_raw_user_info); }
	
	void setRawUserInfo(const std::string& user_info);
	
	const std::string& getHost() const { return m_host; }
	
	void setHost(const std::string& host);
	
	int getPort() const { return m_port; }
	
	int guessPort() const;
	
	void setPort(int port);
	
	const std::string& getRawPath() const { return m_raw_path; }
	
	std::string getDecodedPath() const { return unescape(m_raw_path); }
	
	void setAbsoluteRawPath(const std::string& path);
	
	const std::string& getRawQuery() const { return m_raw_query; }
	
	std::string getDecodedQuery() const { return unescape(m_raw_query, true); }
	
	void setRawQuery(const std::string& query);
	
	void removeQuery();
	
	std::list<NameValuePair> getRawQueryComponents(char separator='&') const;
	
	std::list<NameValuePair> getDecodedQueryComponents(char separator='&') const;
	
	const std::string& getRawFragment() const { return m_raw_fragment; }
	
	std::string getDecodedFragment() const { return unescape(m_raw_fragment); }
	
	void setRawFragment(const std::string& fragment);
	
	void removeFragment();
	
	void makeRelative(); // remove scheme and authority
	
	std::string toString() const;
	
	void toStream(std::ostream& strm) const;
private:
	void parseSchemeSpecificPart(const std::string& str);
	
	void updateSchemeSpecificPart();
	
	void parseAuthority(const std::string& str);
	
	void updateAuthority();
	
	bool validateScheme(const std::string& str) const;
	
	bool validateHost(const std::string& str) const;
private:
	static const std::string ALNUM;
	bool m_is_good;
	bool m_is_absolute;
	bool m_is_relative_path;
	bool m_has_authority;
	bool m_has_user_info;
	bool m_has_query;
	bool m_has_fragment;
	std::string m_scheme;
	std::string m_raw_scheme_specific_part;
	std::string m_raw_authority;
	std::string m_raw_user_info;
	std::string m_host;
	int m_port;
	std::string m_raw_path;
	std::string m_raw_query;
	std::string m_raw_fragment;
};


inline std::ostream&
operator<<(std::ostream& strm, const URI& uri)
{
	uri.toStream(strm);
	return strm;
}

#endif

